/**@addtogroup debug
 * @{*/

/**@addtogroup debug_ext  Interface 
 * @{*/

/**@file
 * @brief Provides configurable logging support.
 * @version 1.6
 * @date Last modified 2003-02-06 */


#ifndef _FL1DEBUG_H
#define _FL1DEBUG_H

#include <assert.h>
#include "csim/csim.h"
#include "ttpc/fp1const.h"


/*************************************************************************************/

#define LOG_LEVEL_NONE        0x0000  /**< No log level (logging disabled). @see LOG_LEVEL*/
#define LOG_LEVEL_CTRL1       0x0001  /**< TTP/C communication controller functional layer log level. @see LOG_LEVEL*/
#define LOG_LEVEL_CTRL2       0x0002  /**< TTP/C communication controller simulation layer log level. @see LOG_LEVEL*/
#define LOG_LEVEL_CHANNEL     0x0004  /**< Transmission channel log level. @see LOG_LEVEL*/
#define LOG_LEVEL_BG          0x0008  /**< Bus guardian log level. @see LOG_LEVEL*/
#define LOG_LEVEL_OS          0x0010  /**< OS specific log level. @see LOG_LEVEL*/
#define LOG_LEVEL_CRITICAL    0x0100  /**< Critical failures level. @see LOG_LEVEL*/
#define LOG_LEVEL_FI          0x0200  /**< Fault Injection log level. @see LOG_LEVEL*/
#define LOG_LEVEL_OTHER       0x0400  /**< Unspecified log level. @see LOG_LEVEL*/
#define LOG_LEVEL_APP         0x0800  /**< Application specific log level. @see LOG_LEVEL*/
#define LOG_LEVEL_ALL         0xFFFF  /**< Mask for every log level. @see LOG_LEVEL*/
#define LOG_LEVEL_USER        0xFF00  /**< Mask for every log level not used by TTP/C model itself. @see LOG_LEVEL*/

#ifndef LOG_LEVEL
/** This constant enables/disables the output level-specific logging functions.
 * 
 * This constant can be any combination of the single-level logging flags. You
 * may use @c LOG_LEVEL_ALL to enable all log-levels at once, or @c
 * LOG_LEVEL_NONE to disable loging completely. The default setting is @c
 * LOG_LEVEL_CRITICAL.
 *
 * @see LOG_LEVEL_APP, LOG_LEVEL_OS, LOG_LEVEL_CTRL1, LOG_LEVEL_CTRL2,
 *  LOG_LEVEL_CHANNEL, LOG_LEVEL_BG, LOG_LEVEL_FI LOG_LEVEL_CRITICAL,
 *  LOG_LEVEL_OTHER, LOG_LEVEL_ALL, LOG_LEVEL_USER*/
#ifdef DOXYGEN
#	define LOG_LEVEL     LOG_LEVEL_ALL
#else
#	define LOG_LEVEL     LOG_LEVEL_CRITICAL
#endif
#endif

/*************************************************************************************/

/** Return the name of current log file.
 * 
 * @return  An internally allocated string. The string should be
 * considered read-only and never deallocated. */
char *get_log_file_name(void);

/** Change the name of current log file, close the old log file.
 * 
 * @param [in] name  The new file name used for logging output. */
void set_log_file_name(char *name);

/** Print a message to log-file using the specified format string.
 * 
 * @param format  The parameters of this fuctinon follow the syntax of the
 *  standard @c printf() function. */
void print_logf(char *format, ...);

/** Print a message to log-file using the specified format string, start with
 * current simulation time and process name.
 * 
 * @param format  The parameters of this fuctinon follow the syntax of the
 *  standard @c printf() function. */
void print_logf_csim(char *format, ...);

/** Flush all buffered log-writes.*/
void log_flush(void);

/** Change behavior of log-output functions (performance vs safety).
 * 
 * @param enabled  If set to a non-zero value, the logging engine will close
 *  (flush) the output file after every write operation. This is much
 *  slower, but guarantees that the data are preserved after a crash. */
void set_log_close_after_write(int enabled);

/** Get behavior of log-output functions (performance vs safety).
 * 
 * @return
 * 	- @c 0 when the log-file is buffered and the output is not written
 * 	immediatelly to disk.
 * 	- @c 1 when the log-file is flushed after every write. */
int get_log_close_after_write(void);

/** Change the point (in model time) when logging is started.
 * 
 * @param time  A log output is performed only when the current C-Sim time is
 *  greater than the value specified in @a time.*/
void set_log_start(CSIM_TIME time);

/** Get the point (in model time) when logging is started.
 * 
 * @return  The earliest point in time for which logging is enabled. */
CSIM_TIME get_log_start(void);

/** Change the duration (in model time) of logging process.
 * 
 * @param time  The latest point in time for which logging is performed is the
 *  value returned by get_log_start() plus @a time. */
void set_log_duration(CSIM_TIME time);

/** Get the maximum duration (in model time) of logging process.
 * 
 * @return  The maximum time for which logging is performed since it was
 *  started. */
CSIM_TIME get_log_duration(void);

/** Change the format of printf function for C-Sim enhanced logging.
 * 
 * @param [in] fmt  The format string (as in printf() function) that is used to
 *  add the current time and process name to the @c _csim versions of logging
 *  macros. */
void set_csim_output_format(char *fmt);

/** Get the format of printf function for C-Sim enhanced logging.
 * 
 * @return  The printf-like format string used for printing the C-Sim time and
 *  process name of the extended logging macros. */
char *get_csim_output_format(void);


/*************************************************************************************/
/**@name Logging Macros
 * @{*/

/**@def log_app
 * Plain log of application specific message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(("Testing entry number %d (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_APP, log_app_csim */

/**@def log_app_csim
 * Log of application specific message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_APP, log_app*/

#if LOG_LEVEL & LOG_LEVEL_APP
	#define log_app(x) print_logf x
	#define log_app_csim(x) print_logf_csim x
#else
	#define log_app(x) {}
	#define log_app_csim(x) {}
#endif


/**@def log_os
 * Plain log of TTPos specific message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_OS, log_os_csim*/

/**@def log_os_csim
 * Log of TTPos specific message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_OS, log_os*/

#if LOG_LEVEL & LOG_LEVEL_OS
	#define log_os(x) print_logf x
	#define log_os_csim(x) print_logf_csim x
#else
	#define log_os(x) {}
	#define log_os_csim(x) {}
#endif

/**@def log_ctrl1
 * Plain log of TTP/C communication controller lower level message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CTRL1, log_os_ctrl1*/

/**@def log_ctrl1_csim
 * Log of TTP/C communication controller lower level message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CTRL1, log_ctrl1*/

#if LOG_LEVEL & LOG_LEVEL_CTRL1
	#define log_ctrl1(x) print_logf x
	#define log_ctrl1_csim(x) print_logf_csim x
#else
	#define log_ctrl1(x) {}
	#define log_ctrl1_csim(x) {}
#endif


/**@def log_ctrl2
 * Plain log of TTP/C communication controller upper level message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CTRL2, log_os_ctrl2*/

/**@def log_ctrl2_csim
 * Log of TTP/C communication controller upper level message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CTRL2, log_ctrl2*/

#if LOG_LEVEL & LOG_LEVEL_CTRL2
	#define log_ctrl2(x) print_logf x
	#define log_ctrl2_csim(x) print_logf_csim x
#else
	#define log_ctrl2(x) {}
	#define log_ctrl2_csim(x) {}
#endif

/**@def log_channel
 * Plain log of transmission channel specific message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CHANNEL, log_channel_csim*/

/**@def log_channel_csim
 * Log of transmission specific message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CHANNEL, log_channel*/

#if LOG_LEVEL & LOG_LEVEL_CHANNEL
	#define log_channel(x) print_logf x
	#define log_channel_csim(x) print_logf_csim x
#else
	#define log_channel(x) {}
	#define log_channel_csim(x) {}
#endif


/**@def log_bg
 * Plain log of bus guardian specific message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_BG, log_bg_csim*/

/**@def log_bg_csim
 * Log of bus guardian specific message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_BG, log_bg*/

#if LOG_LEVEL & LOG_LEVEL_BG
	#define log_bg(x) print_logf x
	#define log_bg_csim(x) print_logf_csim x
#else
	#define log_bg(x) {}
	#define log_bg_csim(x) {}
#endif


/**@def log_fi
 * Plain log of fault injection specific message.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_FI, log_fi_csim*/

/**@def log_fi_csim
 * Log of fault injection specific message enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_FI, log_fi*/

#if LOG_LEVEL & LOG_LEVEL_FI
	#define log_fi(x) print_logf x
	#define log_fi_csim(x) print_logf_csim x
#else
	#define log_fi(x) {}
	#define log_fi_csim(x) {}
#endif


/**@def log_critical
 * Plain log of critical failure messages.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CRITICAL, log_critical_csim*/

/**@def log_critical_csim
 * Log of critical failure messages enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_CRITICAL, log_critical*/

#if LOG_LEVEL & LOG_LEVEL_CRITICAL
	#define log_critical(x) print_logf x
	#define log_critical_csim(x) print_logf_csim x
#else
	#define log_critical(x) {}
	#define log_critical_csim(x) {}
#endif


/**@def log_other
 * Plain log of message that does not fit ubto any other group.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_OTHER, log_other_csim*/

/**@def log_other_csim
 * Log of message that does not fit into anyother group enhanced by current C-Sim state.
 *
 * @param x This argument is special. It should contain the format specifier as in
 *  a call to printf function and optional arguments, but the whole set of parameters
 *  must be enclosed in one additional pair of parenthesis, e.g.: @code
 *    log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 *  @endcode
 *
 * @see LOG_LEVEL_OTHER, log_other*/

#if LOG_LEVEL & LOG_LEVEL_OTHER
	#define log_other(x) print_logf x
	#define log_other_csim(x) print_logf_csim x
#else
	#define log_other(x) {}
	#define log_other_csim(x) {}
#endif

/**@def log_assert
 * This macro provides the same functionality as @c assert macro, but closes the log file
 * first. 
 *
 * It replaces the standard macro from @c <assert.h>. Additionaly it outputs a
 * log-message if @c LOG_LEVEL_CRITICAL is enabled and closes the log file (if open).
 * The function of this macro is also controled by the standard macro @c NDEBUG (If this
 * constant is defined, then the program will not terminate on failed asserions, but it
 * will still output a message  to the log using the @c log_critical_csim macro).
 *
 * @param cond An expression that evaluates to an integer value. If the evaluation of this
 *  expression is 0, the macro terminates operation of the program.
 *
 * @see LOG_LEVEL_CRITICAL, log_critical_csim */
#if LOG_LEVEL & LOG_LEVEL_CRITICAL
	#if defined NDEBUG
		#define log_assert(cond) do { \
			if (!(cond)) \
				log_critical_csim(("Assertion failed in \""__FILE__"\", line %d: ("#cond")\n", __LINE__)); \
		} while (0)
	#else
		#define log_assert(cond) do { \
			if (!(cond)) { \
				log_critical_csim(("Assertion failed in \""__FILE__"\", line %d: ("#cond")\n", __LINE__)); \
				log_flush(); \
				assert(cond); \
			} \
		} while (0)
	#endif
#else
	#if defined NDEBUG
		#define log_assert(cond)
	#else
		#define log_assert(cond) do { \
				log_flush(); \
				assert(cond); \
		} while (0)
	#endif
#endif

#endif

/*@}*/
/*************************************************************************************/

/*@}*/
/*@}*/

