/**@addtogroup misc_func
 * @{*/

/**@addtogroup misc_func_ext  Interface
 * @{*/


/**@file
 * @brief Various small and usefull functions and macros (interface)
 * @version 1.5
 * @date Last modified 2002-11-16 */


#ifndef _FL1FUNC_H
#define _FL1FUNC_H


/************************************************************************************/

/** Pointer to a function with two arguments - named by the argument type.*/
typedef void (*PTR_INT_CALLBACK)(void *, unsigned);

/** Pointer to a function with one argument - named by the argument type.*/
typedef void (*PTR_CALLBACK)(void *);


/************************************************************************************/

/** Returns a binary mask with number @a n of lower bits set to 1.
 * This macro internaly works with 32-bit integers, so do not use arguments
 * larger than 32.
 *
 * @param n  Number of bits */
#define bit_mask(n) ( ((n) < 32) ? ((1 << (n)) - 1) : 0xFFFFFFFF )

/** Extracts @a n bits from @a x staring with @a lo bit, shifts the result by 
 * @a lo bits to right. This macro uses internaly @c bit_mask macro, so its use is 
 * also limited to 32-bit integers. */
#define bit_range(x, lo, n) \
	((x >> lo) & bit_mask(n))

/** Returns an integer with only bit @a n set - bits are numbered from 0.
 * This macro internaly works with 32-bit integers, so do not use arguments larger 
 * than 31.*/
#define bit_pos(n) (((n) < 32) ? (1 << (n)) : 0x80000000)


/************************************************************************************/
/**@name Code Generation Macros
 * @{*/

/** Creates a function prototype that takes pointer to a structure as argument and 
 * returns an unsigned value read from this structure.
 * 
 * The macro creates a function prototype with the name composed by this formulae:
 * @code function_name + _get_ + field_name @endcode. The function arguments will be 
 * as follows: @code (struct_name *_structure)@endcode. The return value is 
 * @c unsigned.
 * 
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and 
 *  the last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @return Unsigned value read from the structure.
 *
 * @note For example, this macro call: @code
 *   get_function_prototype(FRAME, frame, type);
 * @endcode expands to: @code
 *   unsigned frame_get_type(FRAME *_structure);
 * @endcode 
 *
 * @see get_function_body, get_sig_function_prototype, get_index_function_prototype */
#define get_function_prototype(struct_name, function_name, field_name) \
	/** Return the value of a single unsigned integer field of the structure. */ \
	unsigned function_name##_get_##field_name(struct_name *_structure)


/** Defines a function that returns a particular unsigned field from a structure 
 * that is passed as an argument.
 *
 * This macro uses @c get_function_protype macro to create the actual function header
 * and only appends a program that performs the reading. It reads the 
 * content of @a field_name from @a _structure, truncates the value if it is
 * larger then @a bit_size bits and returns it.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and the 
 *  last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @return An unsigned value read from the structure.
 *
 * @note For example, this macro call: @code 
 *   get_function_body(FRAME, frame, type, 3)
 * @endcode expands to:	@code
 *   unsigned frame_get_type(FRAME *_structure)
 *   {
 *   	return _structure->type & bit_mask(3);
 *   }
 * @endcode
 *
 * @see get_function_prototype, get_sig_function_body, get_index_function_body */
#define get_function_body(struct_name, function_name, field_name, bit_size)  \
	get_function_prototype(struct_name, function_name, field_name)       \
	{                                                                    \
		return _structure->field_name & bit_mask(bit_size);          \
	}


/*----------------------------------------------------------------------------------*/

/** Creates a function prototype that takes pointer to a structure as argument and 
 * returns a signed value read from this structure.
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _get_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure). The return value is @c signed.
 * 
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and the 
 *  last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @return Signed value read from the structure.
 *
 * @note For example, this macro call: @code
 *   get_sig_function_prototype(FRAME, frame, type);
 * @endcode expands to: @code
 *   signed frame_get_type(FRAME *_structure);
 * @endcode
 * 
 * @see get_sig_function_body, get_function_body, get_index_function_prototype */
#define get_sig_function_prototype(struct_name, function_name, field_name)       \
	/** Get the value of a single signed integer field of the structure. */ \
	signed function_name##_get_##field_name(struct_name *_structure)


/** Defines a function that returns a particular signed field from a structure 
 * that is passed as an argument.
 *
 * This macro uses @c get_sig_function_protype macro to create the actual function 
 * header and only appends a program that performs the reading. It reads the 
 * content of @a field_name from @a _structure, truncates the value if it is
 * larger then @a bit_size bits and returns it. The sign bit is replicated in all
 * unused bits of the result, so the number can be used normaly for arithmetics.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and the 
 *  last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @return A signed value read from the structure.
 *
 * @note For example, this macro call: @code 
 *   get_sig_function_body(FRAME, frame, type, 3)
 * @endcode expands to:	@code
 *   signed frame_get_type(FRAME *_structure)
 *   {
 *   	signed x = _structure->type;
 *   		if ((x & (1 << (3 - 1))) == 0)
 *   			return x;
 *   		else
 *   			return x | ~bit_mask(3);
 *   }
 * @endcode
 *
 * @see get_sig_function_prototype, get_function_body, get_index_function_body */
#define get_sig_function_body(struct_name, function_name, field_name, bit_size)  \
	get_sig_function_prototype(struct_name, function_name, field_name)       \
	{                                                                    \
		signed x = _structure->field_name; \
		if ((x & (1 << (bit_size - 1))) == 0) \
			return x; \
		else \
			return x | ~bit_mask(bit_size); \
	}

/*----------------------------------------------------------------------------------*/


/** Creates a function prototype that takes pointer to a structure and unsigned
 * index as arguments and returns an unsigned value read from a vector in
 * the structure.
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _get_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure, @c unsigned @a _index). The return 
 * value is @c unsigned.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and the
 *   last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _index (fn) Integer index of one item within the @a field_name vetor.
 * @return Unsigned value read from the structure.
 *
 * @note For example, this macro call: @code
 *   get_index_function_prototype(PEOPLE, people, age);
 * @endcode expands to: @code
 *   unsigned people_get_age(PEOPLE *_structure, unsigned _index);
 * @endcode
 *
 * @see get_index_function_body, get_function_prototype, get_sig_function_prototype */
#define get_index_function_prototype(struct_name, function_name, field_name) \
	/** Get the value of a single unsigned array element within the structure. */ \
	unsigned function_name##_get_##field_name(struct_name *_structure, unsigned _index)


/** Defines a function that returns a chosen unsigned field from a particular vector
 * within a structure that is passed as an argument.
 *
 * This macro uses @c get_index_function_protype macro to create the actual function 
 * header and only appends a program that performs the reading. It reads the 
 * content of @a field_name [@a _index] from @a _structure, truncates the value 
 * if it is larger then @a bit_size bits and returns it.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and the 
 *  last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _index (fn) Integer index of one item within the @a field_name vetor.
 * @return An unsigned value read from the structure.
 *
 * @note For example, this macro call: @code 
 *   get_index_function_body(PEOPLE, people, age, 8)
 * @endcode expands to: @code	
 *   unsigned people_get_age(PEOPLE *_structure, unsigned _index)
 *   {
 *   	return _structure->age[_index] & bit_mask(8);
 *   }
 * @endcode
 *
 * @see get_function_prototype, get_sig_function_body, get_index_function_body*/
#define get_index_function_body(struct_name, function_name, field_name, bit_size)  \
	get_index_function_prototype(struct_name, function_name, field_name)       \
	{                                                                          \
		return _structure->field_name[_index] & bit_mask(bit_size);        \
	}

/*----------------------------------------------------------------------------------*/


/** Creates a function prototype that takes pointer to a structure and an unsigned
 * value as arguments and returns an unsigned value. 
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _set_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure, @c unsigned @a _value). The return 
 * value is @c unsigned.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and 
 *  the last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) An unsigned value that will be written into the @a field_name item.
 * @return The unsigned value that has been written.
 * 
 * @note For example, this macro call: @code
 *   set_function_prototype(FRAME, frame, type);
 * @endcode expands to:	@code
 *   unsigned frame_set_type(FRAME *_structure, unsigned _value);
 * @endcode
 *
 * @see set_function_body, set_sig_function_prototype, set_index_function_prototype*/
#define set_function_prototype(struct_name, function_name, field_name)                    \
	/** Set the value of a single unsigned integer field within the structure. */ \
	unsigned function_name##_set_##field_name(struct_name *_structure, unsigned _value)


/** Defines a function that writes a new unsigned value to a particular field of
 * structure that is passed to the function; returns the written value.
 *
 * This macro uses @c set_function_protype macro to create the actual function header
 * and only appends a program that performs the write opearion. It first truncates
 * the content of @a _value to the maximum of @a bit_size bits and then writes the
 * modified value to @a field_name within the structure @a _structure.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and the
 *  last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) An unsigned value that will be written into the @a field_name item.
 * @return The unsigned value that has been written.
 *
 * @note For example, this macro call: @code 
 *   set_function_body(FRAME, frame, type, 3)
 * @endcode expands to:	@code
 *   unsigned frame_set_type(FRAME *_structure, unsigned _value)
 *   {
 *   	return _structure->type = _value & bit_mask(3);
 *   }
 * @endcode
 *
 * @see set_function_prototype, set_sig_function_body, set_index_function_body*/
#define set_function_body(struct_name, function_name, field_name, bit_size)  \
	set_function_prototype(struct_name, function_name, field_name)       \
	{                                                                    \
		return _structure->field_name = _value & bit_mask(bit_size); \
	}

/*----------------------------------------------------------------------------------*/


/** Creates a function prototype that takes pointer to a structure and a signed
 * value as arguments and returns a signed value. 
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _set_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure, @c signed @a _value). The return 
 * value is @c signed.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new
 *  function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure 
 *  and the last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) A signed value that will be written into the @a field_name item.
 * @return The signed value that has been written.
 *
 * @note For example, this macro call: @code
 *   set_sig_function_prototype(FRAME, frame, type);
 * @endcode expands to:	@code
 *   signed frame_set_type(FRAME *_structure, signed _value);
 * @endcode
 *
 * @see set_sig_function_body, set_function_prototype, set_index_function_prototype*/
#define set_sig_function_prototype(struct_name, function_name, field_name)                    \
	/** Set the value of a single signed integer field within the structure. */ \
	signed function_name##_set_##field_name(struct_name *_structure, signed _value)


/** Defines a function that writes a new signed value to a particular field of
 * structure that is passed to the function; returns the written value.
 *
 * This macro uses @c set_function_protype macro to create the actual function header
 * and only appends a program that performs the write opearion. It first truncates
 * the content of @a _value to the maximum of @a bit_size bits and then writes the
 * modified value to @a field_name within the structure @a _structure.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new 
 *  function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure 
 *  and the last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits 
 *  are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) A signed value that will be written into the @a field_name item.
 * @return The signed value that has been written.
 *
 * @note For example, this macro call: @code 
 *   set_sig_function_body(FRAME, frame, type, 3)
 * @endcode expands to:	@code
 *   signed frame_set_type(FRAME *_structure, signed _value)
 *   {
 *   	_structure->field_name = _value & bit_mask(bit_size);
 *   	return _value;
 *   }
 * @endcode
 *
 * @see set_sig_function_prototype, set_function_body, set_index_function_body*/
#define set_sig_function_body(struct_name, function_name, field_name, bit_size)  \
	set_sig_function_prototype(struct_name, function_name, field_name)       \
	{                                                                    \
		_structure->field_name = _value & bit_mask(bit_size); \
		return _value; \
	}

/*----------------------------------------------------------------------------------*/


/** Creates a function prototype that takes pointer to a structure, unsigned index 
 * and an unsigned	value as arguments and returns an unsigned value. 
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _set_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure, @c unsigned @a _index, @c unsigned @a _value). 
 * The return value is @c unsigned.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a vector inside the @a struct_name structure 
 *  and the last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _index (fn) Integer index of one item within the @a field_name vetor.
 * @param _value (fn) An unsigned value that will be written into @a field_name [@a _index ].
 * @return The unsigned value that has been written.
 *
 * @note For example, this macro call: @code
 *   set_index_function_prototype(PEOPLE, people, age);
 * @endcode expands to:	@code
 *   unsigned people_set_age(PEOPLE *_structure, unsigned _index, unsigned _value);
 * @endcode
 *
 * @see set_index_function_body, set_function_prototype, set_sig_function_prototype*/
#define set_index_function_prototype(struct_name, function_name, field_name)                                 \
	/** Set the value of a single unsigned array element within the structure. */ \
	unsigned function_name##_set_##field_name(struct_name *_structure, unsigned _index, unsigned _value)



/** Defines a function that writes a new unsigned value to chosen item of a particular
 * vector within a given structure; returns the written value.
 *
 * This macro uses @c set_index_function_protype macro to create the actual function header
 * and only appends a program that performs the write opearion. It first truncates
 * the content of @a _value to the maximum of @a bit_size bits and then writes the
 * modified value to @a field_name [@a _index ] within the structure @a _structure.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a vector inside the @a struct_name structure 
 *  and the last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits 
 *  are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _index (fn) Integer index of one item within the @a field_name vetor.
 * @param _value (fn) An unsigned value that will be written into @a field_name [@a _index ].
 * @return The unsigned value that has been written.
 *
 * @note For example, this macro call: @code 
 *   set_index_function_body(PEOPLE, people, age, 8)
 * @endcode expands to: @code
 *   unsigned people_set_age(PEOPLE *_structure, unsigned _index, unsigned _value)
 *   {
 *   	return _structure->age[_index] = _value & bit_mask(8);
 *   }
 * @endcode
 *
 * @see set_function_prototype, set_sig_function_body, set_index_function_body*/
#define set_index_function_body(struct_name, function_name, field_name, bit_size)         \
	set_index_function_prototype(struct_name, function_name, field_name)              \
	{                                                                                 \
		return _structure->field_name[_index] = _value & bit_mask(bit_size);      \
	}

/*----------------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------------*/


/** Creates a function prototype that takes pointer to a structure and a signed
 * value as arguments and returns an unsigned value. 
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _increment_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure, @c int @a _value). The return 
 * value is @c unsigned.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure 
 *  and the last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) A signed value that will be added to the @a field_name item.
 * @return The result of the addition.
 *
 * @note For example, this macro call: @code
 *   increment_function_prototype(QUEUE, queue, count);
 * @endcode expands to:	@code
 *   unsigned queue_increment_count(QUEUE *_structure, int _value);
 * @endcode
 *
 * @see increment_function_body, increment_sig_function_prototype*/
#define increment_function_prototype(struct_name, function_name, field_name)                      \
	/** Increment the value of a single unsigned integer field within the structure by a given value. */ \
	unsigned function_name##_increment_##field_name(struct_name *_structure, int _value)


/** Defines a function that adds a signed value to a particular field of
 * structure that is passed to the function; returns the resultant sum.
 *
 * This macro uses @c increment_function_protype macro to create the actual function header
 * and only appends a program that performs the add opearion. For this purpose it
 * a \em _get_ and \em _set_ functions for the particular @a struct_type and @a field_name. Such
 * functions can be created using the @c get_function_body and @c set_body_function
 * macros.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure
 *  and the last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. 
 *  Extra bits are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) A signed value that will be added to the @a field_name item.
 * @return The resulting sum.
 *
 * @note For example, this macro call: @code 
 *   increment_function_body(QUEUE, queue, count, 16)
 * @endcode expands to: @code
 *   unsigned queue_increment_count(QUEUE *_structure, int _value)
 *   {
 *   	unsigned _u = (queue_get_count(_structure) + _value) & bit_mask(16);
 *   	queue_set_count(_structure, _u);
 *   	return _u;
 *   }
 * @endcode
 * 
 * @see increment_function_prototype, get_function_body, set_function_body*/
#define increment_function_body(struct_name, function_name, field_name, bit_size)                           \
	increment_function_prototype(struct_name, function_name, field_name)                                \
	{                                                                                                   \
		unsigned _u = (function_name##_get_##field_name(_structure) + _value) & bit_mask(bit_size); \
		function_name##_set_##field_name(_structure, _u);                                           \
		return _u;                                                                                  \
	}

/*----------------------------------------------------------------------------------*/


/** Creates a function prototype that takes pointer to a structure and a signed
 * value as arguments and returns a signed value. 
 *
 * The macro creates a function prototype with the name composed by this formulae:
 * @a function_name + @c _increment_ + @a field_name. The function arguments will be as 
 * follows: (@c struct_name @a *_structure, @c int @a _value). The return 
 * value is @c signed.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure
 *  and the last part of function name.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) A signed value that will be added to the @a field_name item.
 * @return The result of the addition.
 * 
 * @note For example, this macro call: @code
 *   increment_sig_function_prototype(QUEUE, queue, count);
 * @endcode expands to:	@code
 *   signed queue_increment_count(QUEUE *_structure, int _value);
 * @endcode
 *
 * @see increment_sig_function_body, increment_function_prototype*/
#define increment_sig_function_prototype(struct_name, function_name, field_name)                      \
	/** Increment the value of a single signed integer field within the structure by a given value. */ \
	signed function_name##_increment_##field_name(struct_name *_structure, int _value)


/** Defines a function that adds a signed value to a particular field of
 * structure that is passed to the function; returns the resultant sum.
 *
 * This macro uses @c increment_function_protype macro to create the actual function header
 * and only appends a program that performs the add opearion. For this purpose it
 * a \em _get_ and \em _set_ functions for the particular @a struct_type and @a field_name. Such
 * functions can be created using the @c get_sig_function_body and @c set_sig_body_function
 * macros.
 *
 * @param struct_name (macro) Type name of the structure we want to access.
 * @param function_name (macro) Identifier that forms the first part of the new function name.
 * @param field_name (macro) The name of a field inside the @a struct_name structure and 
 *  the last part of funtion name.
 * @param bit_size (macro) Number of bits used by the field @a field_name. Extra bits are truncated.
 * @param _structure (fn) Pointer to the structure that will be accessed by the function.
 * @param _value (fn) A signed value that will be added to the @a field_name item.
 * @return The resulting sum.
 *
 * @note For example, this macro call: @code 
 *   increment_sig_function_body(QUEUE, queue, count, 16)
 * @endcode expands to: @code
 *   signed queue_increment_count(QUEUE *_structure, int _value)
 *   {
 *   	signed _u = (queue_get_count(_structure) + _value);
 *   	queue_set_count(_structure, _u);
 *   	return _u;
 *   }
 * @endcode
 *
 * @see increment_sig_function_prototype, get_sig_function_body, set_sig_function_body*/
#define increment_sig_function_body(struct_name, function_name, field_name, bit_size)                           \
	increment_sig_function_prototype(struct_name, function_name, field_name)                                \
	{                                                                                                   \
		signed _u = function_name##_get_##field_name(_structure) + _value; \
		function_name##_set_##field_name(_structure, _u);                                           \
		return _u;                                                                                  \
	}

/*----------------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------------*/


/** Create function prototypes for reading, writing and incrementing of a particular
 * unsigned field within a given structure.
 *
 * The macro calls in sequence three other macros: @c get_function_prototype,
 * @c set_function_prototype and @c increment_function_prototype. This way
 * a complete set of access functions prototypes to an unsigned field @a field_name
 * within the structure @a struct_name is created.
 *
 * @param struct_name (macro) Type name of the structure to access.
 * @param function_name (macro) Identifier that forms the first part of the generated 
 *  function names.
 * @param field_name (macro) The name of a field inside the @a struct_name structure
 *  and the last part of funtion names.
 * 
 * @attention A semicolon after this macro is mandatory.
 * 
 * @see implement_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype*/
#define declare_access_functions(struct_name, function_name, field_name)    \
	extern get_function_prototype(struct_name, function_name, field_name);     \
	extern set_function_prototype(struct_name, function_name, field_name);     \
	extern increment_function_prototype(struct_name, function_name, field_name)

/** Define functions for reading, writing and incrementing of a particular
 * unsigned field within a given structure.
 *
 * The macro calls in sequence three other macros: @c get_function_body,
 * @c set_function_body and @c increment_function_body. This way a complete set 
 * of access functions to an unsigned field @a field_name within the structure 
 * @a struct_name is defined.
 *
 * @param struct_name (macro) Type name of the structure to access.
 * @param function_name (macro) Identifier that forms the first part of the 
 *  generated function names.
 * @param field_name (macro) The name of a field inside the @a struct_name structure 
 *  and the last part of funtion names.
 * @param bit_size (macro) Number of bits used by the field @a field_name.
 *  Extra bits are truncated.
 *  
 * @attention Do not put a semicolon after this macro.
 * 
 * @see declare_access_functions, get_function_body, set_function_body, 
 *  increment_function_body*/
#define implement_access_functions(struct_name, function_name, field_name, bit_size)  \
	get_function_body(struct_name, function_name, field_name, bit_size)           \
	set_function_body(struct_name, function_name, field_name, bit_size)           \
	increment_function_body(struct_name, function_name, field_name, bit_size)

/*----------------------------------------------------------------------------------*/

/** Create function prototypes for reading, writing and incrementing of a particular
 * signed field within a given structure.
 *
 * The macro calls in sequence three other macros: @c get_sig_function_prototype,
 * @c set_sig_function_prototype and @c increment_sig_function_prototype. This way
 * a complete set of access function prototypes to a signed field @a field_name 
 * within the structure @a struct_name is created.
 *
 * @param struct_name (macro) Type name of the structure to access.
 * @param function_name (macro) Identifier that forms the first part of the
 *  generated function names.
 * @param field_name (macro) The name of a field inside the @a struct_name 
 *  structure and the last part of funtion names.
 *
 * @attention A semicolon after this macro is mandatory.
 *
 * @see implement_sig_access_functions, get_sig_function_prototype, 
 *  set_sig_function_prototype, increment_sig_function_prototype*/
#define declare_sig_access_functions(struct_name, function_name, field_name)    \
	extern get_sig_function_prototype(struct_name, function_name, field_name);     \
	extern set_sig_function_prototype(struct_name, function_name, field_name);     \
	extern increment_sig_function_prototype(struct_name, function_name, field_name)


/** Define functions for reading, writing and incrementing of a particular
 * signed field within a given structure.
 *
 * The macro calls in sequence three other macros: @c get_sig_function_body,
 * @c set_sig_function_body and @c increment_sig_function_body. This way
 * a complete set of access functions to a signed field @a field_name within the
 * structure @a struct_name is defined.
 *
 * @param struct_name (macro) Type name of the structure to access.
 * @param function_name (macro) Identifier that forms the first part of the 
 *  generated function names.
 * @param field_name (macro) The name of a field inside the @a struct_name 
 *  structure and the last part of funtion names.
 * @param bit_size (macro) Number of bits used by the field @a field_name. 
 *  Extra bits are truncated.
 *
 * @attention Do not put a semicolon after this macro.
 *
 * @see declare_sig_access_functions, get_sig_function_body, 
 *  set_sig_function_body, increment_sig_function_body*/
#define implement_sig_access_functions(struct_name, function_name, field_name, bit_size)  \
	get_sig_function_body(struct_name, function_name, field_name, bit_size)           \
	set_sig_function_body(struct_name, function_name, field_name, bit_size)           \
	increment_sig_function_body(struct_name, function_name, field_name, bit_size)

/*----------------------------------------------------------------------------------*/

/** Create function prototypes for reading and writing of a chosen unsigned item 
 * in a particular vector within a given structure.
 *
 * The macro calls in sequence two other macros: @c get_index_function_prototype and
 * @c set_index_function_prototype. This way a complete set of access function prototypes 
 * to an unsigned vector @a field_name within the structure @a struct_name is created.
 *
 * @param struct_name (macro) Type name of the structure to access.
 * @param function_name (macro) Identifier that forms the first part of the 
 *  generated function names.
 * @param field_name (macro) The name of a vector inside the @a struct_name 
 *  structure and the last part of funtion names.
 *
 * @attention A semicolon after this macro is mandatory.
 *
 * @see implement_index_access_functions, get_function_prototype, 
 *  set_function_prototype*/
#define declare_index_access_functions(struct_name, function_name, field_name) \
	extern get_index_function_prototype(struct_name, function_name, field_name);  \
	extern set_index_function_prototype(struct_name, function_name, field_name)


/** Define functions for reading and writing of a chosen unsigned item in
 * a particular vector within a given structure.
 *
 * The macro calls in sequence two other macros: @c get_index_function_body and
 * @c set_index_function_body. This way a complete set of access functions to an 
 * unsigned vector @a field_name within the structure @a struct_name is defined.
 *
 * @param struct_name (macro) Type name of the structure to access.
 * @param function_name (macro) Identifier that forms the first part of the 
 *  generated function names.
 * @param field_name (macro) The name of a vector inside the @a struct_name
 *  structure and the last part of funtion names.
 * @param bit_size (macro) Number of bits used by one item of the vector @a 
 *  field_name. Extra bits are truncated.
 * 
 * @attention Do not put a semicolon after this macro.
 *
 * @see declare_index_access_functions, get_index_function_body, 
 *  set_index_function_body*/
#define implement_index_access_functions(struct_name, function_name, field_name, bit_size) \
	get_index_function_body(struct_name, function_name, field_name, bit_size)          \
	set_index_function_body(struct_name, function_name, field_name, bit_size)

/*@}*/
/************************************************************************************/


/** Performs continous calculation of 16-bit CRC code using the provided buffer.
 * 
 * The function may be used to calculate CRC from data of any size. Individial
 * items of the data are feeded sequentialy into this function. The size of
 * inidividual data items can be up to 32 bits. Before a new CRC is calculated,
 * the content of @a crc variable should be initialized.
 * 
 * @param [in,out] crc  A user allocated buffer, where the CRC is stored. It must be
 *  initialized or contain an older CRC value.
 * @param polynomial  The polynomial used for CRC calculations.
 * @param data  A piece of data to be included into the CRC code.
 * @param bits  Number of bits to be taken from the @a data argument.
 * @return  The current value of CRC; the same as the content of @a crc
 *  argument.
 *
 * @todo Optimize using a byte table*/
extern unsigned get_crc(unsigned *crc, unsigned polynomial, unsigned data, unsigned bits);

/** Returns the fractional part of a floating point number.
 *
 * @param x A floating point number.
 * @return The fractional part of @a x.*/
extern double fract(double x);


/************************************************************************************/

#endif /* #ifndef _FL1FUNC_H */

/*@}*/
/*@}*/

