/**@addtogroup bg_fun
 * @{*/

/**@addtogroup busg_fun_ext  Interface
 * @{*/

/**@file
 * @brief Local bus guardian functionality model (interface).
 * @version 1.5
 * @date Last modified 2002-11-16 */

#ifndef _FP1BUSG_H
#define _FP1BUSG_H

#include "ttpc/fp1const.h"
#include "ttpc/fl1func.h"
#include "ttpc/fp1medl.h"


/************************************************************************************/

/** Bus guardian tick length - controls the internal clock of BG.*/
#define BUS_GUARDIAN_CLK_SPD	(BUS_GUARDIAN_FREQUENCY / BUS_GUARDIAN_DIVISOR)


/************************************************************************************/

struct bus_guardian_t;

/** Pointer to function that can set a single unsigned integer value within BG.
 * The BG contains hooks that can be used to implement something like virtual methods.
 * Functions that have to be implemented by a module that encapsulates BG are implemented
 * as pointers.
 * 
 * @param bg pointer to the instance of BG. This instance must be either of type
 * 	@c struct @c bus_guardian or a type correctly derived.
 * @param value The new content of the modified field.
 * @return Returns the written value.
 *
 * @see d_bus_guardian, BUS_GUARDIAN */
typedef unsigned (*BUS_GUARDIAN_SET_FN)(struct bus_guardian_t *bg, unsigned value);


/** State codes used for the internal state machine of BG.*/
typedef enum {
	BGS_INITIALIZE = 0, /**< Initialize - the startup default.*/
	BGS_CONFIGURE,      /**< Configure - BG parameters may be modified.*/
	BGS_ARM,            /**< Arm - Awaiting the BA signal.*/
	BGS_START_TIMER,    /**< Start timer - Bus disabled for slot offset phase.*/
	BGS_ENABLE,         /**< Enable - Bus enabled for slot duration.*/
	BGS_DISABLE,        /**< Disable - Bus disabled for remainding round phase.*/
	BGS_NODE_FAILURE    /**< Node Failure - Final state for detected errors.*/
} BUS_GUARDIAN_STATES;


/** Error codes that can be found in the @c error_code field of BG.
If an error is detected inside the BG unit, the @c er flag is set and one of
these codes is writen into @c error_code field.*/
typedef enum {
	BGEC_NO_ERROR = 0,                        /**< This is the value used when no error occurs.*/
	BGEC_HALT = 1,                            /**< BG operation was manualy stopped.*/ 
	BGEC_BUS_ACCESS_VIOLATION = 2,            /**< Attempting to send while bus access is disabled.*/
	BGEC_ARM_DURING_BUS_ACCESS_CYCLE = 3,     /**< BA signal received during the wrong phase.*/
	BGEC_MISSING_ARM = 4,                     /**< Arm signal was not received during the ARM phase.*/
	BGEC_ARM_DURING_TRANSMISSION_DISABLED = 5,/**< BA signal received while BG was explicitly disabled.*/
	BGEC_UNKNOWN_ERROR                        /**< Any other error was detected.*/
} BUS_GUARDIAN_ERROR_CODES;


/** A list of all BG protocol-defined data fields.
 * This macro is used to simplify the type derivation used within the model. It lists
 * all items that define the state of a BG unit in a format that can be used to define
 * the content of a @c struct type. All @c struct types that include this macro at the
 * beginning of their definition are then backwards compatible (we call them derived
 * structures. */
#define d_bus_guardian \
	/* Configuration Registers */ \
	unsigned tap;           /**< Arm phase duration.*/ \
	unsigned tso;           /**< Slot offset duration.*/ \
	unsigned tsd;           /**< Slot duration.*/ \
	unsigned trr;           /**< Remaining round duration.*/ \
	/* Control Registers */ \
	unsigned am;            /**< Arm signal.*/ \
	unsigned ht;            /**< Halt flag.*/ \
	unsigned rc;            /**< Reconfiguration flag.*/ \
	unsigned cmod;          /**< Controller mode - enabled/disabled channels.*/ \
	/* Status Registers */ \
	unsigned error_code;    /**< Numerical identification of last error.*/ \
	unsigned state_code;    /**< Current bus guardian state (image).*/ \
	unsigned lf;            /**< Lifesign flag.*/ \
	unsigned er;            /**< Error flag.*/ \
	/* Internal state */ \
	unsigned saved_cmod;    /**< Temporary storage for @c cmod value.*/ \
	void *controller;       /**< Pointer to the attached controller instance.*/ \
	unsigned bg_state;      /**< Current state of internal BG state machine.*/ \
	BUS_GUARDIAN_SET_FN arm_signal;  /**< Virtual function pointer.*/;


/** Internal state of the BG. */
typedef struct bus_guardian_t {
	d_bus_guardian
} BUS_GUARDIAN;


/************************************************************************************/
/**@name Generated Access Functions
 * Access functions for @c BUS_GUARDIAN that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 * increment_function_prototype 
 * @{*/
declare_access_functions(BUS_GUARDIAN, bus_guardian, tap);
declare_access_functions(BUS_GUARDIAN, bus_guardian, tso);
declare_access_functions(BUS_GUARDIAN, bus_guardian, tsd);
declare_access_functions(BUS_GUARDIAN, bus_guardian, trr);
declare_access_functions(BUS_GUARDIAN, bus_guardian, cmod);
declare_access_functions(BUS_GUARDIAN, bus_guardian, error_code);
declare_access_functions(BUS_GUARDIAN, bus_guardian, state_code);
declare_access_functions(BUS_GUARDIAN, bus_guardian, lf);
declare_access_functions(BUS_GUARDIAN, bus_guardian, er);
get_function_prototype(BUS_GUARDIAN, bus_guardian, am);
set_function_prototype(BUS_GUARDIAN, bus_guardian, am);
get_function_prototype(BUS_GUARDIAN, bus_guardian, ht);
set_function_prototype(BUS_GUARDIAN, bus_guardian, ht);
get_function_prototype(BUS_GUARDIAN, bus_guardian, rc);
set_function_prototype(BUS_GUARDIAN, bus_guardian, rc);
/*@}*/

/************************************************************************************/

/** Switch the BG into node-failure state and set the correct error code.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param error_code This value will be written into the @c error_code field of BG. */
void bus_guardian_raise_error(BUS_GUARDIAN *bg, unsigned error_code);

/** Check whether transmission on the specified channel is enabled (1) or disabled (0).
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param channel The number of a channel - either 0 (channel A) or 1 (channel B).
 * @return Returns 1 if the channel is open or 0 if it is closed. */
int bus_guardian_is_enabled(BUS_GUARDIAN *bg, unsigned channel);

/** Set the initial power-on values of BG - called after creation of BG.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.*/
void bus_guardian_init(BUS_GUARDIAN *bg);

/** Perform a reconfiguration cycle for BG and writes a new paramet set.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param tap  bus arm phase duration.
 * @param tso  slot open phase duration.
 * @param tsd  slot disabled phase duration.
 * @param trr  remaining round phase duration. */
void bus_guardian_set_params(BUS_GUARDIAN *bg, MEDL_BGP *params);

/** Perform a reconfiguration cycle for BG and writes a new paramet set 
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param params Pointer to a structure that contains the new parameters. The parameters
 * 	are expected in a format that is used within the MEDL table. */
void bus_guardian_set_params_ext(BUS_GUARDIAN *bg, unsigned tap, unsigned tso,
		unsigned tsd, unsigned trr);

/** Perform all actions appropiate for BG init state.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_init_state(BUS_GUARDIAN *bg);

/** Perform all actions appripiate for BG configure state.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_configure_state(BUS_GUARDIAN *bg);

/** Perform all actions appropiate for BG arm state.  This function is called
 * when the arm phase of BG ends and the unit should change its state. It
 * checks for a possible reconfiguration request and monitors the arm signal.
 * If an arm signal occurs, a new bus access cycle is started.
 * 
 * @note Modification of TTP/C C1 specification: when no arm signal is received
 * during the bus arm phase, an error is raised.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_arm_state(BUS_GUARDIAN *bg);

/** Perform all actions appropiate for BG start_timer state.  This function is
 * called when the BG transits from  start_timer into enable state.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_start_timer_state(BUS_GUARDIAN *bg);

/** Perform all actions appropiate for BG enable state.  This function is
 * called when the BG transits from enable into disable state. If the arm
 * signal is still active at this point, an error is raised.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_enable_state(BUS_GUARDIAN *bg);

/** Perform all actions appropiate for BG disable state.  This function is
 * called when the BG transits from \e disable into arm state.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.*/
void bus_guardian_disable_state(BUS_GUARDIAN *bg);

/************************************************************************************/

#endif /* #ifndef _FP1BUSG_H */

/*@}*/
/*@}*/

