/**@addtogroup chan_fun
 * @{*/

/**@addtogroup chan_fun_ext  Interface
 * @{*/

/**@file
 * @brief Functional model of a bus transmission channel (interface).
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */

#ifndef _FP1CHAN_H
#define _FP1CHAN_H

#include "ttpc/fl1debug.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1frame.h"
#include "ttpc/fl1func.h"

/************************************************************************************/

/** The internal state of a channel.
 *
 * @note This structure should be in fact derived using the @c d_channel macro.
 * It was intentionally defined explicitly to allow the documentation generator
 * (doxygen) document all its fields. */
struct channel_t;

/** Describes current transmission state of channel.*/
typedef enum {
	CHANNEL_TS_START,  /**< Transmission started and isn't complete yet.*/
	CHANNEL_TS_FINISH  /**< Transmission already finished - no active transmission.*/
} CHANNEL_TS;


/** Pointer to a notification function that is called for every registered
 * receiver after a transmission starts/ends.
 *
 * A controller that is listening on the channel must register using the
 * appropiate function call (@c channel_receive). After a frame transmission is
 * started/finished, a registered controller is immediatelly notified using a
 * stored function of this type.
 *
 * @param arg User defined argument that is stored on registration and will
 *  passed every time the function is called (e.g. pointer to the called
 *  controller).
 * @param channel The ID number of this channel. Can be 0 (for channel A) or 1
 *  (channel B).
 * @param ts Current transmission state of the channel (active/inactive).
 * @param frame Pointer to a channel-internal image of the transmitted frame.
 *
 * @see FRAME, CONTROLLER, CHANNEL_TS */
typedef void (*CHANNEL_ON_RCV_FN)(void *arg, unsigned channel, CHANNEL_TS ts, FRAME *frame);


/** Pointer to a function that sends a frame - called usually by a controller.
 * 
 * The channel contains hooks that can be used to implement something like
 * virtual methods.  Functions that have to be implemented by a module that
 * encapsulates the basic channel defined here are implemented as pointers.
 * This specific function type enables to create own functions that send a
 * frame using the channel. The calling controller registrates using a callback
 * function to receive notification about the completion of sending operation.
 *
 * @param channel Pointer to a @c CHANNEL instance or a derived structure.
 * @param arg User defined argument that is stored on function call and will
 *  passed to the callback function later @a fn.
 * @param fn Pointer to a callback function that is used to inform the sender
 *  about job	comletion.
 * @return The function is expected to return 0 if no error is detected, and
 *  user defined non-zero code when the frame could not be sent.
 *
 * @see PTR_CALLBACK, CHANNEL */
typedef int (*CHANNEL_SEND_FRAME_FN)(struct channel_t *chnl, void *arg, PTR_CALLBACK fn);


/** Pointer to a notification function that is called for every transmission
 * and enables to selectively modify the sent frame.
 *
 * A controller that is interested the posibility to monitor/alter all sent
 * frames must register using the appropiate function call (@c
 * channel_set_transmission_callback).  This function is then called with the
 * stored parameters.
 *
 * @param arg User defined argument that is stored on registration and will
 *  passed every time the function is called (e.g. pointer to the called
 *  controller).
 * @param frame Pointer to a channel-internal image of the transmitted frame.
 * @param ts Current transmission state of the channel (active/inactive).
 *
 * @see FRAME, CHANNEL_TS */
typedef void (*CHANNEL_TRANSMISSION_FN)(void *arg, FRAME *frame, CHANNEL_TS ts);


/** A list of all channel protocol-required data fields.
 *
 * This macro is used to simplify the type derivation used within the model. It
 * lists all items that define the state of a single transmission channel in a
 * format that can be used to define the content of a @c struct type. All @c
 * struct types that include this macro at the beginning of their definition
 * are then backwards compatible (we call them derived structures. */
#define d_channel \
	unsigned bus_id; 	/**< Unique identification (@c 0 or @c 1) */ \
	int active; 		/**< Transmission activity flag */ \
	unsigned collisions;    /**< Number of collisions */ \
	FRAME frame;		/**< Currently transmitted frame copy */ \
	void *sender; 		/**< Pointer to the sending controller */ \
	unsigned receiver_count;        /**< Number of receiving controllers */ \
	void *receiver[MAX_SRU_COUNT];  /**< List of all receiving controllers */ \
	\
	/* Callback functions */ \
	PTR_CALLBACK send_callback;     /**< Called when my transmission is finished */ \
	CHANNEL_ON_RCV_FN receive_callback[MAX_SRU_COUNT];  /**< Called when incoming frame transmission starts/stops */ \
	PTR_CALLBACK bg_error_callback;  /**< Called when BG prohibits transmissions */ \
	CHANNEL_TRANSMISSION_FN transmission_callback;  /**< Monitoring callback executed on every transmission */ \
	void *transmission_callback_arg;   /**< argument to @c transmission_callback function */ \
	/* Virtual functions */ \
	CHANNEL_SEND_FRAME_FN channel_send_frame;  /**< Send a frame on this channel */ ;


/** The internal state of a channel. */
typedef struct channel_t {
	d_channel
} CHANNEL;


/************************************************************************************/

/**@name Generated Access Functions
 *
 * Access functions for @c CHANNEL that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a
 * _structure argument and either write a new value (@a _value argument) or
 * return the current content of a particular field. For access to vectors an
 * index must be also passed using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype,
 *  set_function_prototype, increment_function_prototype
 * @{ */

declare_access_functions(CHANNEL, channel, bus_id);
declare_access_functions(CHANNEL, channel, active);
/*@}*/

/************************************************************************************/

/** Set the initial power-on values of @c CHANNEL - called after creation.
 * 
 * @param channel Pointer to a @c CHANNEL instance or a derived structure.
 * @return Return the pointer passed in @a channel. */
CHANNEL *channel_init(CHANNEL *channel);

/** Registers the caller to receive the next sent frame.
 * 
 * The caller specifies two arguments with this funtion, that identify him:
 * @a ctrl and \a callback. These parameters are stored and when a frame is received,
 * the folowing function call is made:\code
 * callback(ctrl, channel->id, transmission_state, channel->frame);\endcode
 * If the receiver is already registered, the function returns with error code.
 * A correct registration is valid until a frame is completely transmitted on the channel.
 * 
 * @param channel Pointer to a channel that receives the registration.
 * @param arg User defined argument that is stored and later passed to the @a callback 
 * 	function.
 * @param callback Pointer to a function that should be called after a frame is received.
 * @return Returns 0 on success, nonzero otherwise. 
 *
 * @see channel_cancel_receive */
int channel_receive(CHANNEL *channel, void *arg, CHANNEL_ON_RCV_FN callback);

/** Unregisters a controller so it no more listens on the channel.
 *
 * @param channel Pointer to a channel that should cancel the registration.
 * @param ctrl This argument is used when searching for the registration to cancel.
 *   All registrations with this user-argument are canceled.
 *
 * @see channel_receive */
void channel_cancel_receive(CHANNEL *channel, void *arg);

/** Initializes the sending of a frame - register the sender for notification.
 *
 * This function assumes that the sender is either @c CONTROLLER or a
 * derived structure, because the frame is copied directly from controller's
 * local data structures.  The sender (@a controller argument) is stored and so
 * is the callback function the is provided - after the send operation
 * finished, these stored values are used to nitify the sender. The sending
 * controller is excluded from receiving the frame by explicitly removing him
 * from registered receivers.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param controller Pointer to a @c CONTROLLER or derived structure. This
 *  value is passed to the callback function later.
 * @param callback Function that is used to notify the controller.
 * @return Returns 0 if no error occurs, nonzero code otherwise.
 * 
 * @see channel_send_complete, CHANNEL_SEND_FRAME_FN */
int channel_send_init(CHANNEL *channel, void *controller, PTR_CALLBACK callback);

/** Finishes the sending of a frame - notifies all receivers and cleans up the channel. 
 *
 * Sets Frame State to COMPLETE if there was no noise while transmiting and
 * sends notifications to all receivers and also to the sender. At last clears
 * all transmission fields. 
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 *
 * @see channel_send_init, CHANNEL_SEND_FRAME_FN */
void channel_send_complete(CHANNEL *channel);

/** Registers the caller to receive notifications about transmitted frame - it
 * is possible to modify such frames before sending. 
 * 
 * This function enables the caller to monitor all traffic on the channel and
 * even to modify the frames before transmission. So far it is not easily
 * possible to selectively modify frames for different receivers.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param callback Function that is used to notify the registered controller.
 * @param arg User defined argument that is stored and later passed to the @a
 *  callback function. */
void channel_set_transmission_callback(CHANNEL *channel, CHANNEL_TRANSMISSION_FN callback, void *arg);

/** Registers an error handling function for BG reported failures. 
 * 
 * This function gives an oportunity to catch BG errors immediately as they
 * occur even at higher levels. So far there is no direct reporting from BG to
 * controller, so this is necessary.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param callback The callback function that will be called when any error
 *  from the BG is detected. The @a channel argument will be passed to the
 *  function. */
void channel_set_bg_error_callback(CHANNEL *channel, PTR_CALLBACK callback);

/** Check BG: if it is open return 0, raise an error otherwise and return nonzero code.
 * 
 * Because the BG is currently integrated into the controller, this function
 * has to look into the @a controller structure to find the appropiate BG
 * instance.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param controller Pointer to a @c CONTROLLER or derived structure.
 * @return Returns 0 if BG is open, nonzero code otherwise. */
int channel_test_bg(CHANNEL *channel, void *controller);


/*************************************************************************************/

#endif /* #ifndef _FP1CHAN_H */

/*@}*/
/*@}*/

