/**@addtogroup cni_fun
 * @{*/

/**@addtogroup cni_fun_ext  Interface
 * @{*/

/**@file
 * @brief CNI data structure and timeless functionality (interface).
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#ifndef _FP1CNI_H
#define _FP1CNI_H

#include "ttpc/fl1func.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1cstat.h"
#include "ttpc/fp1lname.h"
#include "ttpc/fp1frame.h"

/************************************************************************************/


/** Maximum reasonable size of the CNI message area. */
#define CNI_MSG_AREA_SIZE (sizeof(CNI_MAE) * 2 * 512)


/** This value (when written into the @c co field in @c CNI structure) turns on the 
 * controller - any other value causes the controller to shut down. */
#define CONTROLLER_ON		0xFFFF


/** This value (when written into the @c ca field in @c CNI structure) initiates the 
 * await state - any other value is ignored.*/
#define CONTROLLER_AWAIT	0xFFFF


/** Predefined constants for Assertion Code generated by BIST-AV error.
 *
 * This value is stored in the @c assertion_code field in the #CNI when 
 * an assertion violation error is raised. */
typedef enum {
	AVC_INVALID_CCF_VALUE = 0,      /**< An invalid value of CCF was found.*/
	AVC_TP_SYNC_ERROR = 1           /**< Clock disagreement with program flow.*/
} AV_CODE;


/** Possible values for Error Indication Field (EIF).*/
typedef enum {
	EIF_NULL_FRAME      = 0x00,  /**< No traffic detected.*/
	EIF_INVALID_FRAME   = 0x02,  /**< No syntactically valid frame received.*/
	EIF_CRC_ERROR       = 0x04,  /**< Frame did not pass the CRC check.*/
	EIF_CORRECT_FRAME   = 0x08,  /**< Frame is coorrect.*/
	EIF_NOISE           = 0x01   /**< There was noise on channel during the transmission.*/
} CNI_EIF_VALUES;


/************************************************************************************/

/** Protocol Status structure (a part of the #CNI structure).*/
typedef struct {
	unsigned failed_counter;  /**< Number of incorrect frames (bad CRC).*/
	unsigned accept_counter;  /**< Number of correctly received frames.*/
	unsigned controller_ready;/**< Contains the #CONTROLLER_READY constant when 
				       the controller is ready to communicate with host.*/
	unsigned invalid_counter; /**< Number of syntacticaly invalid frames.*/
	unsigned i_frame_switch;  /**< 1-bit selector that determines the next channel used for cold start.*/
	unsigned mfc;		  /**< Succesive membership failure count.*/
	unsigned c_state_valid;   /**< Contains the #C_STATE_VALID constant when the 
				       controller is synchronized and has a valid C-State.*/
	unsigned csct;		  /**< Clock State Correction Term - current value of 
				       meassured clock difference.*/

	/**@name Error Flags
	 * @{*/
	unsigned ne;		/**< NBW protocol error */
	unsigned mv;		/**< Mode violation error */
	unsigned so;		/**< Slot occupied error */
	unsigned dc;		/**< Download comleted */
	unsigned cb;		/**< Comunication system blackout */
	unsigned mc;		/**< MEDL CRC error */
	unsigned me;		/**< Membership error */
	unsigned se;		/**< Synchronization error */
	unsigned ae;		/**< Acknowledgement error */
	unsigned be;		/**< BG error */
	unsigned bl;		/**< BG lifesign.*/
	/*@}*/
	unsigned protocol_state;  /**< Current state of the internal state-machine.*/
} CNI_PS;

/**@name Generated Access Functions for CNI_PS
 * Access functions for @c CNI_PS that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_PS, cni_ps, accept_counter);
declare_access_functions(CNI_PS, cni_ps, failed_counter);
declare_access_functions(CNI_PS, cni_ps, invalid_counter);
declare_access_functions(CNI_PS, cni_ps, controller_ready);
declare_access_functions(CNI_PS, cni_ps, mfc);
declare_access_functions(CNI_PS, cni_ps, i_frame_switch);
declare_sig_access_functions(CNI_PS, cni_ps, csct);
declare_access_functions(CNI_PS, cni_ps, c_state_valid);
declare_access_functions(CNI_PS, cni_ps, protocol_state);
declare_access_functions(CNI_PS, cni_ps, bl);
declare_access_functions(CNI_PS, cni_ps, be);
declare_access_functions(CNI_PS, cni_ps, ae);
declare_access_functions(CNI_PS, cni_ps, se);
declare_access_functions(CNI_PS, cni_ps, me);
declare_access_functions(CNI_PS, cni_ps, mc);
declare_access_functions(CNI_PS, cni_ps, cb);
declare_access_functions(CNI_PS, cni_ps, dc);
declare_access_functions(CNI_PS, cni_ps, so);
declare_access_functions(CNI_PS, cni_ps, mv);
declare_access_functions(CNI_PS, cni_ps, ne);
/*@}*/

/**********************************************************************************/

/** Frame Diagnosis field structure (a part of #CNI structure).*/
typedef struct {
	unsigned eif_chb;          /**< EIF (error flags) for transmission on channel B.*/
	FRAME_HEADER header_chb;   /**< Header of frame received on channel B.*/
	unsigned eif_cha;          /**< EIF (error flags) for transmission on channel A.*/
	FRAME_HEADER header_cha;   /**< Header of frame received on channel A.*/
	unsigned mtd_chb;          /**< Measured time difference on channel B in 
				        microticks (used for clock synchronization).*/
	unsigned mtd_cha;          /**< Measured time difference on channel A in 
				        microticks (used for clock synchronization).*/
} CNI_FD;

/**@name Generated Access Functions for CNI_FD
 * Access functions for @c CNI_FD that were generated by preprocessor macros.
 * 
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype 
 * @{*/
declare_access_functions(CNI_FD, cni_fd, eif_cha);
declare_access_functions(CNI_FD, cni_fd, eif_chb);
declare_sig_access_functions(CNI_FD, cni_fd, mtd_cha);
declare_sig_access_functions(CNI_FD, cni_fd, mtd_chb);
/*@}*/

/**********************************************************************************/

/** Personality field structure (a part of #CNI structure).*/
typedef struct {
	unsigned controller_version;  /**< TTP/C Controller version.*/
	unsigned physical_sru_name;   /**< Static name of the SRU (role independent).*/
	unsigned app_version;         /**< Current application (MEDL) version.*/
	unsigned app_id;              /**< Current application (MEDL) name.*/
	LOGICAL_NAME logical_name;    /**< Current role (dynamic name) of this SRU.*/
} CNI_PERS;

/**@name Generated Access Functions for CNI_PERS
 * Access functions for @c CNI_PERS that were generated by preprocessor macros.
 * 
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_PERS, cni_pers, physical_sru_name);
declare_access_functions(CNI_PERS, cni_pers, controller_version);
declare_access_functions(CNI_PERS, cni_pers, app_id);
declare_access_functions(CNI_PERS, cni_pers, app_version);
/*@}*/

/**********************************************************************************/

/** Interrupt Status field and Interrupt Enable field structure (a part of #CNI 
 * structure).
 *
 * The Interrupt Status (@c CNI_IS structure) and Interrupt Enable (#CNI_IE structure)
 * sections of CNI are the same so this structure is used for both of them. 
 * #CNI_IE is a @c typedef of # CNI_IS. */
typedef struct {
	unsigned ti2;		/**< time interrupt \b 2 (programmed with @c timer2 field).*/
	unsigned ti1;		/**< Time interrupt \b 1 (programmed with @c timer1 field).*/
	unsigned ui2;		/**< User defined interrupt \b 2 (defined within @c MEDL).*/
	unsigned ui1;		/**< User defined interrupt \b 1 (defined within @c MEDL).*/
	unsigned mi;		/**< Mode changed.*/
	unsigned mc;		/**< Membership changed (any membership).*/
	unsigned ml;		/**< Membership loss (we talk about our own membership).*/
	unsigned he;		/**< Host error, see @c CNI_PS for more detailed error codes.*/
	unsigned pe;		/**< Protocol error, see @c CNI_PS for more detailed error codes.*/
	unsigned be;		/**< BIST (Build-in Self-test) error, see @c CNI_PS for more detailed error codes.*/
} CNI_IS;

/**@name Generated Access Functions for CNI_IS
 * Access functions for @c CNI_IS that were generated by preprocessor macros.
 *
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_IS, cni_is, be);
declare_access_functions(CNI_IS, cni_is, pe);
declare_access_functions(CNI_IS, cni_is, he);
declare_access_functions(CNI_IS, cni_is, ml);
declare_access_functions(CNI_IS, cni_is, mc);
declare_access_functions(CNI_IS, cni_is, mi);
declare_access_functions(CNI_IS, cni_is, ui1);
declare_access_functions(CNI_IS, cni_is, ui2);
declare_access_functions(CNI_IS, cni_is, ti1);
declare_access_functions(CNI_IS, cni_is, ti2);
/*@}*/


/**********************************************************************************/

/** Interrupt Status field and Interrupt Enable field structure (a part of #CNI 
 * structure).
 * 
 * The Interrupt Status (#CNI_IS structure) and Interrupt Enable (#CNI_IE structure)
 * sections of CNI are the same so this structure is used for both of them. */
typedef CNI_IS CNI_IE;

/**@name Generated Access Functions for CNI_IE
 * Access functions for @c CNI_IE that were generated by preprocessor macros.
 *
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_IE, cni_ie, be);
declare_access_functions(CNI_IE, cni_ie, pe);
declare_access_functions(CNI_IE, cni_ie, he);
declare_access_functions(CNI_IE, cni_ie, ml);
declare_access_functions(CNI_IE, cni_ie, mc);
declare_access_functions(CNI_IE, cni_ie, mi);
declare_access_functions(CNI_IE, cni_ie, ui1);
declare_access_functions(CNI_IE, cni_ie, ui2);
extern get_function_prototype(CNI_IE, cni_ie, ti1);
extern get_function_prototype(CNI_IE, cni_ie, ti2);
/*@}*/

/**********************************************************************************/

/** Build-in self-test (BIST) status field structure (a part of #CNI structure).
 * 
 * The individual fields are set accordingly after the internal self-test is 
 * performed.*/
typedef struct {
	unsigned wd;		/**< Watchdog error.*/
	unsigned rf;		/**< Register file error.*/
	unsigned rm;		/**< RAM error.*/
	unsigned br;		/**< Invalid BIST request.*/
	unsigned av;		/**< Assertion violation.*/
	unsigned sc;		/**< Selftest complete.*/
}  CNI_BS;

/**@name Generated Access Functions for CNI_BS
 * Access functions for #CNI_BS that were generated by preprocessor macros.
 * 
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_BS, cni_bs, sc);
declare_access_functions(CNI_BS, cni_bs, av);
declare_access_functions(CNI_BS, cni_bs, br);
declare_access_functions(CNI_BS, cni_bs, rm);
declare_access_functions(CNI_BS, cni_bs, rf);
declare_access_functions(CNI_BS, cni_bs, wd);
/*@}*/

/**********************************************************************************/

/** Build-In Self-Test (BIST) enable field structure (a part of #CNI structure).
 *
 * This field serves to selectively enable/disable tests performed by the 
 * internal self-test mechanism. Value of zero means, that the corresponding test
 * will not be performed.
 *
 * @see CNI_BS*/
typedef struct {
	unsigned rf;		/**< Register file checking.*/
	unsigned rm;		/**< RAM checking.*/
} CNI_BE;

/**@name Generated Access Functions for CNI_BE
 * Access functions for @c CNI_BE that were generated by preprocessor macros.
 *
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_BE, cni_be, rm);
declare_access_functions(CNI_BE, cni_be, rf);
/*@}*/


/**********************************************************************************/

/** Message Area Entry - the basic element of of #CNI message area.
 *
 * This structure is the only piece of code within the model that uses bit-sized 
 * items. It was necessary to have compatible address space within the message are 
 * with the HW implementation of TTP/C. This way we are able to use MEDL files 
 * generated using TTTech tools.
 *
 * @todo Find a way to remove the bit-items or rewrite everything else to be a 1:1 
 *  memory copy.*/
typedef struct {
	/**@name Message status field
	 * @{*/
	char not_used;     /**< Reserved field.*/
	char ccf:4;        /**< Concurrency Control Field (used by NBW protocol).*/
	char eif:4;        /**< Error Indication Field (see @c CNI_EIF_VALUES definition).*/
	/*@}*/

	char data[16];     /**< Message data field - actual size of this field depends on the message size (this is maximum).*/
} CNI_MAE;


/**@name Generated Access Functions for CNI_MAE
 * Access functions for @c CNI_MAE that were generated by preprocessor macros.
 *
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI_MAE, cni_mae, ccf);
declare_access_functions(CNI_MAE, cni_mae, eif);
declare_index_access_functions(CNI_MAE, cni_mae, data);
/*@}*/

/***********************************************************************************/

/** Comunication Network Interface (CNI) structure.
 *
 * This structure is exactly defined within the written specification of TTP/C. This
 * implementation does not keep 1:1 address mapping - it implements all defined
 * fields as 32-bit integers to speed up access. There is a module that can map
 * the orginal address space to the model space and vice versa. */
typedef struct {
	void *controller;         /**< Pointer to an instance of #CONTROLLER or derived structure that owns this CNI.*/

	/**@name CNI Status Fields
	 * @{*/
	unsigned ccf;             /**< Concurrency Control Field - used by NBW protocol.*/
	C_STATE c_state;          /**< Current Cluster State - synchronized between all nodes.*/
	unsigned null_frame[MAX_SRU_COUNT];   /**< Vector that contains a @c 1 for 
						every SRU that did not send a frame in the last TDMA round.*/
	CNI_PS protocol_status;   /**< Current protocol status.*/
	CNI_FD frame_diagnosis;   /**< Diagnostic information for last received frame.*/
	CNI_PERS personality;     /**< Personalized information for this SRU - name, slot position, etc.*/
	unsigned controller_lifesign;  /**< A field periodically updated by the controller.*/
	unsigned reserved1[8];    /**< Unused reserved field.*/
	unsigned assertion_code;  /**< Numerical identification of assertion 
				       violation errors - see #AV_CODE constants.*/
	/*@}*/
	
	/**@name CNI Control Fields
	 * @{*/
	unsigned host_lifesign;    /**< This field has to periodically updated by the host.*/
	unsigned external_rate_correction;  /**< Any value written here is applied to 
					         the clock correction mechanism.*/
	unsigned mode_change_request[3];    /**< Input field for requesting or canceling a mode change.*/
	unsigned ca;               /**< \e Await state control field (see @c CONTROLLER_AWAIT constant).*/
	CNI_BE bist_enable;        /**< Control flags for the interna self-test mechanism (BIST).*/
	/*@}*/
	

	/**@name CNI Status Fields
	 * @{*/
	unsigned cluster_time;    /**< Current time that is globally synchronized (in macroticks).*/
	CNI_IS interrupt_status;   /**< Status flags that are set when any interrupt 
				        occurs to identify the type of that interrupt.*/
	CNI_BS bist_status;       /**< Results of BIST are written here.*/
	/*@}*/

	
	/**@name CNI Control Fields
	 * @{*/
	unsigned timer1;           /**< Programmable timer 1 - The \e TI1 interrupt 
				        will be raised when time reaches a value stored here.*/
	unsigned timer2;           /**< Programmable timer 2 - The \e TI2 interrupt will 
				        be raised when time reaches a value stored here.*/
	CNI_IE interrupt_enable;   /**< Flags that enable/disable individual TTP/C interrupts.*/
	unsigned time_overflow_counter;  /**< This counter is incremented everytime 
					      the @c cluster_time coutner overflows.*/
	unsigned co;               /**< Power On/Off "Button" of the controller - write 
				        the @c CONTROLLER_ON constant to turn it on.*/
	/*@}*/

	/**@name CNI Message Area
	 * @{*/
	char cni_msg_base[CNI_MSG_AREA_SIZE];   /**< This memory block maps the message area described by 
						    specification in 1:1 address relation.*/
	/*@}*/
} CNI;

/**@name Generated Access Functions for CNI
 * Access functions for @c CNI that were generated by preprocessor macros.
 *
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 *  increment_function_prototype
 * @{*/
declare_access_functions(CNI, cni, ccf);
declare_index_access_functions(CNI, cni, null_frame);
declare_access_functions(CNI, cni, controller_lifesign);
declare_access_functions(CNI, cni, assertion_code);
declare_access_functions(CNI, cni, host_lifesign);
declare_index_access_functions(CNI, cni, mode_change_request);
declare_access_functions(CNI, cni, time_overflow_counter);
extern increment_sig_function_prototype(CNI, cni, external_rate_correction);
extern get_function_prototype(CNI, cni, timer1);
extern increment_function_prototype(CNI, cni, timer1);
extern get_function_prototype(CNI, cni, timer2);
extern increment_function_prototype(CNI, cni, timer2);
extern set_function_prototype(CNI, cni, cluster_time);
extern increment_function_prototype(CNI, cni, cluster_time);
extern get_function_prototype(CNI, cni, co);
extern increment_function_prototype(CNI, cni, co);
extern get_function_prototype(CNI, cni, ca);
extern increment_function_prototype(CNI, cni, ca);
/*@}*/


/***********************************************************************************/
/**@name Mode Change Request Handling
 * @{*/

/** Return the mode number from Mode Change Request (MCR) fields.*/
unsigned cni_mcr_get_mode(CNI *cni);

/** Set the mode number in Mode Change Request (MCR) fields.*/
void cni_mcr_set_mode(CNI *cni, unsigned mode);

/** Check whether the Mode Change Request (MCR) fields contain an immediate mode change.*/
unsigned cni_mcr_get_immediate(CNI *cni);

/** Specifie whether the Mode Change Request (MCR) should be immediate (or deffered).*/
void cni_mcr_set_immediate(CNI *cni, unsigned imm);

/** Clear any pending mode change request.*/
void cni_mcr_clear(CNI *cni);

/*@}*/
/************************************************************************************/

/** Pointer to a function that can set a single unsigned integer value within CNI.
 *
 * This CNI module provides global hooks that can be used to implement something like 
 * virtual methods. Functions that have to be implemented by a module that encapsulates
 * the basic CNI definition are implemented as pointers.
 *
 * @param cni Pointer to an instance of CNI or a derived structure.
 * @param value New content of the modified field.
 * @return Returns the written value.*/
typedef unsigned (*CNI_SET_FN)(CNI *cni, unsigned value);


/** Pointer to a function that can read a single unsigned integer value within CNI.
 *
 * This CNI module provides global hooks that can be used to implement something like 
 * virtual methods. Functions that have to be implemented by a module that encapsulates
 * the basic CNI definition are implemented as pointers.
 *
 * @param cni Pointer to an instance of CNI or a derived structure.
 * @return Returns one unsigned value from CNI.*/
typedef unsigned (*CNI_GET_FN)(CNI *cni);


/************************************************************************************/
/**@name Virtual Function Pointers
 * These variables contain pointers to functions that provide extended access to some
 * fields of the @c CNI structure.
 *
 * Some of them are defined implicitly, but the implicit definition may not be 
 * sufficient for you. In such case you may write a new version and assign these
 * variables to point to your version. It is possible to still use the original
 * access functions - if you store the old value.
 * @{*/

extern CNI_GET_FN cni_get_cluster_time;
extern CNI_SET_FN cni_ie_set_ti1;
extern CNI_SET_FN cni_ie_set_ti2;
extern CNI_SET_FN cni_set_timer1;
extern CNI_SET_FN cni_set_timer2;
extern CNI_SET_FN cni_set_co;
extern CNI_SET_FN cni_set_ca;

/*@}*/
/************************************************************************************/


/* Return pointer to default CNI fault mask (static allocation).
 * Individual CNI areas can be disabled/enabled using arguments. */
unsigned* cni_fault_mask(int status, int control, int message);
    
/** Convert an offset relative to CNI message area to a pointer to @c CNI_MAE structure.*/
CNI_MAE *cni_get_message_entry(CNI *cni, unsigned offset);

/** Clear all fields in @c CNI except the @c co field.*/
void cni_clear_mem(CNI *cni);

/** Return a string containing the name of given state.*/
char *cni_get_protocol_state_name(unsigned state_number);

/** Copy message data from a @c FRAME structure to the @c CNI_MAE entry.*/
void cni_mae_extract_frame_data(CNI_MAE* mae, FRAME *frame, unsigned data_length);


/************************************************************************************/
/* Special Field Access Functions */

/** Return the value of External Rate Correction field from @c CNI structure.*/
int cni_get_external_rate_correction(CNI *cni);

/** Change the value of External Rate Correction field in @c CNI structure.*/
int cni_set_external_rate_correction(CNI *cni, int val);


/************************************************************************************/

#endif /* #ifndef _FP1CNI_H */

/*@}*/
/*@}*/

