/**@addtogroup misc
 * @{*/

/**@addtogroup misc_const  Project Configuration
 * 
 * This module is included from every other file in the TTP/C model. It serves
 * as a central source for parametrization of the model.
 *
 * @{ */


/**@file
 * @brief TTP/C protocol specific global constants (interface).
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */


#ifndef _FP1CONST_H
#define _FP1CONST_H

/*****************************************************************************/

#ifndef OK
/** Symbolic representation of a correct return value.*/
#define OK      0
#endif

#ifndef ERROR
/** Symbolic representation of an erroneous return value.*/
#define ERROR  -1
#endif

#ifndef USER_ERROR
/** Users of the model can safely use error numbers above this value for their
 * own purpose. */
#define USER_ERROR 1000
#endif

#ifndef TRUE
/** Symbolic name for Boolean 'true' value.*/
#define TRUE 1
#endif

#ifndef FALSE
/** Symbolic name for Boolean 'false' value. */
#define FALSE 0
#endif

/** Time scale used in all time functions within the simulation.
 * The meaning of the value is the number of simulation time units in 1 second.
 * All time units within the model use this time scale - unless stated explicitly 
 * otherwise. 
 *
 * @see csim_time, CSIM_TIME in the C-Sim library manual*/
#define TIME_SCALE                      (1.0e+6)

/** Duration of 1 macrotick in chosen \c TIME_SCALE (not in real time).
 * The current version of model uses one \em macrotick as the basic time unit. 
 * The duration of smaller units (e.g. \em microtick) is computed relatively to
 * this value.
 *
 * @note The actual duration of a particular macrotick may is different to simulate
 * the effect of clock drift or random flcutuations.*/
#define MACROTICK_DURATION              1.0

/** Number of microticks in one macrotick (floating point value).*/
#define MICRO_PER_MACRO                 20.0

/** Maximum time difference of any two synchronized clocks.*/
#define CLOCK_PRECISION                 (0.9 * MACROTICK_DURATION)

/** Channel transmission speed in bits per \em time \em unit.*/
#define CHANNEL_SPEED                   (2.0 * 1024 * 1024 / TIME_SCALE)

/** The duration of IFG phase in macroticks. */
#define IFG_DURATION                    101

/** Source MEDL file name definition, %d will be substituted by an index.  The
 * model contains a tool for loading the contents of text-based MEDL files into
 * memory. This tool expects to find the MEDLs in files, whose names are
 * generated using this pattern.
 *
 * @see medl_read_file*/
#define MEDL_READ_FILE_NAME             "medl%02d.txt"

/** CRC generator polynomial 1 - used for short messages.  This polynomial is
 * used to compute the CRC code for messages with length up to \c
 * MAX_LENGTH_FOR_CRC_POLY_1 bytes.
 * 
 * @see MAX_LENGTH_FOR_CRC_POLY_1*/
#define CRC_POLYNOMIAL_1                0233023

/** CRC generator polynomial 2 - used for long messages.  This polynomial is
 * used to compute the CRC code for messages with length greater than \c
 * MAX_LENGTH_FOR_CRC_POLY_1 bytes.  @see MAX_LENGTH_FOR_CRC_POLY_1*/
#define CRC_POLYNOMIAL_2                0214461

/** Maximum frame length to use CRC polynomial 1.
 *
 * @see CRC_POLYNOMIAL_1, CRC_POLYNOMIAL_2*/
#define MAX_LENGTH_FOR_CRC_POLY_1       151

/** CRC calculator initialization value.  This value is stored into the CRC
 * generator prior to sending the actual message.*/
#define CRC_STARTING_VALUE              0xFFFF

/** The maximum number of bytes sent in N-Frames.*/
#define MAX_APPLICATION_DATA_SIZE       16

/** Size of the stack used by the clock synchronization algorithm.*/
#define SYNC_STACK_SIZE                 4

/** TTP/C protocol states.  This enum is used to control the internal state
 * machine of the protocol.  The states copy exactly the definition from
 * written TTP/C specification.*/
typedef enum {
	PS_INIT        = 1,  /**< Init.*/
	PS_LISTEN      = 2,  /**< Listen.*/
	PS_COLD_START  = 3,  /**< Cold Start.*/
	PS_STARTUP     = 4,  /**< Startup.*/
	PS_APPLICATION = 5,  /**< Application.*/
	PS_READY       = 6,  /**< Ready.*/
	PS_PASSIVE     = 7,  /**< Passive.*/
	PS_AWAIT       = 8,  /**< Await - Not implemented*/
	PS_SELFTEST    = 9,  /**< Self-test - Not implemented*/
	PS_FREEZE      = 10, /**< Freeze.*/
	PS_DOWNLOAD    = 12  /**< Download - Not implemented*/
} PROTOCOL_STATE;

/** Denotes that a message should not be stored in the CNI.*/
#define CNI_MAE_NO_ADDRESS              0x3FFF

/** This value, when written to \c c_state_valid field, denotes that current
 * C-State is valid.*/
#define C_STATE_VALID                   0x0A

/** This value, when written to \c controller_ready field, denotes that current
 * controller is ready.*/
#define CONTROLLER_READY                0x0A

/** Denotes an invalid (unused) entry in mode control entry.*/
#define INVALID_DEFERRED_MODE			0x1F

/** Maximum mode count in MEDL.*/
#define MAX_MODE_COUNT                  30

/** First application mode number.
 *
 * All modes from this number up to \c LAST_APP_MODE_NUMBER are considered to
 * be application modes.
 *
 * @see LAST_APP_MODE_NUMBER*/
#define FIRST_APP_MODE_NUMBER           2

/** Last application mode number.
 *
 * All modes from \c FIRST_APP_MODE_NUMBER up to this number are considered to
 * be application modes.
 *
 * @see FIRST_APP_MODE_NUMBER*/
#define LAST_APP_MODE_NUMBER            30

/** Cold start mode number.
 *
 * This mode does not need to have a valid mode entry in MEDL.*/
#define COLD_START_MODE_NUMBER          0

/** Startup mode number.
 *
 * The first mode entered after startup. Only I-frames are sent.*/
#define STARTUP_MODE_NUMBER             1

/** Download mode number - no real MEDL mode entry.
 *
 * This mode does not need to have a valid mode entry in MEDL.*/
#define DOWNLOAD_MODE_NUMBER            31

/** Maximum slot count in one transmission cycle.*/
#define MAX_TRANSMISSION_CYCLE_LENGTH   512

/** Maximum number of SRUs in cluster.*/
#define MAX_SRU_COUNT                   64

/** Bus guardian clock speed in MHz.*/
#define BUS_GUARDIAN_FREQUENCY          16

/** Bus guardian frquency divisor for values in MEDL.
 *
 * MEDL parameters for BG are coded using multiples of this value.*/
#define BUS_GUARDIAN_DIVISOR            16

/** The delay between physicaly receiving a frame and recognition of the event.
 *
 * In the model, this value may be zero and in fact, no other value has been
 * tested so far. HW implementations have a small latency.*/
#define RECEIVE_CALLBACK_DELAY			0.0

/** The duration of receive window is used to determine if a frame was received
 * completely.
 *
 * The window is centered on action time for the given slot and spans the given
 * number of macroticks to both sides.*/
#define RECEIVE_WINDOW_LEN              1.0


#ifdef DOXYGEN  /* Documentaton generator */

/** This constant disables the integrated (local) BG unit of TTP/C.
 *
 * When \c DISABLE_BG is defined, the local BG is disabled and the controller
 * has unlimited access to the transmission channels.*/
#define DISABLE_BG

/** This constant enables the proposed modification of internal BG.
 *
 * When \c ENABLE_BG_MOD is defined, the arm phase of BG is restricted which
 * makes the operation safer vs multiple faults. The modification requires
 * exteded MEDL files (ARM phase duration for all BG modes).*/
#define ENABLE_BG_MOD

/** This constant enables dynamic BG parameter computation during the
 * synchronization phase of the controller. */
#define DYN_BG

#endif /* DOXYGEN */


#endif /* #ifdef _FP1CONST_H */

/*@}*/
/*@}*/

