/**@addtogroup ctrl_fun
 * @{*/

/**@addtogroup ctrl_fun_ext  Interface
 * @{*/

/**@file
 * @brief TTP/C Communication controller functionality (interface).
 * @version 1.6
 * @date Last modified 2003-02-15 
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */

#ifndef _FP1CTRL_H
#define _FP1CTRL_H

#include "ttpc/fp1cni.h"
#include "ttpc/fp1medl.h"
#include "ttpc/fp1ldata.h"
#include "ttpc/fp1chan.h"
#include "ttpc/fp1busg.h"

/************************************************************************************/

struct controller_t;

/** Pointer to a function that operates upon a controller instance with no additional 
 * arguments.
 * 
 * The #CONTROLLER structure provides function hooks that can be used to implement 
 * something like virtual methods. Functions that have to be implemented by a module 
 * that encapsulates the basic #CONTROLLER definition are implemented as pointers.
 * 
 * @param ctrl Pointer to an instance of #CONTROLLER or a derived structure. */
typedef void (*CONTROLLER_FN)(struct controller_t *ctrl);


/** Pointer to a function that operates upon a controller instance with one @c int 
 * argument.
 *
 * The #CONTROLLER structure provides function hooks that can be used to implement 
 * something like virtual methods. Functions that have to be implemented by a module that 
 * encapsulates the basic #CONTROLLER definition are implemented as pointers.
 *
 * @param ctrl Pointer to an instance of #CONTROLLER or a derived structure.
 * @param i Any integer argument. */
typedef void (*CONTROLLER_INT_FN)(struct controller_t *ctrl, int i);


/** Pointer to a function that operates upon a controller instance with one additional 
 * <code>void *</code> argument.
 *
 * The #CONTROLLER structure provides function hooks that can be used to implement 
 * something like virtual methods. Functions that have to be implemented by a module that 
 * encapsulates the basic #CONTROLLER definition are implemented as pointers.
 * 
 * @param ctrl Pointer to an instance of #CONTROLLER or a derived structure.
 * @param arg Untyped pointer to anything (<code>void *</code>). */
typedef void (*CONTROLLER_PTR_FN)(struct controller_t *ctrl, void *arg);


/**A list of all protocol-defined data fields for a communication controller.
 *
 * This macro is used to simplify the type derivation used within the model. It lists
 * all items that define the state of a controller unit in a format that can be used to 
 * define the content of a @c struct type. All @c struct types that include this macro at 
 * the beginning of their definition are then backwards compatible (we call them derived
 * structures. */
#define d_controller \
	CNI cni;                        /**< CNI image.*/ \
	MEDL medl;                      /**< MEDL image.*/ \
	PLOC_DATA ploc_data;            /**< Protocol specific internal data.*/ \
	\
	BUS_GUARDIAN *bus_guardian;     /**< Pointer to controller's bus guardian.*/ \
	CHANNEL *channel[2];            /**< Pointer to the bus (2 replicas).*/ \
	\
	CONTROLLER_INT_FN synchronize_time; /**< VF: synchronize time on a frame.*/ \
	CONTROLLER_FN reset_clk_correction; /**< VF: start new clock synchronization interval.*/ \
	CONTROLLER_FN update_cluster_time;  /**< VF: update the cluster time field.*/ \
	CHANNEL_ON_RCV_FN receive_callback; /**< VF: frame received.*/ \
	PTR_CALLBACK send_callback;         /**< VF: frame sent.*/ \
	\
	PTR_CALLBACK host_interrupt;        /**< Host interrupt function address.*/ \
	void *host_arg;                     /**< Argument passed to the host interrupt function.*/;


/** Communication Controller data structure - protocol level fields.
 * This structure the the base for all other #controller derived structures. To derive 
 * a new structure from #controller it is best to use the macro @c d_controller, that
 * is in fact a list of all necessary fields. */
typedef struct controller_t {
	d_controller
} CONTROLLER;


/** Change the current state of the controller.
 *
 * @param ctrl  Pointer to an instance of #CONTROLLER or derived structure.
 * @param state  Desired state of the controller.*/
void controller_set_state(CONTROLLER *ctrl, unsigned state);

/** Copy the local CNI status area into the public CNI, update cluster time.*/
void controller_update_cni(CONTROLLER *ctrl);

/** Return pointer to the controller's current MEDL Slot Control Entry (SCE).*/
MEDL_SCE *controller_get_current_sce(CONTROLLER *ctrl);

/** Check whether the controller is sender in the current slot.*/
unsigned controller_is_current_sender(CONTROLLER *ctrl);

/** Check whether cold-start is allowed for the given controller.*/
unsigned controller_is_cold_start_allowed(CONTROLLER *ctrl);

/** Raise TP synchronous interrupt if there is any pending.*/
void controller_raise_tp_sync_interrupt(CONTROLLER *ctrl);

/** Check if timer(x) interrupt (TIx) is enabled, if so then raise it.*/
void controller_raise_timer_interrupt(CONTROLLER *ctrl, unsigned x);

/** Send a prepared frame using the specified channel.*/
int controller_send_frame(CONTROLLER *ctrl, unsigned channel);

/** Start to listen for incoming frames on both channels.*/
int controller_start_receive(CONTROLLER *ctrl);

/** Return the listen timeout length (see specification) in macroticks.*/
unsigned controller_get_listen_timeout(CONTROLLER *ctrl);

/** Return the cold start timeout length (see specification) in macroticks.*/
unsigned controller_get_cold_start_timeout(CONTROLLER *ctrl);

/** Set new BG params if there was no bus activity.*/
int controller_cold_start_bg_arm(CONTROLLER *ctrl);

/** Trigger BG Arm signal if current slot has the BA flag enabled in MEDL.*/
int controller_bg_arm(CONTROLLER *ctrl);

/** Perform all actions for controller's init state.*/
void controller_init_state(CONTROLLER *ctrl);

/** Prepare the controller's listen state (phase 1 of 2).*/
int controller_prepare_listen_state(CONTROLLER *ctrl);

/** Finish the controller's listen state (phase 2 of 2).*/
void controller_finish_listen_state(CONTROLLER *ctrl);

/** Prepare the controller's cold start state (phase 1 of 3).*/
int controller_prepare_cold_start_state(CONTROLLER *ctrl);

/** Perform all actions for a specified duration wihin the cold start state 
	(phase 2 of 3).*/
void controller_run_cold_start_state(CONTROLLER *ctrl, int macroticks);

/** Finish the controller's cold start state (phase 3 of 3).*/
int controller_finish_cold_start_state(CONTROLLER *ctrl);

/** Perform all actions for controller's freeze state.*/
unsigned controller_freeze_state(CONTROLLER *ctrl);

/** Perform all actions for controller's download state.*/
void controller_download_state(CONTROLLER *ctrl);

/** Controller's Pre Send Phase (PSP) - the preparation of frames.*/
unsigned controller_psp(CONTROLLER *ctrl);

/** Controller's Post Receive Phase (PRP) - evaluation of received data.*/
void controller_prp(CONTROLLER *ctrl);

/** Initialize a @c CONTROLLER type data structure.*/
CONTROLLER *controller_init(CONTROLLER *ctrl);

/** Bind host part of a node and communication controller together. */
void controller_attach_host(CONTROLLER *ctrl, void *arg, PTR_CALLBACK int_fn);

/** Bind the controller to channel(x) - necessary to call twice for channels A and B.*/
void controller_attach_channel(CONTROLLER *ctrl, CHANNEL *channel, unsigned x);

/** This function should be called by the channel after controller's transmission is 
	completed.*/
void controller_send_callback(void *ctrl);

/** This function should be called by the BG when it detects any error.*/
void controller_bg_error_callback(void *ctrl);

/** Perform all actions that are neccessary at the transmission phase (TP) start.*/
int controller_tp_start(CONTROLLER *ctrl);

/** Update MEDL position (current slot) and cluster time in C-State field.*/
void controller_set_next_slot(CONTROLLER *ctrl);

/** Perform the clock synchronization algorithm if enabled for the current slot.*/
unsigned controller_perform_cs(CONTROLLER *ctrl);

/** Check the global @c CCF field in the CNI and if valid then lock the status area.*/
int controller_enter_nbw(CONTROLLER *ctrl);

/** Checks the global CCF field and if it is valid unlocks the CNI status area.*/
int controller_leave_nbw(CONTROLLER *ctrl);

/** Act when the Watchdog timeout expires - force a shutdown.*/
void controller_run_watch_dog(CONTROLLER *ctrl);

/** Check whether all conditions expected at TP start are met.*/
int controller_tp_sync_assertion(CONTROLLER *ctrl);

/** Verify that the protocol state is either unchanged or the change was the result 
	of an enabled state transition.*/
int controller_check_state(CONTROLLER *ctrl);

/** Translate the numerical value of an implicit acknowledgement (IACK) state into 
	string.*/
char *controller_get_iack_state_name(unsigned iack);

/** Cancel any pending or active receive operation on both channels.*/
void controller_stop_receive(CONTROLLER *ctrl);

#endif /* #ifndef _FP1CTRL_H */

/*@}*/
/*@}*/
/*@}*/

