/**@addtogroup frame
 * @{*/

/**@addtogroup frame_ext  Interface
 * @{*/

/**@file
 * @brief Definition of N-frame and I-frame (interface).
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */


#ifndef _FP1FRAME_H
#define _FP1FRAME_H

#include "ttpc/fp1const.h"
#include "ttpc/fp1cstat.h"
#include "ttpc/fl1func.h"

/************************************************************************************/

/** This enum is used to numericaly define the type of frame.*/
typedef enum {
	FRAME_TYPE_N = 0,
	FRAME_TYPE_I
} FRAME_TYPE;


/** Defines the header of every frame.*/
typedef struct {
	/** Frame type flag: for I-frames is set to @c 1, for N-frames to @c 0.*/
	unsigned i_n_bit;   

	/** Mode change request field.
	 * The mode change request field consits of three bits:
	 * 	- bit 0 is immediate mode change request flag.
	 * 	- bit 1 is the lower bit for mode number index.
	 * 	- bit 2 is the higher bit for mode number index. */
	unsigned mode_change[3];  
} FRAME_HEADER;


/** Defines the type-specific content of an I-Frame.*/
typedef struct {
	C_STATE c_state;         /**< Copy of C-State vector.*/
	unsigned c_state_length; /**< Length of stored C-State in words.*/
} FRAME_I_DATA;


/** Defines the type-specific content of a N-Frame.*/
typedef struct {
	unsigned char app_data[MAX_APPLICATION_DATA_SIZE];   /**< Storage for application data.*/
	unsigned char _padding0[sizeof(C_STATE) - MAX_APPLICATION_DATA_SIZE];  /**< Zero fields to match the size of C-State 
										 (needed when we use a union with I-Frame)*/
	unsigned app_data_length;         /**< Length of stored application data in bytes (-1)*/
} FRAME_N_DATA;


/** This union encapsulates all frame sub-types into a single type, but only the
 * type-specific data.*/
typedef union {
	FRAME_N_DATA n;      /**< N-Frame specific part.*/
	FRAME_I_DATA i;      /**< I-Frame specific part.*/
} FRAME_UNION;


/** Describes all possible states of a frame during its lifetime.*/
typedef enum {
	FRAME_STATE_NULL,         /**< The frame is empty -  nothing was received or written yet.*/
	FRAME_STATE_RECEIVING,    /**< The frame is currently beeing updated - it is already partially written.*/
	FRAME_STATE_COMPLETE      /**< The frame has been completelly defined - either received or written manualy.*/
} FRAME_STATES;


/** A single frame type that defines every sub-type completelly.*/
typedef struct {
	FRAME_HEADER header;       /**< Type independent frame header.*/
	FRAME_UNION type;          /**< Type specific data for all types (union).*/
	unsigned crc;              /**< Type independent 16-bit CRC.*/

	FRAME_STATES state;        /**< Current state of this frame.*/
	unsigned noise;            /**< This value is set by the channel if there was any disturbance during transmission.*/
} FRAME;


/************************************************************************************/

/** Calculate the CRC of frame header structure (without initialization).*/
void frame_header_chain_crc(FRAME_HEADER *header, unsigned *crc, unsigned polynomial);

/** Return the requested mode number stored within frame header.*/
unsigned frame_header_get_mode(FRAME_HEADER *header);

/** Return the immediate mode change flag stored within frame header.*/
unsigned frame_header_get_immediate(FRAME_HEADER *header);

/** Write the requested mode number into a frame header.*/
void frame_header_set_mode(FRAME_HEADER *header, unsigned mode);

/** Write the immediate mode change flag into a frame header.*/
void frame_header_set_immediate(FRAME_HEADER *header, unsigned imm);


/************************************************************************************/

/** Return the physical frame size in bits.*/
unsigned frame_get_length(FRAME *frame);

/** Check whether the frame is a complete I-frame with correct CRC.*/
unsigned frame_is_valid_i_frame(FRAME *frame);

/** Check whether the frame is a complete N-frame with correct CRC.*/
unsigned frame_is_valid_n_frame(FRAME *frame, C_STATE *c_state, unsigned cs_length);

/** Check whether the frame is a complete and with correct CRC.*/
unsigned frame_is_valid(FRAME *frame, C_STATE *c_state, unsigned cs_length);

/** Calculate and return the CRC of given frame (for N-Frames include also the C-State).*/
unsigned frame_calculate_crc(FRAME *frame, C_STATE *c_state, unsigned cs_length);


/*************************************************************************************/
/*************************************************************************************/
/**@name Field Access Macros
 * The following macros are defined to ease the access to some fields of the FRAME type -
 * typically they call one of the exported access functions for some of the frame types.
 *
 * @todo Implement as ordinary functions to make the interface more transparent.
 * @{*/

/** Return byte @a i from the application data of N-Frame.*/
#define frame_get_app_data(frame, i)        frame_n_data_get_app_data(&(frame)->type.n, i)

/** Write value @a x into byte at position @a i in the application data of N-Frame.*/
#define frame_set_app_data(frame, i, x)     frame_n_data_set_app_data(&(frame)->type.n, i, x)

/** Return the length of application data (in bytes - 1) stored within N-Frame.*/
#define frame_get_app_data_length(frame)    frame_n_data_get_app_data_length(&(frame)->type.n)

/** Set the length of application data stored within N-Frame to @a x (in bytes - 1).*/
#define frame_set_app_data_length(frame, x) frame_n_data_set_app_data_length(&(frame)->type.n, x)

/** Return pointer to the C-State stored within an I-frame.*/
#define frame_get_c_state(frame)            (&((frame)->type.i.c_state))

/** Stores C-State @a x into the given I-Frame.*/
#define frame_set_c_state(frame, x)         ((frame)->type.i.c_state = x)

/** Return the length of C-State (in words) stored within an I-frame.*/
#define frame_get_c_state_length(frame)     frame_i_data_get_c_state_length(&(frame)->type.i)

/** Store the length @a x of C-State (in words) into an I-frame.*/
#define frame_set_c_state_length(frame, x)  frame_i_data_set_c_state_length(&(frame)->type.i, (x))

/** Return the value of frame type bit stored within frame header.*/
#define frame_get_i_n_bit(frame)            frame_header_get_i_n_bit((frame)->header)

/** Store the value @a x of frame type bit into frame header.*/
#define frame_set_i_n_bit(frame, x)         frame_header_set_i_n_bit((frame)->header, x)

/** Return the mode number from mode change request field within frame header.*/
#define frame_get_mode(frame)               frame_header_get_mode((frame)->header)

/** Store the mode number @a x into the mode change request field within frame header.*/
#define frame_set_mode(frame, x)            frame_header_set_mode((frame)->header, x)

/** Return the immediate flag value from mode change request field within frame header.*/
#define frame_get_immediate(frame)          frame_header_get_immediate((frame)->header)

/** Store the immediate flag @a x into the mode change request field within frame header.*/
#define frame_set_immediate(frame, x)       frame_header_set_immediate((frame)->header, x)

/*@}*/

/************************************************************************************/
/**@name Generated Access Functions
 * @{*/
/************************************************************************************/

declare_access_functions(FRAME_HEADER, frame_header, i_n_bit);
declare_index_access_functions(FRAME_HEADER, frame_header, mode_change);
declare_access_functions(FRAME_N_DATA, frame_n_data, app_data_length);
declare_index_access_functions(FRAME_N_DATA, frame_n_data, app_data);
declare_access_functions(FRAME_I_DATA, frame_i_data, c_state_length);
declare_access_functions(FRAME, frame, crc);

/*@}*/
/************************************************************************************/

#endif /* #ifndef _FP1FRAME_H */

/*@}*/
/*@}*/

