/**@addtogroup ldata_fun
 * @{*/

/**@addtogroup ldata_fun_ext  Interface 
 * @{*/


/**@file
 * @brief Protocol defined controller's local data structure and access functions.
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 * 
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */
 
#ifndef _FP1LDATA_H
#define _FP1LDATA_H

#include "ttpc/fp1frame.h"
#include "ttpc/fp1lname.h"


/************************************************************************************/

/** Describes all possible implicit acknowledgement (IACK) algorithm states.*/
typedef enum {
	IACK_OK,     		/**< No check needs to be performed.*/
	IACK_SEND,		/**< Initial state after sending a frame.*/
	IACK_CHECK1, 		/**< Check I is about to be performed.*/
	IACK_CHECK2, 		/**< Check II is about to be performed.*/
	IACK_ERROR,  		/**< Both checks were performed and the result is that this controller had failed.*/
	IACK_DOUBLE_FAILURE	/**< Double failure (see specification for a better explanation).*/
} IACK_STATES;


/** Defines the internal state of a controller (at the protocol level).*/
typedef struct {
	FRAME frame[2];                  /**< Storage for transmitted frames (interface to channel).*/
	unsigned old_ccf;                /**< Current value of CNI status area - stored by the locking function.*/
	unsigned old_msg_ccf[2];         /**< Current value of MAE CCF - stored by the locking function for appropiate channel.*/
	IACK_STATES iack_state;          /**< Current implicit acknowledgement algorithm state.*/
	unsigned successor;              /**< Membership position of controller's direct successor (not counting failed SRUs).*/
	unsigned membership_pos;         /**< Controller's own membership position.*/
	int sync_stack[SYNC_STACK_SIZE]; /**< Push-down stack for meassured time differences (clock synchronization).*/
	int sync_stack_top;              /**< Index that points above the most recent item in sync_stack.*/
	int mtd_a;                       /**< Measured time difference of last transmission on channel A.*/
	int mtd_b;                       /**< Measured time difference of last transmission on channel B.*/
	int tp_sync_interrupt;           /**< Flag that signals a pending TP-synchronous interrupt.*/
	int immediate_send_permission;   /**< Set while BG contains synchronization period parameters and BA signal was nat raised yet.*/
	int sync_period_params;          /**< Set while BG contains synchronization period parameters.*/
	int psp_phase;                   /**< This flag signals, that the current slot is a sending slot.*/

	/**\name Fields mirrored in the CNI
	The following fields are redundant and improve fault tolerance of the protocol - they
	were added to achieve an implementation closer to the HW.*/
	/*\{*/
	unsigned protocol_state;         /**< Current protocol state - redundant copy of a CNI field.*/
	int csct;                        /**< A redundant copy of clock state correction term from CNI.*/
	LOGICAL_NAME logical_name;       /**< A redundant copy of logical name from CNI.*/
	/*\}*/
} PLOC_DATA;


/************************************************************************************/

/** Initialize controller's local data with default values.*/
PLOC_DATA *ploc_data_initialize(PLOC_DATA *ldata);

/** Store a new value at the top of @c sync_stack and increment the pointer.*/
int ploc_data_push_sync_stack(PLOC_DATA *ldata, int value);


#endif /* #ifndef _FP1LDATA_H */

/*@}*/
/*@}*/

