/**@addtogroup medl
 * @{*/

/**@addtogroup medl_ext  Interface
 * @{*/

/**@file
 * @brief Defines the MEDL structure and access functions (implementation).
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 * @version 1.5
 * @date Last modified 2002-11-16
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#ifndef _FP1MEDL_H
#define _FP1MEDL_H

#include "ttpc/fl1func.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1lname.h"

/************************************************************************************/

/** Defines cotroller configuration parameters (CCP) part of MEDL.*/
typedef struct {
	unsigned physical_sru_name;       /**< Numerical identification of SRU - static name.*/
	unsigned controller_version;      /**< Version of TTP/C controller this MEDL is designed for.*/
	unsigned app_id;                  /**< Numerical identification of this application and MEDL.*/
	unsigned app_version;             /**< Version of the application and MEDL.*/
	unsigned cf;                      /**< Cold start enabled/disabled flag (1 = enabled).*/
	unsigned c_state_length;          /**< Length of the C-State vector in words.*/
	unsigned tdma_round_length;       /**< Duration of one TDMA round in macroticks.*/
	unsigned startup_timeout;         /**< Duration of startup timeout.*/
	unsigned free_running_macroticks; /**< Minimum period of clock corection application.*/
	unsigned mmfc;                    /**< Maximum successive membership failure count.*/
	unsigned microtick_int;           /**< Number of microticks per macrotick - integral part.*/
	unsigned microtick_frac;          /**< Number of microticks per macrotick - fractional part (1/2^16).*/
	unsigned max_cold_start_entry;    /**< Maximum successive attempts to send a cold start frame.*/
	unsigned crc;                     /**< CRC for this part of MEDL.*/
} MEDL_CCP;

/**@name Generated Access Functions for MEDL_CCP
 * 
 * Access functions for @c MEDL_CCP structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, 
 * set_function_prototype, increment_function_prototype
 * @{*/
declare_access_functions(MEDL_CCP, medl_ccp, physical_sru_name);
declare_access_functions(MEDL_CCP, medl_ccp, controller_version);
declare_access_functions(MEDL_CCP, medl_ccp, app_id);
declare_access_functions(MEDL_CCP, medl_ccp, app_version);
declare_access_functions(MEDL_CCP, medl_ccp, cf);
declare_access_functions(MEDL_CCP, medl_ccp, c_state_length);
declare_access_functions(MEDL_CCP, medl_ccp, tdma_round_length);
declare_access_functions(MEDL_CCP, medl_ccp, startup_timeout);
declare_access_functions(MEDL_CCP, medl_ccp, free_running_macroticks);
declare_access_functions(MEDL_CCP, medl_ccp, mmfc);
declare_access_functions(MEDL_CCP, medl_ccp, microtick_int);
declare_access_functions(MEDL_CCP, medl_ccp, microtick_frac);
declare_access_functions(MEDL_CCP, medl_ccp, max_cold_start_entry);
declare_access_functions(MEDL_CCP, medl_ccp, crc);
/*@}*/

/************************************************************************************/

/** Defines bus guardian configuration parameters (BGP) part of MEDL.*/
typedef struct {
	unsigned bg_arm_phase;        /**< Duration of bus arm phase - transmission disabled.
					@note This is a modification of the original TTP/C protocol.*/
	unsigned bg_slot_offset;      /**< Duration of slot offset phase - transmission disabled.*/
	unsigned bg_slot_duration;    /**< Sending slot duration (transmission phase) - transmission enabled.*/
	unsigned bg_remaining_round;  /**< Remaining round phase duration - transmission disabled.*/
	unsigned crc;                 /**< CRC for this part of MEDL.*/
} MEDL_BGP;

/**@name Generated Access Functions for MEDL_BGP
 *
 * Access functions for @c MEDL_BGP structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 * increment_function_prototype
 * @{*/
declare_access_functions(MEDL_BGP, medl_bgp, bg_arm_phase);
declare_access_functions(MEDL_BGP, medl_bgp, bg_slot_offset);
declare_access_functions(MEDL_BGP, medl_bgp, bg_slot_duration);
declare_access_functions(MEDL_BGP, medl_bgp, bg_remaining_round);
declare_access_functions(MEDL_BGP, medl_bgp, crc);
/*@}*/

/******************************************************/

/** Defines role list (RL) part of MEDL.
 * A single SRU may perform different roles within a TTP/C cluster. This structure defines
 * one single role of the SRU (the MEDL may contain a list of such roles).*/
typedef struct {
	LOGICAL_NAME logical_name;   /**< Logical name for this particular role (dynamic name).*/
	MEDL_BGP bus_guardian;       /**< BG parameter set for particular this role.*/
	unsigned crc;                /**< CRC for this part of MEDL.*/
} MEDL_RL;

/**@name Generated Access Functions for MEDL_RL
 *
 * Access functions for @c MEDL_RL structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 * increment_function_prototype
 * @{*/
declare_access_functions(MEDL_RL, medl_rl, crc);
/*@}*/

/******************************************************/

/** Defines mode control entry (MCE) part of MEDL.
 * The host may request an immediate mode change into modes 1-3. A deffered mode change
 * is translated using this entry. For every mode there is one entry of this type that
 * translates the requested relative mode number value (in range 0-2) into an absolute mode 
 * number (in range 1-30).*/
typedef struct {
	unsigned deferred_mode[3];   /**< Translation table for deffered mode changes - contains absolute mode numbers.*/
	unsigned crc;                /**< CRC for this part of MEDL.*/
} MEDL_MCE;

/**@name Generated Access Functions for MEDL_MCE
 *
 * Access functions for @c MEDL_MCE structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype, 
 * increment_function_prototype
 * @{*/
declare_index_access_functions(MEDL_MCE, medl_mce, deferred_mode);
declare_access_functions(MEDL_MCE, medl_mce, crc);
/*@}*/

/******************************************************/

/** Defines implementation specific parameters (ISP) part of MEDL.*/
typedef struct {
	unsigned entries;                    /**< Maximum number of MAEs within MEDL is 2^(6+<i>entries</i>).*/
	unsigned rom_speed;                  /**< Not used.*/
	unsigned i_frame_time;               /**< Not used.*/
	unsigned ifg_time;                   /**< Duration of IFG in macroticks*/
	unsigned listen_delay;               /**< Not used.*/
	unsigned watchdog_slot_length;       /**< Watchdog timout duration in microticks.*/
	unsigned listen_timeout_extension;   /**< Not used.*/
	unsigned listen_timeout_restart_time;/**< Not used.*/
	unsigned oversampling_rate;          /**< Not used.*/
	unsigned sync_interval;              /**< Not used.*/
	unsigned crc;                        /**< CRC for the previous group of fields.*/
	
	MEDL_BGP synchronization_period_params;    /**< BG parameter used during synchronization period.*/
	MEDL_BGP listen_cold_start_period_params;  /**< BG parameter set used in listen state by cold start enabled nodes.*/
} MEDL_ISP;

/**@name Generated Access Functions for MEDL_ISP
 *
 * Access functions for @c MEDL_CCP structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype,
 * increment_function_prototype
 * @{*/
declare_access_functions(MEDL_ISP, medl_isp, entries);
declare_access_functions(MEDL_ISP, medl_isp, rom_speed);
declare_access_functions(MEDL_ISP, medl_isp, i_frame_time);
declare_access_functions(MEDL_ISP, medl_isp, ifg_time);
declare_access_functions(MEDL_ISP, medl_isp, listen_delay);
declare_access_functions(MEDL_ISP, medl_isp, watchdog_slot_length);
declare_access_functions(MEDL_ISP, medl_isp, listen_timeout_extension);
declare_access_functions(MEDL_ISP, medl_isp, listen_timeout_restart_time);
declare_access_functions(MEDL_ISP, medl_isp, oversampling_rate);
declare_access_functions(MEDL_ISP, medl_isp, sync_interval);
declare_access_functions(MEDL_ISP, medl_isp, crc);
/*@}*/

/******************************************************/

/** Defines slot control entry (SCE) part of MEDL.
 * The MEDL contains one SCE for every transmission slot, but every operation mode
 * shares the same SCE.*/
typedef struct {
	unsigned sru_slot_duration; /**< Duration of this slot in macroticks.*/
	unsigned mo;                /**< Flag that denotes a multiplexed SRU with own membership.*/
	unsigned membership_pos;    /**< Bit position of the sending SRU for this slot within membership vector.*/
	unsigned medl_pos;          /**< Poisition of this slot within MEDL table (absolute slot number).*/
	unsigned mm;                /**< Flag that denotes a multiplexed SRU with multiplexed membership.*/
	unsigned eor;               /**< Flag that marks the end of current TDMA round.*/
	unsigned ra;                /**< The sending node may reintegrate within this slot if the flag is set.*/
	unsigned eot;               /**< Flag that marks the end of transmission cycle (return to the first slot again).*/
	unsigned cs;                /**< This flag enables the execution of clock synchronization algorithm in this slot.*/
	unsigned syf;               /**< Frames sent in this slot shall be used for clock synchronization.*/
	unsigned ba;                /**< This flag tells the sending node to raise its BG arm signal.*/
	unsigned delay_correction;  /**< Latency of message transmission from the sender node in microticks.*/
	LOGICAL_NAME logical_name;  /**< Logical name of the sender node for this slot.*/
} MEDL_SCE;

/**@name Generated Access Functions for MEDL_SCE
 *
 * Access functions for @c MEDL_SCE structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype,
 * increment_function_prototype
 * @{*/
declare_access_functions(MEDL_SCE, medl_sce, sru_slot_duration);
declare_access_functions(MEDL_SCE, medl_sce, mo);
declare_access_functions(MEDL_SCE, medl_sce, membership_pos);
declare_access_functions(MEDL_SCE, medl_sce, medl_pos);
declare_access_functions(MEDL_SCE, medl_sce, mm);
declare_access_functions(MEDL_SCE, medl_sce, eor);
declare_access_functions(MEDL_SCE, medl_sce, ra);
declare_access_functions(MEDL_SCE, medl_sce, eot);
declare_access_functions(MEDL_SCE, medl_sce, cs);
declare_access_functions(MEDL_SCE, medl_sce, syf);
declare_access_functions(MEDL_SCE, medl_sce, ba);
declare_access_functions(MEDL_SCE, medl_sce, delay_correction);
/*@}*/

/************************************************************************************/

/** Defines mode address entry (MAE) part of MEDL.*/
typedef struct {
	unsigned ru1;            /**< If 1 then raise user interrupt 1 in this slot.*/
	unsigned ia;             /**< If 1 then expect an I-frame on channel A, otherwise 
				   N-Frame should be sent.*/
	unsigned cni_address_cha;/**< Word offset into CNI message are where N-Frame 
				   data from channel A should be stored.*/
	unsigned ru2;            /**< If 1 then raise user interrupt 2 in this slot.*/
	unsigned ib;             /**< If 1 then expect an I-frame on channel B, otherwise 
				   N-Frame should be sent.*/
	unsigned cni_address_chb;/**< Word offset into CNI message are where N-Frame data 
				   from channel B should be stored.*/

	/**@name Mode change permission fields
	 * @{*/
	unsigned mcp_i[3];       /**< Value of 1 here means that the appropiate deffered
				   mode change request is permitted.*/
	unsigned mcp_d[3];       /**< Value of 1 here means that the appropiate immediate
				   mode change request is permitted.*/
	unsigned mcp_c;          /**< When this flag is set, request for clear pending
				   mode change is permitted.*/
	/*@}*/

	unsigned eoc;            /**< This flag denotes the end of the current cluster cycle.*/
	unsigned dafla;          /**< Application data field length on channel A 
				   (in bytes and one less than the actual length).*/
	unsigned daflb;          /**< Application data field length on channel B 
				   (in bytes and one less than the actual length).*/
	unsigned crc;            /**< CRC for this part of MEDL.*/
} MEDL_MAE;

/**@name Generated Access Functions for MEDL_MAE
 * Access functions for @c MEDL_MAE structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype,
 * increment_function_prototype
 * @{*/
declare_access_functions(MEDL_MAE, medl_mae, ru1);
declare_access_functions(MEDL_MAE, medl_mae, ia);
declare_access_functions(MEDL_MAE, medl_mae, cni_address_cha);
declare_access_functions(MEDL_MAE, medl_mae, ru2);
declare_access_functions(MEDL_MAE, medl_mae, ib);
declare_access_functions(MEDL_MAE, medl_mae, cni_address_chb);
declare_index_access_functions(MEDL_MAE, medl_mae, mcp_i);
declare_index_access_functions(MEDL_MAE, medl_mae, mcp_d);
declare_access_functions(MEDL_MAE, medl_mae, mcp_c);
declare_access_functions(MEDL_MAE, medl_mae, eoc);
declare_access_functions(MEDL_MAE, medl_mae, dafla);
declare_access_functions(MEDL_MAE, medl_mae, daflb);
declare_access_functions(MEDL_MAE, medl_mae, crc);
/*@}*/

/******************************************************/

/** Message descriptor list (MEDL) content definition.*/
typedef struct {
	unsigned revision;     /**< The revision number of this particular MEDL.*/

	/**@name Configuration Parameter Block
	 * @{*/
	MEDL_CCP controller_cfg_params;        /**< Controller configuration parameters.*/
	MEDL_BGP bus_guardian_cold_start;      /**< BG parameter set used during the cold start protocol state.*/
	MEDL_BGP bus_guardian_download;        /**< BG parameter set used during the download protocol state (not used).*/
	MEDL_RL role_list;                     /**< List of all possible roles for this SRU (currently limited to 1).*/
	MEDL_MCE mode_control_entry[MAX_MODE_COUNT + 1];  /**< Mode control entries for every operation mode.*/
	MEDL_ISP implementation_params;        /**< Implementation specific parameter set.*/
	/*@}*/

	/**@name Transmission Parameter Block
	 * @{*/
	MEDL_SCE slot_control_entry[MAX_TRANSMISSION_CYCLE_LENGTH];    
		/**< Slot control entries for every slot in the transmission cycle.*/
	MEDL_MAE mode_address_entry[MAX_MODE_COUNT + 1][MAX_TRANSMISSION_CYCLE_LENGTH];  
		/**< Mode address entries for every operation mode and every slot in the transmission cycle.*/
	/*@}*/
} MEDL;

/**@name Generated Access Functions for MEDL
 *
 * Access functions for @c MEDL structure that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see declare_access_functions, get_function_prototype, set_function_prototype,
 * increment_function_prototype
 * @{*/
declare_access_functions(MEDL, medl, revision);
/*@}*/

/************************************************************************************/

/** Compose a single floating point value from microticks per macrotick 
 * integer and fractional parts.*/
double medl_ccp_get_microticks(MEDL_CCP *medl_ccp);

/** Check CRC for all protected sections of the MEDL table.*/
int medl_validity_check(MEDL *medl);

/************************************************************************************/

#endif /* #ifndef _FP1MEDL_H */

/*@}*/
/*@}*/

