/**@addtogroup os
 * @{*/

/**@addtogroup os_ext  Interface
 * @{*/

/**@file
 * @brief TTPOs simulation based on C-Sim (interface)
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, University of West Bohemia, Pilsen, Czech Republic
 * @author Michal Jun, Czech Technical University, Prague, Czech Republic
 * @author Tomas Krejzek, Czech Technical University, Prague, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1,
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#ifndef _SA1OS_H
#define _SA1OS_H

#include "csim/csim.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1mdlrd.h"
#include "ttpc/fp1cni.h"
#include "ttpc/sp1ctrl.h"

/*************************************************************************************/

/** Maximum number of tasks that the OS can run (used for static array allocation).*/
#define OS_MAX_TASK_NUM   32

/** Number of bits needed to store a taks ID (must be large enough to contain
 * #OS_MAX_TASK_NUM mins @c 1.*/
#define OS_TASK_NUM_BITS  5

/** Specification of mode change request.*/
#define IMMEDIATE         0
/** Specification of mode change request.*/
#define DEFERRED          1
/** Specification of mode change request.*/
#define CLEAR             2


/** This constant is used to enable a task to run in any operation mode.*/
#define OS_ALL_MODES      0xFFFFFFFF

/** This constant is used to enable a task to run in any application mode.*/
#define OS_APP_MODES      0xFFFFFFFE

/*************************************************************************************/

struct os_process;

/** Task function - supplied by application programmer.
@param arg This argument is supplied by definition of the task.*/
typedef void (*OS_TASK_FN)(void *arg);


/** Individual fields of this structure are used to count occurances of different interrupts.*/
typedef struct {
	/**@name Non-Error Interrupts 
	 * @{*/
	unsigned mi;         /**< Mode changed interrupt.*/
	unsigned ml;         /**< Membership lost interrupt.*/
	unsigned mc;         /**< Membership changed interrupt.*/
	unsigned ui1;        /**< User interrupt 1.*/
	unsigned ui2;        /**< User interrupt 2.*/
	unsigned ti1;        /**< Time interrupt 1.*/
	unsigned ti2;        /**< Time interrupt 2.*/
	/*@}*/

	/**@name Protocol Errors
	 * @{*/
	unsigned pe;         /**< Any type of protocol error.*/
	unsigned pe_be;      /**< Bus guardian error.*/
	unsigned pe_ae;      /**< Acknowledgment error.*/
	unsigned pe_se;      /**< Clock synchronization error.*/
	unsigned pe_me;      /**< Membership error - maximum successive membership failures.*/
	unsigned pe_mc;      /**< MEDL CRC error.*/
	unsigned pe_cb;      /**< Communication blackout error.*/
	unsigned pe_dc;      /**< Download completed.*/
	/*@}*/

	/**@name BIST Errors
	 * @{*/
	unsigned be;         /**< Any type of BIST error.*/
	unsigned be_sc;      /**< Selftest completed.*/
	unsigned be_av;      /**< Assertion violation.*/
	unsigned be_br;      /**< Invalid BIST request.*/
	unsigned be_rm;      /**< RAM error.*/
	unsigned be_rf;      /**< Register file error.*/
	unsigned be_wd;      /**< Watchdog error.*/
	/*@}*/

	/**@name Host Errors
	 * @{*/
	unsigned he;         /**< Any type of host error.*/
	unsigned he_so;      /**< Slot occupied error.*/
	unsigned he_mv;      /**< Mode violation error.*/
	unsigned he_ne;      /**< NBW protocol error or message EIF contains invalid value.*/
	/*@}*/

	/**@name OS Errors
	 * @{*/
	unsigned os_error;   /**< OS deteceted error.*/
	unsigned os_active;  /**< Transition into the appliacation state.*/
	/*@}*/
} OS_EDM_COUNTERS;


/** OS task internal state and configuration.*/
typedef struct {
	char name[64];     /**< Identification for logging and debugging (needs not be unique).*/
	unsigned time;     /**< Schedule time - in macroticks since transmission cycle start.*/
	unsigned modes;    /**< This field contains one bit for every operation mode - if the bit is set, then the task may run the mode.*/
	OS_TASK_FN prog;   /**< Pointer to function that executes the task.*/
	void *prog_arg;    /**< Argument passed to the task function when it is called.*/
} OS_TASK;


/** Interrupt entry definition for the OS interrupt table.*/
typedef struct {
	unsigned enabled;  /**< This field masks the interrupt enable field in CNI - the application must not used CNI_IE directly.*/
	OS_TASK_FN prog;   /**< Pointer to service function for the interrupt.*/
	void *prog_arg;    /**< Argument passed to the interrupt function when it is called.*/
} OS_INTERRUPT;


/** One entry for every interrupt raised by the OS.*/
typedef struct {
	OS_INTERRUPT be;          /** BIST error.*/
	OS_INTERRUPT pe;          /** Protocol error.*/
	OS_INTERRUPT he;          /** Host error.*/
	OS_INTERRUPT ml;          /** Membership lost.*/
	OS_INTERRUPT mc;          /** Membership changed.*/
	OS_INTERRUPT mi;          /** Mode changed.*/
	OS_INTERRUPT ui1;         /** User interrupt 1.*/
	OS_INTERRUPT ui2;         /** User interrupt 2.*/
	OS_INTERRUPT ti1;         /** Time interrupt 1.*/
	OS_INTERRUPT ti2;         /** Time interrupt 2.*/
	OS_INTERRUPT os_error;    /** OS error.*/
	OS_INTERRUPT os_active;   /** Transition into active state.*/
} OS_INTERRUPT_TABLE;


/** List of all possible OS state (used by the internal state machine).*/
typedef enum {
	OS_STATE_OFF,           /** The controller is switched off.*/
	OS_STATE_INIT,          /** Initialization after power-up.*/
	OS_STATE_INIT_TASK,     /** Execution of the initialization task (application initialization).*/
	OS_STATE_READY,         /** Controller is ready - register interrupts.*/
	OS_STATE_SYNCHRONISED,  /** Controller is synchronized - get cluster state, schedule first task.*/
	OS_STATE_STARTUP,       /** Cluster is in startup mode - lifesign update.*/
	OS_STATE_APPLICATION,   /** Cluster is in application mode - normal operation.*/
	OS_STATE_ERROR          /** A fatal error was detected and was not handled yet.*/
} OS_STATES;


/** Detailed error codes for OS failures.*/
typedef enum {
	OS_ERROR_NONE,            /**< No error happended.*/
	OS_ERROR_EDM_BE,          /**< BIST error detected.*/
	OS_ERROR_EDM_PE,          /**< Protocol error detected.*/
	OS_ERROR_INT_TIMEOUT,     /**< Interrupt timeout exceeded.*/
	OS_ERROR_FREEZE,          /**< Controller was unexpectedly switched off.*/
	OS_ERROR_ALREADY_ON,      /**< Controller is already switched on.*/
	OS_ERROR_NOT_READY,       /**< Controller is not ready.*/
	OS_ERROR_SYNC_TIMEOUT,    /**< Synchronization state timout exceeded.*/
	OS_ERROR_RESTART,         /**< Application requests restart (not an error).*/
	OS_ERROR_NO_TASK,         /**< No task to schedule.*/
	OS_ERROR_BAD_TASK,        /**< Invalid task function pointer.*/
	OS_ERROR_BAD_SLOT_COUNT,  /**< Invalid number of SRU slots detected (read from MEDL).*/
	OS_ERROR_BAD_SCHEDULE     /**< The scheduled task does not exist.*/
} OS_ERROR_CODE;


/** Internal state of a C-Sim process simulating TTP/OS.*/
typedef struct os_process {
	/**@name Inherited from C-Sim
	 * @{*/
	csim_d_process;
	/*@}*/
	
	CONTROLLER *controller;          /**< Pointer to the attached controller model instance.*/
	CNI *cni;                        /**< Pointer CNI interface to the controller.*/

	OS_TASK init_task;               /**< Application initialization task.*/
	OS_TASK tasks[OS_MAX_TASK_NUM];  /**< List of all tasks.*/
	unsigned slot_time[MAX_TRANSMISSION_CYCLE_LENGTH + 1]; /**< Action time (relative to start of transmission cycle) of every slot.*/
	OS_INTERRUPT_TABLE int_table;    /**< List of all interrupts - enable-flags, handlers.*/

	/* Status flags */
	OS_ERROR_CODE error_code;        /**< Last error code - set before the OS enters error state.*/
	OS_STATES os_state;              /**< Current execution state of OS - used by internal state machine.*/
	unsigned interrupted;            /**< Set when by interrupt from controller and cleared after the interrupt is serviced.*/
	unsigned old_msg_ccf;            /**< Value of CCF stored at beginning of protected CNI access.*/

	/* Provided statistics */
	OS_EDM_COUNTERS edm_counters;    /**< Counters for every interrupt eccurance (must be enabled).*/
	unsigned startup_count;          /**< Every time the OS tries to start the controller this field is incremented.*/
	unsigned transmission_cycle;     /**< Number of current transmission cycle (?since last restart).*/

	/* Control Fields */
	CSIM_TIME error_delay;           /**< Time to spend in the error state before restarting (or stopping).*/
	CSIM_TIME last_timer;            /**< Time of last timer interrupt arrival.*/
	unsigned auto_life_sign;         /**< If nonzero then schedule life-sign update task automatically.*/
	unsigned min_node_count;         /**< Minimum number of nodes required to request a mode change.*/
	unsigned first_app_mode;         /**< Number of application mode to switch to after request number of nodes is detected.*/
	unsigned start_allowed;          /**< If nonzero then the OS will try to start the controller again after failure.*/
	unsigned int_log_enabled;        /**< When set the OS will count every interrupt occurance.*/
	unsigned next_task;              /**< Number of next scheduled task.*/
	unsigned next_task_time;         /**< Activation time of next scheduled task.*/
	unsigned task_count;             /**< Current number of tasks wihin the @c tasks vector.*/
	unsigned iteration;              /**< Current of iteration number in the wait-for-sychronization loop.*/
	unsigned old_controller_lifesign;/**< Previous value of controller's lifesign field.*/
	unsigned transmission_cycle_len; /**< Duration of one transmission cycle (in macroticks).*/
	unsigned round_len;              /**< Duration of one TDMA round (in macroticks).*/
	unsigned slot_count;             /**< Number of slots within one transmission cycle.*/
	unsigned round_count;            /**< Number of TDMA rounds within one transmission cycle.*/
	unsigned send_time;              /**< The start of IFG preceding my first sending slot.*/
} OS_PROCESS;

/**@name Generated Access functions
 * @{*/
declare_access_functions(OS_PROCESS, os, auto_life_sign);
/*@}*/


/************************************************************************************/

/** Return pointer to the default OS fault mask
 * @return  The returned pointer is to a statically allocated memory area that
 * should not be changed. */
extern unsigned* os_fault_mask(void);

/** Dynamically create and initialize TTPOs C-Sim process.*/
OS_PROCESS *os_create(char *name);

/** Bind the operating system to a specific controller.*/
void os_attach_controller(OS_PROCESS *os, CONTROLLER *cntl);

/** Create new task with specified parameters.*/
unsigned os_add_task(OS_PROCESS *os, char *name, unsigned time, unsigned modes, OS_TASK_FN prog, void *prog_arg);

/** Create a special task that is executed prior to any other task - just after
   the controller is ready.*/
void os_add_init_task(OS_PROCESS *os, OS_TASK_FN prog, void *prog_arg);

/** Schedule the lifesign update task to the specified time.*/
unsigned os_add_lifesign_update(OS_PROCESS *os, unsigned time);

/** Removes all tasks with specified name from the scheduler.*/
unsigned os_remove_task_name(OS_PROCESS *os, char *name);

/** Remove task specified by number from the scheduler.*/
unsigned os_remove_task_id(OS_PROCESS *os, unsigned id);

/** Check controller lifesign and update host lifesign as well.*/
unsigned os_test_controller_lifesign(OS_PROCESS *os);

/** Lock a message area entry at @a addr by incrementing its CCF field - pair
	function to @c os_ccf_leave().*/
unsigned os_ccf_enter(OS_PROCESS *os, unsigned addr);

/** Release the lock on message area entry at @a addr by incrementing its CCF field 
	again - pair function to @c os_ccf_enter().*/
unsigned os_ccf_leave(OS_PROCESS *os, unsigned addr);

/** Check whether the global CCF field in the CNI enables reading.*/
unsigned os_check_ccf(OS_PROCESS *os);

/** Set the EIF field in a message area entry at @a addr to a valid value (use before sending).*/
void os_validate_eif(OS_PROCESS *os, unsigned addr);

/** Set the EIF field in a message area entry at @a addr to an invalid value.*/
void os_invalidate_eif(OS_PROCESS *os, unsigned addr);

/** Return pointer to the CNI interface.*/
CNI *os_get_cni(OS_PROCESS *os);

/** Return pointer to MEDL table of the controller. */
MEDL *os_get_medl(OS_PROCESS *os);

/** Return current time (relative to the begining of transmission cycle).*/
unsigned os_get_time(OS_PROCESS *os);

/** Return the sending slot number (module TDMA round length) of the node (0-based).*/
unsigned os_get_my_slot(OS_PROCESS *os);

/** Return current slot number (modulo transmission cycle length), starting with 0.*/
unsigned os_get_current_slot(OS_PROCESS *os);

/** Check whether the specified mode change request is permitted.*/
unsigned os_get_mode_change_permission(OS_PROCESS *os, unsigned mode, unsigned slot, unsigned bit, unsigned req_type);

/** Return the EIF of message area entry at given offset.*/
CNI_EIF_VALUES os_get_eif(OS_PROCESS *os, unsigned addr);

/** Return current operation mode (read from as in C-State).*/
unsigned os_get_current_mode(OS_PROCESS *os);

/** Request immediate or deferred mode change.*/
unsigned os_request_mode_change(OS_PROCESS *os, unsigned mode, unsigned immediate);

/** Immediately cancel any pending (deferred) mode change.*/
unsigned os_cancel_mode_change_request(OS_PROCESS *os);

/** Return count of currently active nodes (within membership vector).*/
unsigned os_get_node_count(OS_PROCESS *os);

/** Return pointer to the internal EDM counters.*/
OS_EDM_COUNTERS *os_get_edm_counters(OS_PROCESS *os);

/** Clear the internal EDM counters.*/
void os_clear_edm_counters(OS_PROCESS *os);

/** Return the number of node startups.*/
unsigned os_get_startup_count(OS_PROCESS *os);

/** Return the number of current transmission cycle.*/
unsigned os_get_transmission_cycle(OS_PROCESS *os);

/** Enable the startup of the controller (does not mean it must be down).*/
void os_start_controller(OS_PROCESS *os);

/** Enable/Disable counting of interrupt occurances.*/
void os_set_int_log(OS_PROCESS *os, unsigned enable);

/** Set the triggering condition for issuing a mode change request from 
	startup into application mode.*/
void os_set_mode_trigger(OS_PROCESS *os, unsigned min_cnt, unsigned mode);

/** Restart the controller and the TTPos.*/
void os_restart(OS_PROCESS *os);

/** Translate the contents of given counters into a readable string.*/
char *os_counters2str(OS_EDM_COUNTERS *cnt);

/** Return the string representation od internal EDM counters of the OS.*/
char *os_get_edm_str(OS_PROCESS *os);

/** Return the start of TP of the specified slot in macroticks (relative 
	to transmission cycle start).*/
unsigned os_get_slot_time(OS_PROCESS *os, unsigned slot);

/** Return the duration of a single TDMA round in macroticks.*/
unsigned os_get_round_length(OS_PROCESS *os);

/** Return the start time of IFG preceding my sending slot modified by round number
	and an offset.*/
unsigned os_get_send_time(OS_PROCESS *os, unsigned round, int offset);


/*************************************************************************************/

/** Return pending deferred mode change request (stored within C-State).
@param os Pointer to an instance of OS_PROCESS.
@return Zero if no mode change is requested, relative mode number otherwise.*/
#define os_get_dmc(os) \
	c_state_get_dmc(&(os)->cni->c_state)


/** Return the status of specified interrupt.
@param os Pointer to OS_PROCESS instance.
@param int_id Identification of the interrupt - the shortcut version with no quotes is
	expected, i.e. one of: be, pe, he, ml, mc, mi, ui1, ui2, ti1, ti2, os_error, os_active.
@return Nonzero value if the interrupt has been raised and not handle yet, zero otherwise.*/
#define os_get_int_status(os, int_id) \
	(cni_is_get_##int_id(&(os)->cni->interrupt_status))


/** Validate two replicated messages at once.
@param os Pointer to OS_PROCESS instance.
@param addr Vector of two integers that contains word offsets of two message area entries.*/
#define os_validate_rep_eif(os, addr) do { \
	os_validate_eif(os, addr[0]); \
	os_validate_eif(os, addr[1]); \
} while (0)


/** Invalidate two replicated messages at once.
@param os Pointer to OS_PROCESS instance.
@param addr Vector of two integers that contains word offsets of two message area entries.*/
#define os_invalidate_rep_eif(os, addr) do { \
	os_invalidate_eif(os, addr[0]); \
	os_invalidate_eif(os, addr[1]); \
} while (0)


/*************************************************************************************/
/**@name Interrupt Service Registration
 * 
 * This group of functions can be used to install a separate application defined handler
 * for any interrupt raised by the OS - only one routine may be registered for one 
 * interrupt though.
 * 
 * All of the functions use the same arguments, so they will be explained only once here.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param prog Pointer to interrupt service function.
 * @param arg Argument passed to the service function every time the interrupt is raised.
 *
 * @{*/

/** Register interrupt service for any BIST error.*/
void os_add_be_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for any protocol error.*/
void os_add_pe_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for any host error.*/
void os_add_he_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for membership lost event.*/
void os_add_ml_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for membership changed event.*/
void os_add_mc_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for mode changed event.*/
void os_add_mi_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for user interrupt 1.*/
void os_add_ui1_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for user interrupt 2.*/
void os_add_ui2_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for time interrupt 1.*/
void os_add_ti1_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for time interrupt 2.*/
void os_add_ti2_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for os error.*/
void os_add_os_error_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/** Register interrupt service for transition int oactive state event.*/
void os_add_os_active_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg);

/*@}*/
/*************************************************************************************/
/*************************************************************************************/
/**@name Interrupt Enable/Disable
 * All of the functions use the same arguments, so they will be explained only once here.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param state Requested state of the interrupt - 1 = enabled, 0 = disabled.
 * @{*/

/** Enable/disable the raising of BIST error interrupt.*/
void os_enable_be_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of host error interrupt.*/
void os_enable_he_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of protocol error interrupt.*/
void os_enable_pe_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of membership lost interrupt.*/
void os_enable_ml_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of membership changed interrupt.*/
void os_enable_mc_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of mode changed interrupt.*/
void os_enable_mi_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of user interrupt 1.*/
void os_enable_ui1_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of user interrupt 2.*/
void os_enable_ui2_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of time interrupt 2.*/
void os_enable_ti2_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of OS error interrupt.*/
void os_enable_os_error_int(OS_PROCESS *os, unsigned state);
/** Enable/disable the raising of tranition to active state interrupt.*/
void os_enable_os_active_int(OS_PROCESS *os, unsigned state);

/*@}*/
/*************************************************************************************/
/*************************************************************************************/
/**@name Single Message Writing
 * The folowing group of functions is used to write into the data field of a single 
 * message area entry.
 *
 * Most of the functions use the same arguments, so they will be explained only once here.
 * All data changing routines call internally @c os_set_ofs_data() routine.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param val Value that will be written into the message.
 * @{*/

/** Copy data of any size into the message data field at specified byte offset.
 * 
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param buf Pointer to data that should be written into the message.
 * @param size Number of bytes to write into the message.*/
void os_set_ofs_data(OS_PROCESS *os, unsigned addr, unsigned ofs, char *buf, unsigned size);

/** Copy one @c char into the message data field at specified byte offset.*/
void os_set_ofs_char(OS_PROCESS *os, unsigned addr, unsigned ofs, unsigned char val);

/** Copy one @c short into the message data field at specified byte offset.*/
void os_set_ofs_short(OS_PROCESS *os, unsigned addr, unsigned ofs, unsigned short val);

/** Copy one @c int into the message data field at specified byte offset.*/
void os_set_ofs_int(OS_PROCESS *os, unsigned addr, unsigned ofs, unsigned int val);

/** Copy one @c long into the message data field at specified byte offset.*/
void os_set_ofs_long(OS_PROCESS *os, unsigned addr, unsigned ofs, unsigned long val);

/** Copy one @c float into the message data field at specified byte offset.*/
void os_set_ofs_float(OS_PROCESS *os, unsigned addr, unsigned ofs, float val);

/** Copy one @c double into the message data field at specified byte offset.*/
void os_set_ofs_double(OS_PROCESS *os, unsigned addr, unsigned ofs, double val);

/************ (ofs = 0) **************/

/** Copy data of any size into the message data field at offset 0.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param buf Pointer to data that should be written into the message.
 * @param size Number of bytes to write into the message.*/
#define os_set_data(os, addr, buf, size)	os_set_ofs_data(os, addr, 0, buf, size)

/** Copy one @c char into the message data field at offset 0.*/
#define os_set_char(os, addr, val)			os_set_ofs_char(os, addr, 0, val)

/** Copy one @c short into the message data field at offset 0.*/
#define os_set_short(os, addr, val)	        os_set_ofs_short(os, addr, 0, val)

/** Copy one @c int into the message data field at offset 0.*/
#define os_set_int(os, addr, val)			os_set_ofs_int(os, addr, 0, val)

/** Copy one @c long into the message data field at offset 0.*/
#define os_set_long(os, addr, val)			os_set_ofs_long(os, addr, 0, val)

/** Copy one @c float into the message data field at offset 0.*/
#define os_set_float(os, addr, val)			os_set_ofs_float(os, addr, 0, val)

/** Copy one @c double into the message data field at offset 0.*/
#define os_set_double(os, addr, val)		os_set_ofs_double(os, addr, 0, val)

/*@}*/
/*************************************************************************************/
/*************************************************************************************/
/**@name Replicated Message Writing
 * The folowing group of functions is used to write into the data field of two replicated
 * message area entries at once.
 *
 * Most of the functions use the same arguments, so they will be explained only once here.
 * All data changing routines call internally @c os_set_ofs_rep_data() routine.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Vector of two word offset for two replicated message area entries within 
 *  CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param val Value that will be written into the messages.
 * @{*/

/** Copy data of any size into two replicated messages (start at specified byte offset).
 * 
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Vector of two word offset for two replicated message area entries within 
 *  CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param buf Pointer to data that should be written into the message.
 * @param size Number of bytes to write into the message.*/
#define os_set_ofs_rep_data(os, addr, ofs, buf, size) do { \
	os_set_ofs_data(os, addr[0], ofs, buf, size); \
	os_set_ofs_data(os, addr[1], ofs, buf, size); \
} while (0)

/** Copy one @c char into two replicated messages (start at specified byte offset).*/
#define os_set_ofs_rep_char(os, addr, ofs, val) do { \
	os_set_ofs_char(os, addr[0], ofs, val); \
	os_set_ofs_char(os, addr[1], ofs, val); \
} while (0)

/** Copy one @c short into two replicated messages (start at specified byte offset).*/
#define os_set_ofs_rep_short(os, addr, ofs, val) do { \
	os_set_ofs_short(os, addr[0], ofs, val); \
	os_set_ofs_short(os, addr[1], ofs, val); \
} while (0)

/** Copy one @c int into two replicated messages (start at specified byte offset).*/
#define os_set_ofs_rep_int(os, addr, ofs, val) do { \
	os_set_ofs_int(os, addr[0], ofs, val); \
	os_set_ofs_int(os, addr[1], ofs, val); \
} while (0)

/** Copy one @c long into two replicated messages (start at specified byte offset).*/
#define os_set_ofs_rep_long(os, addr, ofs, val) do { \
	os_set_ofs_long(os, addr[0], ofs, val); \
	os_set_ofs_long(os, addr[1], ofs, val); \
} while (0)

/** Copy one @c float into two replicated messages (start at specified byte offset).*/
#define os_set_ofs_rep_float(os, addr, ofs, val) do { \
	os_set_ofs_float(os, addr[0], ofs, val); \
	os_set_ofs_float(os, addr[1], ofs, val); \
} while (0)

/** Copy one @c double into two replicated messages (start at specified byte offset).*/
#define os_set_ofs_rep_double(os, addr, ofs, val) do { \
	os_set_ofs_double(os, addr[0], ofs, val); \
	os_set_ofs_double(os, addr[1], ofs, val); \
} while (0)

/************ (ofs = 0) **************/

/** Copy data of any size into two replicated messages (start at offset 0).
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Vector of two word offset for two replicated message area entries within 
 *  CNI (as specified in MEDL).
 * @param buf Pointer to data that should be written into the message.
 * @param size Number of bytes to write into the message.*/
#define os_set_rep_data(os, addr, buf, size) \
	os_set_ofs_rep_data(os, addr, 0, buf, size)

/** Copy one @c char into two replicated messages (start at 0 offset).*/
#define os_set_rep_char(os, addr, val) \
	os_set_ofs_rep_char(os, addr, 0, val)

/** Copy one @c short into two replicated messages (start at 0 offset).*/
#define os_set_rep_short(os, addr, val) \
	os_set_ofs_rep_short(os, addr, 0, val)

/** Copy one @c int into two replicated messages (start at 0 offset).*/
#define os_set_rep_int(os, addr, val) \
	os_set_ofs_rep_int(os, addr, 0, val)

/** Copy one @c long into two replicated messages (start at 0 offset).*/
#define os_set_rep_long(os, addr, val) \
	os_set_ofs_rep_long(os, addr, 0, val)

/** Copy one @c float into two replicated messages (start at 0 offset).*/
#define os_set_rep_float(os, addr, val) \
	os_set_ofs_rep_float(os, addr, 0, val)

/** Copy one @c double into two replicated messages (start at 0 offset).*/
#define os_set_rep_double(os, addr, val) \
	os_set_ofs_rep_double(os, addr, 0, val)

/*@}*/
/*************************************************************************************/
/*************************************************************************************/
/*************************************************************************************/
/**@name Single Message Reading
 * The folowing group of functions is used to read from the data field of a single message 
 * area entry.
 *
 * Most of the functions use the same arguments, so they will be explained only once here.
 * All data changing routines call internally @c os_get_ofs_data() routine.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param val Pointer to a field that will store the read message item.
 * @return
 * 	- @c 0 if the message was correctly received.
 * 	- @c 1 otherwise.
 *
 * @{*/

/** Read data of any size from the message data field (at specified byte offset).
 * 
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param buf Pointer to an allocated buffer that should contain the read data.
 * @param size Number of bytes to read from the message.
 * @return
 * 	- @c 0 if the message was correctly received.
 * 	- @c 1 otherwise.*/
unsigned os_get_ofs_data(OS_PROCESS *os, unsigned addr, unsigned ofs, char *buf, unsigned size);

/** Read one @c char from the message data field (at specified byte offset).*/
#define os_get_ofs_char(os, addr, ofs, val) \
	os_get_ofs_data(os, addr, ofs, (char *)(val), sizeof(char))

/** Read one @c short from the message data field (at specified byte offset).*/
#define os_get_ofs_short(os, addr, ofs, val) \
	os_get_ofs_data(os, addr, ofs, (char *)(val), sizeof(short))

/** Read one @c int from the message data field (at specified byte offset).*/
#define os_get_ofs_int(os, addr, ofs, val) \
	os_get_ofs_data(os, addr, ofs, (char *)(val), sizeof(int))

/** Read one @c long from the message data field (at specified byte offset).*/
#define os_get_ofs_long(os, addr, ofs, val) \
	os_get_ofs_data(os, addr, ofs, (char *)(val), sizeof(long))

/** Read one @c float from the message data field (at specified byte offset).*/
#define os_get_ofs_float(os, addr, ofs, val) \
	os_get_ofs_data(os, addr, ofs, (char *)(val), sizeof(float))

/** Read one @c double from the message data field (at specified byte offset).*/
#define os_get_ofs_double(os, addr, ofs, val) \
	os_get_ofs_data(os, addr, ofs, (char *)(val), sizeof(double))

/************ (ofs = 0) **************/

/** Read data of any size from the message data field (at offset 0).
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param buf Pointer to an allocated buffer that should contain the read data.
 * @param size Number of bytes to read from the message.
 * @return
 * 	- @c 0 if the message was correctly received.
 * 	- @c 1 otherwise.*/
#define os_get_data(os, addr, buf, size) \
	os_get_ofs_data(os, addr, 0, buf, size)

/** Read one @c char from the message data field (at offset 0).*/
#define os_get_char(os, addr, val) \
	os_get_ofs_char(os, addr, 0, val)

/** Read one @c short from the message data field (at offset 0).*/
#define os_get_short(os, addr, val) \
	os_get_ofs_short(os, addr, 0, val)

/** Read one @c int from the message data field (at offset 0).*/
#define os_get_int(os, addr, val) \
	os_get_ofs_int(os, addr, 0, val)

/** Read one @c long from the message data field (at offset 0).*/
#define os_get_long(os, addr, val) \
	os_get_ofs_long(os, addr, 0, val)

/** Read one @c float from the message data field (at offset 0).*/
#define os_get_float(os, addr, val) \
	os_get_ofs_float(os, addr, 0, val)

/** Read one @c double from the message data field (at offset 0).*/
#define os_get_double(os, addr, val) \
	os_get_ofs_double(os, addr, 0, val)

/*@}*/
/*************************************************************************************/
/*************************************************************************************/
/**@name Replicated Message Reading
 * The folowing group of functions is used to read the data fields of two replicated
 * message area entries at once. No voting is performed - if both replicated messages were
 * correctly received only the first is returned.
 *
 * Most of the functions use the same arguments, so they will be explained only once here.
 * All data changing routines call internally @c os_get_ofs_rep_data() routine.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Vector of two word offset for two replicated message area entries within 
 *  CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param val Pointer to variable that should store the read messages item.
 * @return
 * 	- @c 0 if at least one message was correctly received.
 * 	- @c 1 otherwise.
 *
 * @{*/

/** Read data of any size from specified byte offset from the first correct message from two replicas.
 *
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param buf Pointer to an allocated buffer that should contain the read data.
 * @param size Number of bytes to read from the message.
 * @return
 * 	- @c 0 if the message was correctly received.
 * 	- @c 1 otherwise.*/
unsigned os_get_ofs_rep_data(OS_PROCESS *os, const unsigned addr[], unsigned ofs, char *buf, unsigned size);

/** Read one @c char from specified byte offset from the first correct message from two replicas.*/
unsigned os_get_ofs_rep_char(OS_PROCESS *os, const unsigned addr[], unsigned ofs, unsigned char *val);

/** Read one @c short from specified byte offset from the first correct message from two replicas.*/
unsigned os_get_ofs_rep_short(OS_PROCESS *os, const unsigned addr[], unsigned ofs, unsigned short *val);

/** Read one @c int from specified byte offset from the first correct message from two replicas.*/
unsigned os_get_ofs_rep_int(OS_PROCESS *os, const unsigned addr[], unsigned ofs, unsigned int *val);

/** Read one @c long from specified byte offset from the first correct message from two replicas.*/
unsigned os_get_ofs_rep_long(OS_PROCESS *os, const unsigned addr[], unsigned ofs, unsigned long *val);

/** Read one @c float from specified byte offset from the first correct message from two replicas.*/
unsigned os_get_ofs_rep_float(OS_PROCESS *os, const unsigned addr[], unsigned ofs, float *val);

/** Read one @c double from specified byte offset from the first correct message from two replicas.*/
unsigned os_get_ofs_rep_double(OS_PROCESS *os, const unsigned addr[], unsigned ofs, double *val);

/************ (ofs = 0) **************/

/** Read data of any size from offset 0 from the first correct message from two replicas.
 * 
 * @param os Pointer to OS_PROCESS instance.
 * @param addr Word offset of an message area entry within CNI (as specified in MEDL).
 * @param ofs Byte offset within the message data field. If not specified, zero is assumed.
 * @param buf Pointer to an allocated buffer that should contain the read data.
 * @param size Number of bytes to read from the message.
 * @return
 * 	- @c 0 if the message was correctly received.
 * 	- @c 1 otherwise.*/
#define os_get_rep_data(os, addr, buf, size) \
	os_get_ofs_rep_data(os, addr, 0, buf, size)

/** Read one @c char from offset 0 from the first correct message from two replicas.*/
#define os_get_rep_char(os, addr, val) \
	os_get_ofs_rep_char(os, addr, 0, val)

/** Read one @c short from offset 0 from the first correct message from two replicas.*/
#define os_get_rep_short(os, addr, val) \
	os_get_ofs_rep_short(os, addr, 0, val)

/** Read one @c int from offset 0 from the first correct message from two replicas.*/
#define os_get_rep_int(os, addr, val) \
	os_get_ofs_rep_int(os, addr, 0, val)

/** Read one @c long from offset 0 from the first correct message from two replicas.*/
#define os_get_rep_long(os, addr, val) \
	os_get_ofs_rep_long(os, addr, 0, val)

/** Read one @c float from offset 0 from the first correct message from two replicas.*/
#define os_get_rep_float(os, addr, val) \
	os_get_ofs_rep_float(os, addr, 0, val)

/** Read one @c double from offset 0 from the first correct message from two replicas.*/
#define os_get_rep_double(os, addr, val) \
	os_get_ofs_rep_double(os, addr, 0, val)

/*@}*/
/*************************************************************************************/


#endif

/*@}*/
/*@}*/

