/**@addtogroup ctrl_sim
 * @{*/

/**@addtogroup ctrl_sim_ext  Interface
 * @{*/


/**@file
 * @brief C-Sim simulation wrapper for the communication controller (interface).
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */

#ifndef _SP1CTRL_H
#define _SP1CTRL_H

#include "csim/csim.h"
#include "ttpc/fp1ctrl.h"
#include "ttpc/sp1ldata.h"

/*************************************************************************************/

struct controller_model;

/** Function of this type is used for monitoring the execution of controller.
 *
 * Such function is called after every csim_hold() call, i.e. it is called in every
 * time-slice of controller exection.
 *
 * @param ctrl Pointer to the monitore instance of communication controller model.
 * @param state Current execution state (the state that is just beginning).*/
typedef void (*MONITOR_FN)(struct controller_model *ctrl, CONTROLLER_EXE_STATE state);


/** C-Sim process internal structure for an independant timer.*/
typedef struct timer_process {
	/**@name Inherited from C-Sim
	 * @{*/
	csim_d_process;
	/*@}*/
	void *controller;		/**< Pointer to the attached controller.*/
	int timer;				/**< Identification number of this timer (0 or 1).*/
} TIMER_PROCESS;


/** C-Sim process internal structure for an indepentent watch dog timer.*/
typedef struct watch_dog_process {
	/**@name Inherited from C-Sim
	 * @{*/
	csim_d_process;
	/*@}*/
	struct controller_model *controller;     /**< Pointer to the attached controller.*/
} WATCH_DOG_PROCESS;


/** C-Sim process that includes controller model definition into its internal state.*/
typedef struct controller_process {
	/**@name Inherited from C-Sim
	 * @{*/
	csim_d_process;
	/*@}*/
	CONTROLLER *protocol;            /**< Pointer to protocol defined fields.*/
	struct controller_model *model;  /**< Pointer to simulation model specific fields.*/
} CONTROLLER_PROCESS;


/** Internal state of controller enhanced with simulation specific fields. */
typedef struct controller_model {
	/**@name Inherited fields
	 * @{*/
	d_controller
	/*@}*/

	/**@name Simulation Specific Fields
	 * @{*/
	double clock_drift;              /**< Controller's internal clock drift rate [sec/sec].*/
	MLOC_DATA mloc_data;             /**< Encapsulation of simulation model specific internal data.*/
	MONITOR_FN monitor;              /**< Pointer to a monitoring function - called after every 
					   csim_hold() operation.*/
	CONTROLLER_PTR_FN mtd_monitor[2];    /**< Pointer to a function that is called for 
					       every meassured time difference on separate channels.*/
	void *mtd_monitor_arg[2];        /**< User defined argument passed to the @c mtd_monitor() functions.*/

	CONTROLLER_PROCESS *process;     /**< Pointer to the C-Sim process executing this controller model.*/
	TIMER_PROCESS *timer[2];         /**< Pointers to two internal timer processes (for timer interrupts).*/
	WATCH_DOG_PROCESS *watch_dog;    /**< Pointer to internal watch dog process.*/
	/*@}*/
} CONTROLLER_MODEL;


/*************************************************************************************/

/** Dynamically create and initialize a controller C-Sim process.*/
CONTROLLER_MODEL *controller_create(char *name);

/** Call this after every change of @c cni.co field (used internally).*/
void controller_co_changed(CONTROLLER_MODEL *model);

/** Call this after every change of @c cni.ca field (used internally).*/
void controller_ca_changed(CONTROLLER_MODEL *model);

/** Call this after every change of the @c timer or @c ti fields in CNI (used internally).*/
void controller_timer_changed(CONTROLLER_MODEL *model, int timer, int cs);

/** Register a monitoring function to watch controllers execution progress.*/
void controller_set_monitor(CONTROLLER_MODEL *model, MONITOR_FN monitor_fn);

/** Register a function to catch every meassured time difference on specified channel.*/
void controller_set_mtd_monitor(CONTROLLER_MODEL *model, CONTROLLER_PTR_FN mtd_fn, void *arg, int channel);

/** Set the controller's local clock drift rate [sec/sec].*/
void controller_set_clock_drift(CONTROLLER_MODEL *model, double drift);

/*************************************************************************************/

#endif /* #ifndef _SP1CTRL_H */

/*@}*/
/*@}*/

