/**@addtogroup debug
 * @{*/

/**@addtogroup debug_int  Implementation 
 * @{ */

/**@file
 * @brief Provides configurable logging support.
 * @version 1.1
 * @date Last modified 2003-02-06 
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project \b FIT - Fault Injection for TTA, IST-1999-10748
 */


#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>

#include "csim/csim.h"
#include "ttpc/fl1debug.h"

/*************************************************************************************/
/* Preprocessor defines */


/** If nonzero, then all logging is sent to console.*/
#define ENABLE_CONSOLE        0

/** Defines the target file name for console output (possible are @c stdin or @c stderr).*/
#define CONSOLE_TARGET        stderr

/** The default name of log-file for non-console output.
It is possible to change the file name later using @c set_log_file_name() function.*/
#define LOG_FILE_NAME         "#ttpc.log"

/**@def open_log_file
Opens the log-file specified in @c log_file_name.
For console output only assigns the correct file descriptor.*/                              

/**@def close_log_file
Closes the log-file specified in @c log_file_name.
For console output only writes @c NULL to the variable and does nothing.*/ 

#if (ENABLE_CONSOLE) > 0
	#define open_log_file() { \
		log_file = CONSOLE_TARGET; \
		strcpy(log_file_name, #ENABLE_CONSOLE); \
	} while (0)

	#define close_log_file() log_file = NULL
#else
	#define open_log_file() do { \
		if (log_file == NULL) \
			log_file = fopen(log_file_name, "at"); \
		assert(log_file != NULL); \
	} while (0)

	#define close_log_file() do { \
		if (log_file != NULL) { \
			fclose(log_file); \
			log_file = NULL; \
		} \
	} while (0)
#endif


/*************************************************************************************/
/* Static Variables */

/** Current name of log-file.
Contains either a file name or the values @c "stdin" or @c "stderr".*/
static char log_file_name[1024] = LOG_FILE_NAME;

/** Log file descriptor. If the file is closed, this variable contains @c NULL.
For console output it points either to @c stdin or to @c stderr (These two special file
descriptors must NOT be closed). */
static FILE *log_file = NULL;

/** If nonzero, then the log-file is closed after every write.*/
static int close_after_write = 0;

/** The point (in model time) when logging is started.
Value of zero means, that logging starts immediately.
@see LOG_DURATION*/
static CSIM_TIME log_start = 0.0;

/** The duration (in model time) of logging process.
The logging will be enabled from @c LOG_START until the time specified here elapses.
A value of zero means that the logging will be performed infinitely.
@see LOG_START*/
static CSIM_TIME log_duration = 0.0;

/** Used when printing log info with current C-Sim state.*/
static char csim_output_format[64] = "%10.4f: %s: ";


/*************************************************************************************/
/* Exported Functions */


                        
/* Return the name of current log file.
 * @return  An internally allocated string. The string should be
 * 	considered read-only and never deallocated. */
char *get_log_file_name()
{
	return log_file_name;	
}
                         
                         
/**.
The old file is closed and left on disk and the new file will be opened by the next 
write operation in \em append mode. 
@param name New name of log-file.*/
void set_log_file_name(char *name)
{                                                           
	close_log_file();
	strncpy(log_file_name, name, sizeof(log_file_name));
}


/**.
Use this function before any abnormal termination of program (e.g. before @c assert 
statements) to ensure that no logged information is lost.*/
void log_flush(void)
{
	close_log_file();
}


/**.
No additional formating of the string is porformed, so the output is comletely user 
specified. Parameters to this function are the same as for standard @c printf function.
\attention Do not use directly, use @c log_xxx macros instead.*/
void print_logf(char *format, ...)
{
	va_list args;
	CSIM_TIME t;

	t = csim_time();
	if (t >= log_start) {
		if (log_duration <= 0 || t - log_start <= log_duration) {
			open_log_file();
			va_start(args, format);
			vfprintf(log_file, format, args);
			va_end(args);
			if (close_after_write) {
				close_log_file();
			}
		}
	}
}


/**.
Adds C-Sim related information to the output. Curently it writes the current simulation
time followed by currently active process name in front of the actual output.  Parameters 
to this function are the same as for standard @c printf function.
\attention Do not use directly, use @c log_xxx macros instead.*/
void print_logf_csim(char *format, ...)
{
	va_list args;
	CSIM_PROCESS *p = csim_current();
	CSIM_TIME t;

	t = csim_time();
	if (t >= log_start) {
		if (log_duration <= 0 || t - log_start <= log_duration) {
			open_log_file();
			fprintf(log_file, csim_output_format, csim_time(), (p) ? p->name : "NULL");
			va_start(args, format);
			vfprintf(log_file, format, args);
			va_end(args);
			if (close_after_write)
				close_log_file();
		}
	}
}


/**.
This function can force the log-output functions to close the log-file after every
write operation. This ensures, that no logging information is lost even when the
program crashes, but the closing and reopening of a file is very time-consuming.
Default setting is not to close the file everytime, it is closed when the program
exits cleanly or when the @c log_flush() or @c set_log_name() functions are called.
When this funtion is called with nonzero parameter, the log file is closed immediately.

@param enabled Set to a nonzero value to force closing of the log file everytime
  it is changed.*/
void set_log_close_after_write(int enabled)
{
	close_after_write = enabled;
	if (enabled)
		log_flush();
}


/**.
This functions is used to turn off all logging until the specified time is
reached.
@param time The minimum value of model time that is required to perform logging.
  The default value is '0.0'.
@see set_log_duration()*/
void set_log_start(CSIM_TIME time)
{
	log_start = time;
}


/**.
This functions is used to enable logging only for a limited time (the time is 
counted from the value set by @c set_log_start() ).
@param time A value of zero means that the logging will be performed infinitely 
  (default), any other value limits the logging duration.
@see set_log_start()*/
void set_log_duration(CSIM_TIME time)
{
	log_duration = time;
}


/**.
This function sets the output format for C-Sim based logging (i.e. log output that
is automatically prepended by current C-Sim state). The format string must contain
specify two values: a @c CSIM_TIME type and a string (for process name) in this order.
The default value for C-Sim output format is <code> "%10.4f: %s: "</code>.
Maximum length of the format string is 64 bytes including the 0 terminator.
@param fmt Format speicifier for as defined in the
  printf function. The string is copied into local memory.*/
void set_csim_output_format(char *fmt)
{
	strncpy(csim_output_format, fmt, sizeof(csim_output_format));
}


/**.*/
int get_log_close_after_write(void)
{
	return close_after_write;
}


/**.*/
CSIM_TIME get_log_start(void)
{
	return log_start;
}


/**.*/
CSIM_TIME get_log_duration(void) 
{
	return log_duration;
}


/**.*/
char *get_csim_output_format(void)
{
	return csim_output_format;
}


/*************************************************************************************/

/*@}*/
/*@}*/

