/**@addtogroup bg_fun
 * @{*/

/**@addtogroup bg_fun_int  Implementation
 * @{*/
 

/**@file
 * @brief Bus Guardian functionality and data structures (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 *
 * @note This module is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#include "ttpc/fl1func.h"
#include "ttpc/fp1busg.h"
#include "ttpc/fl1debug.h"
#include "ttpc/fp1ctrl.h"


/************************************************************************************/

/** A shortcut that makes the logging process throughout the whole project 
 * transparent and unified.
 *
 * @param msg  This argument is special. It should contain the format specifier as in 
 * 	a call to printf function and optional arguments, but the whole set of 
 * 	parameters must be enclosed in one additional pair of parenthesis.
 * 	
 * An example how to call such macro:
 * @code 
 * 	log(( "This is testing entry number %d for this log (made by %s)\n", 1, "me" ));
 * @endcode */
#define log(msg) log_bg_csim(msg)


/************************************************************************************/
/* Exported Generated Functions */
/************************************************************************************/

implement_access_functions(BUS_GUARDIAN, bus_guardian, tap, 16)
implement_access_functions(BUS_GUARDIAN, bus_guardian, tso, 16)
implement_access_functions(BUS_GUARDIAN, bus_guardian, tsd, 16)
implement_access_functions(BUS_GUARDIAN, bus_guardian, trr, 16)
implement_access_functions(BUS_GUARDIAN, bus_guardian, cmod, 2)
implement_access_functions(BUS_GUARDIAN, bus_guardian, error_code, 3)
implement_access_functions(BUS_GUARDIAN, bus_guardian, state_code, 4)
implement_access_functions(BUS_GUARDIAN, bus_guardian, lf, 1)
implement_access_functions(BUS_GUARDIAN, bus_guardian, er, 1)
get_function_body(BUS_GUARDIAN, bus_guardian, ht, 1)
get_function_body(BUS_GUARDIAN, bus_guardian, rc, 1)
get_function_body(BUS_GUARDIAN, bus_guardian, am, 1)

	
/************************************************************************************/
/* Exported Functions */
/************************************************************************************/

/* Switch the BG into node-failure state and set the correct error code.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param error_code This value will be written into the @c error_code field of BG. */
void bus_guardian_raise_error(BUS_GUARDIAN *bg, unsigned error_code)
{
	bus_guardian_set_state_code(bg, bg->bg_state);
	bus_guardian_set_error_code(bg, error_code);
	bg->am = 0;
	bg->lf = 0;
	bg->er = 1;
	bg->bg_state = BGS_NODE_FAILURE;
#ifndef DISABLE_BG
	controller_bg_error_callback(bg->controller);
#endif
}


/* Check whether transmission on the specified channel is enabled (1) or disabled (0).
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param channel The number of a channel - either 0 (channel A) or 1 (channel B).
 * @return Returns 1 if the channel is open or 0 if it is closed. */
int bus_guardian_is_enabled(BUS_GUARDIAN *bg, unsigned channel)
{
#ifndef DISABLE_BG
	unsigned ch_bit = 1u << channel;
	return bus_guardian_get_lf(bg) && (ch_bit & bg->saved_cmod) == ch_bit;
#else
	return 1;
#endif
}


/* Set the initial power-on values of BG - called after creation of BG.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.*/
void bus_guardian_init(BUS_GUARDIAN *bg)
{
	log(("Initializing BG\n"));

	bg->bg_state = BGS_INITIALIZE;
	bg->tap = 0;
	bg->tso = 0;
	bg->tsd = 0;
	bg->trr = 0;
	bg->am = 0;
	bg->ht = 0;
	bg->rc = 1;
	bg->cmod = 0;
	bg->saved_cmod = 0;
	bg->lf = 0;
	bg->er = 0;
	bg->error_code = BGEC_NO_ERROR;
	bg->state_code = BGS_INITIALIZE;
}


/* Perform a reconfiguration cycle for BG and writes a new paramet set.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param tap  bus arm phase duration.
 * @param tso  slot open phase duration.
 * @param tsd  slot disabled phase duration.
 * @param trr  remaining round phase duration. */
void bus_guardian_set_params_ext(BUS_GUARDIAN *bg, unsigned tap, unsigned tso,
		unsigned tsd, unsigned trr)
{
	log(("Setting new BG parameters\n"));
	bus_guardian_init(bg);
	bus_guardian_set_rc(bg, 1);
	bus_guardian_set_tap(bg, tap);
	bus_guardian_set_tso(bg, tso);
	bus_guardian_set_tsd(bg, tsd);
	bus_guardian_set_trr(bg, trr);
	bus_guardian_set_cmod(bg, 3);
	bus_guardian_set_rc(bg, 0);
}


/* Perform a reconfiguration cycle for BG and writes a new paramet set 
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param params Pointer to a structure that contains the new parameters. The parameters
 * 	are expected in a format that is used within the MEDL table. */
void bus_guardian_set_params(BUS_GUARDIAN *bg, MEDL_BGP *params)
{
	bus_guardian_set_params_ext(bg,
			medl_bgp_get_bg_arm_phase(params),
			medl_bgp_get_bg_slot_offset(params),
			medl_bgp_get_bg_slot_duration(params),
			medl_bgp_get_bg_remaining_round(params)
			);
}


/* Perform all actions appropiate for BG init state.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_init_state(BUS_GUARDIAN *bg)
{
	/* Set power-on values. */
	bus_guardian_init(bg);
	/* Transit to configure state. */
	bg->bg_state = BGS_CONFIGURE;
}


/* Perform all actions appripiate for BG configure state.
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_configure_state(BUS_GUARDIAN *bg)
{
	/* Wait until the rc bit is reset. */
	if (bus_guardian_get_rc(bg) == 0)
		bg->bg_state = BGS_ARM;
}


/* Perform all actions appropiate for BG arm state.  This function is called
 * when the arm phase of BG ends and the unit should change its state. It
 * checks for a possible reconfiguration request and monitors the arm signal.
 * If an arm signal occurs, a new bus access cycle is started.
 * 
 * @note Modification of TTP/C C1 specification: when no arm signal is received
 * during the bus arm phase, an error is raised.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_arm_state(BUS_GUARDIAN *bg)
{
	if (bus_guardian_get_rc(bg) == 1) {
		/* reconfiguration request */
		log(("Reconfiguration request granted\n"));
		bg->bg_state = BGS_CONFIGURE;
	} else if (bus_guardian_get_am(bg) == 1) {
		/* bus arm triggered */
		log(("Arm signal correctly received\n"));
		bg->bg_state = BGS_START_TIMER;
		bg->saved_cmod = bus_guardian_get_cmod(bg);
	}
#ifdef ENABLE_BG_MOD
	else {
		/* controller failure - no arm signal */
		log_critical_csim(("No Arm signal occured during the BUS_ARM phase\n"));
		bus_guardian_raise_error(bg, BGEC_MISSING_ARM);
	}
#else
	/* Wait for the arm signal. */
#endif
}


/* Perform all actions appropiate for BG start_timer state.  This function is
 * called when the BG transits from  start_timer into enable state.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_start_timer_state(BUS_GUARDIAN *bg)
{
	/* Wait the slot-offset-duration time and transit to enable state. */
	bg->bg_state = BGS_ENABLE;
	bus_guardian_set_lf(bg, 1);
}


/* Perform all actions appropiate for BG enable state.  This function is
 * called when the BG transits from enable into disable state. If the arm
 * signal is still active at this point, an error is raised.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure. */
void bus_guardian_enable_state(BUS_GUARDIAN *bg)
{
	bg->bg_state = BGS_DISABLE;
	bus_guardian_set_lf(bg, 0);
	if (bus_guardian_get_am(bg) != 0) {
		log(("BG-ARM signal is active until the DISABLE phase\n")); 
		bus_guardian_raise_error(bg, BGEC_UNKNOWN_ERROR);
	}
}


/* Perform all actions appropiate for BG disable state.  This function is
 * called when the BG transits from \e disable into arm state.
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.*/
void bus_guardian_disable_state(BUS_GUARDIAN *bg)
{
	/* Wait the remaining-round-duration time and transit back to arm state. */
	bg->bg_state = BGS_ARM;
}


/************************************************************************************/
/* Special field access functions */

/** Change the value of @c ht flag (halt the bus guardian).
 *
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param value The desired value of the @c ht flag.
 * @return The new value of the @c ht flag. */
unsigned bus_guardian_set_ht(BUS_GUARDIAN *bg, unsigned value)
{
	value &= bit_mask(1);
	bg->ht = value;
	if (value != 0) {
		log(("BG-Halt Signal received\n"));
		bus_guardian_raise_error(bg, BGEC_HALT);
	}
	return value;
}


/** Change the value of @c rc flag (start/stop the reconfiguration mode).
 * 
 * @param bg Pointer to a BUS_GUARDIAN or derived structure.
 * @param value The desired value of the @c rc flag.
 * @return The new value of the @c rc flag. */
unsigned bus_guardian_set_rc(BUS_GUARDIAN *bg, unsigned value)
{
	if (bg->rc == value)
		return bg->rc;
	bg->rc = value & bit_mask(1);

	if (bg->rc == 1) {
		bus_guardian_set_lf(bg, 0);
	}
	else
		bg->bg_state = BGS_ARM;

	return bg->rc;
}


/** Change value if the @c am flag (start/stop arm signal).
 * The function attempts to set/clear the @c am flag of BG. If a protocol violation
 * is detected, nonzero code is returned and the node_failure state is entered.
 *  
 * @param bg Pointer to a @c BUS_GUARDIAN or derived structure.
 * @param value The desired value of the @c am flag.
 * @return Returns zero code if no error occurs, nonzero code otherwise:
 * 	- @c 1 = Transmission disabled.
 * 	- @c 2 = Arm during bus access cycle.
 * 	- @c 3 = Arm signal lost during other state than enabled. */
unsigned bus_guardian_set_am(BUS_GUARDIAN *bg, unsigned value)
{
	if (value == bg->am)
		return value;
	bg->am = value & bit_mask(1);

	if (bg->am == 1) {
		if (bus_guardian_get_cmod(bg) == 0) {
			log(("BG received ARM signal while transmission was disabled\n"));
			bus_guardian_raise_error(bg, BGEC_ARM_DURING_TRANSMISSION_DISABLED);
			return 1;
		}
		if (bg->bg_state != BGS_ARM) {
			log(("BG received ARM signal during bus access cycle\n"));
			bus_guardian_raise_error(bg, BGEC_ARM_DURING_BUS_ACCESS_CYCLE);
			return 2;
		}
	}
	else if (bg->bg_state != BGS_ENABLE) {
		log(("ARM signal lost during other state then ENABLED\n"));
		bus_guardian_raise_error(bg, BGEC_UNKNOWN_ERROR);
		return 3;
	}
		
	return 0;
}

/************************************************************************************/

/*@}*/
/*@}*/

