/**@addtogroup chan_fun
 * @{*/

/**@addtogroup chan_fun_int  Implementation
 * @{*/

/**@file
 * @brief Transmission channel timeless functionality (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic 
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 * 
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, created as
 *  part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */


#include <string.h>

#include "ttpc/fl1debug.h"
#include "ttpc/fp1chan.h"
#include "ttpc/fp1busg.h"
#include "ttpc/sp1ctrl.h"
#include "ttpc/fl1func.h"


/*************************************************************************************/

/** A shortcut that makes the logging process throughout the whole project transparent 
 * and unified.
 *
 * @param msg This argument is special. It should contain the format specifier as in
 *  a call to printf() function and optional arguments, but the whole set of 
 *  parameters must be enclosed in one additional pair of parenthesis. */
#define log(msg) log_channel_csim(msg)


/** Print logging information regarding a transmission collision.
 *
 * @param channel  Pointer to the channel that observed the collision.
 * @param ctrl  Pointer to the controller that caused the collision.
 * @param msg  Message printed before oher details of the event. */
static void log_collision(CHANNEL *channel, void *ctrl, char *msg)
{
#define controller  ((CONTROLLER *)ctrl)
#define bg  ( (BUS_GUARDIAN *)controller->bus_guardian )
	
	log_critical_csim(("%s on channel %d, slot %d\n",
				msg,
				channel->bus_id, 
				logical_name_get_slot_position(
					&controller->medl.role_list.logical_name)
				));
	log_critical_csim(("BG params: tap=%u, tso=%u, tsd=%u, trr=%u\n",
				bg->tap, bg->tso, bg->tsd, bg->trr));
#undef bg
#undef controller
}


/*************************************************************************************/
/* Exported Generated Functions */
/*************************************************************************************/

implement_access_functions(CHANNEL, channel, bus_id, 16);
implement_access_functions(CHANNEL, channel, active, 1);

/*************************************************************************************/
/* Exported Functions */
/*************************************************************************************/

/* Set the initial power-on values of @c CHANNEL - called after creation.
 *
 * @param channel Pointer to a @c CHANNEL instance or a derived structure.
 * @return Return the pointer passed in @a channel. */
CHANNEL *channel_init(CHANNEL *channel)
{
	memset(&channel->frame, 0, sizeof(FRAME));
	channel->frame.state = FRAME_STATE_NULL;
	channel->active = 0;
	channel->receiver_count = 0;
	channel->bus_id = 0;
	channel->sender = NULL;
	channel->bg_error_callback = NULL;
	channel->send_callback = NULL;
	channel->channel_send_frame = NULL;
	channel->transmission_callback = NULL;
	channel->collisions = 0;

	return channel;
}


/* Registers the caller to receive the next sent frame.
 * 
 * The caller specifies two arguments with this funtion, that identify him:
 * @a ctrl and \a callback. These parameters are stored and when a frame is received,
 * the folowing function call is made:\code
 * callback(ctrl, channel->id, transmission_state, channel->frame);\endcode
 * If the receiver is already registered, the function returns with error code.
 * A correct registration is valid until a frame is completely transmitted on the channel.
 * 
 * @param channel Pointer to a channel that receives the registration.
 * @param ctrl User defined argument that is stored and later passed to the @a callback 
 * 	function.
 * @param callback Pointer to a function that should be called after a frame is received.
 * @return Returns 0 on success, nonzero otherwise. 
 *
 * @see channel_cancel_receive */
int channel_receive(CHANNEL *channel, void *ctrl, CHANNEL_ON_RCV_FN callback)
{
	FRAME *frame = ((CONTROLLER *)ctrl)->ploc_data.frame;

	log(("Registering %d. receiver: %s\n",
		channel->receiver_count + 1, ((CONTROLLER_MODEL *)ctrl)->process->name));
	log_assert(channel->receiver_count < MAX_SRU_COUNT);

	frame[channel->bus_id] = channel->frame;

	if (!channel->active) {
		channel->receiver[channel->receiver_count] = ctrl;
		channel->receive_callback[channel->receiver_count] = callback;
		channel->receiver_count++;
		frame[channel->bus_id].state = FRAME_STATE_NULL;
	} else
		frame[channel->bus_id].state = FRAME_STATE_RECEIVING;

	return channel->active;
}


/* Unregisters a controller so it no more listens on the channel.
 *
 * @param channel Pointer to a channel that should cancel the registration.
 * @param ctrl This argument is used when searching for the registration to cancel.
 *   All registrations with this user-argument are canceled.
 *
 * @see channel_receive */
void channel_cancel_receive(CHANNEL *channel, void *ctrl)
{
	unsigned i = 0;

	while (i < channel->receiver_count && channel->receiver[i] != ctrl)
		i++;

	if (i < channel->receiver_count) {
		log(("Removing controller from receive queue.\n"));
		channel->receiver_count--;
		while (i < channel->receiver_count) {
			channel->receiver[i] = channel->receiver[i + 1];
			channel->receive_callback[i] = channel->receive_callback[i + 1];
			i++;
		}
	}
}


/* Initializes the sending of a frame - register the sender for notification.
 *
 * This function assumes that the sender is either @c CONTROLLER or a
 * derived structure, because the frame is copied directly from controller's
 * local data structures.  The sender (@a controller argument) is stored and so
 * is the callback function the is provided - after the send operation
 * finished, these stored values are used to nitify the sender. The sending
 * controller is excluded from receiving the frame by explicitly removing him
 * from registered receivers.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param controller Pointer to a @c CONTROLLER or derived structure. This
 *  value is passed to the callback function later.
 * @param callback Function that is used to notify the controller.
 * @return Returns 0 if no error occurs, nonzero code otherwise.
 * 
 * @see channel_send_complete, CHANNEL_SEND_FRAME_FN */
int channel_send_init(CHANNEL *channel, void *controller, PTR_CALLBACK callback)
{
	unsigned id = channel->bus_id;
	FRAME *frame = &((CONTROLLER *)controller)->ploc_data.frame[id];
	unsigned i;

	if (channel_test_bg(channel, controller))
		return 1;	/* BG Closed */

	if (channel->active) {
		channel->frame.noise = 1;
		if (channel->sender == controller) {
			log(("Controller restarted sending, hmm, making noise\n"));
		} else {
			channel->collisions++;
			log_collision(channel, channel->sender, "Interrupted transmission");
			log_collision(channel, controller, "Offender");
		}
		return 2;	/* Silence Violation */
	}

	channel->active = 1;
	channel->sender = controller;
	channel->send_callback = callback;
	frame->noise = 0;
	frame->state = FRAME_STATE_COMPLETE;

	/* Handle sender separately */
	channel_cancel_receive(channel, controller);

	/* Store frame to transmitt */
	channel->frame = *frame;

	/* Notify monitor */
	if (channel->transmission_callback)
		channel->transmission_callback(channel->transmission_callback_arg, 
				&channel->frame, CHANNEL_TS_START);

	/* Notify all receivers */
	for (i = 0; i < channel->receiver_count; i++) {
		if (channel->receive_callback[i])
			channel->receive_callback[i](channel->receiver[i], 
					channel->bus_id, CHANNEL_TS_START, &channel->frame);
	}

	return 0;
}


/* Finishes the sending of a frame - notifies all receivers and cleans up the channel. 
 *
 * Sets Frame State to COMPLETE if there was no noise while transmiting and
 * sends notifications to all receivers and also to the sender. At last clears
 * all transmission fields. 
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 *
 * @see channel_send_init, CHANNEL_SEND_FRAME_FN */
void channel_send_complete(CHANNEL *channel)
{
	unsigned i;

	/* Notify monitor */
	if (channel->transmission_callback)
		channel->transmission_callback(channel->transmission_callback_arg, &channel->frame, CHANNEL_TS_FINISH);
	
	if (channel->frame.noise) {
		log(("Noise on the channel.\n"));
		channel->frame.state = FRAME_STATE_RECEIVING;
	}

	log(("Receiver count %d.\n", channel->receiver_count));
	for (i = 0; i < channel->receiver_count; i++)
		channel->receive_callback[i](channel->receiver[i], channel->bus_id, CHANNEL_TS_FINISH, &channel->frame);
	channel->receiver_count = 0;

	if (channel->send_callback != NULL)
		channel->send_callback(channel->sender);

	channel->active = 0;
	channel->frame.noise = 0;
	channel->frame.state = FRAME_STATE_NULL;
}


/* Registers the caller to receive notifications about transmitted frame - it
 * is possible to modify such frames before sending. 
 *
 * This function enables the caller to monitor all traffic on the channel and
 * even to modify the frames before transmission. So far it is not easily
 * possible to selectively modify frames for different receivers.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param callback Function that is used to notify the registered controller.
 * @param arg User defined argument that is stored and later passed to the @a
 *  callback function. */
void channel_set_transmission_callback(CHANNEL *channel, CHANNEL_TRANSMISSION_FN callback, void *arg)
{
	channel->transmission_callback = callback;
	channel->transmission_callback_arg = arg;
}


/* Registers an error handling function for BG reported failures. 
 *
 * This function gives an oportunity to catch BG errors immediately as they
 * occur even at higher levels. So far there is no direct reporting from BG to
 * controller, so this is necessary.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 *
 * @param callback The callback function that will be called when any error
 * from the BG is detected. The @a channel argument will be passed to the
 * function. */
void channel_set_bg_error_callback(CHANNEL *channel, PTR_CALLBACK callback)
{
	channel->bg_error_callback = callback;
}


/* Check BG: if it is open return 0, raise an error otherwise and return
 * nonzero code.
 * 
 * Because the BG is currently integrated into the controller, this function
 * has to look into the @a controller structure to find the appropiate BG
 * instance.
 *
 * @param channel Pointer to a @c CHANNEL or derived structure.
 * @param controller Pointer to a @c CONTROLLER or derived structure.
 * @return Returns 0 if BG is open, nonzero code otherwise. */
int channel_test_bg(CHANNEL *channel, void *controller)
{
#ifndef DISABLE_BG
	if (!bus_guardian_is_enabled((BUS_GUARDIAN *)((CONTROLLER *)controller)->bus_guardian, channel->bus_id)) {
		if (channel->bg_error_callback != NULL)
			channel->bg_error_callback(controller);
		return 1;
	}
#endif
	return 0;
}


/*************************************************************************************/

/*@}*/
/*@}*/

