/**@addtogroup cni_fun
 * @{*/

/**@addtogroup cni_fun_int  Implementation
 * @{*/


/**@file
 * @brief CNI data structure and timeless functionality (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project @b FIT - Fault Injection for TTA, IST-1999-10748
 */

#include <string.h>

#include "ttpc/fl1debug.h"
#include "ttpc/fp1cni.h"
#include "ttpc/fl1func.h"

/***************************************************************************/

/* Fill the argument with default CNI status area fault mask.
 * DOES NOT clear the mask first. */
static void cni_status_fault_mask(unsigned* mask)
{
	CNI *p = (CNI*)mask;
	int i;

	/* Status/Control Area CCF Field: */
	p->ccf = bit_mask(2);
	/* C-State Time Field: */
	c_state_fault_mask((unsigned*)&p->c_state);
	/* Null Frame Vector: */
	for (i = 0; i <= MAX_SRU_COUNT; i++)
		p->null_frame[i] = 1;
	/* Protocol Status Field: */
	p->protocol_status.accept_counter = bit_mask(8);
	p->protocol_status.failed_counter = bit_mask(8);
	p->protocol_status.invalid_counter = bit_mask(8);
	p->protocol_status.controller_ready = bit_mask(8);
	p->protocol_status.mfc = bit_mask(8);
	p->protocol_status.i_frame_switch = bit_mask(8);	
	p->protocol_status.csct = bit_mask(8);
	p->protocol_status.c_state_valid = bit_mask(8);
	p->protocol_status.protocol_state = bit_mask(4);
	p->protocol_status.bl = 1;
	p->protocol_status.be = 1;
	p->protocol_status.ae = 1;
	p->protocol_status.se = 1;
	p->protocol_status.me = 1;
	p->protocol_status.mc = 1;
	p->protocol_status.cb = 1;
	p->protocol_status.dc = 1;
	p->protocol_status.so = 1;
	p->protocol_status.mv = 1;
	p->protocol_status.ne = 1;
	/* Frame Diagnosis Field: */
	p->frame_diagnosis.eif_cha = bit_mask(4);
	p->frame_diagnosis.eif_chb = bit_mask(4);
	p->frame_diagnosis.mtd_cha = bit_mask(8);
	p->frame_diagnosis.mtd_chb = bit_mask(8);
	p->frame_diagnosis.header_cha.i_n_bit = 1;
	for (i = 0; i < 3; i++) {
		p->frame_diagnosis.header_cha.mode_change[i] = 1;
		p->frame_diagnosis.header_chb.mode_change[i] = 1;
	}
	/* Personality Field: */
	p->personality.physical_sru_name = bit_mask(8);
	p->personality.controller_version = bit_mask(8);
	p->personality.app_id = bit_mask(8);
	p->personality.app_version = bit_mask(8);
	logical_name_fault_mask((unsigned*)&p->personality.logical_name);
	/* other fields */
	p->controller_lifesign = bit_mask(16);
	p->assertion_code = bit_mask(16);
	for (i = 0; i < 8; i++) 
	        p->reserved1[i] = bit_mask(16);
	p->cluster_time = bit_mask(16);
	/* BIST Status Field: */
	p->bist_status.sc = 1;
	p->bist_status.av = 1;
	p->bist_status.br = 1;
	p->bist_status.rm = 1;
	p->bist_status.rf = 1;
	p->bist_status.wd = 1;
}

/* Fill the argument with default CNI control area fault mask.
 * DOES NOT clear the mask first. */
static void cni_control_fault_mask(unsigned* mask)
{
	int i;
	CNI* p = (CNI*)mask;

	p->host_lifesign = bit_mask(16);
	p->external_rate_correction = bit_mask(16);
	for (i = 0; i < 3; i++)
		p->mode_change_request[i] = 1;
	p->ca = bit_mask(16);
	/* BIST Enable Field: */
	p->bist_enable.rm = 1;
	p->bist_enable.rf = 1;
	/* other fields */
	p->timer1 = bit_mask(16);
	p->timer2 = bit_mask(16);
	/* Interrupt Enable Field: */
	p->interrupt_enable.be = 1;
	p->interrupt_enable.pe = 1;
	p->interrupt_enable.he = 1;
	p->interrupt_enable.ml = 1;
	p->interrupt_enable.mc = 1;
	p->interrupt_enable.mi = 1;
	p->interrupt_enable.ui1 = 1;
	p->interrupt_enable.ui2 = 1;
	p->interrupt_enable.ti1 = 1;
	p->interrupt_enable.ti2 = 1;
	/* Interrupt Status Field: */
	p->interrupt_status.be = 1;
	p->interrupt_status.pe = 1;
	p->interrupt_status.he = 1;
	p->interrupt_status.ml = 1;
	p->interrupt_status.mc = 1;
	p->interrupt_status.mi = 1;
	p->interrupt_status.ui1 = 1;
	p->interrupt_status.ui2 = 1;
	p->interrupt_status.ti1 = 1;
	p->interrupt_status.ti2 = 1;
	/* other fields */
	p->co = bit_mask(16);
	p->time_overflow_counter = bit_mask(16);
}

/* Fill the argument with default CNI message area fault mask.
 * DOES NOT clear the mask first. */
static void cni_message_fault_mask(unsigned* mask)
{
	CNI* p = (CNI*)mask;
	memset(p->cni_msg_base, 0xFF, CNI_MSG_AREA_SIZE);
}


/************************************************************************************/

/**@name Generated Access Functions for CNI_PS
 * 
 * Access functions for @c CNI_PS that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @todo cni_ps_get function may be buggy: test last_clock_access
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body 
 * @{ */
implement_access_functions(CNI_PS, cni_ps, accept_counter, 8)
implement_access_functions(CNI_PS, cni_ps, failed_counter, 8)
implement_access_functions(CNI_PS, cni_ps, invalid_counter, 8)
implement_access_functions(CNI_PS, cni_ps, controller_ready, 8)
implement_access_functions(CNI_PS, cni_ps, mfc, 8)
implement_access_functions(CNI_PS, cni_ps, i_frame_switch, 8)
implement_sig_access_functions(CNI_PS, cni_ps, csct, 8)
implement_access_functions(CNI_PS, cni_ps, c_state_valid, 8)
implement_access_functions(CNI_PS, cni_ps, protocol_state, 4)
implement_access_functions(CNI_PS, cni_ps, bl, 1)
implement_access_functions(CNI_PS, cni_ps, be, 1)
implement_access_functions(CNI_PS, cni_ps, ae, 1)
implement_access_functions(CNI_PS, cni_ps, se, 1)
implement_access_functions(CNI_PS, cni_ps, me, 1)
implement_access_functions(CNI_PS, cni_ps, mc, 1)
implement_access_functions(CNI_PS, cni_ps, cb, 1)
implement_access_functions(CNI_PS, cni_ps, dc, 1)
implement_access_functions(CNI_PS, cni_ps, so, 1)
implement_access_functions(CNI_PS, cni_ps, mv, 1)
implement_access_functions(CNI_PS, cni_ps, ne, 1)
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_FD
 * 
 * Access functions for @c CNI_FD that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 * 
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body 
 * @{ */
implement_access_functions(CNI_FD, cni_fd, eif_cha, 4)
implement_access_functions(CNI_FD, cni_fd, eif_chb, 4)
implement_sig_access_functions(CNI_FD, cni_fd, mtd_cha, 8)
implement_sig_access_functions(CNI_FD, cni_fd, mtd_chb, 8)
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_PERS
 * 
 * Access functions for @c CNI_PERS that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body
 * @{ */
implement_access_functions(CNI_PERS, cni_pers, physical_sru_name, 8)
implement_access_functions(CNI_PERS, cni_pers, controller_version, 8)
implement_access_functions(CNI_PERS, cni_pers, app_id, 8)
implement_access_functions(CNI_PERS, cni_pers, app_version, 8)
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_IS
 *
 * Access functions for @c CNI_IS that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body
 * @{ */
implement_access_functions(CNI_IS, cni_is, be, 1);
implement_access_functions(CNI_IS, cni_is, pe, 1);
implement_access_functions(CNI_IS, cni_is, he, 1);
implement_access_functions(CNI_IS, cni_is, ml, 1);
implement_access_functions(CNI_IS, cni_is, mc, 1);
implement_access_functions(CNI_IS, cni_is, mi, 1);
implement_access_functions(CNI_IS, cni_is, ui1, 1);
implement_access_functions(CNI_IS, cni_is, ui2, 1);
implement_access_functions(CNI_IS, cni_is, ti1, 1);
implement_access_functions(CNI_IS, cni_is, ti2, 1);
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_IE
 * 
 * Access functions for @c CNI_IE that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body 
 * @{ */
implement_access_functions(CNI_IE, cni_ie, be, 1);
implement_access_functions(CNI_IE, cni_ie, pe, 1);
implement_access_functions(CNI_IE, cni_ie, he, 1);
implement_access_functions(CNI_IE, cni_ie, ml, 1);
implement_access_functions(CNI_IE, cni_ie, mc, 1);
implement_access_functions(CNI_IE, cni_ie, mi, 1);
implement_access_functions(CNI_IE, cni_ie, ui1, 1);
implement_access_functions(CNI_IE, cni_ie, ui2, 1);
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_BS
 *
 * Access functions for @c CNI_BS that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body
 * @{ */
implement_access_functions(CNI_BS, cni_bs, sc, 1);
implement_access_functions(CNI_BS, cni_bs, av, 1);
implement_access_functions(CNI_BS, cni_bs, br, 1);
implement_access_functions(CNI_BS, cni_bs, rm, 1);
implement_access_functions(CNI_BS, cni_bs, rf, 1);
implement_access_functions(CNI_BS, cni_bs, wd, 1);
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_BE
 *
 * Access functions for @c CNI_BE that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 * 
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body
 * @{ */
implement_access_functions(CNI_BE, cni_be, rm, 1);
implement_access_functions(CNI_BE, cni_be, rf, 1);
/*@}*/

/**********************************************************************************/

/**@name Generated Access Functions for CNI_MAE
 *
 * Access functions for @c CNI_MAE that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body
 * @{ */
implement_access_functions(CNI_MAE, cni_mae, ccf, 4);
implement_access_functions(CNI_MAE, cni_mae, eif, 4);
implement_index_access_functions(CNI_MAE, cni_mae, data, 8);
/*@}*/

/**********************************************************************************/
/**@name Generated Access Functions for CNI
 *
 * Access functions for @c CNI that were generated by preprocessor macros.
 * All of these functions access the structure passed to them in the @a _structure
 * argument and either write a new value (@a _value argument) or return the current
 * content of a particular field. For access to vectors an index must be also passed
 * using the @a _index argument.
 *
 * @see implement_access_functions, get_function_body, set_function_body, 
 *  increment_function_body
 * @{ */
implement_access_functions(CNI, cni, ccf, 16)
implement_index_access_functions(CNI, cni, null_frame, 1)
implement_access_functions(CNI, cni, controller_lifesign, 16)
implement_access_functions(CNI, cni, assertion_code, 16)
implement_access_functions(CNI, cni, host_lifesign, 16)
implement_index_access_functions(CNI, cni, mode_change_request, 1)
implement_access_functions(CNI, cni, time_overflow_counter, 16)
increment_sig_function_body(CNI, cni, external_rate_correction, 16)
get_function_body(CNI, cni, timer1, 16)
increment_function_body(CNI, cni, timer1, 16)
get_function_body(CNI, cni, timer2, 16)
increment_function_body(CNI, cni, timer2, 16)
get_function_body(CNI_IE, cni_ie, ti1, 1)
get_function_body(CNI_IE, cni_ie, ti2, 1)
set_function_body(CNI, cni, cluster_time, 16)
increment_function_body(CNI, cni, cluster_time, 16)
get_function_body(CNI, cni, co, 16)
increment_function_body(CNI, cni, co, 16)
get_function_body(CNI, cni, ca, 16)
increment_function_body(CNI, cni, ca, 16)
/*@}*/
	
/***********************************************************************************/
/***********************************************************************************/
/**@name Mode Change Request Handling
 * @{ */

/*.
 * @param cni Pointer to a CNI instance.
 * @return The mode number extracted from MCR fields. */
unsigned cni_mcr_get_mode(CNI *cni)
{
	return (cni_get_mode_change_request(cni, 2) << 1)
			+ cni_get_mode_change_request(cni, 1);
}


/*.
 * @param cni Pointer to a CNI instance.
 * @param mode Number of the requested operation mode. */
void cni_mcr_set_mode(CNI *cni, unsigned mode)
{
	cni_set_mode_change_request(cni, 1, mode & 1);
	cni_set_mode_change_request(cni, 2, (mode >> 1) & 1);
}


/*.
@param cni Pointer to a CNI instance.
@return Return nonzero code if an immediate mode change is requesed.*/
unsigned cni_mcr_get_immediate(CNI *cni)
{
	return cni_get_mode_change_request(cni, 0);
}


/*.
@param cni Pointer to a CNI instance.
@param imm If this argument is non-zero, then an immediate mode change is requested,
	otherwise a deffered MCR is set.*/
void cni_mcr_set_immediate(CNI *cni, unsigned imm)
{
	cni_set_mode_change_request(cni, 0, imm);
}


/*.
@param cni Pointer to a CNI instance.*/
void cni_mcr_clear(CNI *cni)
{
	cni_set_mode_change_request(cni, 0, 0);
	cni_set_mode_change_request(cni, 1, 0);
	cni_set_mode_change_request(cni, 2, 0);
}

/*@}*/
/***********************************************************************************/

/***********************************************************************************/
/**@name Virtual Functions Pointers
 * 
 * These variables contain pointers to functions that provide extended access
 * to some fields of the @c CNI structure. Some of them are defined implicitly,
 * but the implicit definition may not be sufficient for you. In such case you
 * may write a new version and assign these variables to point to your version.
 * It is possible to still use the original access functions - if you store the
 * old value.
 * @{ */
CNI_GET_FN cni_get_cluster_time;
CNI_SET_FN cni_ie_set_ti1;
CNI_SET_FN cni_ie_set_ti2;
CNI_SET_FN cni_set_timer1;
CNI_SET_FN cni_set_timer2;
CNI_SET_FN cni_set_co;
CNI_SET_FN cni_set_ca;

/*@}*/
/***********************************************************************************/

/*Exported Functions*/

/* Return pointer to default CNI fault mask (static allocation).
 * Individual CNI areas can be disabled/enabled using arguments. */
unsigned* cni_fault_mask(int status, int control, int message)
{
    static CNI mask;
    
    memset(&mask, 0, sizeof(CNI));
    if (status)
        cni_status_fault_mask((unsigned*)&mask);
    if (control)
        cni_control_fault_mask((unsigned*)&mask);
    if (message)
        cni_message_fault_mask((unsigned*)&mask);
    return (unsigned*)&mask;
}


/**. 
@param cni Pointer to a @c CNI instance or a derived structure.
@param offset Word offset within the message area (must point exactly at the start of
	some message, otherwise the returned value will not make sense).
@return Returns a pointer to @c CNI_MAE structure that points to the same location
	as @a offset value.*/
CNI_MAE *cni_get_message_entry(CNI *cni, unsigned offset)
{
	return (CNI_MAE *)( (char *)(&cni->cni_msg_base[0]) + 2 * offset);
}


/**.
The @c co field is set to \b 0xFFFF, i.e. the controller is switched on.
@param cni Pointer to an instance of @c CNI or a derive structure.*/
void cni_clear_mem(CNI *cni)
{
	memset(&cni->ccf, 0, sizeof(cni->ccf));
	memset(&cni->c_state, 0, sizeof(cni->c_state));
	memset(&cni->protocol_status, 0, sizeof(cni->protocol_status));
	memset(&cni->frame_diagnosis, 0, sizeof(cni->frame_diagnosis));
	memset(&cni->personality, 0, sizeof(cni->personality));
	memset(&cni->controller_lifesign, 0, sizeof(cni->controller_lifesign));
	memset(&cni->assertion_code, 0, sizeof(cni->assertion_code));
	memset(&cni->cluster_time, 0, sizeof(cni->cluster_time));
	memset(&cni->interrupt_status, 0, sizeof(cni->interrupt_status));
	memset(&cni->bist_status, 0, sizeof(cni->bist_status));
	memset(&cni->host_lifesign, 0, sizeof(cni->host_lifesign));
	memset(&cni->external_rate_correction, 0, sizeof(cni->external_rate_correction));
	memset(&cni->mode_change_request[0], 0, sizeof(cni->mode_change_request));
	memset(&cni->ca, 0, sizeof(cni->ca));
	memset(&cni->bist_enable, 0, sizeof(cni->bist_enable));
	memset(&cni->timer1, 0, sizeof(cni->timer1));
	memset(&cni->timer2, 0, sizeof(cni->timer2));
	memset(&cni->interrupt_enable, 0, sizeof(cni->interrupt_enable));
	memset(&cni->time_overflow_counter, 0, sizeof(cni->time_overflow_counter));
	memset(&cni->cni_msg_base[0], 0, sizeof(cni->cni_msg_base));
}


/**.
This function should be used for N-frames only.
@param mae Pointer to the target entry in the message area.
@param frame Pointer to a frame with data we want to place in the CNI.
@param data_length Length of the data part of the message (in bytes).*/
void cni_mae_extract_frame_data(CNI_MAE* mae, FRAME *frame, unsigned data_length)
{
	FRAME_N_DATA *n_data = &frame->type.n;
	unsigned i;

	for (i = 0; i < data_length; i++)
		cni_mae_set_data(mae, i, frame_n_data_get_app_data(n_data, i));
}


/**.
@param state_number The state number to be transalted (see @c PROTOCOL_STATE definition).
@return A statically allocated string that describes the given state.*/
char *cni_get_protocol_state_name(unsigned state_number)
{
	switch (state_number) {
		case PS_INIT:        return "INIT       ";
		case PS_LISTEN:      return "LISTEN     ";
		case PS_COLD_START:  return "COLDSTART  ";
		case PS_STARTUP:     return "STARTUP    ";
		case PS_APPLICATION: return "APPLICATION";
		case PS_READY:       return "READY      ";
		case PS_PASSIVE:     return "PASSIVE    ";
		case PS_AWAIT:       return "AWAIT      ";
		case PS_SELFTEST:    return "SELFTEST   ";
		case PS_FREEZE:      return "FREEZE     ";
		case PS_DOWNLOAD:    return "DOWNLOAD   ";
	}
	return "Unknown!";
}


/***********************************************************************************/
/* Special Field Access Functions */

/**.
The value is stored at in 16-bits cell although only 9 lower bits are used + the MSB 
as a sign in the current HW implementation. This function masks the unused bits to 
achieve the same behavior.
@param cni Pointer to a @c CNI instance or a derived structure.
@return Returns the signed value of @c csct field truncated to 10 bits.
@see cni_set_external_rate_correction*/
int cni_get_external_rate_correction(CNI *cni)
{
	int x = cni->external_rate_correction & 0x81FF;

	if (x & bit_pos(15))
		return x | ~bit_mask(9);
	else
		return x;
}


/**.
The value is stored at in 16-bits cell although only 9 lower bits are used + the MSB 
	as a sign in the current HW implementation. This function masks the unused bits to 
	achieve the same behavior.
@param cni Pointer to a @c CNI instance or a derived structure.
@param val The new value of the @c csct field.
@return Returns the signed value written to @c CNI truncated to 10 bits.
@see cni_get_external_rate_correction*/
int cni_set_external_rate_correction(CNI *cni, int val)
{
	cni->external_rate_correction = val & bit_mask(16);
	return val;
}

/*@}*/
/*@}*/


