/**@addtogroup cstate
 * @{*/

/**@addtogroup cstate_int  Implementation
 * @{*/


/**@file 
 * @brief C-State field definition (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 * 
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project \b FIT - Fault Injection for TTA, IST-1999-10748
 */

#include "ttpc/fl1debug.h"
#include "ttpc/fp1cstat.h"
#include "ttpc/fl1func.h"
#include "ttpc/fp1const.h"
#include <string.h>

/**********************************************************************************/
/* Generated Access Functions for C_STATE */
/**********************************************************************************/

implement_access_functions(C_STATE, c_state, time, 16)
implement_access_functions(C_STATE, c_state, medl_position, 9)
implement_access_functions(C_STATE, c_state, mode, 5)
implement_access_functions(C_STATE, c_state, dmc, 2)
implement_index_access_functions(C_STATE, c_state, membership, 1)

/**********************************************************************************/

/* Fills the argument with default fault mask for C-State. */
void c_state_fault_mask(unsigned* mask)
{
    int i;
    C_STATE *p = (C_STATE*)mask;
    memset(mask, 0, sizeof(C_STATE));

    p->time = bit_mask(16);
    p->medl_position = bit_mask(9);
    p->mode = bit_mask(5);
    p->dmc = bit_mask(2);
    for (i = 0; i < MAX_SRU_COUNT; i++)
        p->membership[i] = 1;
}

/** Calculate the CRC of whole @c C_STATE structure.
 * 
 * This function computes the CRC over all fields of the given @a c_state instance. 
 * From the membership vector only relevant fields are used (according to the
 * @a length argument.	No initialization of the @a crc buffer is performed. 
 *
 * @param c_state Pointer to an instance of @c C_STATE structure.
 * @param crc Pointer to an allocated CRC buffer. The result is returned in this field.
 * @param polynomial Numerial expression of he input polynomial for CRC generator.
 * @param length C-State length as written within the MEDL table.
 *
 * @see get_crc*/
void c_state_chain_crc(C_STATE *c_state, unsigned *crc, unsigned polynomial, int length)
{
	int i;

	get_crc(crc, polynomial, c_state_get_time(c_state), 16);
	get_crc(crc, polynomial, c_state_get_medl_position(c_state), 9);
	get_crc(crc, polynomial, c_state_get_mode(c_state), 5);
	get_crc(crc, polynomial, c_state_get_dmc(c_state), 2);
	length = (length - 2) * 16;
	for (i = 0; i < length; i++)
		get_crc(crc, polynomial, c_state_get_membership(c_state, i), 1);
}


/** Compares two @c C_STATE structures and returns nonzero code if they are identical.
The function does not compare parts of memebrship vector that are not used - this is
	deermined using the @a length argument.
@param cs1 First compared instance of @c C_STATE structure.
@param cs2 SEcond compared instance of @c C_STATE structure.
@param length C-State length as written within the MEDL table.
@return Nonzero code if both instances are identical, zero otherwise.*/
unsigned c_state_cmp(C_STATE *cs1, C_STATE *cs2, int length)
{
	int i = 0;
	int result =
		(c_state_get_time(cs1) == c_state_get_time(cs2))
		&& (c_state_get_medl_position(cs1) == c_state_get_medl_position(cs2))
		&& (c_state_get_mode(cs1) == c_state_get_mode(cs2))
		&& (c_state_get_dmc(cs1) == c_state_get_dmc(cs2));
	while (result && i < (length - 2) * 16) {
		result = result && (c_state_get_membership(cs1, i) == c_state_get_membership(cs2, i));
		i++;
	}
	return result;
}

/*@}*/
/*@}*/

