/**@addtogroup ctrl_fun
 * @{*/

/**@addtogroup ctrl_fun_int Implementation
 * @{*/

/**@file
 * @brief TTP/C Communication controller functionality (implementation).
 * @version 1.6
 * @date Last modified 2003-02-15 
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1,
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#include <string.h>
#include <math.h>

#include "ttpc/fl1debug.h"
#include "ttpc/fp1ctrl.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1cni.h"
#include "ttpc/fp1medl.h"
#include "ttpc/fp1frame.h"
#include "ttpc/fp1chan.h"

#include "ttpc/sp1chan.h"

/*****************************************************************************/

/** A shortcut that makes the logging process throughout the whole project transparent 
 * and unified. This version logs the current model time and process name.
 *
 * @param msg This argument is special. It should contain the format specifier as in
 * a call to printf function and optional arguments, but the whole set of parameters
 * must be enclosed in one additional pair of parenthesis, e.g.: @code 
 *   log(("Testing entry number %d (made by %s)\n", 1, "me" )); @endcode */
#define log(msg) log_ctrl1_csim(msg)


/** A shortcut that makes the logging process throughout the whole project
 * transparent and unified. This version does not log the current model time
 * and process name - only the actual parameters in @a msg are printed.
 * 
 * @param msg This argument is special. It should contain the format specifier
 * as in a call to printf function and optional arguments, but the whole set of
 * parameters must be enclosed in one additional pair of parenthesis, e.g.:
 * @code
 *   log(("Testing entry number %d (made by %s)\n", 1, "me" ));	@endcode */
#define log_asis(msg) log_ctrl1(msg)

/*****************************************************************************/
/* Static Variables */
/*****************************************************************************/

/** This 2D array describes which protocol state transition are allowed and
 * which are not. 
 *
 * Transition from state A to state B is allowed if and only if this condition 
 * is true:
 *   <br><center><code>state_transitions[A][B] != 0</code></center><br>
 * The numerical value of protocol state is used to idx this table. */
static int state_transitions[16][16] = {
/*                = 0  */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_INIT        = 1  */ {0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_LISTEN      = 2  */ {0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0},
/* PS_COLD_START  = 3  */ {0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_STARTUP     = 4  */ {0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_APPLICATION = 5  */ {0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_READY       = 6  */ {0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_PASSIVE     = 7  */ {0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0},
/* PS_AWAIT       = 8  */ {0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0},
/* PS_SELFTEST    = 9  */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0},
/* PS_FREEZE      = 10 */ {0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/*                = 11 */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/* PS_DOWNLOAD    = 12 */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0},
/*                = 13 */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/*                = 14 */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0},
/*                = 15 */ {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0}
};


/*****************************************************************************/
/* Static Functions */
/*****************************************************************************/

/** Check the CCF field of a particular message (MAE) and if it is valid then lock the message.*/
static int enter_msg_nbw(CONTROLLER *ctrl, CNI_MAE *mae, int i);

/** Check the CCF field of a particular message (MAE) and if it is valid then unlock the message.*/
static int leave_msg_nbw(CONTROLLER *ctrl, CNI_MAE *mae, int i);

/** Changes the operation mode of a ctrl, does not check permissions.*/
static void set_mode(CONTROLLER *ctrl,  unsigned mode);

/** Check the permission for a given mode change request.*/
static unsigned check_mode_change_request(CONTROLLER *ctrl, unsigned mode, unsigned immediate);

/** Perform a given mode change request - regardless whether it is permitted or not.*/
static void perform_mode_change_request(CONTROLLER *ctrl, unsigned mode, unsigned immediate);

/** Determine and return the EIF of the specified locally stored frame.*/
static CNI_EIF_VALUES get_frame_eif(CONTROLLER *ctrl, unsigned idx);

/** Store the specified frame into the given CNI message area entry.*/
static unsigned store_frame_data(CONTROLLER *ctrl, CNI_MAE *cni_mae, unsigned idx);

/** Return pointer to a message area entry specified by a word offset.*/
static CNI_MAE *get_cni_mae(CONTROLLER *ctrl, unsigned ofs);

/** Clear all host error flags - used when leaving the passive protocol state.*/
static void clear_host_error_flags(CONTROLLER *ctrl);

/** Check whether commmunication blackout error (CB) conditions are met and possibly raise the error.*/
static unsigned test_communication_blackout(CONTROLLER *ctrl);

/** Check whether acknowledgement error (AE) conditions are met and possibly raise the error.*/
static unsigned test_ack_error(CONTROLLER *ctrl);

/** Check whether clock synchronization error (SE) conditions are met and possibly raise the error.*/
static unsigned test_synchronization_error(CONTROLLER *ctrl, unsigned cor_term);

/** Check whether membership error (ME) conditions are met and possibly raise the error.*/
static unsigned test_membership_error(CONTROLLER *ctrl);

/** Raises a membership changed interrupt (MC) if enabled.*/
static void raise_membership_changed_interrupt(CONTROLLER *ctrl);

/** Raises a mode changed interrupt (MI) if enabled.*/
static void raise_mode_changed_interrupt(CONTROLLER *ctrl);

/** Raises user interrupt @a idx (UIx) if enabled.*/
static void raise_user_interrupt(CONTROLLER *ctrl, unsigned idx);

/** Raise a membership lost interrupt (ML) if enabled.*/
static void raise_membership_loss_error(CONTROLLER *ctrl);

/** Raise a protocol error interrupt (PE) if enabled and swithes to freeze state.*/
static void raise_protocol_error(CONTROLLER *ctrl);

/** Set mode vioaltion (MV) status flag and trigger a host error.*/
static void raise_mode_violation_error(CONTROLLER *ctrl);

/** Raise a host error interrupt (HE) if enabled and swithes to passive state.*/
static void raise_host_error(CONTROLLER *ctrl);

/** Raise a BIST error interrupt (BE) if enabled and swithes to passive state.*/
static void raise_bist_error(CONTROLLER *ctrl);

/** Trigger the interrupt line to the host processor.*/
static void raise_interrupt(CONTROLLER *ctrl);

/** Return pointer to the ctrl's current Mode Address Entry (in MEDL).*/
static MEDL_MAE *get_current_mae(CONTROLLER *ctrl);

/** Check whether the current operation mode is application mode.*/
static unsigned in_application_mode(CONTROLLER *ctrl);

/** Check whether the current operation mode is startup mode.*/
static unsigned in_startup_mode(CONTROLLER *ctrl);

/** Check whether the ctrl is a multiplexed node.*/
static unsigned is_multiplexed(CONTROLLER *ctrl);

/** Check whether the specified slot control entry (SCE) is the ctrl's sending slot.*/
static unsigned is_my_slot(CONTROLLER *ctrl, MEDL_SCE *sce);

/** Check whether bus activity was observed since listening was started.*/
static unsigned bus_activity(CONTROLLER *ctrl);

/** Check whether the ctrl is allowed to reconfigure into the cluster in the given slot.*/
static unsigned is_reconfiguration_allowed(CONTROLLER *ctrl, MEDL_SCE *sce);

/** Return the number of the first valid I-frame stored in local data.*/
static unsigned get_valid_i_frame(CONTROLLER *ctrl);

/** Check whether the host life-sign is up-to-date and clear the host life-sign.*/
static unsigned test_host_lifesign(CONTROLLER *ctrl);

/** Synchronize the ctrl with cluster using an I-frame received on the specified channel.*/
static void synchronize_on_i_frame(CONTROLLER *ctrl, unsigned channel);

/** Return the MEDL position of ctrl's first sending slot.*/
static unsigned get_first_sending_slot(CONTROLLER *ctrl);

/** Return the MEDL position of ctrl's next sending slot.*/
static unsigned get_next_sending_slot(CONTROLLER *ctrl);

/** Set-up C-State field for the Cold Start state.*/
static void setup_cold_start_c_state(CONTROLLER *ctrl);

/** Set-up the specified frame to be a cold start I-Frame.*/
static void setup_cold_start_i_frame(CONTROLLER *ctrl, unsigned channel);

/** Returns the number of first I-Frame that is a correct answer to a previously sent cold start frame.*/
static unsigned cold_start_get_valid_i_frame(CONTROLLER *ctrl);

/** Prepare the appropiate frames for sending (according to MEDL) within ctrl's local data.*/
static unsigned prepare_frame(CONTROLLER *ctrl);

/** Prepare an I-frame for sending using the specified channel in the ctrl's local data.*/
static unsigned prepare_i_frame(CONTROLLER *ctrl, unsigned channel);

/** Prepare a N-frame for sending using the specified channel in the ctrl's local data.*/
static unsigned prepare_n_frame(CONTROLLER *ctrl, unsigned channel);

/** Return the result of implicit acknowledgement (IACK) algorithm check on specified channel.*/
static void check_iack(CONTROLLER *ctrl, unsigned channel, IACK_STATES *state, unsigned *valid);

/** Perform implicit acknoledgement (IACK) algorithm for the current slot transmission.*/
static unsigned perform_iack(CONTROLLER *ctrl);


/*************************************************************************/
/* Static functions - implementation */

/** Configure the BG with dynamically computed parameters during the
 * synchronization phase.
 *
 * @param ctrl  Pointer to an instance of #CONTROLLER or derived structure. */
static void conf_dyn_bg(CONTROLLER *ctrl)
{
	int i_max = get_next_sending_slot(ctrl),
	    i = c_state_get_medl_position(&ctrl->cni.c_state),
	    tso = c_state_get_time(&ctrl->cni.c_state) - cni_get_cluster_time(&ctrl->cni);
	
	log(("DYNBG: i=%d, imax=%d, tso=%d\n", i, i_max, tso));
	while (i < i_max) {
		MEDL_SCE *sce = &ctrl->medl.slot_control_entry[i];
		tso += medl_sce_get_sru_slot_duration(sce);
		if (medl_sce_get_eot(sce))
			i = 0;
		else
			i++;
	}
	
	log(("DYNBG: tso = %d\n", tso));
	BUS_GUARDIAN *bg = (BUS_GUARDIAN *)ctrl->bus_guardian;
	bus_guardian_set_params_ext(bg,
			1,
			tso - 1,  /* to be safe */
			medl_bgp_get_bg_slot_duration(&ctrl->medl.role_list.bus_guardian),
			0xFFFF);
	bg->arm_signal(bg, 1);  /*TODO: check ret val */
	ctrl->ploc_data.sync_period_params = 1;
}


/**.
The function first checks whether the CCF field is even - if not, then a NBW error  
(host error) is raised and nonzero code is returned. If the CCF is
valid, it is incremented by 1 (locked), the new value is stored within an internal
variable and zero is returned.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param mae Pointer to a message area entry within CNI.
@param i The is the number of the channel that is used to transmit the message.
@return Zero if everything works, nonzero otherwise - in such case a an error is raised.*/
static int enter_msg_nbw(CONTROLLER *ctrl, CNI_MAE *mae, int i)
{
	CNI *cni = &ctrl->cni;

	log(("Entering NBW protected section of Message area.\n"));
	if (cni_mae_get_ccf(mae) & 1) {
		cni_ps_set_ne(&cni->protocol_status, 1);
		raise_host_error(ctrl);
		cni_mae_set_ccf(mae, 0);
		return 1;
	}

	ctrl->ploc_data.old_msg_ccf[i] = cni_mae_increment_ccf(mae, 1);
	return 0;
}


/**.
The function first compares the current value of CCF with a stored value and if it is 
different, it raises a NBW error (host error) and returns nonzero code. Otherwise the 
value is incremented by one (unlocked) and zero code is returned.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param mae Pointer to a message area entry within CNI.
@param i The is the number of the channel that is used to transmit the message.
@return Zero if everything works, nonzero otherwise - in such case a state an error is raised.*/
static int leave_msg_nbw(CONTROLLER *ctrl, CNI_MAE *mae, int i)
{
	CNI *cni = &ctrl->cni;

	log(("Leaving NBW protected section of Message area.\n"));
	if (cni_mae_get_ccf(mae) != ctrl->ploc_data.old_msg_ccf[i]) {
		cni_ps_set_ne(&cni->protocol_status, 1);
		raise_host_error(ctrl);
		cni_mae_set_ccf(mae, 0);
		return 1;
	}

	cni_mae_increment_ccf(mae, 1);
	return 0;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param mode New operation mode number.*/
static void set_mode(CONTROLLER *ctrl,  unsigned mode)
{
	C_STATE *c_state = &ctrl->cni.c_state;

	if (mode != 0) {
		controller_set_state(ctrl, (mode == STARTUP_MODE_NUMBER) ? PS_STARTUP : PS_APPLICATION);
		log(("Transition into mode %d.\n", mode));
		c_state_set_mode(c_state, mode);
		raise_mode_changed_interrupt(ctrl);
	} else
		log(("Error: Mode 0 not valid.\n"));
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param mode Requested operation mode number.
@param immediate If set to zero a deffered mode change is requested, otherwise an immediate change.
@return Nonzero code if the mode change is permitted, zero otherwise.*/
static unsigned check_mode_change_request(CONTROLLER *ctrl, unsigned mode, unsigned immediate)
{
	MEDL_MAE *mae = get_current_mae(ctrl);

	log(("Checking mode change request (m=%u, i=%u).\n", mode, immediate));
	if (immediate) {
		if (mode == 0) { /* immediate mode change */
			if (medl_mae_get_mcp_c(mae) == 0)
				return 0;
		} else {
			if (medl_mae_get_mcp_i(mae, mode) == 0)
				return 0;
		}
	} else {			/* deffered mode change */
		if (mode != 0) {
			if (medl_mae_get_mcp_d(mae, mode - 1) == 0)
				return 0;
		}
	}
	return 1;
}


/**.
\note This function does not change the mode itself, only requests a mode change.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param mode Requested operation mode number.
@param immediate If set to zero a deffered mode change is requested, otherwise an immediate change.*/
static void perform_mode_change_request(CONTROLLER *ctrl, unsigned mode, unsigned immediate)
{
	C_STATE *c_state = &ctrl->cni.c_state;

	log(("Performing mode change request (m=%u, i=%u).\n",
		mode, immediate));
	if (immediate)
		if (mode == 0)	/* immediate mode change */
			c_state_set_dmc(c_state, 0);
		else
			set_mode(ctrl, mode);
	else				/* deffered mode change */
		if (mode != 0)
			c_state_set_dmc(c_state, mode);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param idx Number of the frame to check - either 0 or 1 (channels A and B).
@return The EIF value determined from frame @a idx.*/
static CNI_EIF_VALUES get_frame_eif(CONTROLLER *ctrl, unsigned idx)
{
	FRAME *frame = &ctrl->ploc_data.frame[idx];
	unsigned c_state_len = medl_ccp_get_c_state_length(&ctrl->medl.controller_cfg_params);
	CNI_EIF_VALUES eif;

	if (frame->state == FRAME_STATE_NULL)
		eif = EIF_NULL_FRAME;
	else
		if (frame->state != FRAME_STATE_COMPLETE)
			eif = EIF_INVALID_FRAME;
		else
			if (!frame_is_valid(frame, &ctrl->cni.c_state, c_state_len))
				eif = EIF_CRC_ERROR;
			else
				eif = EIF_CORRECT_FRAME;

	if (frame->noise != 0)
		eif = (CNI_EIF_VALUES)(eif | EIF_NOISE);

	return eif;
}


/**.
This function firsts checks if the current message should be stored within the message
area of CNI (N-Frames only that have a valid address within MEDL). If yes, it copies the
data oart of the message and returns the number of bytes stored.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param cni_mae Pointer to the message area entry that should store the message data.
@param idx Number of the frame to check - either 0 or 1 (channels A and B).
@return The size of stored message or zero when the message was not stored.*/
static unsigned store_frame_data(CONTROLLER *ctrl, CNI_MAE *cni_mae, unsigned idx)
{
	MEDL_MAE *medl_mae = get_current_mae(ctrl);
	unsigned len;

	if (cni_mae == NULL)
		return 0;

	len = medl_mae_get_dafla(medl_mae) + 1;
	cni_mae_extract_frame_data(cni_mae, &ctrl->ploc_data.frame[idx], len);

	return len;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param idx Channel selection - there are two replicated channels A (0) and B(1).
@return Pointer to the chosen CNI message area entry within CNI.*/
static CNI_MAE *get_cni_mae(CONTROLLER *ctrl, unsigned idx)
{
	MEDL_MAE *medl_mae = get_current_mae(ctrl);
	unsigned cni_ofs;

	if (idx == 0)
		cni_ofs = medl_mae_get_cni_address_cha(medl_mae);
	else
		cni_ofs = medl_mae_get_cni_address_chb(medl_mae);
	if (cni_ofs == CNI_MAE_NO_ADDRESS)
		return NULL;

	return cni_get_message_entry(&ctrl->cni, cni_ofs);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void clear_host_error_flags(CONTROLLER *ctrl)
{
	CNI_PS *ps = &ctrl->cni.protocol_status;

	cni_ps_set_mv(ps, 0);
	cni_ps_set_ne(ps, 0);
	cni_ps_set_so(ps, 0);
	controller_update_cni(ctrl);	
}


/**.
\note May switch the state to freeze.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero if CB error is raised, zero otherwise.*/
static unsigned test_communication_blackout(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;

	if (	(cni_ps_get_failed_counter(&cni->protocol_status)
			+ cni_ps_get_invalid_counter(&cni->protocol_status)
			+ cni_ps_get_accept_counter(&cni->protocol_status)) == 0) {
		cni_ps_set_cb(&cni->protocol_status, 1);
		raise_protocol_error(ctrl);
		return 1;
	}
	return 0;
}


/**.
\note May switch the state to freeze.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero if AE error is raised, zero otherwise.*/
static unsigned test_ack_error(CONTROLLER *ctrl)
{
	CNI_PS *ps = &ctrl->cni.protocol_status;

	if (	(cni_ps_get_failed_counter(ps)
			+ cni_ps_get_invalid_counter(ps))
			>= cni_ps_get_accept_counter(ps)) {
		cni_ps_set_ae(ps, 1);
		raise_protocol_error(ctrl);
		return 1;
	}
	return 0;
}


/**.
\note May switch the state to freeze.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param cor_term Clock correction term that will be applied - value in microtics.
@return Nonzero if SE error is raised, zero otherwise.*/
static unsigned test_synchronization_error(CONTROLLER *ctrl, unsigned cor_term)
{
	double correction = (double)cor_term / medl_ccp_get_microticks(&ctrl->medl.controller_cfg_params);

	if (correction > CLOCK_PRECISION / 2.0) {
		cni_ps_set_se(&ctrl->cni.protocol_status, 1);
		raise_protocol_error(ctrl);
		return 1;
	}
	return 0;
}


/**.
\note May switch the state to freeze.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero if ME error is raised, zero otherwise.*/
static unsigned test_membership_error(CONTROLLER *ctrl)
{
	CNI_PS *ps = &ctrl->cni.protocol_status;
	MEDL_CCP *ccp = &ctrl->medl.controller_cfg_params;

	if (medl_ccp_get_mmfc(ccp) > 0 && cni_ps_get_mfc(ps) >= medl_ccp_get_mmfc(ccp)) {
		cni_ps_set_me(ps, 1);
		raise_protocol_error(ctrl);
		return 1;
	}
	return 0;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void raise_membership_changed_interrupt(CONTROLLER *ctrl)
{
	log(("Membership change detected.\n"));
/*	if (cni_is_get_mc(&ctrl->cni.interrupt_status) == 1)
		return;*/
	cni_is_set_mc(&ctrl->cni.interrupt_status, 1);
	if (cni_ie_get_mc(&ctrl->cni.interrupt_enable) == 1)
		ctrl->ploc_data.tp_sync_interrupt = 1;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void raise_mode_changed_interrupt(CONTROLLER *ctrl)
{
	log(("Mode change detected.\n"));
/*	if (cni_is_get_mi(&ctrl->cni.interrupt_status) == 1)
		return;*/
	cni_is_set_mi(&ctrl->cni.interrupt_status, 1);
	if (cni_ie_get_mi(&ctrl->cni.interrupt_enable) == 1)
		ctrl->ploc_data.tp_sync_interrupt = 1;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param idx The number of the user interrupt - currently nubers 0 and 1 are supported.*/
static void raise_user_interrupt(CONTROLLER *ctrl, unsigned idx)
{
	switch (idx) {
	case 0:
		log(("UI1 detected.\n"));
/*		if (cni_is_get_ui1(&ctrl->cni.interrupt_status) == 1)
			return;*/
		cni_is_set_ui1(&ctrl->cni.interrupt_status, 1);
		if (cni_ie_get_ui1(&ctrl->cni.interrupt_enable) == 1)
			ctrl->ploc_data.tp_sync_interrupt = 1;
		break;
	case 1:
		log(("UI2 detected.\n"));
/*		if (cni_is_get_ui2(&ctrl->cni.interrupt_status) == 1)
			return;*/
		cni_is_set_ui2(&ctrl->cni.interrupt_status, 1);
		if (cni_ie_get_ui2(&ctrl->cni.interrupt_enable) == 1)
			ctrl->ploc_data.tp_sync_interrupt = 1;
		break;
	}
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void raise_membership_loss_error(CONTROLLER *ctrl)
{
	log(("Membership loss error.\n"));
/*	if (cni_is_get_ml(&ctrl->cni.interrupt_status) == 1)
		return;*/
	cni_is_set_ml(&ctrl->cni.interrupt_status, 1);
	if (cni_ie_get_ml(&ctrl->cni.interrupt_enable) == 1)
		ctrl->ploc_data.tp_sync_interrupt = 1;
}


/**.
\note The caller of this function should set at least of the detailed PE flags
in the protocol state field in CNI.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void raise_protocol_error(CONTROLLER *ctrl)
{
	CNI_PS *ps = &ctrl->cni.protocol_status;

	log(("Protocol error: "));
		if (cni_ps_get_ae(ps))
				log_asis(("Acknowlegment Error "));
		if (cni_ps_get_be(ps))
				log_asis(("Bus Guardian Error "));
		if (cni_ps_get_se(ps))
				log_asis(("Clock Synchronization Error "));
		if (cni_ps_get_mc(ps))
				log_asis(("Medl CRC Error "));
		if (cni_ps_get_cb(ps))
				log_asis(("Communication Blackout Error "));
		if (cni_ps_get_dc(ps))
				log_asis(("Download Completed "));
		if (cni_ps_get_me(ps))
				log_asis(("Membership Error "));
	log_asis(("\n"));

	/* FIXME: should be the same interrupt raised twice? */
	cni_is_set_pe(&ctrl->cni.interrupt_status, 1);
	if (cni_ie_get_pe(&ctrl->cni.interrupt_enable) == 1)
		raise_interrupt(ctrl);
	controller_set_state(ctrl, PS_FREEZE);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void raise_mode_violation_error(CONTROLLER *ctrl)
{
	cni_ps_set_mv(&ctrl->cni.protocol_status, 1);
	raise_host_error(ctrl);
}


/**.
\note The caller of this function should set at least of the detailed HE flags
in the protocol state field in CNI.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
\todo Should be the same type of interrupt raised mupltiple times?*/
static void raise_host_error(CONTROLLER *ctrl)
{
	CNI_PS *ps = &ctrl->cni.protocol_status;

	log(("Host error: "));
		if (cni_ps_get_so(ps))
			log_asis(("Slot Occupied Error "));
		if (cni_ps_get_mv(ps))
			log_asis(("Mode Violation Error "));
		if (cni_ps_get_ne(ps))
			log_asis(("NBW Protocol Error "));
	log_asis(("\n"));

	cni_is_set_he(&ctrl->cni.interrupt_status, 1);
	cni_set_host_lifesign(&ctrl->cni, 0);
	if (cni_ie_get_he(&ctrl->cni.interrupt_enable) == 1)
		raise_interrupt(ctrl);
	controller_set_state(ctrl, PS_PASSIVE);
}


/**. 
\note The caller of this function should set at least of the detailed BE flags
in the protocol state field in CNI.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
\todo Should be the same type of interrupt raised mupltiple times?*/
static void raise_bist_error(CONTROLLER *ctrl)
{
	CNI_BS *bs = &ctrl->cni.bist_status;

	log(("BIST error: "));
		if (cni_bs_get_sc(bs))
			log_asis(("Selftest Completed "));
		if (cni_bs_get_av(bs))
			log_asis(("Assertion Violation (%d) ", ctrl->cni.assertion_code));
		if (cni_bs_get_br(bs))
			log_asis(("Invalid BIST request "));
		if (cni_bs_get_rm(bs))
			log_asis(("RAM error "));
		if (cni_bs_get_rf(bs))
			log_asis(("Register File Error "));
		if (cni_bs_get_wd(bs))
			log_asis(("Watchdog error "));
	log_asis(("\n"));

	cni_is_set_be(&ctrl->cni.interrupt_status, 1);
	if (cni_ie_get_be(&ctrl->cni.interrupt_enable) == 1)
		raise_interrupt(ctrl);
	controller_set_state(ctrl, PS_FREEZE);
}


/**.
The function copies the internal interrupt status flags to the CNI status area and
then calls the interrupt handler of the host. Afterwards it clears the internal 
interrupt status.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void raise_interrupt(CONTROLLER *ctrl)
{
	log(("Raising an interrupt.\n"));
	if (ctrl->host_interrupt != NULL)
		ctrl->host_interrupt(ctrl->host_arg);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
return Pointer to the current MAE entry (determined from C-State) in MEDL table.*/
static MEDL_MAE *get_current_mae(CONTROLLER *ctrl)
{
	C_STATE *c_state = &ctrl->cni.c_state;

	return &ctrl->medl.mode_address_entry
		[c_state_get_mode(c_state)]
		[c_state_get_medl_position(c_state)];
}


/**.
The numerical range of application modes is defined by two constants in
fp1const.h file (currently the range is 2-30).
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if the current mode is not an application mode, nonzero otherwise.*/
static unsigned in_application_mode(CONTROLLER *ctrl)
{
	C_STATE *c_state = &ctrl->cni.c_state;

	return (c_state_get_mode(c_state) >= FIRST_APP_MODE_NUMBER
		&& c_state_get_mode(c_state) <= LAST_APP_MODE_NUMBER);
}


/**.
The number of startup mode is defined within the fp1const.h file.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if the current mode is not startup mode, nonzero otherwise.*/
static unsigned in_startup_mode(CONTROLLER *ctrl)
{
	return (c_state_get_mode(&ctrl->cni.c_state) == STARTUP_MODE_NUMBER);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero if the node is multiplexed, zero otherwise.*/
static unsigned is_multiplexed(CONTROLLER *ctrl)
{
	return (ctrl->ploc_data.logical_name.multiplex_id > 0);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param sce Pointer to one slot control entry within the MEDL table.
@return Nonzero if the ctrl is sender in the given SCE, zero otherwise.*/
static unsigned is_my_slot(CONTROLLER *ctrl, MEDL_SCE *sce)
{
	return (ctrl->ploc_data.logical_name.slot_position
		== logical_name_get_slot_position(&sce->logical_name));
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero if any bus activity has been observed, zero otherwise.*/
static unsigned bus_activity(CONTROLLER *ctrl)
{
	FRAME *frame = ctrl->ploc_data.frame;

	return frame[0].state != FRAME_STATE_NULL || frame[1].state != FRAME_STATE_NULL;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param sce Pointer to a slot control entry within the MEDL table (presumably the next).
@return Nonzero if reintegration is allowed, zero otherwise.*/
static unsigned is_reconfiguration_allowed(CONTROLLER *ctrl, MEDL_SCE *sce)
{
	CNI *cni = &ctrl->cni;
	unsigned memb_pos = medl_sce_get_membership_pos(sce);

	if (medl_sce_get_ra(sce) != 1)
		return 0;
	if (cni_get_null_frame(cni, memb_pos) != 1)
		return 0;
	if (c_state_get_membership(&cni->c_state, memb_pos) == 1) {
		cni_ps_set_so(&cni->protocol_status, 1);
		raise_host_error(ctrl);
		return 0;
	}
	return 1;
}


/**.
This funtion checks the received frames stored in ctrl's local data (the EIF
field) and returns the number of the first correct frame (also the number of the channel
that transmitted the frame).
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return The number of first correct I-frame or 0xFFFF if no correct I-frame is found.*/
static unsigned get_valid_i_frame(CONTROLLER *ctrl)
{
	FRAME *frame = ctrl->ploc_data.frame;

	if (frame_is_valid_i_frame(&frame[0]))
		return 0;
	if (frame_is_valid_i_frame(&frame[1]))
		return 1;

	return 0xFFFF;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero if host is alive, zero otherwise (a host error is raised).*/
static unsigned test_host_lifesign(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;
	unsigned result;

	log(("Updating ctrl lifesign\n"));
	result = ((~cni_get_host_lifesign(cni) & bit_mask(16)) == cni_get_controller_lifesign(cni));
	cni_set_host_lifesign(cni, 0);
	cni_set_controller_lifesign(cni, c_state_get_time(&cni->c_state) & bit_mask(15));

	if (!result)
		log(("Host not alive\n"));

	return result;
}


/**.
The functions read a new C-State from the frame, updates the C-State valid flag and
changes state into passive or ready.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel The number of channel to read the frame from (A is 0, B is 1).*/
static void synchronize_on_i_frame(CONTROLLER *ctrl, unsigned channel)
{
	CNI *cni = &ctrl->cni;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	FRAME *frame = &ploc_data->frame[channel];
	C_STATE *frame_c_state = frame_get_c_state(frame);

	if (c_state_get_mode(frame_c_state) == COLD_START_MODE_NUMBER && is_multiplexed(ctrl))
		return;          /* multiplexed node: discard frame */

	cni_ps_increment_accept_counter(&cni->protocol_status, 1);

	/* Copy C-State */
	cni->c_state = *frame_c_state;
	if (c_state_get_mode(frame_c_state) == COLD_START_MODE_NUMBER)
		/* Cold Start I-Frame: Modify Mode Number */
		c_state_set_mode(&cni->c_state, STARTUP_MODE_NUMBER); 

	cni_ps_set_c_state_valid(&cni->protocol_status, C_STATE_VALID);
	cni_set_null_frame(cni, medl_sce_get_membership_pos(controller_get_current_sce(ctrl)), 0);

	/* Set Cluster Time */
	ctrl->synchronize_time(ctrl, channel);

	/*TODO Check mode change request here.*/
	
#ifdef DYN_BG
	conf_dyn_bg(ctrl);
#else
	/* Set Synchronization Params for BG */
	bus_guardian_set_params((BUS_GUARDIAN *)ctrl->bus_guardian,
		&ctrl->medl.implementation_params.synchronization_period_params);
	ploc_data->immediate_send_permission = 1;
	ploc_data->sync_period_params = 1;
#endif

	if (in_application_mode(ctrl)) {
		test_host_lifesign(ctrl);
		log(("Synchronized on APP I-Frame\n"));
		controller_set_state(ctrl, PS_PASSIVE);
	}
	else if (in_startup_mode(ctrl)) {
		log(("Synchronized on Startup I-Frame\n"));
		controller_set_state(ctrl, PS_READY);
	}
	else if (test_host_lifesign(ctrl)) {
		log(("Synchronized on ? I-Frame\n"));
		controller_set_state(ctrl, PS_READY);
	}
	else {
		log(("Synchronized on ? I-Frame\n"));
		controller_set_state(ctrl, PS_PASSIVE);
	}

	controller_set_next_slot(ctrl);

	/* Interrupts */
	raise_mode_changed_interrupt(ctrl);
	raise_membership_changed_interrupt(ctrl);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return The number of ctrl's first sending slot (0 is the first MEDL position)*/
static unsigned get_first_sending_slot(CONTROLLER *ctrl)
{
	MEDL_SCE *sce = ctrl->medl.slot_control_entry;
	LOGICAL_NAME *lname = &ctrl->ploc_data.logical_name;
	unsigned i = 0;

	while ((i < MAX_TRANSMISSION_CYCLE_LENGTH) &&
		(logical_name_cmp(&sce[i].logical_name, lname) == 0))
		i++;
	log_assert(i < MAX_TRANSMISSION_CYCLE_LENGTH); /* Bad MEDL design */
	return i;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return The number of ctrl's next sending slot (0 is the first MEDL position)*/
static unsigned get_next_sending_slot(CONTROLLER *ctrl)
{
	MEDL_SCE *sce = ctrl->medl.slot_control_entry;
	LOGICAL_NAME *lname = &ctrl->ploc_data.logical_name;
	unsigned i = c_state_get_medl_position(&ctrl->cni.c_state),
		 cnt = 0;

	while ((cnt < MAX_TRANSMISSION_CYCLE_LENGTH) 
			&& (logical_name_cmp(&sce[i].logical_name, lname) == 0)) {
		if (medl_sce_get_eot(&sce[i]))
			i = 0;
		else
			i++;
		cnt++;
	}
	log_assert(cnt < MAX_TRANSMISSION_CYCLE_LENGTH); /* Bad MEDL design */
	return i;
}



/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
static void setup_cold_start_c_state(CONTROLLER *ctrl)
{
	PLOC_DATA *ploc = &ctrl->ploc_data;
	C_STATE *c_state = &ctrl->cni.c_state;

	c_state_set_mode(c_state, COLD_START_MODE_NUMBER);
	c_state_set_medl_position(c_state, get_first_sending_slot(ctrl));

	/* set Membership vector */
	memset(&c_state->membership, 0, sizeof(c_state->membership));
	c_state_set_membership(c_state, ploc->membership_pos, 1);

	c_state_set_dmc(c_state, 0);
	c_state_set_time(c_state, cni_get_cluster_time(&ctrl->cni));
}


/**.
This function pepares a frame in the ctrl's local data for transmission
in the cold start state.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel Number of the frame (and channel) that will be used.*/
static void setup_cold_start_i_frame(CONTROLLER *ctrl, unsigned channel)
{
	C_STATE *c_state = &ctrl->cni.c_state;
	unsigned cs_length = medl_ccp_get_c_state_length(&ctrl->medl.controller_cfg_params);
	FRAME *frame = &ctrl->ploc_data.frame[channel];
	FRAME_HEADER *header = &frame->header;

	frame->state = FRAME_STATE_NULL;

	/* set header */
	frame_header_set_i_n_bit(header, FRAME_TYPE_I);
	frame_header_set_mode(header, 0);
	frame_header_set_immediate(header, 0);
	/* set C-State */
	frame_set_c_state(frame, *c_state);
	frame_set_c_state_length(frame, cs_length);
	/* set CRC */
	frame_set_crc(frame, frame_calculate_crc(frame, c_state, cs_length));
}


/**.
This function checks whether there is a correct reply to a previously sent cold start
I-frame. If a correct reply is found, the frame (and channel) number is returned,
otherwise the value 0xFFFF is returned.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return The number of first correct I-frame or 0xFFFF if no correct I-frame is found.*/
static unsigned cold_start_get_valid_i_frame(CONTROLLER *ctrl)
{
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	C_STATE *c_state = &ctrl->cni.c_state;
	C_STATE *frame_state;
	unsigned mbp1, mbp2;
	unsigned medl_position;
	unsigned c;

	if ((c = get_valid_i_frame(ctrl)) > 1)
		return 0xFFFF;
	frame_state = frame_get_c_state(&ploc_data->frame[c]);

	mbp1 = ploc_data->membership_pos;
	mbp2 = medl_sce_get_membership_pos(controller_get_current_sce(ctrl));
	medl_position = c_state_get_medl_position(frame_state);

	if ( (c_state_get_mode(frame_state) == STARTUP_MODE_NUMBER)
		&& (c_state_get_time(frame_state) == c_state_get_time(c_state))
		&& (medl_position == c_state_get_medl_position(c_state))
		&& (c_state_get_membership(frame_state, mbp1) == 1)
		&& (c_state_get_membership(frame_state, mbp2) == 1) )
		return c;
	else
		return 0xFFFF;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if successfull, nonzero otherwise.*/
static unsigned prepare_frame(CONTROLLER *ctrl)
{
	FRAME *frame = ctrl->ploc_data.frame;
	CNI *cni = &ctrl->cni;
	C_STATE *c_state = &cni->c_state;
	unsigned mode = cni_mcr_get_mode(cni);
	unsigned immediate = cni_mcr_get_immediate(cni);
	unsigned membership_pos = ctrl->ploc_data.membership_pos;
	unsigned membership = c_state_get_membership(&ctrl->cni.c_state, membership_pos);
	unsigned ret = 0;

	log(("Preparing frames for TP.\n"));
	if (!check_mode_change_request(ctrl, mode, immediate)) {
		cni_mcr_clear(cni);
		raise_mode_violation_error(ctrl);
		return 1;
	}

	c_state_set_membership(c_state, membership_pos, 1);

	if (medl_mae_get_ia(get_current_mae(ctrl)) == 1)
		ret += prepare_i_frame(ctrl, 0);
	else
		ret += prepare_n_frame(ctrl, 0);
	frame[0].state = FRAME_STATE_COMPLETE;

	if (medl_mae_get_ib(get_current_mae(ctrl)) == 1)
		ret += prepare_i_frame(ctrl, 1);
	else
		ret += prepare_n_frame(ctrl, 1);

	cni_mcr_clear(cni);
	frame[1].state = FRAME_STATE_COMPLETE;

	/* Restore origin Membership Vector status */
	c_state_set_membership(c_state, membership_pos, membership);
	return ret;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel Number of a channel (and also frame replica) - can be 0 for A or 1 for B.
@return Zero if successfull, nonzero otherwise.*/
static unsigned prepare_i_frame(CONTROLLER *ctrl, unsigned channel)
{
	CNI *cni = &ctrl->cni;
	unsigned cs_length = medl_ccp_get_c_state_length(&ctrl->medl.controller_cfg_params);
	FRAME *frame = &ctrl->ploc_data.frame[channel];
	FRAME_HEADER *header = &frame->header;

	frame_header_set_i_n_bit(header, FRAME_TYPE_I);
	frame_header_set_mode(header, cni_mcr_get_mode(cni));
	frame_header_set_immediate(header, cni_mcr_get_immediate(cni));

	frame_set_c_state(frame, cni->c_state);
	frame_set_c_state_length(frame, cs_length);

	frame_set_crc(frame, frame_calculate_crc(frame, &cni->c_state, cs_length));

	return 0;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel Number of a channel (and also frame replica) - can be 0 for A or 1 for B.
@return Zero if successfull, nonzero otherwise.*/
static unsigned prepare_n_frame(CONTROLLER *ctrl, unsigned channel)
{
	CNI *cni = &ctrl->cni;
	FRAME *frame = &ctrl->ploc_data.frame[channel];
	FRAME_HEADER *header = &frame->header;
	FRAME_N_DATA *n_data = &frame->type.n;
	MEDL_MAE *mae = get_current_mae(ctrl);
	CNI_MAE *cni_mae;
	unsigned data_len, i;

	if (channel == 0) {
		 cni_mae = cni_get_message_entry(cni, medl_mae_get_cni_address_cha(mae));
		 data_len = medl_mae_get_dafla(mae);
	}
	else {
		 cni_mae = cni_get_message_entry(cni, medl_mae_get_cni_address_chb(mae));
		 data_len = medl_mae_get_daflb(mae);
	}

	if ((cni_mae_get_eif(cni_mae) & EIF_CORRECT_FRAME) == 0) {
		cni_ps_set_ne(&cni->protocol_status, 1);
		log(("Invalid EIF value!\n"));
		raise_host_error(ctrl);
		return 1;
	}

	frame_header_set_i_n_bit(header, FRAME_TYPE_N);
	frame_header_set_mode(header, cni_mcr_get_mode(cni));
	frame_header_set_immediate(header, cni_mcr_get_immediate(cni));

	frame_set_app_data_length(frame, data_len);
	for (i = 0; i < data_len + 1; i++)
		frame_n_data_set_app_data(n_data, i, cni_mae_get_data(cni_mae, i));

	frame_set_crc(frame, frame_calculate_crc(frame, &cni->c_state,
		medl_ccp_get_c_state_length(&ctrl->medl.controller_cfg_params)));

	return 0;
}


/**.
\note The internal state of the ctrl is not changed.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel Number of a channel (and also frame replica) - can be 0 for A or 1 for B.
\retval state New IACK state - determined from the last transmission on the 
	given channel.
\retval valid Validity flag for the @a state field (zero of the new state could 
	not be determined, nonzero otherwise).*/
static void check_iack(CONTROLLER *ctrl, unsigned channel, IACK_STATES *state, unsigned *valid)
{
	C_STATE *c_state = &ctrl->cni.c_state;
	C_STATE l_state = *c_state;
	unsigned c_state_len = medl_ccp_get_c_state_length(&ctrl->medl.controller_cfg_params);
	PLOC_DATA *ploc_data = &ctrl->ploc_data;

	*valid = 1;
	switch (ploc_data->iack_state) {
		/* Assumming that I am node A, node B is my successor and node C is B's successor. */
		case IACK_SEND:
			/* Send phase */
			*state = IACK_CHECK1;
			break;
		case IACK_OK:
			/* Nothing to check. */
			*state = IACK_OK;
			break;
		case IACK_CHECK1:
			/* Did node B say than I was OK? */
			if (frame_is_valid(&ploc_data->frame[channel], c_state, c_state_len)) {
				*state = IACK_OK;
				break;
			}
			/* Did node B say that I was not OK? */
			c_state_set_membership(&l_state, ploc_data->membership_pos, 0);
			if (frame_is_valid(&ploc_data->frame[channel], &l_state, c_state_len)) {
				*state = IACK_CHECK2;
				break;
			}
			/* Neither is the case, we must wait for another node B (successor). */
			*state = IACK_CHECK1;
			*valid = 0;
			break;
		case IACK_CHECK2:
			/* Did node C say that I was OK and B had failed? */
			c_state_set_membership(&l_state, ploc_data->successor, 0);
			if (frame_is_valid(&ploc_data->frame[channel], &l_state, c_state_len)) {
				*state = IACK_OK;
				break;
			}
			/* Did node C say that I had failed and B was OK? */
			l_state = *c_state;
			c_state_set_membership(&l_state, ploc_data->membership_pos, 0);
			c_state_set_membership(&l_state, ploc_data->successor, 1);
			if (frame_is_valid(&ploc_data->frame[channel], &l_state, c_state_len)) {
				*state = IACK_ERROR;
				break;
			}
			/* Neither is the case, double failure occured. */
			*state = IACK_DOUBLE_FAILURE;
			*valid = 0;
			break;
		case IACK_ERROR:
		case IACK_DOUBLE_FAILURE:
			break;
	}
}


/**.
The function determines a new IACK state from the last transmission and
writes the state in ctrl's local_data. Depending on the result it may modify
the membership vector and modify one or more counters in the protocol status field
in CNI. If the ctrl loses membership, state transistion may follow.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if the IACK algorithm decision is final, nonzero otherwise.*/
static unsigned perform_iack(CONTROLLER *ctrl)
{
	IACK_STATES iack_a, iack_b;
	unsigned valid_a, valid_b;
	CNI *cni = &ctrl->cni;
	CNI_PS *cni_ps = &cni->protocol_status;
	C_STATE *c_state = &cni->c_state;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	MEDL_SCE *sce = controller_get_current_sce(ctrl);
	PROTOCOL_STATE ps = (PROTOCOL_STATE)cni_ps_get_protocol_state(cni_ps);

	log(("Performing IAck: "));

	if (ps == PS_READY || ps == PS_PASSIVE || ps == PS_COLD_START)
		/* In READY, PASSIVE, COLD START states no checking is done */
		iack_a = IACK_OK;
	else {
		/* get IACK results for both channels */
		check_iack(ctrl, 0, &iack_a, &valid_a);
		check_iack(ctrl, 1, &iack_b, &valid_b);

		/* if one result is invalid, then replace it by the other */
		if (!valid_a && valid_b)
			iack_a = iack_b;
		else
			if (!valid_b && valid_a)
				iack_b = iack_a;

		if (iack_a != iack_b) {
			/* iack results differ -> BIST error */
			cni_bs_set_av(&cni->bist_status, 1);
			raise_bist_error(ctrl);
			return 1;
		}
	}

	log_asis(("%s -> %s\n",
		controller_get_iack_state_name(ploc_data->iack_state), controller_get_iack_state_name(iack_a)));

	/* switch according to previous IACK state */
	/* assume:	A is sender waiting for acknoldgement,
			B is working successor of A
			C is direct succesor of B */
	switch (ploc_data->iack_state) {
		case IACK_OK:
			return 0;
		case IACK_SEND:
			break;
		case IACK_CHECK1:
			if (iack_a == IACK_OK)
				/* B sends correct ack. */
				cni_ps_increment_accept_counter(cni_ps, 1);
			if (iack_a == IACK_CHECK1 && bus_activity(ctrl)) {
				/* B failed */
				cni_ps_increment_failed_counter(cni_ps, 1);
				c_state_set_membership(c_state, medl_sce_get_membership_pos(sce), 0);
				raise_membership_changed_interrupt(ctrl);
				log(("Node B failed.\n"));
			}
			if (iack_a == IACK_CHECK2) {
				/* B sent negative ack. */
				ploc_data->successor = medl_sce_get_membership_pos(sce);
			}
			break;
		case IACK_CHECK2:
			if (iack_a == IACK_ERROR) {
				/* B and C agree that A failed */
				cni_ps_increment_accept_counter(cni_ps, 2);
				c_state_set_membership(c_state, ploc_data->membership_pos, 0);
				c_state_set_membership(c_state, ploc_data->successor, 1);
				c_state_set_membership(c_state, medl_sce_get_membership_pos(sce), 1);
				cni_ps_increment_mfc(cni_ps, 1);

				raise_membership_changed_interrupt(ctrl);
				raise_membership_loss_error(ctrl);
				if (test_membership_error(ctrl))
					return 1;
				log(("B and C agree that A (me) failed.\n"));
				controller_set_state(ctrl, PS_READY);
			}
			if (iack_a == IACK_DOUBLE_FAILURE) {
				/* B says A is wrong, C failed */
				if (bus_activity(ctrl))
					cni_ps_increment_failed_counter(cni_ps, 2);
				else
					cni_ps_increment_failed_counter(cni_ps, 1);
				c_state_set_membership(c_state, ploc_data->successor, 0);
				c_state_set_membership(c_state, medl_sce_get_membership_pos(sce), 0);
				raise_membership_changed_interrupt(ctrl);
				controller_set_state(ctrl, PS_READY);
				log(("B says A (me) is wrong, C failed.\n"));
			}
			if (iack_a == IACK_OK) {
				/* C says B failed, A is OK */
				cni_ps_increment_accept_counter(cni_ps, 1);
				cni_ps_increment_failed_counter(cni_ps, 1);
				c_state_set_membership(c_state, ploc_data->successor, 0);
				raise_membership_changed_interrupt(ctrl);
				log(("C says B failed, A (me) is OK.\n"));
			}
			break;
		case IACK_ERROR:
		case IACK_DOUBLE_FAILURE:
			break;
	}

	ploc_data->iack_state = iack_a;

	return 1;
}


/*****************************************************************************/
/* Exported Functions */
/*****************************************************************************/

/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Pointer to the initialized structure (the same as @a ctrl argument).*/
CONTROLLER *controller_init(CONTROLLER *ctrl)
{
	memset(&ctrl->cni, 0, sizeof(CNI));
	memset(&ctrl->medl, 0, sizeof(MEDL));
	ploc_data_initialize(&ctrl->ploc_data);

	ctrl->bus_guardian = NULL;
	ctrl->channel[0] = NULL;
	ctrl->channel[1] = NULL;
	ctrl->cni.controller = (void *)ctrl;

	ctrl->synchronize_time = NULL;
	ctrl->reset_clk_correction = NULL;
	ctrl->update_cluster_time = NULL;
	ctrl->receive_callback = NULL;
	ctrl->send_callback = controller_send_callback;

	ctrl->host_interrupt = NULL;
	ctrl->host_arg = NULL;

	controller_set_state(ctrl, PS_FREEZE);

	return ctrl;
}


/**.
This function binds the host part of a node to the communication controller using
an interrupt function and one user-defined argument. To bind the CNI interface is
so far in the care of the host - it should use the one allocated by the ctrl.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param arg User defined argument that is passed to the interrupt function every time it is called. 
@param int_fn Interrupt function that is called from the ctrl. Serves to notify the host
	that an interrupt condition was met.*/
void controller_attach_host(CONTROLLER *ctrl, void *arg, PTR_CALLBACK int_fn)
{
	ctrl->host_arg = arg;
	ctrl->host_interrupt = int_fn;
}


/**.
The controller does not create its own instances of channel, it shares a channel
with the other controllers.
\note This function must be called before the ctrl is started!
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel Pointer to an instance of @c CHANNEL or derived stricture.
@param idx Number of the channel: 0 for channel A, 1 for channel B.*/
void controller_attach_channel(CONTROLLER *ctrl, CHANNEL *channel, unsigned idx)
{
	ctrl->channel[idx] = channel;
}


/**.
Pointer to this function is passed to both channels and it is then used to notify the
ctrl that its own transmission finished. When activated, the ctrl stops
the BA signal.
\note The ctrl function depends on this notification, so do not forget to
	register this function by both channels.
@param p_ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_send_callback(void *p_ctrl)
{
	CONTROLLER *ctrl = (CONTROLLER *)p_ctrl;

	ctrl->bus_guardian->arm_signal(ctrl->bus_guardian, 0);
}


/**.
Pointer to this function is passed to the bus guardian and it is then used to notify the
ctrl that it has violated the bus access cycle. When activated, the ctrl
sets appropiate error flags and transits into freeze state.
@param p_ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_bg_error_callback(void *p_ctrl)
{
	CONTROLLER *ctrl = (CONTROLLER *)p_ctrl;

	cni_ps_set_be(&ctrl->cni.protocol_status, 1);
	raise_protocol_error(ctrl);
}


/** Change the current state of the controller.
 * 
 * @param ctrl  Pointer to an instance of #CONTROLLER or derived structure.
 * @param state  Desired state of the controller.*/
void controller_set_state(CONTROLLER *ctrl, unsigned state)
{
	CNI *cni = &ctrl->cni;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;

	if (state_transitions[ploc_data->protocol_state][state])
		ploc_data->protocol_state = state;
	else 
		return;

	cni_ps_set_protocol_state(&cni->protocol_status, state);
	if (state == PS_FREEZE)
		cni_set_co(cni, 0);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
\note This function currently does nothing as the CNI is not mirrored anymore.*/
void controller_update_cni(CONTROLLER *ctrl)
{
    
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
MEDL_SCE *controller_get_current_sce(CONTROLLER *ctrl)
{
	return &ctrl->medl.slot_control_entry[c_state_get_medl_position(&ctrl->cni.c_state)];
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero value if @a ctrl is the current sender, zero otherwise.*/
unsigned controller_is_current_sender(CONTROLLER *ctrl)
{
	MEDL_SCE *sce = controller_get_current_sce(ctrl);

	return logical_name_cmp(&ctrl->ploc_data.logical_name, &sce->logical_name);
}


/**.
This depends on the value of CF flag and I-Frame switch counter.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero value if @a ctrl can enter \e cold-start, zero otherwise.*/
unsigned controller_is_cold_start_allowed(CONTROLLER *ctrl)
{
	MEDL_CCP *ccp = &ctrl->medl.controller_cfg_params;

	return (medl_ccp_get_cf(ccp) == 1)
		&& ((cni_ps_get_i_frame_switch(&ctrl->cni.protocol_status)
			< medl_ccp_get_max_cold_start_entry(ccp))
			|| medl_ccp_get_max_cold_start_entry(ccp) == 0);
}


/**.
TP-synchronous interrupts may be raised only at the begining of a transmission
phase (TP), although the ctrl often detects the cause for such an interrupt 
earlier. In such case, a flag is set within the local data and at the start of next
TP this flag is checked.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_raise_tp_sync_interrupt(CONTROLLER *ctrl)
{
	PLOC_DATA *ploc_data = &ctrl->ploc_data;

	if (ploc_data->tp_sync_interrupt) {
		raise_interrupt(ctrl);
		ploc_data->tp_sync_interrupt = 0;
	}
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param channel Number of the channel - currently there are two channels (A and B) with numbers 0 and 1.
@return Zero if sending is succesfull, nozero otherwise.*/
int controller_send_frame(CONTROLLER *ctrl, unsigned channel)
{
	CHANNEL *chann = ctrl->channel[channel];
	int result;

	ctrl->ploc_data.mtd_a = 0;
	ctrl->ploc_data.mtd_b = 0;

	result = chann->channel_send_frame(chann, (void *)ctrl, ctrl->send_callback);

	if (result != 0)  /* Send error - stop ARM signal */
		ctrl->bus_guardian->arm_signal(ctrl->bus_guardian, 0);

	return result;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if successfull, nonzero otherwise (in such case listening is canceled).*/
int controller_start_receive(CONTROLLER *ctrl)
{
	CHANNEL **channel = (CHANNEL **)ctrl->channel;
	int result;

	result = channel_receive(channel[0], (void *)ctrl, ctrl->receive_callback)
		| channel_receive(channel[1], (void *)ctrl, ctrl->receive_callback);

	if (result)
		controller_stop_receive(ctrl);

	return result;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Duration of listen timeout in macroticks.*/
unsigned controller_get_listen_timeout(CONTROLLER *ctrl)
{
	return medl_ccp_get_startup_timeout(&ctrl->medl.controller_cfg_params)
		+ 2 * medl_ccp_get_tdma_round_length(&ctrl->medl.controller_cfg_params);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Duration of cold start timeout in macroticks.*/
unsigned controller_get_cold_start_timeout(CONTROLLER *ctrl)
{
	return medl_ccp_get_tdma_round_length(&ctrl->medl.controller_cfg_params) +
			medl_ccp_get_startup_timeout(&ctrl->medl.controller_cfg_params);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero code if bus activity was detected, zero otherwise. */
int controller_cold_start_bg_arm(CONTROLLER *ctrl)
{
	if (bus_activity(ctrl)) {
		controller_update_cni(ctrl);
		return 1;
	}

	bus_guardian_set_params((BUS_GUARDIAN *)ctrl->bus_guardian, &ctrl->medl.bus_guardian_cold_start);
	ctrl->bus_guardian->arm_signal(ctrl->bus_guardian, 1);

	controller_update_cni(ctrl);
	return 0;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if successfull, nonzero otherwise (in such case listening is canceled).*/
int controller_bg_arm(CONTROLLER *ctrl)
{
	MEDL_SCE *sce = controller_get_current_sce(ctrl);
	BUS_GUARDIAN *bg = ctrl->bus_guardian;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;

	if (medl_sce_get_ba(sce) == 0)
		return 0;

#ifndef DYN_BG
	if (ploc_data->immediate_send_permission)
		return 0;
#endif

	if (ploc_data->sync_period_params) {
		ploc_data->sync_period_params = 0;
		bus_guardian_set_params((BUS_GUARDIAN *)ctrl->bus_guardian, 
				&ctrl->medl.role_list.bus_guardian);
	}

	return bg->arm_signal(bg, 1);
}


/**.
During the initialization, the following actions are performed:
	- clear CNI, 
	- set null frame vector in CNI
	- set personality field in CNI
	- check MEDL table (CRC)
	- transit into LISTEN state

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_init_state(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;
	unsigned i;

	/* clear DPRAM */
	cni_clear_mem(cni);
	for (i = 0; i < MAX_SRU_COUNT; i++)
		cni->null_frame[i] = 1;
	ploc_data_initialize(&ctrl->ploc_data);
	ctrl->ploc_data.protocol_state = PS_INIT;

	/* Set Personality Field */
	ctrl->ploc_data.logical_name = ctrl->medl.role_list.logical_name;;
	cni->personality.logical_name = ctrl->medl.role_list.logical_name;
	if (medl_validity_check(&ctrl->medl) == 0) {
		cni_ps_set_mc(&cni->protocol_status, 1);
		raise_protocol_error(ctrl);
		return;
	}

	/* store ctrl's membership position */
	i = get_first_sending_slot(ctrl);
	i = medl_sce_get_membership_pos(&ctrl->medl.slot_control_entry[i]);
	ctrl->ploc_data.membership_pos = i;

	cni_ps_set_controller_ready(&cni->protocol_status, CONTROLLER_READY);
	controller_set_state(ctrl, PS_LISTEN);
}


/**.
In the first phase of the listen state these actions are performed:
	- check ctrl's state
	- set BG params for cold start enabled nodes
	- set wake-up conditions for the ctrl (on receive)

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition 
	follows.*/
int controller_prepare_listen_state(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;

	/* test CO field: enter FREEZE state */
	if (cni_get_co(cni) != CONTROLLER_ON) {
		controller_set_state(ctrl, PS_FREEZE);
		return 1;
	}
	/* test CA field: enter AWAIT state */
	if (cni_get_ca(cni) == CONTROLLER_AWAIT) {
		controller_set_state(ctrl, PS_AWAIT);
		return 2;
	}
	/* test BIST enable field: enter SELFTEST state */
	if (cni_be_get_rm(&cni->bist_enable) != 0 || cni_be_get_rf(&cni->bist_enable) != 0) {
		controller_set_state(ctrl, PS_SELFTEST);
		return 3;
	}

	/* Test whether this is a cold-start-allowed node. */
	if (controller_is_cold_start_allowed(ctrl)) {
		/* Set BG Cold Start params */
		bus_guardian_set_params(ctrl->bus_guardian, &ctrl->medl.implementation_params.listen_cold_start_period_params);
		ctrl->bus_guardian->arm_signal(ctrl->bus_guardian, 1);
	}

	return 0;
}


/**.
In the last phase of the listen state these actions are performed:
	- stop listening for frames on both channels
	- check whether received data is an I-Frame
	- synchronize on a valid I-Frame
	- enter the cold start state if cold start is enabled and the bus 
	is silent 

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_finish_listen_state(CONTROLLER *ctrl)
{
	unsigned channel;

	controller_stop_receive(ctrl);

	if (cni_get_co(&ctrl->cni) != CONTROLLER_ON) {
		controller_set_state(ctrl, PS_FREEZE);
		return;
	}

	if (bus_activity(ctrl)) {
		/* bus active: check for a valid I-frame on both channels */
		channel = get_valid_i_frame(ctrl);
		if (channel == 0 || channel == 1) {
			log(("Got valid I-Frame on channel %d.\n", channel));
			synchronize_on_i_frame(ctrl, channel);
			return;		/* Synchronized (if not multiplexed) */
		}
	} else
		if (controller_is_cold_start_allowed(ctrl) && test_host_lifesign(ctrl)) {
			controller_set_state(ctrl, PS_COLD_START);
			return;		/* Enter Cold Start mode */
		}

	log(("No valid I-Frame Received\n"));
}


/**.
In the first phase of the cold start state these actions are performed:
	- check ctrl's state
	- initialize C-State
	- send one cold start frame and increment the I-Frame switch

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition follows.*/
int controller_prepare_cold_start_state(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;
	unsigned channel;

	/* test CO field: enter FREEZE state */
	if (cni_get_co(cni) != CONTROLLER_ON) {
		controller_set_state(ctrl, PS_FREEZE);
		return 1;
	}
	/* Determine which channel to use - LSB in I-frame switch. */
	channel = cni_ps_get_i_frame_switch(&cni->protocol_status) & 1;
	/* Prepare C-State. */
	setup_cold_start_c_state(ctrl);
	setup_cold_start_i_frame(ctrl, channel);

	log(("Cold Start I-Frame: Cluster Time = %u, C-State Time = %u\n",
		cni_get_cluster_time(cni), c_state_get_time(&cni->c_state)));

	/* Send I-Frame on appropriate channel. */
	if (controller_send_frame(ctrl, channel))
		;   /* FIXME: handle error scenario */
	cni_ps_increment_i_frame_switch(&cni->protocol_status, 1);

	/* Wait for next slot and setup BG */
	controller_update_cni(ctrl);

	log(("Starting cold start timeout.\n"));
	return 0;
}


/**.
This function is called to update the internal state of the ctrl after it
	wakes from a waiting state within the cold start state. The duration of the 
	sleep is given in @a macroticks. The following fields are updated:
	- MEDL position.
	- C-State - MEDL position and Time fields

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param macroticks The duration of cold start state that should be executed in 
	macroticks.*/
void controller_run_cold_start_state(CONTROLLER *ctrl, int macroticks)
{
	unsigned i;
	unsigned duration;

	i = c_state_get_medl_position(&ctrl->cni.c_state);
	while (macroticks > 0) {
		duration = medl_sce_get_sru_slot_duration(&ctrl->medl.slot_control_entry[i]);
		macroticks -= duration;
		controller_set_next_slot(ctrl);
		/* Advance through medl slot control entries */
		if (medl_sce_get_eot(&ctrl->medl.slot_control_entry[i]))
			i = 0;
		else
			i++;
	}
}


/**.
In the last phase of the cold start state these actions are performed:
	- stop listening on both channels
	- disable all wake-up condition
	- update time field and MEDL position in the C-State
	- check whether a correct I-Frame was received and possibly use it to synchronize with the cluster

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if an I-Frame was received and used for synchronization, nonzero otherwise.*/
int controller_finish_cold_start_state(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;
	CNI_PS *cni_ps = &cni->protocol_status;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	unsigned channel;

	controller_stop_receive(ctrl);

	if (cni_get_co(cni) != CONTROLLER_ON) {
		controller_set_state(ctrl, PS_FREEZE);
		return 1;
	}

	if (bus_activity(ctrl)) {
		channel = cold_start_get_valid_i_frame(ctrl);
		if (channel == 0 || channel == 1) {
			/* received correct I-Frame */
			cni_ps_set_i_frame_switch(cni_ps, 0);
			cni->c_state = *frame_get_c_state(&ploc_data->frame[channel]);
			cni_ps_set_c_state_valid(cni_ps, C_STATE_VALID);
			cni_ps_increment_accept_counter(cni_ps, 1);
			cni_set_null_frame(cni, medl_sce_get_membership_pos(
						controller_get_current_sce(ctrl)), 0);
			cni_set_null_frame(cni, ploc_data->membership_pos, 0);

#ifdef DYN_BG
			conf_dyn_bg(ctrl);
#else
			/* Set Synchronization Params for BG */
			bus_guardian_set_params((BUS_GUARDIAN *)ctrl->bus_guardian,
				&ctrl->medl.implementation_params.synchronization_period_params);
			ctrl->ploc_data.immediate_send_permission = 1;
			ctrl->ploc_data.sync_period_params = 1;
#endif

			controller_set_state(ctrl, PS_STARTUP);
			/* Interrupts */
			raise_mode_changed_interrupt(ctrl);
			raise_membership_changed_interrupt(ctrl);
			return 0;	/* synchronized */
		}
		else {
			/* no valid I-Frame received */
			log(("No valid I-Frame received\n"));
			controller_set_state(ctrl, PS_LISTEN); /* back to listen */
		}
	}
	else {
		/* no bus activity detected */
		log(("No bus activity detected\n"));
		if (!controller_is_cold_start_allowed(ctrl))
			controller_set_state(ctrl, PS_LISTEN);
	}
	return 2;
}


/**.
The function first clears the @c controller_ready \c c_state_valid flags. Then it
continuously checks (or uses a wake-up condition) the value of the @c co field in the
CNI. If this field is set to @c CONTROLLER_ON, a state transition is performed into the
init state.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if ctrl is switched on (transition into init state), nonzero code 
	otherwise (ctrl remains in the freeze state).*/
unsigned controller_freeze_state(CONTROLLER *ctrl)
{
	CNI_PS *cni_ps = &ctrl->cni.protocol_status;

	controller_stop_receive(ctrl);
	channel_cancel_send(ctrl->channel[0], ctrl);
	channel_cancel_send(ctrl->channel[1], ctrl);
	if (cni_get_co(&ctrl->cni) == CONTROLLER_ON) {
		controller_set_state(ctrl, PS_INIT);
		return 0;
	}

	cni_ps_set_controller_ready(cni_ps, 0);
	cni_ps_set_c_state_valid(cni_ps, 0);
	controller_update_cni(ctrl);
	return 1;
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
\todo Download is not implemented.*/
void controller_download_state(CONTROLLER *ctrl)
{
	cni_ps_set_dc(&ctrl->cni.protocol_status, 1);
	controller_set_state(ctrl, PS_FREEZE);
	raise_protocol_error(ctrl);
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@param idx Number of the timer - currently there are two timers with numbers 0 and 1.*/
void controller_raise_timer_interrupt(CONTROLLER *ctrl, unsigned idx)
{
	CNI *cni = &ctrl->cni;

	switch (idx) {
	case 0:
		log(("TI1 detected.\n"));
		cni_is_set_ti1(&cni->interrupt_status, 1);
		if (cni_ie_get_ti1(&cni->interrupt_enable) == 1)
			raise_interrupt(ctrl);
		break;
	case 1:
		log(("TI2 detected.\n"));
		cni_is_set_ti2(&cni->interrupt_status, 1);
		if (cni_ie_get_ti2(&cni->interrupt_enable) == 1)
			raise_interrupt(ctrl);
	}
}


/**.
This function is called during the second part of IFG, just before the transmission
phase (TP). When the ctrl is not the current sender, zero is returned. If
the ctrl is the current sender however, the following actions are performed
before a nonzero code is returned:
	- Check frame counters (clique avoidance, blackout detection) and clear them
	- Check host life-sign
	- Prepare frames to be sent in ctrl's local data

\note The function may switch the internal state fo the ctrl (to freeze, passive, ready).
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Nonzero code if a frame is prepared and should be sent, zero code if 
	the ctrl should not transmitt in this slot.*/
unsigned controller_psp(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;
	MEDL_SCE *sce = controller_get_current_sce(ctrl);
	CNI_PS *cni_ps = &cni->protocol_status;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;

	if (is_my_slot(ctrl, sce)) {
		if (test_communication_blackout(ctrl))
			return 1;
		if (test_ack_error(ctrl))
			return 2;
		/* reset frame counters */
		cni_ps_set_failed_counter(cni_ps, 0);
		cni_ps_set_invalid_counter(cni_ps, 0);
		cni_ps_set_accept_counter(cni_ps, 0);
	}

	if (!controller_is_current_sender(ctrl))
		return 3;

	if (!test_host_lifesign(ctrl)) {
		if (c_state_get_membership(&cni->c_state, ploc_data->membership_pos)) {
			c_state_set_membership(&cni->c_state, ploc_data->membership_pos, 0);
			log(("Membership lost due to missing host lifesign\n"));
			raise_membership_loss_error(ctrl);
			raise_membership_changed_interrupt(ctrl);
		}
		log(("No Host Lifesign\n"));
		controller_set_state(ctrl, PS_PASSIVE);
		cni_mcr_clear(cni);
		return 4;
	} else {
		if (cni_ps_get_protocol_state(cni_ps) == PS_PASSIVE) {
			clear_host_error_flags(ctrl);
			log(("Host is active.\n"));
			controller_set_state(ctrl, PS_READY);
		}
	}

	if (cni_ps_get_protocol_state(cni_ps) == PS_READY && !is_reconfiguration_allowed(ctrl, sce)) {
		cni_mcr_clear(cni);
		return 5;
	}

	if (prepare_frame(ctrl))
		return 6;
	ploc_data->iack_state = IACK_SEND;

	if (cni_ps_get_protocol_state(cni_ps) == PS_READY) {
		log(("Reintegration from READY state.\n"));
		if (in_startup_mode(ctrl))
			controller_set_state(ctrl, PS_STARTUP);
		else
			controller_set_state(ctrl, PS_APPLICATION);
	}
#ifndef DYN_BG
	if (ploc_data->immediate_send_permission) {
		log(("Immediate send permission: Raise BA signal.\n"));
		ploc_data->immediate_send_permission = 0;
		if (ctrl->bus_guardian->arm_signal(ctrl->bus_guardian, 1))
			return 7;
	}
#endif
	return 0;
}


/**.
This function is called during the first part of IFG regardless whether a frame was
received or not. The following actions are performed:
	- cancel listening for both channels
	- check received frames (null, invalid, valid, correct) and update the EIF fields in CNI
	- execute the implicit ackownledgement algorithm
	- update membership vector and null frame vector
	- eventualy store data from N-frames into CNI

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_prp(CONTROLLER *ctrl)
{
	FRAME *frame = ctrl->ploc_data.frame;
	MEDL_MAE *medl_mae = get_current_mae(ctrl);
	MEDL_SCE *sce = controller_get_current_sce(ctrl);
	unsigned membership_pos = medl_sce_get_membership_pos(sce); /* memb.pos. of actual sender */
	CNI *cni = &ctrl->cni;
	C_STATE *c_state = &cni->c_state;
	unsigned old_membership = c_state_get_membership(c_state, membership_pos);
	CNI_MAE *cni_mae;
	CNI_EIF_VALUES eif_a, eif_b;
	unsigned mode = 0;
	unsigned immediate = 0;
	unsigned iack_in_progress;

	/* Cancel pending receive. */
	controller_stop_receive(ctrl);

	/* We assume that the current sender is valid. */
	c_state_set_membership(c_state, membership_pos, 1);
	iack_in_progress = perform_iack(ctrl);

	/* Handle frame on channel A */
	eif_a = get_frame_eif(ctrl, 0);

	if ((eif_a & EIF_CORRECT_FRAME) == EIF_CORRECT_FRAME) {
		immediate = frame_header_get_immediate(&frame[0].header);
		mode = frame_header_get_mode(&frame[0].header);
	}
	if (medl_mae_get_ia(medl_mae) == 0) {	/* N-Frame */
		cni_mae = get_cni_mae(ctrl, 0);
		if (!enter_msg_nbw(ctrl, cni_mae, 0)) {
			if ((eif_a & EIF_CORRECT_FRAME) == EIF_CORRECT_FRAME)
				store_frame_data(ctrl, cni_mae, 0);
			cni_mae_set_eif(cni_mae, eif_a);
			leave_msg_nbw(ctrl, cni_mae, 0);
		}
	}
	cni_fd_set_eif_cha(&cni->frame_diagnosis, eif_a);

	/* Handle frame on channel B */
	eif_b = get_frame_eif(ctrl, 1);
	if ((eif_b & EIF_CORRECT_FRAME) == EIF_CORRECT_FRAME) {
		immediate = frame_header_get_immediate(&frame[1].header);
		mode = frame_header_get_mode(&frame[1].header);
	}
	if (medl_mae_get_ib(medl_mae) == 0) {	/* N-Frame */
		cni_mae = get_cni_mae(ctrl, 1);
		if (!enter_msg_nbw(ctrl, cni_mae, 1)) {
			if ((eif_b & EIF_CORRECT_FRAME) == EIF_CORRECT_FRAME)
				store_frame_data(ctrl, cni_mae, 1);
			cni_mae_set_eif(cni_mae, eif_b);
			leave_msg_nbw(ctrl, cni_mae, 1);
		}
	}
	cni_fd_set_eif_chb(&cni->frame_diagnosis, eif_b);

	/* update membership and null-frame vector */
	cni_set_null_frame(cni, membership_pos, 0);
	if (((eif_a | eif_b) & EIF_CORRECT_FRAME) == 0) {
		/* No correct frame received -> sending SRU loses it's membership */
		c_state_set_membership(&cni->c_state, membership_pos, 0);
		log(("Node %u loses its membership - no correct frame arrived\n", membership_pos));
		if (((eif_a | eif_b) & EIF_CRC_ERROR) == 0) {
			if (((eif_a | eif_b) & EIF_INVALID_FRAME) == 0) {
				/* Null frame on both channels */
				cni_set_null_frame(cni, membership_pos, 1);
			} else {
				/* No valid frame received */
				if (!iack_in_progress)
					cni_ps_increment_invalid_counter(&cni->protocol_status, 1);
			}
		}
		else {
			if (!iack_in_progress)
				cni_ps_increment_failed_counter(&cni->protocol_status, 1);
		}
	}
	else {
		if (!iack_in_progress)
			cni_ps_increment_accept_counter(&cni->protocol_status, 1);
	}

	if (old_membership != c_state_get_membership(c_state, membership_pos))
		raise_membership_changed_interrupt(ctrl);

	if (check_mode_change_request(ctrl, mode, immediate))
		perform_mode_change_request(ctrl, mode, immediate);
}


/**.
This function is called at every transmission phase start (IFG end) and performs
the following tasks:
	- send frames in case the ctrl is current sender (see @c ::controller_psp function)
	- handle mode change requests

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition follows.*/
int controller_tp_start(CONTROLLER *ctrl)
{
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	C_STATE *c_state = &ctrl->cni.c_state;
	int result;

	log(("TP Start.\n"));
	if (controller_leave_nbw(ctrl))
		return 1;

	if (c_state_get_membership(c_state, ploc_data->membership_pos) == 0) {
		log(("Setting membership vector.\n"));
		c_state_set_membership(c_state, ploc_data->membership_pos, 1);
		raise_membership_changed_interrupt(ctrl);
	}

	result = controller_send_frame(ctrl, 0);
	result = result || controller_send_frame(ctrl, 1);

	return result;
}


/**.
This function is called during the idle phase of every IFG and performs these tasks:
	- increment MEDL position modulo transmission cycle length
	- increment cluster time by slot duration (so that it will point to next action time)
	- perform deferred mode changes when end of cluster cycle is detected
	- raise User interrupts when the appropiate condition is met

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_set_next_slot(CONTROLLER *ctrl)
{
	C_STATE *c_state = &ctrl->cni.c_state;
	MEDL_SCE *old_sce = controller_get_current_sce(ctrl);
	MEDL_MAE *old_mae = get_current_mae(ctrl);
	unsigned dmc = c_state_get_dmc(c_state);
	unsigned mode;

	if (dmc > 0 && medl_mae_get_eoc(old_mae)) {
		/* Perform deffered mode change */
		mode = medl_mce_get_deferred_mode(
			&ctrl->medl.mode_control_entry[c_state_get_mode(c_state)],
			dmc - 1);
		if (mode != INVALID_DEFERRED_MODE) {
			log(("Deferred mode change, mode %d.\n", mode));
			set_mode(ctrl, mode);
		} /* FIXME: What happens if invalid mode change request is stored? */
		c_state_set_dmc(c_state, 0);
		c_state_set_medl_position(c_state, 0);
	} else
		/* Increment medl_position and test End Of Transmission flag: */
		if (medl_sce_get_eot(old_sce) == 0)
			c_state_increment_medl_position(c_state, 1);
		else
			c_state_set_medl_position(c_state, 0);

	/* Increment C-State time */
	c_state_set_time(c_state, bit_mask(16) & (
		c_state_get_time(c_state) + medl_sce_get_sru_slot_duration(old_sce)));

	/* Raise user interrupts */
	if (medl_mae_get_ru1(get_current_mae(ctrl)))
		raise_user_interrupt(ctrl, 0);
	if (medl_mae_get_ru2(get_current_mae(ctrl)))
		raise_user_interrupt(ctrl, 1);
}


/**.
The function first checks the MEDL table and when the @c cs flag is not set, it
returns with zero code. If clock symchronization is permitted it uses the stored 
time differences to calculate a new clock correction term. This term is stored in
the @c csct field in \c CNI for further use by the clock correction mechanism.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition follows.*/
unsigned controller_perform_cs(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;
	CNI_FD *cni_fd = &cni->frame_diagnosis;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	MEDL_SCE *sce = controller_get_current_sce(ctrl);
	int i, min, max, dif_a, dif_b, sum;

	if (medl_sce_get_syf(sce) == 1) {
		/* Clock synchronization frame */
		i = sum = 0;
		if (cni_fd_get_eif_cha(cni_fd) & EIF_CORRECT_FRAME) {
			dif_a = ploc_data->mtd_a;
			dif_a -= medl_sce_get_delay_correction(sce);
			cni_fd_set_mtd_cha(cni_fd, dif_a);
			sum += dif_a;
			log(("SYF on channel A: MTD = %d\n", dif_a));
			i++;
		}
		if (cni_fd_get_eif_chb(cni_fd) & EIF_CORRECT_FRAME) {
			dif_b = ploc_data->mtd_b;
			dif_b -= medl_sce_get_delay_correction(sce);
			cni_fd_set_mtd_chb(cni_fd, dif_b);
			sum += dif_b;
			log(("SYF on channel B: MTD = %d\n", dif_b));
			i++;
		}

		if (i > 0) /* At least one correct frame */
			ploc_data_push_sync_stack(ploc_data, sum / i);

		log(("CSA Stack: [%+3d -> %+3d -> %+3d -> %+3d]\n",
			ploc_data->sync_stack[(ploc_data->sync_stack_top + 3) % 4],
			ploc_data->sync_stack[(ploc_data->sync_stack_top + 2) % 4],
			ploc_data->sync_stack[(ploc_data->sync_stack_top + 1) % 4],
			ploc_data->sync_stack[(ploc_data->sync_stack_top + 0) % 4]
		));
	}
	
	if (medl_sce_get_cs(sce) == 1) {
		/* Perform clock correction - compute clock state correction term (CSCT) */
		min = max = 0;
		sum = ploc_data->sync_stack[0];
		for (i = 1; i < 4; i++) {
			if (ploc_data->sync_stack[min] > ploc_data->sync_stack[i])
				min = i;
			else if (ploc_data->sync_stack[max] < ploc_data->sync_stack[i])
				max = i;
			sum += ploc_data->sync_stack[i];
		}

		/* throw maximum and minum difference away */
		sum = sum - ploc_data->sync_stack[min] - ploc_data->sync_stack[max];
		sum = - sum / 2;
		sum += cni_get_external_rate_correction(cni);

		log(("External rate correction = %d\n",	cni_get_external_rate_correction(cni)));
		log(("Clock synchronization: CSCT = %d\n", sum));
		if (test_synchronization_error(ctrl, abs(sum)))
			return 1;

        ploc_data->csct = sum;
		cni_ps_set_csct(&cni->protocol_status, sum);
		ctrl->reset_clk_correction(ctrl);
	}
	return 0;
}


/**.
The function first checks whether the CCF field is even - if not, then an assertion 
violation error (BIST error) is raised and a nonzero code is returned. If the CCF is
valid, it is incremented by 1 (locked), the new value is stored within an internal
variable and zero is returned.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition follows.*/
int controller_enter_nbw(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;

	log(("Entering NBW protected section\n"));
	if (cni_get_ccf(cni) & 1) {
		cni_bs_set_av(&cni->bist_status, 1);
		cni_set_assertion_code(cni, AVC_INVALID_CCF_VALUE);
		raise_bist_error(ctrl);
		return 1;
	}

	ctrl->ploc_data.old_ccf = cni_increment_ccf(cni, 1);
	return 0;
}


/**.
The function first compares the current value of CCF with a stored value and if 
it is different, it raises an assertion violation error (BIST error) and returns 
nonzero code. Otherwise the value is incremented by one (unlocked) and zero code 
is returned.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition 
	follows.*/
int controller_leave_nbw(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;

	log(("Leaving NBW protected section\n"));
	if (cni_get_ccf(cni) != ctrl->ploc_data.old_ccf) {
		cni_bs_set_av(&cni->bist_status, 1);
		cni_set_assertion_code(cni, AVC_INVALID_CCF_VALUE);
		raise_bist_error(ctrl);
		return 1;
	}

	cni_increment_ccf(cni, 1);
	return 0;
}


/**.
This function raises a watchdog error (BIST error) and switches the ctrl into
the freeze state. It is called by an independent process that checks whether the
ctrl is alive.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_run_watch_dog(CONTROLLER *ctrl)
{
	log(("Watchdog activated: Raising hell :-)\n"));
	cni_bs_set_wd(&ctrl->cni.bist_status, 1);
	raise_bist_error(ctrl);
}


/**.
It is assumed, that this function is called at transmission phase start. It verifies
that the internal state of the ctrl is consistent and agrees with this 
assumption - if not an assertion violation (BIST) error is raised and the ctrl
transits into the freeze state.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if everything works, nonzero otherwise - in such case a state transition follows.*/
int controller_tp_sync_assertion(CONTROLLER *ctrl)
{
	CNI *cni = &ctrl->cni;

	if (cni_get_cluster_time(cni) == c_state_get_time(&cni->c_state))
		return 0;

	cni_bs_set_av(&cni->bist_status, 1);
	cni_set_assertion_code(cni, AVC_TP_SYNC_ERROR);
	raise_bist_error(ctrl);
	return 1;
}


/**.
This function is called everytime the current protocol state field is accessed to
verify that is contains a valid value. For this verification the ctrl contains
an internal copy of this field that can be used for comparison. Three different 
scenarios are possible:
	- The protocol state is unchanged: do nothing and return 0.
	- The protocol state changed and the transition is valid: update the internal copy and return 0.
	- The protocol state changed but the transition is invalid: restore the protocol state and return 1.

@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.
@return Zero if no problem was found, nonzero otherwise.*/
int controller_check_state(CONTROLLER *ctrl)
{
	unsigned old_state = ctrl->ploc_data.protocol_state;
	unsigned csim_state = cni_ps_get_protocol_state(&ctrl->cni.protocol_status);

	if (state_transitions[old_state][csim_state]) {
		ctrl->ploc_data.protocol_state = csim_state;
		return 0;
	}

	cni_ps_set_protocol_state(&ctrl->cni.protocol_status, old_state);
	return 1;
}


/**.
@param iack Numerical value of an IACK (implicit acknowledgement) state.
@return A descriptive string for the given IACK state or NULL when the state is invalid.*/
char *controller_get_iack_state_name(unsigned iack)
{
	switch (iack) {
		case IACK_OK:             return "OK         ";
		case IACK_SEND:           return "SEND       ";
		case IACK_CHECK1:         return "CHECK1     ";
		case IACK_CHECK2:         return "CHECK2     ";
		case IACK_ERROR:          return "ERROR      ";
		case IACK_DOUBLE_FAILURE: return "DBL_FAILURE";
	}
	return "Unknown!";
}


/**.
@param ctrl Pointer to an instance of @c CONTROLLER or derived structure.*/
void controller_stop_receive(CONTROLLER *ctrl)
{
	log(("Receiving canceled\n"));
	channel_cancel_receive((CHANNEL *)ctrl->channel[0], (void *)ctrl);
	channel_cancel_receive((CHANNEL *)ctrl->channel[1], (void *)ctrl);
}


/*****************************************************************************/

/*@}*/
/*@}*/

