/**@addtogroup frame
 * @{*/

/**@addtogroup frame_int  Implementation
 * @{*/


/**@file
 * @brief Definition of N-frame and I-frame.
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */


#include "ttpc/fl1debug.h"
#include "ttpc/fp1frame.h"
#include "ttpc/fp1const.h"
#include "ttpc/fl1func.h"
#include "ttpc/fp1cstat.h"


/************************************************************************************/
/* Static Functions */
/************************************************************************************/

/** Calculate and return the CRC of an I-Frame.*/
static unsigned i_frame_calculate_crc(FRAME *frame);

/** Return the physical I-Frame size in bits.*/
static unsigned i_frame_get_length(FRAME *frame);

/** Calculate and return the CRC of N-frame concated with given C-state.*/
static unsigned n_frame_calculate_crc(FRAME *frame, C_STATE *c_state, unsigned cs_length);

/** Return the physical N-Frame size in bits.*/
static unsigned n_frame_get_length(FRAME *frame);


/************************************************************************************/

/**.
The CRC calculation in this function is completely standalone - it uses the default
initialization value and determines the right polynimial to use by the physical length
of frame.
@param frame Pointer to a frame instance.
@return 16-bit CRC calculated over all used fields of the frame.*/
static unsigned i_frame_calculate_crc(FRAME *frame)
{
	unsigned crc_poly, crc;
	FRAME_I_DATA *i_data = &frame->type.i;

	crc_poly = ((8 + 16 * i_data->c_state_length) < MAX_LENGTH_FOR_CRC_POLY_1)
		? CRC_POLYNOMIAL_1 : CRC_POLYNOMIAL_2;
	crc = CRC_STARTING_VALUE;

	frame_header_chain_crc(&frame->header, &crc, crc_poly);
	c_state_chain_crc(&i_data->c_state, &crc, crc_poly, i_data->c_state_length);
	get_crc(&crc, crc_poly, 0, 16);

	return crc;
}


/**.
@param frame Pointer to a frame instance that contains an I-Frame.
@return The physical size of an I-frame in bits.*/
static unsigned i_frame_get_length(FRAME *frame)
{
	FRAME_I_DATA *i_data = &frame->type.i;

	return 8 + 16 * i_data->c_state_length + 16;
}


/**.
The CRC calculation in this function is completely standalone - it uses the default
initialization value and determines the right polynimial to use by the physical length
of frame.
@param frame Pointer to a frame instance that contains a N-Frame.
@param c_state Pointer to the C-State instance that should be added to the frame before CRC calculation.
@param cs_length Length of the C-State in words.
@return 16-bit CRC calculated over the frame and C-State.*/
static unsigned n_frame_calculate_crc(FRAME *frame, C_STATE *c_state, unsigned cs_length)
{
	FRAME_N_DATA *n_data = &frame->type.n;
	unsigned crc_poly, crc;
	unsigned i;

	/* Determine which polynom to use. */
	crc_poly = ((8 + 16 * cs_length + (n_data->app_data_length + 1) * 8) < MAX_LENGTH_FOR_CRC_POLY_1)
		? CRC_POLYNOMIAL_1 : CRC_POLYNOMIAL_2;
	crc = CRC_STARTING_VALUE;

	frame_header_chain_crc(&frame->header, &crc, crc_poly);
	for (i = 0; i < n_data->app_data_length + 1; i++)
		get_crc(&crc, crc_poly, n_data->app_data[i], 8);
	c_state_chain_crc(c_state, &crc, crc_poly, cs_length);
	get_crc(&crc, crc_poly, 0, 16);

	return crc;
}


/**.
@param frame Pointer to a frame instance that contains a N-Frame.
@return Physical size of N-frame in bits.*/
static unsigned n_frame_get_length(FRAME *frame)
{
	FRAME_N_DATA *n_data = &frame->type.n;

	return 8 + 8 * (n_data->app_data_length + 1) + 16;
}


/************************************************************************************/
/* Exported Functions */
/************************************************************************************/

/**.
The CRC code is calculated over all fields within the header structure in 
the order in which the field are physically stored.
@param header Pointer to a frame header.
@param crc Pointer to an integer that contains either a partial CRC or the initial constant.
@param poly Numerical representation of the polynomial to use.
\retval crc Pointer to the CRC code generated for this frame header.*/ 
void frame_header_chain_crc(FRAME_HEADER *header, unsigned *crc, unsigned poly)
{
	get_crc(crc, poly, frame_header_get_i_n_bit(header), 1);
	get_crc(crc, poly, frame_header_get_mode_change(header, 0), 1);
	get_crc(crc, poly, frame_header_get_mode_change(header, 1), 1);
	get_crc(crc, poly, frame_header_get_mode_change(header, 2), 1);
	get_crc(crc, poly, 0, 4);
}


/**.
@param header Pointer to a frame header.
@return The mode number stored within the mode change request field in frame header.*/
unsigned frame_header_get_mode(FRAME_HEADER *header)
{
	return (frame_header_get_mode_change(header, 2) << 1)
		+ frame_header_get_mode_change(header, 1);
}


/**.
@param header Pointer to a frame header.
@param mode Mode number that will be stored within the mode chage request field.*/
void frame_header_set_mode(FRAME_HEADER *header, unsigned mode)
{
	frame_header_set_mode_change(header, 1, mode & 1);
	frame_header_set_mode_change(header, 2, (mode >> 1) & 1);
}


/**.
@param header Pointer to a frame header.
@return The value of immediate flag within mode change request field.*/
unsigned frame_header_get_immediate(FRAME_HEADER *header)
{
	return frame_header_get_mode_change(header, 0);
}


/**.
@param header Pointer to a frame header.
@param imm Value of immediate flag that will be stored within the mode change request field.*/
void frame_header_set_immediate(FRAME_HEADER *header, unsigned imm)
{
	frame_header_set_mode_change(header, 0, imm);
}


/**.
@param frame Pointer to a frame instance that contains any frame type.
@return Physical size of the frame in bits.*/
unsigned frame_get_length(FRAME *frame)
{
	unsigned type = frame_header_get_i_n_bit(&frame->header);

	if (type == FRAME_TYPE_N)
		return n_frame_get_length(frame);
	else
		return i_frame_get_length(frame);
}


/**.
@param frame Pointer to a frame instance that contains I-Frame.
@return Nonzero code if the frame is complete I-Frame with a correct CRC, nonzero otherwise.*/
unsigned frame_is_valid_i_frame(FRAME *frame)
{
	unsigned type = frame_header_get_i_n_bit(&frame->header);

	if (type != FRAME_TYPE_I || frame->state != FRAME_STATE_COMPLETE)
		return 0;
	else
		return (frame_get_crc(frame) == i_frame_calculate_crc(frame));
}


/**.
@param frame Pointer to a frame instance that contains N-Frame.
@param c_state Pointer to the C-State instance that should be added to the frame before CRC calculation.
@param cs_length Length of the C-State in words.
@return Nonzero code if the frame is complete N-Frame with a correct CRC, nonzero otherwise.*/
unsigned frame_is_valid_n_frame(FRAME *frame, C_STATE *c_state, unsigned cs_length)
{
	unsigned type = frame_header_get_i_n_bit(&frame->header);

	if (type != FRAME_TYPE_N || frame->state != FRAME_STATE_COMPLETE)
		return 0;
	else
		return (frame_get_crc(frame) == n_frame_calculate_crc(frame, c_state, cs_length));
}


/**.
@param frame Pointer to a frame instance that contains any type of frame.
@param c_state Pointer to the C-State instance that should be added to an N-Frame before CRC calculation.
@param cs_length Length of the C-State in words.
@return Nonzero code if the frame is complete with a correct CRC, nonzero otherwise.*/
unsigned frame_is_valid(FRAME *frame, C_STATE *c_state, unsigned cs_length)
{
	unsigned type = frame_header_get_i_n_bit(&frame->header);

	if (frame->state != FRAME_STATE_COMPLETE)
		return 0;

	if (frame_calculate_crc(frame, c_state, cs_length) != frame_get_crc(frame))
		return 0;

	if (type == FRAME_TYPE_I)
		return c_state_cmp(&frame->type.i.c_state, c_state, cs_length);

	return 1;
}

/**.
@param frame Pointer to a frame instance that contains any type of frame.
@param c_state Pointer to the C-State instance that should be added to an N-Frame before CRC calculation.
@param cs_length Length of the C-State in words.
@return 16-bit CRC calculated over the frame and C-State (implicit for I-Frames, explicit for N-Frames).*/
unsigned frame_calculate_crc(FRAME *frame, C_STATE *c_state, unsigned cs_length)
{
	unsigned type = frame_header_get_i_n_bit(&frame->header);

	if (type == FRAME_TYPE_N)
		return n_frame_calculate_crc(frame, c_state, cs_length);
	else
		return i_frame_calculate_crc(frame);
}


/************************************************************************************/
/* Generated Functions */
/************************************************************************************/

implement_access_functions(FRAME_HEADER, frame_header, i_n_bit, 1)
implement_index_access_functions(FRAME_HEADER, frame_header, mode_change, 1)
implement_access_functions(FRAME_I_DATA, frame_i_data, c_state_length, 4)
implement_access_functions(FRAME_N_DATA, frame_n_data, app_data_length, 4)
implement_index_access_functions(FRAME_N_DATA, frame_n_data, app_data, 8)
implement_access_functions(FRAME, frame, crc, 16)

/************************************************************************************/

/*@}*/
/*@}*/

