/**@addtogroup medlrd
 * @{*/

/**@addtogroup medlrd_int  Implementation
 * @{*/


/**@file
 * @brief Provides a function that can read MEDL stored in a text file
 *  (implementation).
 * @author Dan Trescak, University of West Bohemia, Czech Republic
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @version 1.5
 * @date Last modified 2002-11-16
 * 
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project \b FIT - Fault Injection for TTA, IST-1999-10748
 */


#include <stdio.h>
#include <string.h>

#include "ttpc/fl1debug.h"
#include "ttpc/fp1mdlrd.h"
#include "ttpc/fp1medl.h"
#include "ttpc/fp1const.h"

/*************************************************************************************/

/** Maximum accepted line length when reading the input file.*/
#define MAX_LINE_LEN		1024

/** Maximum accepted file name length (when generating actual name from pattern).*/
#define MAX_FILE_NAME_LEN	128

/** Definition of possible values read from the file as pointers to different types.*/
typedef enum {
	UNSIGNED_PTR = 1,     /**< Pointer to \c unsigned.*/
	DOUBLE_PTR = 2        /**< Pointer to \c double.*/
} PTR_TYPE;


/** Return with given errocode from function body and close open files.
\param error Error code returned as function value.
\attention Tailored for one specific function.*/
#define error(code) do { \
	if (f != NULL) \
		fclose(f); \
	return (code); \
} while(0)

/** Check given condition and if not true then return with the specified error code.
\param code Error code returned when the check fails.
\param cond Expression that evaluates to an integer value.
\attention Tailored for one specific function.*/
#define CHECK(code, cond) if (!(cond)) error(code)


/*************************************************************************************/
/**\name Static Variables*/
/*\{*/

/** Copy of the last found valid indentifier (used for name completiotion).*/
static char last_valid_id[MAX_LINE_LEN];

/*\}*/
/*************************************************************************************/
/*************************************************************************************/
/**\name Static Functions*/
/*\{*/

/** Expand given field identifier into an absolute field name.*/
static int expand(char *s);

/** Return pointer to the name part found after the first dot.*/
static char *get_second_part(char *str);

/** Remove the name part after the first dot.*/
static char *get_first_part(char *str);

/** Find named field position in MEDL and type of the field.*/
static int get_ptr_MEDL(MEDL *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_BGP and type of the field.*/
static int get_ptr_MEDL_BGP(MEDL_BGP *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_MAE and type of the field.*/
static int get_ptr_MEDL_MAE(MEDL_MAE *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_SCE and type of the field.*/
static int get_ptr_MEDL_SCE(MEDL_SCE *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_ISP and type of the field.*/
static int get_ptr_MEDL_ISP(MEDL_ISP *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_MCE and type of the field.*/
static int get_ptr_MEDL_MCE(MEDL_MCE *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_CCP and type of the field.*/
static int get_ptr_MEDL_CCP(MEDL_CCP *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in MEDL_RL and type of the field.*/
static int get_ptr_MEDL_RL(MEDL_RL *s, void **ptr, PTR_TYPE *type, char *item);

/** Find named field position in LOGICAL_NAME and type of the field.*/
static int get_ptr_LOGICAL_NAME(LOGICAL_NAME *s, void **ptr, PTR_TYPE *type, char *item);


/*************************************************************************************/

/**.
\param id Pointer to a relative field identifier.
\return Zero if the expansion is successfull.*/
static int expand(char *id)
{
	int dot_cnt = 0;
	char *c;

	if (id[0] != '.') {
		/* OK: Does not begin with a dot -> absolute name */
		strcpy(last_valid_id, id);
		return 0;
	};

	if ((id[0] == '.') && (last_valid_id[0] == '\0'))
		/* ERROR: relative field name without parent specification */
		return 1;

	/* Get dot count */
	for (c = id; *c; c++)
		if (*c == '.')
			dot_cnt++;

	/* find the end of string */
	for (c = last_valid_id; *c; c++)
		;
	/* find first matching dot */
	for( ; c >= last_valid_id; c--)
		if (*c == '.' && --dot_cnt == 0)
			break;

	if (c < last_valid_id)
		/* ERROR: relative name contains more dots than absolute name */
		return 2;

	if ((c - last_valid_id + strlen(id)) >= MAX_LINE_LEN)
		/* ERROR: expanded name is too long */
		return 3;

	strcpy(c, id);
	strcpy(id, last_valid_id);

	return 0;
}


/**.
\param str Field identifier with mutiple parts.
\return Pointer to the begining of the second part.*/
static char *get_second_part(char *str)
{
	char *c = str;

	while ((*c != '.') && (*c != '\0'))
		c++;
	if (*c != '\0')
		return c+1;
	return c;
}


/**.
\param str Field identifier with mutiple parts.
\return The argument \a str as it was passed to the function (only the string is shorter).*/
static char *get_first_part(char *str)
{
	char *c = str;

	while ((*c != '.') && (*c != '\0'))
		c++;
	*c = '\0';

	return str;
}


/**.
\param medl Pointer to the target MEDL structure.
\retval ptr	Pointer to the field within given MEDL.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL(MEDL *medl, void **ptr, PTR_TYPE *type, char *id)
{
	char *item, *subitems;

	subitems = get_second_part(id);
	item = get_first_part(id);

	if (strncmp(item, "implementation_params", 21) == 0) {
		return get_ptr_MEDL_ISP(&(medl->implementation_params), ptr, type, subitems);

	} else if (strncmp(item, "bus_guardian_cold_start", 23) == 0) {
		return get_ptr_MEDL_BGP(&(medl->bus_guardian_cold_start), ptr, type, subitems);

	} else if (strncmp(item, "mode_address_entry", 18) == 0) {
		int i1, i2;

		if (sscanf(item, "mode_address_entry[%d][%d]", &i1, &i2) != 2)
			return 2;	/* ERROR */
		return get_ptr_MEDL_MAE(&(medl->mode_address_entry[i1][i2]), ptr, type, subitems);

	} else if (strncmp(item, "slot_control_entry", 18) == 0) {
		int i1;

		if (sscanf(item, "slot_control_entry[%d]", &i1) != 1)
			return 2;	/* ERROR */
		return get_ptr_MEDL_SCE(&(medl->slot_control_entry[i1]), ptr, type, subitems);

	} else if (strncmp(item, "bus_guardian_download", 21) == 0) {
		return get_ptr_MEDL_BGP(&(medl->bus_guardian_download), ptr, type, subitems);

	} else if (strncmp(item, "role_list", 9) == 0) {
		return get_ptr_MEDL_RL(&(medl->role_list), ptr, type, subitems);

	} else if (strncmp(item, "mode_control_entry", 18) == 0) {
		int i1;

		if (sscanf(item, "mode_control_entry[%d]", &i1) != 1)
			return 2;	/* ERROR */
		return get_ptr_MEDL_MCE(&(medl->mode_control_entry[i1]), ptr, type, subitems);

	} else if (strncmp(item, "controller_cfg_params", 21) == 0) {
		return get_ptr_MEDL_CCP(&(medl->controller_cfg_params), ptr, type, subitems);

	} else if (strncmp(item, "revision", 8) == 0) {
		*ptr = (void *)&(medl->revision);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}


/**.
\param bgp Pointer to the target MEDL_BGP structure.
\retval ptr	Pointer to the field within given MEDL_BGP.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_BGP(MEDL_BGP *bgp, void **ptr, PTR_TYPE *type, char *id)
{
	char *item;

	item = get_first_part(id);

	if (strncmp(item, "crc", 3) == 0) {
		*ptr = (void *)&(bgp->crc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "bg_arm_phase", 12) == 0) {
		*ptr = (void *)&(bgp->bg_arm_phase);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "bg_slot_offset", 14) == 0) {
		*ptr = (void *)&(bgp->bg_slot_offset);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "bg_remaining_round", 18) == 0) {
		*ptr = (void *)&(bgp->bg_remaining_round);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "bg_slot_duration", 16) == 0) {
		*ptr = (void *)&(bgp->bg_slot_duration);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}


/**.
\param lname Pointer to the target LOGICAL_NAME structure.
\retval ptr	Pointer to the field within given LOGICAL_NAME.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_LOGICAL_NAME(LOGICAL_NAME *lname, void **ptr, PTR_TYPE *type, char *id)
{
	char *item;

	item = get_first_part(id);

	if (strncmp(item, "multiplex_id", 12) == 0) {
		*ptr = (void *)&(lname->multiplex_id);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ps", 2) == 0) {
		*ptr = (void *)&(lname->ps);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "slot_position", 13) == 0) {
		*ptr = (void *)&(lname->slot_position);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}


/**.
\param mae Pointer to the target MEDL_MAE structure.
\retval ptr	Pointer to the field within given MEDL_MAE.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_MAE(MEDL_MAE *mae, void **ptr, PTR_TYPE *type, char *id)
{
	char *item;

	item = get_first_part(id);

	if (strncmp(item, "mcp_c", 5) == 0) {
		*ptr = (void *)&(mae->mcp_c);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "mcp_d", 5) == 0) {
		int i1;

		if (sscanf(item, "mcp_d[%d]", &i1) != 1)
			return 2;	/* ERROR */
		*ptr = (void *)&(mae->mcp_d[i1]);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "eoc", 3) == 0) {
		*ptr = (void *)&(mae->eoc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "crc", 3) == 0) {
		*ptr = (void *)&(mae->crc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "dafla", 5) == 0) {
		*ptr = (void *)&(mae->dafla);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ru1", 3) == 0) {
		*ptr = (void *)&(mae->ru1);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "mcp_i", 5) == 0) {
		int i1;

		if (sscanf(item, "mcp_i[%d]", &i1) != 1)
			return 2;	/* ERROR */
		*ptr = (void *)&(mae->mcp_i[i1]);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "daflb", 5) == 0) {
		*ptr = (void *)&(mae->daflb);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ru2", 3) == 0) {
		*ptr = (void *)&(mae->ru2);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ia", 2) == 0) {
		*ptr = (void *)&(mae->ia);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "cni_address_cha", 15) == 0) {
		*ptr = (void *)&(mae->cni_address_cha);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ib", 2) == 0) {
		*ptr = (void *)&(mae->ib);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "cni_address_chb", 15) == 0) {
		*ptr = (void *)&(mae->cni_address_chb);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;
}


/**.
\param sce Pointer to the target MEDL_SCE structure.
\retval ptr	Pointer to the field within given MEDL_SCE.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_SCE(MEDL_SCE *sce, void **ptr, PTR_TYPE *type, char *id)
{
	char *item, *subitems;

	subitems = get_second_part(id);
	item = get_first_part(id);

	if (strncmp(item, "eot", 3) == 0) {
		*ptr = (void *)&(sce->eot);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "sru_slot_duration", 17) == 0) {
		*ptr = (void *)&(sce->sru_slot_duration);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "syf", 3) == 0) {
		*ptr = (void *)&(sce->syf);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "mm", 2) == 0) {
		*ptr = (void *)&(sce->mm);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "logical_name", 12) == 0) {
		return get_ptr_LOGICAL_NAME(&(sce->logical_name), ptr, type, subitems);

	} else if (strncmp(item, "ba", 2) == 0) {
		*ptr = (void *)&(sce->ba);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ra", 2) == 0) {
		*ptr = (void *)&(sce->ra);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "medl_pos", 8) == 0) {
		*ptr = (void *)&(sce->medl_pos);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "mo", 2) == 0) {
		*ptr = (void *)&(sce->mo);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "membership_pos", 14) == 0) {
		*ptr = (void *)&(sce->membership_pos);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "delay_correction", 16) == 0) {
		*ptr = (void *)&(sce->delay_correction);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "cs", 2) == 0) {
		*ptr = (void *)&(sce->cs);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "eor", 3) == 0) {
		*ptr = (void *)&(sce->eor);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}


/**.
\param isp Pointer to the target MEDL_ISP structure.
\retval ptr	Pointer to the field within given MEDL_ISP.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_ISP(MEDL_ISP *isp, void **ptr, PTR_TYPE *type, char *id)
{
	char *item, *subitems;

	subitems = get_second_part(id);
	item = get_first_part(id);
		
	if (strncmp(item, "sync_interval", 13) == 0) {
		*ptr = (void *)&(isp->sync_interval);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "listen_timeout_extension", 24) == 0) {
		*ptr = (void *)&(isp->listen_timeout_extension);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "i_frame_time", 12) == 0) {
		*ptr = (void *)&(isp->i_frame_time);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "listen_cold_start_period_params", 31) == 0) {
		return get_ptr_MEDL_BGP(&(isp->listen_cold_start_period_params), ptr, type, subitems);

	} else if (strncmp(item, "oversampling_rate", 17) == 0) {
		*ptr = (void *)&(isp->oversampling_rate);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "watchdog_slot_length", 20) == 0) {
		*ptr = (void *)&(isp->watchdog_slot_length);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "ifg_time", 8) == 0) {
		*ptr = (void *)&(isp->ifg_time);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "synchronization_period_params", 29) == 0) {
		return get_ptr_MEDL_BGP(&(isp->synchronization_period_params), ptr, type, subitems);

	} else if (strncmp(item, "listen_timeout_restart_time", 27) == 0) {
		*ptr = (void *)&(isp->listen_timeout_restart_time);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "crc", 3) == 0) {
		*ptr = (void *)&(isp->crc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "entries", 7) == 0) {
		*ptr = (void *)&(isp->entries);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "listen_delay", 12) == 0) {
		*ptr = (void *)&(isp->listen_delay);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "rom_speed", 9) == 0) {
		*ptr = (void *)&(isp->rom_speed);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}


/**.
\param mce Pointer to the target MEDL_MCE structure.
\retval ptr	Pointer to the field within given MEDL_MCE.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_MCE(MEDL_MCE *mce, void **ptr, PTR_TYPE *type, char *id)
{
	char *item;

	item = get_first_part(id);

	if (strncmp(item, "crc", 3) == 0) {
		*ptr = (void *)&(mce->crc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "deferred_mode", 13) == 0) {
		int i1;

		if (sscanf(item, "deferred_mode[%d]", &i1) != 1)
			return 2;	/* ERROR */
		*ptr = (void *)&(mce->deferred_mode[i1]);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}


/**.
\param rl Pointer to the target MEDL_RL structure.
\retval ptr	Pointer to the field within given MEDL_RL.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_RL(MEDL_RL *rl, void **ptr, PTR_TYPE *type, char *id)
{
	char *item, *subitems;

	subitems = get_second_part(id);
	item = get_first_part(id);

	if (strncmp(item, "crc", 3) == 0) {
		*ptr = (void *)&(rl->crc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "logical_name", 12) == 0) {
		return get_ptr_LOGICAL_NAME(&(rl->logical_name), ptr, type, subitems);

	} else if (strncmp(item, "bus_guardian", 12) == 0) {
		return get_ptr_MEDL_BGP(&(rl->bus_guardian), ptr, type, subitems);

	} else
		return 1;	/* ERROR */
}


/**.
\param ccp Pointer to the target MEDL_CCP structure.
\retval ptr	Pointer to the field within given MEDL_CCP.
\retval type Found field type.
\param id Absolute field name.
\return Zero on success, nonzero otherwise.*/
static int get_ptr_MEDL_CCP(MEDL_CCP *ccp, void **ptr, PTR_TYPE *type, char *id)
{
	char *item;

	item = get_first_part(id);
		
	if (strncmp(item, "free_running_macroticks", 23) == 0) {
		*ptr = (void *)&(ccp->free_running_macroticks);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "mmfc", 4) == 0) {
		*ptr = (void *)&(ccp->mmfc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "c_state_length", 14) == 0) {
		*ptr = (void *)&(ccp->c_state_length);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "max_cold_start_entry", 20) == 0) {
		*ptr = (void *)&(ccp->max_cold_start_entry);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "app_version", 11) == 0) {
		*ptr = (void *)&(ccp->app_version);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "physical_sru_name", 17) == 0) {
		*ptr = (void *)&(ccp->physical_sru_name);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "crc", 3) == 0) {
		*ptr = (void *)&(ccp->crc);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "controller_version", 18) == 0) {
		*ptr = (void *)&(ccp->controller_version);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "cf", 2) == 0) {
		*ptr = (void *)&(ccp->cf);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "startup_timeout", 15) == 0) {
		*ptr = (void *)&(ccp->startup_timeout);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "microtick_int", 13) == 0) {
		*ptr = (void *)&(ccp->microtick_int);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "microtick_frac", 14) == 0) {
		*ptr = (void *)&(ccp->microtick_frac);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "app_id", 6) == 0) {
		*ptr = (void *)&(ccp->app_id);
		*type = UNSIGNED_PTR;
		return 0;

	} else if (strncmp(item, "tdma_round_length", 17) == 0) {
		*ptr = (void *)&(ccp->tdma_round_length);
		*type = UNSIGNED_PTR;
		return 0;

	} else
		return 1;	/* ERROR */
}

/*\}*/
/*************************************************************************************/
/**\name Exported Variables*/
/*\{*/

/** Number of the line where an error stopped execution.*/
int medl_read_error_line;

/*\}*/
/*************************************************************************************/
/*************************************************************************************/
/**\name Exported Functions*/
/*\{*/

/**.
The name is derived using a pattern defined within \c fp1const.h file. By default
the geenrated file name will look like this: \c medl01.txt (for \a index equal to 1).
Read data are stored into the MEDL given as argument. Fields not present in the source 
file are by default set to zero.
\param medl Pointer to a MEDL instance that will be filled with new data.
\param index Numerical index that specifies the node for which to load the MEDL.
\return Zero means successfull completion, any other code is described in 
	MEDL_READ_ERROR_CODES.*/
MEDL_READ_ERROR_CODES medl_read_from_file(MEDL *medl, unsigned index) {
	char fname[MAX_FILE_NAME_LEN];
	char line[MAX_LINE_LEN];
	FILE *f = NULL;

	CHECK(MEDL_READ_INDEX_ERROR, index <= 99);
	CHECK(MEDL_READ_NULL_MEDL_ERROR, medl != NULL);

	sprintf(fname, MEDL_READ_FILE_NAME, index);
	f = fopen(fname, "r");
	CHECK(MEDL_READ_FILE_OPEN_ERROR, f != NULL);

	memset(medl, 0, sizeof(MEDL));

	last_valid_id[0] = '\0';

	medl_read_error_line = 0;
	while(!feof(f)) {
		char value[MAX_LINE_LEN], name[MAX_LINE_LEN];
		void *ptr;
		PTR_TYPE type;

		medl_read_error_line++;
		if (fgets(line, MAX_LINE_LEN, f) == NULL)
			if (!feof(f))
				error(MEDL_READ_FILE_READ_ERROR);

		if (line[0] == '#')
			continue;
		if (sscanf(line, "%s %s", value, name) != 2)
			continue;

		CHECK(MEDL_READ_INVALID_ID_ERROR, expand(name) == 0);

		if (get_ptr_MEDL(medl, &ptr, &type, name) == 0) {
			switch (type) {
			case UNSIGNED_PTR:
				sscanf(value, "%d", (unsigned *)ptr);
				break;

			case DOUBLE_PTR:
				sscanf(value, "%lf", (double *)ptr);
				break;

			default:
				error(MEDL_READ_UNKNOWN_TYPE_ERROR);
			}
		} else
			error(MEDL_READ_UNKNOWN_ID_ERROR);
	}

	fclose(f);

	return MEDL_READ_NO_ERROR;
}

/*\}*/
/***********************************************************************/

/*@}*/
/*@}*/

