/**@addtogroup os
 * @{*/

/**@addtogroup os_int  Implementation
 * @{*/


/**@file
 * @brief TTPOs simulation based on C-Sim.
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, University of West Bohemia, Pilsen, Czech Republic
 * @author Michal Jun, Czech Technical University, Prague, Czech Republic
 * @author Tomas Krejzek, Czech Technical University, Prague, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1,
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#include <stdio.h>
#include <string.h>

#include "csim/csim.h"
#include "ttpc/fl1func.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1cni.h"
#include "ttpc/fp1lname.h"
#include "ttpc/fp1medl.h"
#include "ttpc/sa1os.h"
#include "ttpc/sl1func.h"
#include "ttpc/sp1ctrl.h"
#include "ttpc/sp1cni.h"
#include "ttpc/fl1debug.h"

/************************************************************************************/

/** Initial value of @c error_delay attribute of OS.*/
#define INIT_ERROR_DELAY         250.0

/** Initial value of @c start_allowed attribute of OS.*/
#define INIT_START_ALLOWED       1

/** Initial value of @c int_log_enabled attribute of OS.*/
#define INIT_INT_LOG_ENABLED     0

/** Initial value of @c auto_life_sign attribute of OS.*/
#define INIT_AUTO_LIFE_SIGN      0

/** Time distance between host lifesign update and the next IFG start (in macroticks).
For the lifesign mechanism to work, this value should be greater then 1 and 
less then duration of the SRU slot.*/
#define LIFE_SIGN_UPDATE_RESERVE 50

/** Time to wait before the controller becomes ready after power on.*/
#define OS_INIT_TIME            20.0

/** Maximum number of transmission cycles to wait for synchronization.*/
#define SYNC_WAIT_CYCLES        2

/** Lantency of interrupt service execution.*/
#define INTERRUPT_LATENCY       0.0


/** This makes the logging process throughout the whole project transparent and unified.
 *
 * @param msg This argument is special. It should contain the format specifier as in
 * a call to printf() function and optional arguments, but the whole set of parameters
 * must be enclosed in one additional pair of parenthesis, e.g.: @code
 *   log(("Testing entry number %d (made by %s)\n", 1, "me" ));	@endcode */
#define log(msg) log_os_csim(msg)


/** Suspend the execution of calling C-Sim process for given number of simulation
	time units.
@param t Duration that the process will spend sleeping.*/
#define os_hold(t) do { \
	log(("Holding: %0.4lf\n", (t))); \
	csim_hold(t); \
} while(0)


/** Check whether an application registered service is enabled, if yes call it (in
	any case clear the interrupt status flag for this kind of interrupt.
@param id Shortcut name for the interrupt with no quotes and all lowercase.*/
#define resolve_int(id) do { \
	log(("Received '%s' interrupt\n", #id)); \
	if (int_tab->id.enabled && int_tab->id.prog != NULL) \
		int_tab->id.prog(int_tab->id.prog_arg); \
	cni_is_set_##id(is, (unsigned)0); \
} while (0)


/** Clear the interrupt enable bit in CNI if the interrupt is disabled in the OS.
@param id Shortcut name for the interrupt with no quotes and all lowercase.*/
#define test_clear(id) \
	if (!os->int_table.id.enabled) \
		cni_ie_set_##id(ie, 0); \
	else 


/*************************************************************************************/
/**@name Code Generation Macros
 * @{*/

/** Preprocessor macro for implementing an interrupt enable/disable function.
 * This macro expands into a function definition, e.g. when called like this:
 * @code
 * 	implement_enable_int_fn(mc)
 * @endcode 
 *
 * The preprocessor defines this function:
 * @code
 * 	void os_enable_mc_int(OS_PROCESS *os, unsigned state)
 * 	{
 * 		if (state == os->int_table.mc.enabled)
 * 			return;
 * 		os->int_table.mc.enabled = state;
 * 		if (!state && os->int_log_enabled)
 * 			return;
 * 		cni_ie_set_mc(&os->cni->interrupt_enable, state);
 * 	}
 * @endcode
 *
 * @param id Shortcut name for the interrupt with no quotes and all lowercase. */
#define implement_enable_int_fn(id) \
	void os_enable_##id##_int(OS_PROCESS *os, unsigned state) \
	{ \
		if (state == os->int_table.id.enabled) \
			return; \
		os->int_table.id.enabled = state; \
		if (!state && os->int_log_enabled) \
			return; \
		cni_ie_set_##id(&os->cni->interrupt_enable, state); \
	}


/** Preprocessor macro for implementing an interrupt handler adding function.
 * This macro expands into a function definition, e.g. when called like this:
 * @code
 *   implement_add_int_fn(mc)
 * @endcode the preprocessor defines this function: @code
 *   void os_add_mc_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg)
 *   {
 *   	os->int_table.mc.prog = prog;
 *   	os->int_table.mc.prog_arg = arg;
 *   }
 * @endcode
 *
 * @param id Shortcut name for the interrupt with no quotes and all lowercase. */
#define implement_add_int_fn(id) \
	void os_add_##id##_interrupt(OS_PROCESS *os, OS_TASK_FN prog, void *arg) \
	{ \
		os->int_table.id.prog = prog; \
		os->int_table.id.prog_arg = arg; \
	}


/** Macro for implementing all data seting functions. 
This macro expands into a function definition, e.g. when called like this:
@code
	implement_set_ofs_fn(integer, int)
@endcode 
the preprocessor defines this function:
@code
	void os_set_ofs_integer(OS_PROCESS *os, unsigned addr, unsigned ofs, int val)
	{
		os_set_ofs_data(os, addr, ofs, (char *)&val, sizeof(int));
	}
@endcode
@param name Last component of function name (no quotes).
@param type Type of item that will be written (any type).*/
#define implement_set_ofs_fn(name, type) \
	void os_set_ofs_##name(OS_PROCESS *os, unsigned addr, unsigned ofs, type val) \
	{ \
		os_set_ofs_data(os, addr, ofs, (char *)&val, sizeof(type)); \
	}


/** Macro for implementing all data getting functions.
This macro expands into a function definition, e.g. when called like this:
@code
	implement_get_ofs_fn(integer, int)
@endcode 
the preprocessor defines this function:
@code
	unsigned os_get_ofs_rep_integer(OS_PROCESS *os, const unsigned addr[], unsigned ofs, int *val)
	{
		if (!os_get_ofs_integer(os, addr[0], ofs, val))
			return 0;
		return os_get_ofs_integer(os, addr[1], ofs, val);
	}
@endcode
@note The new function assumes that a function @c os_get_ofs_ + @a name is defined within
	available scope.
@param name Last component of function name (no quotes).
@param type Type of item that will be read.*/
#define implement_get_ofs_rep_fn(name, type) \
	unsigned os_get_ofs_rep_##name(OS_PROCESS *os, const unsigned addr[], unsigned ofs, type *val) \
	{ \
		if (!os_get_ofs_##name(os, addr[0], ofs, val)) \
			return 0; \
		return os_get_ofs_##name(os, addr[1], ofs, val); \
	}


/*@}*/
/*************************************************************************************/

/************************************************************************************/
/**@name Interrupt Counting Macros
 * All macros defined here perform one task - they increment the EDM counter for one
 * particular interrupt. 
 * Since they all take the same arguments, they are defined only once here:
 *
 * @param os Pointer to an instance of OS_PROCESS.
 * @{ */

/** Increment @c ml EDM counter by 1.*/
#define log_ml_interrupt(os) ((os)->edm_counters.ml++)

/** Increment @c mc EDM counter by 1.*/
#define log_mc_interrupt(os) ((os)->edm_counters.mc++)

/** Increment @c mi EDM counter by 1.*/
#define log_mi_interrupt(os) ((os)->edm_counters.mi++)

/** Increment @c ui1 EDM counter by 1.*/
#define log_ui1_interrupt(os) ((os)->edm_counters.ui1++)

/** Increment @c ui2 EDM counter by 1.*/
#define log_ui2_interrupt(os) ((os)->edm_counters.ui2++)

/** Increment @c ti1 EDM counter by 1.*/
#define log_ti1_interrupt(os) ((os)->edm_counters.ti1++)

/** Increment @c ti2 EDM counter by 1.*/
#define log_ti2_interrupt(os) ((os)->edm_counters.ti2++)

/** Increment @c os_error EDM counter by 1.*/
#define log_os_error_interrupt(os) ((os)->edm_counters.os_error++)

/** Increment @c os_active EDM counter by 1.*/
#define log_os_active_interrupt(os) ((os)->edm_counters.os_active++)

/*@}*/
/************************************************************************************/

/************************************************************************************/
/* Generated Functions */
/************************************************************************************/

implement_access_functions(OS_PROCESS, os, auto_life_sign, 1)

/*************************************************************************************/
/* Local Functions */
/*************************************************************************************/

/** Change the current state of internal state machine of OS.*/
static void os_set_state(OS_PROCESS *os, OS_STATES csim_state);

/** Swich state to OS_STATE_ERROR and store the given error code.*/
static void raise_os_error_int(OS_PROCESS *os, OS_ERROR_CODE code);

/** Raise transition-into-activei-state interrupt.*/
static void raise_os_active_int(OS_PROCESS *os);

/** Check if the received message is correct (noise flag is ignored).*/
static int eif_is_valid(OS_PROCESS *os, unsigned addr);

/** Interrupt service function (called only by controller) - set the interrupted
 * flag and activates the OS.*/
static void controller_interrupt(void *ptr);

/** Unified interrupt handler called from the OS process after any interrupt is 
 * detected.*/
static void handle_interrupt(OS_PROCESS *os);

/** Return the difference between @a t2 and @a t1 modulo @a cc_time.*/
static unsigned get_time_gap(unsigned t1, unsigned t2, unsigned cc_time);

/** Lifesign update function.*/
static void lifesign_update_prog(void *arg);

/** Find current position withing transmission cycle and schedule first task.*/
static void schedule_first_task(OS_PROCESS *os);

/** Checks the interrupt timeout, returns nozero if the controller failed to raise
 * interrupt in specified time period, 0 if everything works fine. */
static unsigned check_int_timeout(OS_PROCESS *os);

/** Count controller interrupt occurance and store it into EDM structure */
static void log_controller_interrupt(OS_PROCESS *os);

/** Checks mode change conditions and if they are fulfilled performs a
 * deferred mode change request.
 * Returns the mode number if the request is performed, 0 otherwise */
static unsigned mode_change_trigger(OS_PROCESS *os);

/** Loads the neccessary scheduling data from the MEDL table loaded into the
 * attached controller. All time information is valid only after this function
 * is executed properly. The affected fields include:
 * slot_count, slot_time, round_len, transmission_cycle_len */
static void os_load_scheduling_data(OS_PROCESS *os);

/** Initializes the OS specific memory in the OS_PROCESS structure */
static OS_PROCESS *os_init(OS_PROCESS *os);

/** Schedules the life-sign update tasks to a fixed time before every sending
 * slot of the node. */
static void schedule_life_sign_tasks(OS_PROCESS *os);


/************************************************************************************/

/* Swiches state to OS_STATE_ERROR and stores the error code. */
static void raise_os_error_int(OS_PROCESS *os, OS_ERROR_CODE code)
{
	log(("Raising OS-Error: code = %d\n", code));
	os->error_code = code;
	os_set_state(os, OS_STATE_ERROR);
}

/* Raises the OS_ACTIVE interrupt - after transition into the App. state. */
static void raise_os_active_int(OS_PROCESS *os)
{
	OS_INTERRUPT_TABLE *int_tab = &os->int_table;

	log(("Raising OS-Active interrupt\n"));
	log_os_active_interrupt(os);

	if (int_tab->os_active.enabled)
		int_tab->os_active.prog(int_tab->os_active.prog_arg);
}

/* Transits into chosen state of operation */
static void os_set_state(OS_PROCESS *os, OS_STATES csim_state)
{
	os->os_state = csim_state;
}

/* Test if the received message is correct.
   Returns 0 if the message is invalid. */
static int eif_is_valid(OS_PROCESS *os, unsigned addr)
{
	CNI_EIF_VALUES eif = (CNI_EIF_VALUES)cni_mae_get_eif(cni_get_message_entry(os->cni, addr));

	return (eif & EIF_CORRECT_FRAME) == EIF_CORRECT_FRAME;
}


/** Interrupt service function - called only by the asociated controller */
static void controller_interrupt(void *ptr)
{
	OS_PROCESS *os = (OS_PROCESS *) ptr;
	log(("Interrupt handler - activating OS.\n"));
	os->interrupted++;
	process_activate_delay((CSIM_PROCESS *)os, INTERRUPT_LATENCY);
}

/* Lifesign update function. */
static void lifesign_update_prog(void *arg)
{
	CNI *cni = ((OS_PROCESS *)arg)->cni;

	log(("Udating host lifesign.\n"));
	cni_set_host_lifesign(cni, ~cni_get_controller_lifesign(cni) & bit_mask(16));
}

/* Returns the difference between t2 and t1, when negative adds cc_time. */
static unsigned get_time_gap(unsigned t1, unsigned t2, unsigned cc_time)
{
	if (t2 <= t1)
		return cc_time + t2 - t1;
	else
		return t2 - t1;
}

/* Returns if the asked task is enabled to run in current mode */
static int enabled_in_mode(unsigned mode, unsigned enable_field)
{
	return (enable_field & (0x00000001 << (mode-1))) != 0;
}

/* Schedules the given task to be executed next. Sets the Timer-1 interrupt
   according to its exe-time and updates iternal state correctly. */
static void schedule_task(OS_PROCESS *os, int task_id)
{
	OS_TASK *old_task = &os->tasks[os->next_task];
	OS_TASK *new_task = &os->tasks[task_id];

	if (new_task->time < old_task->time)
		os->transmission_cycle++;
	log(("Transmission cycle %u\n", os->transmission_cycle));

	os->next_task_time = bit_mask(16) & (os->next_task_time + get_time_gap(
		old_task->time,
		new_task->time,
		os->transmission_cycle_len));
	os->next_task = task_id;

	log(("Timer for task %d (%s) set at %u.\n", task_id, new_task->name, os->next_task_time));
	cni_set_timer1(os->cni, os->next_task_time);
}

/* Run next scheduled task. Called from TI1 interrupt. */
static void run_task(OS_PROCESS *os)
{
	unsigned mode = os_get_current_mode(os);
	unsigned next = os->next_task;
	OS_TASK *task = &os->tasks[next];
	unsigned time = task->time;
	OS_TASK *aux = task;
	unsigned count = 0;

	if (os->task_count < 1) {
		raise_os_error_int(os, OS_ERROR_NO_TASK);
		return;
	}
	if (os->task_count > OS_MAX_TASK_NUM)
		os->task_count = OS_MAX_TASK_NUM;

	if (next >= os->task_count) {
		raise_os_error_int(os, OS_ERROR_BAD_SCHEDULE);
		return;
	}

	do {
		if (aux->prog != NULL) {
			log(("Running task %d. (%s)\n", next, aux->name));
			aux->prog(aux->prog_arg);
		} else {
			raise_os_error_int(os, OS_ERROR_BAD_TASK);
			return;
		}

		do {
			next = (next + 1) % os->task_count;
			aux = &os->tasks[next];
			count++;
		} while (!enabled_in_mode(mode, aux->modes) && count < os->task_count);

	} while (time == aux->time && count < os->task_count);

	if (enabled_in_mode(mode, aux->modes))
		schedule_task(os, next);
	else
		raise_os_error_int(os, OS_ERROR_NO_TASK);
}

/* Log the occurance of Protocol error - detailed cause */
static void log_pe_interrupt(OS_PROCESS *os)
{
	CNI_PS *cni_ps = &os->cni->protocol_status;
	OS_EDM_COUNTERS *edm = &os->edm_counters;

	if (cni_ps_get_be(cni_ps))	/* BG error */
		edm->pe_be++;
	if (cni_ps_get_ae(cni_ps))	/* ack. error */
		edm->pe_ae++;
	if (cni_ps_get_se(cni_ps))	/* clock sync. error */
		edm->pe_se++;
	if (cni_ps_get_me(cni_ps))	/* max. membership failures */
		edm->pe_me++;
	if (cni_ps_get_mc(cni_ps))	/* MEDL CRC error */
		edm->pe_mc++;
	if (cni_ps_get_cb(cni_ps))	/* comm. system blackout */
		edm->pe_cb++;
	if (cni_ps_get_dc(cni_ps))	/* download completed */
		edm->pe_dc++;

	edm->pe++;
}

/* Log the occurance of Host error - detailed cause */
static void log_be_interrupt(OS_PROCESS *os)
{
	CNI_BS *bs = &os->cni->bist_status;
	OS_EDM_COUNTERS *edm = &os->edm_counters;

	if (cni_bs_get_sc(bs))
		edm->be_sc++;
	if (cni_bs_get_av(bs))
		edm->be_av++;
	if (cni_bs_get_br(bs))
		edm->be_br++;
	if (cni_bs_get_rm(bs))
		edm->be_rm++;
	if (cni_bs_get_rf(bs))
		edm->be_rf++;
	if (cni_bs_get_wd(bs))
		edm->be_wd++;

	edm->be++;
}

/* Log the occurance of Host error - detailed cause */
static void log_he_interrupt(OS_PROCESS *os)
{
	CNI_PS *cni_ps = &os->cni->protocol_status;
	OS_EDM_COUNTERS *edm = &os->edm_counters;

	if (cni_ps_get_so(cni_ps))	/* slot occupied */
		edm->he_so++;
	if (cni_ps_get_mv(cni_ps))	/* mode violation */
		edm->he_mv++;
	if (cni_ps_get_ne(cni_ps))	/* NBW protocol error, invalid EIF */
		edm->he_ne++;

	edm->he++;
}

/* Performs interrupt occurance counting. */
static void log_controller_interrupt(OS_PROCESS *os)
{
	CNI_IS *is = &os->cni->interrupt_status;

	if (cni_is_get_be(is)) log_be_interrupt(os);
	if (cni_is_get_pe(is)) log_pe_interrupt(os);
	if (cni_is_get_he(is)) log_he_interrupt(os);
	if (cni_is_get_ml(is)) log_ml_interrupt(os);
	if (cni_is_get_mc(is)) log_mc_interrupt(os);
	if (cni_is_get_mi(is)) log_mi_interrupt(os);
	if (cni_is_get_ui2(is)) log_ui1_interrupt(os);
	if (cni_is_get_ui1(is)) log_ui2_interrupt(os);
	if (cni_is_get_ti2(is)) log_ti2_interrupt(os);
	if (cni_is_get_ti1(is)) log_ti1_interrupt(os);
}


/** Handles any interrupt coming from controller. First handles TI1,
   then calls user defined interrupt handlers. */
static void handle_interrupt(OS_PROCESS *os)
{
	CNI *cni = os->cni;
	CNI_IS *is = &os->cni->interrupt_status;
	OS_INTERRUPT_TABLE *int_tab = &os->int_table;

	log(("Interrupt handler, Cluster Time = %u\n", cni_get_cluster_time(cni)));
	if (os->int_log_enabled)
		log_controller_interrupt(os);

	if (cni_is_get_he(is))          /* Host error - MUST handle EIF & NBW errors !!! */
		resolve_int(he);
	if (cni_is_get_ml(is))          /* Membership lost  */
		resolve_int(ml);
	if (cni_is_get_mc(is))          /* Membership change */
		resolve_int(mc);
	if (cni_is_get_mi(is))          /* Mode change */
		resolve_int(mi);
	if (cni_is_get_ui2(is))
		resolve_int(ui2);
	if (cni_is_get_ui1(is))
		resolve_int(ui1);
	if (cni_is_get_ti1(is))
		resolve_int(ti1);
	if (cni_is_get_ti2(is))
		resolve_int(ti2);


	if (cni_is_get_be(is)) {         /* Bist error */
		resolve_int(be);
		raise_os_error_int(os, OS_ERROR_EDM_BE);
	}

	if (cni_is_get_pe(is)) {         /* Protocol error */
		resolve_int(pe);
		raise_os_error_int(os, OS_ERROR_EDM_PE);
	}

	if (cni_get_cluster_time(cni) == os->next_task_time) {
		log(("Task switching time reached.\n"));
		os->last_timer = csim_time();
		run_task(os);
	}
}

/* Schedules the life-sign update tasks to a fixed time before every sending
   slot of the node. */
static void schedule_life_sign_tasks(OS_PROCESS *os)
{
	unsigned i;

	for (i = 0; i < os->round_count; i++)
		os_add_lifesign_update(os, os_get_send_time(os, i, -LIFE_SIGN_UPDATE_RESERVE));
}

/* Finds current position withing transmission cycle and schedules first task. */
static void schedule_first_task(OS_PROCESS *os)
{
	CNI *cni = os->cni;
	C_STATE *c_state = &cni->c_state;
	unsigned slot = c_state_get_medl_position(c_state);
	unsigned slot_time = os->slot_time[slot];
	unsigned time, i;

	time = slot_time - c_state_get_time(c_state) + cni_get_cluster_time(cni);

	for (i = 0; i < os->task_count && (os->tasks[i].time <= time || !enabled_in_mode(os_get_current_mode(os), os->tasks[i].modes)); i++)
		;
	if (i == os->task_count)
		i = 0;

	os->next_task_time = bit_mask(16) & (
		cni_get_cluster_time(cni)
		+ get_time_gap(time, os->tasks[i].time, os->transmission_cycle_len));
	os->next_task = i;

	cni_is_set_ti1(&cni->interrupt_status, 0);
	cni_set_timer1(cni, os->next_task_time);
	os->last_timer = csim_time();

	log(("Detected slot %u, selected task %u (%s), timer set to %u.\n",
		slot, i, os->tasks[i].name, os->next_task_time));
}

/* Checks the interrupt timeout, returns nonzero if the controller failed to raise
   interrupt in specified time period, 0 if everything works fine. */
static unsigned check_int_timeout(OS_PROCESS *os)
{
	if (os->os_state < OS_STATE_STARTUP)
		return 0;

	if (csim_time() - os->last_timer <= os->round_len)
		return 0;

	log(("Interrupt timeout expired.\n"));
	return 1;
}


/* Performs OS initialization before turning On the controller, loads the time
   schedules from MEDL and turns the controller ON */
static void os_init_state(OS_PROCESS *os)
{
	log(("Loading startup values into memory.\n"));

	memset(&os->int_table, 0, sizeof(os->int_table));
	memset(&os->tasks[0], 0, OS_MAX_TASK_NUM * sizeof(OS_TASK));
	memset(&os->slot_time[0], 0, MAX_TRANSMISSION_CYCLE_LENGTH * sizeof(unsigned));

	os->error_code = OS_ERROR_NONE;
	os->os_state = OS_STATE_OFF;
	os->interrupted = 0;
	os->transmission_cycle = 0;
	os->old_controller_lifesign = 0xFFFF;
	os->next_task_time = 0;
	os->last_timer = csim_time();
	os->transmission_cycle_len = 0;
	os->task_count = 0;
	os->next_task = 0;
	os->slot_count = 0;
	os->round_count = 0;
	os->send_time = 0;

	os->auto_life_sign = INIT_AUTO_LIFE_SIGN;
	os->error_delay = INIT_ERROR_DELAY;
	os->min_node_count = 0;
	os->first_app_mode = 0;
	os->int_log_enabled = INIT_INT_LOG_ENABLED;
	os->start_allowed = INIT_START_ALLOWED;

	os_load_scheduling_data(os);

	os->iteration = SYNC_WAIT_CYCLES * os->slot_count;

	log(("Starting Controller.\n"));
	cni_set_co(os->cni, CONTROLLER_ON);
	os->startup_count++;
}

/* Enables interrupts needed by the TTP/OS during node startup.
   The interrupts will stay enabled until the application disables them. */
static void os_enable_interrupts(OS_PROCESS *os)
{
	log(("Enabling interrupts (TI1, MC, MI).\n"));
	cni_ie_set_ti1(os->cni, 1);
	cni_ie_set_mc(&os->cni->interrupt_enable, 1);
	cni_ie_set_mi(&os->cni->interrupt_enable, 1);
	os->last_timer = csim_time();
}


/* Checks mode change conditions and if they are fulfilled performs a
   deferred mode change request.
   Returns the 0 if the the request is handled, nonzero it is impossible */
static unsigned mode_change_trigger(OS_PROCESS *os)
{
	unsigned mode = os->first_app_mode;

	if (!os->min_node_count || !mode)
		return 1;

	if (os_get_node_count(os) < os->min_node_count)
		return 2;

	if (os_get_dmc(os) != 0)
		return 3;

	log(("The OS triggered an automatic mode change from STARTUP state\n"));
	os_request_mode_change(os, os->first_app_mode, 0);
	return 0;
}


/** OS simulation main program */
static csim_program(OS_PROCESS, os_prog)
	for (;;) {

	if (p_my->interrupted) {
		handle_interrupt(p_my);
		p_my->interrupted = 0;
	}

	if (p_my->os_state != OS_STATE_ERROR) {
		if (check_int_timeout(p_my))
			raise_os_error_int(p_my, OS_ERROR_INT_TIMEOUT);
		else if (p_my->os_state > OS_STATE_INIT && cni_get_co(p_my->cni) != CONTROLLER_ON)
			raise_os_error_int(p_my, OS_ERROR_FREEZE);
	}

	switch (p_my->os_state) {

		case OS_STATE_OFF:
			log(("OS_STATE_OFF\n"));
			if (cni_get_co(p_my->cni) == CONTROLLER_ON) {
				raise_os_error_int(p_my, OS_ERROR_ALREADY_ON);
				break;
			}

			if (!p_my->start_allowed)
				csim_passivate();
			else {
				os_hold(0.0);
				os_set_state(p_my, OS_STATE_INIT);
			}
			break;

		case OS_STATE_INIT:
			log(("OS_STATE_INIT\n"));
			os_init_state(p_my);
			os_hold(OS_INIT_TIME);
			os_set_state(p_my, OS_STATE_INIT_TASK);
			break;

		case OS_STATE_INIT_TASK:
			log(("OS_STATE_INIT_TASK\n"));
			if (cni_ps_get_controller_ready(&p_my->cni->protocol_status) != CONTROLLER_READY) {
				log(("Controller not ready.\n"));
				raise_os_error_int(p_my, OS_ERROR_NOT_READY);
			} else {
				os_enable_interrupts(p_my);
				if (p_my->init_task.prog != NULL)
					p_my->init_task.prog(p_my->init_task.prog_arg);
				else
					p_my->auto_life_sign = 1;
				if (p_my->auto_life_sign)
                    schedule_life_sign_tasks(p_my);
				os_set_state(p_my, OS_STATE_READY);
			}
			break;

		case OS_STATE_READY:
			log(("OS_STATE_READY\n"));
			os_test_controller_lifesign(p_my);

			if (cni_ps_get_c_state_valid(&p_my->cni->protocol_status) == C_STATE_VALID) {
				os_set_state(p_my, OS_STATE_SYNCHRONISED);
				break;
			}

			if (!p_my->iteration) {
				log(("Node did not synchronise within two cluster cycles.\n"));
				raise_os_error_int(p_my, OS_ERROR_SYNC_TIMEOUT);
				break;
			}

			p_my->iteration--;
			log(("Waiting for synchronisation (slot %d).\n", SYNC_WAIT_CYCLES * p_my->slot_count - p_my->iteration));
			if (p_my->slot_count == 0)
				raise_os_error_int(p_my, OS_ERROR_BAD_SLOT_COUNT);
			else
				os_hold((double)p_my->transmission_cycle_len / p_my->slot_count);
			break;

		case OS_STATE_SYNCHRONISED:
			log(("OS_STATE_SYNCHRONISED\n"));
			if (os_get_current_mode(p_my) == STARTUP_MODE_NUMBER)
				os_set_state(p_my, OS_STATE_STARTUP);
			else
				os_set_state(p_my, OS_STATE_APPLICATION);
			raise_os_active_int(p_my);
			schedule_first_task(p_my);
			break;

		case OS_STATE_STARTUP:
			if (os_get_current_mode(p_my) != STARTUP_MODE_NUMBER) {
				os_set_state(p_my, OS_STATE_APPLICATION);
				break;
			}

			log(("OS_STATE_STARTUP\n"));
			mode_change_trigger(p_my);
			os_hold((double)p_my->transmission_cycle_len);
			break;

		case OS_STATE_APPLICATION:
			if (os_get_current_mode(p_my) == STARTUP_MODE_NUMBER) {
				os_set_state(p_my, OS_STATE_STARTUP);
				break;
			}

			log(("OS_STATE_APPLICATION\n"));
			os_hold((double)p_my->transmission_cycle_len);
			break;

		case OS_STATE_ERROR:
			log(("OS_STATE_ERROR\n"));
			log_os_error_interrupt(p_my);

			if (p_my->int_table.os_error.enabled)
				p_my->int_table.os_error.prog(p_my->int_table.os_error.prog_arg);
			else
				cni_set_co(p_my->cni, 0);

			os_hold(p_my->error_delay);
			os_set_state(p_my, OS_STATE_OFF);
			break;
	}

	} /* for (;;) */
	log_assert(0);
csim_end_program

/* Loads the neccessary scheduling data from the MEDL table loaded into the
   attached controller. All time information is valid only after this function
   is executed properly. The affected fields include:
     slot_count, slot_time, round_len, transmission_cycle_len */
static void os_load_scheduling_data(OS_PROCESS *os)
{
	unsigned slot = os_get_my_slot(os);
	MEDL_SCE *sce = &os->controller->medl.slot_control_entry[0];
	unsigned i, sum, round, eor;

	sum = round = i = eor = 0;
	os->slot_count = 0;
	os->round_count = 0;

	do {
		os->slot_count++;
		os->slot_time[i] = sum;
		sum += sce[i].sru_slot_duration;
		if (sce[i].eor)
			os->round_count++;
		if (!eor) {
			round = sum;
			eor = sce[i].eor;
		}
	} while (sce[i++].eot == 0);

	sum += sce[i].sru_slot_duration;
	os->slot_time[i + 1] = sum;

	os->transmission_cycle_len = sum;
	os->round_len = round;

	os->send_time = (os->slot_time[slot] - IFG_DURATION + os->round_len) % os->round_len;
}

/**.
*/
static OS_PROCESS *os_init(OS_PROCESS *os)
{
	os->controller = NULL;
	os->cni = NULL;
	os->startup_count = 0;

	memset(&os->edm_counters, 0, sizeof(os->edm_counters));
	memset(&os->init_task, 0, sizeof(os->init_task));
	memset(&os->int_table, 0, sizeof(os->int_table));

	os->last_timer = csim_time();
	os->os_state = OS_STATE_OFF;
	os->interrupted = 0;
	os->start_allowed = 1;

	return os;
}


/************************************************************************************/

/* Return pointer to the default OS fault mask (static allocation). */
unsigned* os_fault_mask(void)
{
    int i;
    static OS_PROCESS mask;
    memset(&mask, 0, sizeof(OS_PROCESS));

    mask.init_task.time = bit_mask(16);
    mask.init_task.modes = 0xFFFFFFFF;
    for (i = 0; i < OS_MAX_TASK_NUM; i++) {
        mask.tasks[i].time = bit_mask(16);
        mask.tasks[i].modes = 0xFFFFFFFF;
    }
    mask.os_state = (OS_STATES)bit_mask(3);
    mask.old_msg_ccf = bit_mask(2);
    mask.startup_count = bit_mask(16);
    mask.transmission_cycle = bit_mask(16);
    mask.min_node_count = bit_mask(8);
    mask.first_app_mode = bit_mask(5);
    mask.start_allowed = 1;
    mask.next_task = bit_mask(OS_TASK_NUM_BITS);
    mask.next_task_time = bit_mask(16);
    mask.iteration = bit_mask(16);
    mask.old_controller_lifesign = bit_mask(16);
/*  
    mask.slot_time[] = 
    mask.int_table = 
    mask.error_code = bit_mask(16);
    mask.interrupted = 1;
    mask.edm_counters =
    mask.auto_life_sign = 1;
    mask.int_log_enabled = 1;
    mask.task_count = bit_mask(OS_TASK_NUM_BITS);
    mask.transmission_cycle_len = bit_mask(16);
    mask.round_len = bit_mask(16);
    mask.slot_count = bit_mask(9);
    mask.round_count = bit_mask(8);
    mask.send_time = bit_mask(16);
*/
    return (unsigned*)&mask;
}

/**.
@param name Name for the created C-Sim process. This name is used for debugging and
	logging and it needs not be unique.
@return Pointer to a new instance of @c OS_PROCESS.*/
OS_PROCESS *os_create(char *name)
{
	OS_PROCESS *os;

	os = csim_new_process(OS_PROCESS, os_prog);
	csim_set_name((CSIM_DYN_MEM *)os, name);
	return os_init(os);
}


/**.
The OS parses MEDL table found within the controller which is not exactly what the
HW implementation does, but as long as the MEDL is non-faulty, it is correct.
\attention The OS must be attached to a controller to operate correctly.
@param os Pointer to OS_PROCESS instance.
@param ctrl Pointer to an instance of CONTROLLER or derived structure.*/
void os_attach_controller(OS_PROCESS *os, CONTROLLER *ctrl)
{
	os->controller = ctrl;
	os->cni = &ctrl->cni;

	controller_attach_host(ctrl, (void *)os, controller_interrupt);
}


/**.
If multiple tasks are scheduled to the same time they are executed in the order in 
which they were added. The task is considered to be periodical, it will be executed
exactly once in every transmission cycle.
@param os Pointer to OS_PROCESS instance.
@param name Internal name of the task used for logging mainly (needs not be unique).
@param time Schedule time of the task (in macroticks relative to the start of
	transmission cycle).
@param modes 32-bit mask that specifies in which modes the task may be executed - bits 
	set to 1 enable execution in the corresponding mode. Least significant bit is for
	startup mode, application modes follow sequentially.
@param prog Pointer to task function - the function must terminate without waiting 
	for any other tasks.
@param prog_arg Agrument passed to the task function every time it is called.
@return
	- current position of the task in the scheduler (time ordered list)
	- value -1 when an error occurs (maximum number of tasks exceeded.*/
unsigned os_add_task(OS_PROCESS *os, char *name, unsigned time, unsigned modes, OS_TASK_FN prog, void *prog_arg)
{
	unsigned i, j;

        log(("Scheduling task '%s' at time = %d.\n", name, time));
	if (os->task_count == OS_MAX_TASK_NUM)
		return (unsigned) ERROR;

	for (i = 0; i < os->task_count && os->tasks[i].time <= time; i++)
		;

	for (j = os->task_count; j > i; j--)
	{
		strcpy(os->tasks[j].name, os->tasks[j-1].name);
		os->tasks[j].prog = os->tasks[j-1].prog;
		os->tasks[j].prog_arg = os->tasks[j-1].prog_arg;
		os->tasks[j].time = os->tasks[j-1].time;
		os->tasks[j].modes = os->tasks[j-1].modes;
	}
	os->task_count++;

	strcpy(os->tasks[i].name, name);
	os->tasks[i].prog = prog;
	os->tasks[i].prog_arg = prog_arg;
	os->tasks[i].time = time;
	os->tasks[i].modes = modes;

	return i;
}


/**.
@param os Pointer to OS_PROCESS instance.
@param prog Pointer to task function - the function must terminate without waiting 
	for any other tasks.
@param prog_arg Agrument passed to the task function every time it is called.*/
void os_add_init_task(OS_PROCESS *os, OS_TASK_FN prog, void *prog_arg)
{
	strcpy(os->init_task.name, "initialization");
	os->init_task.prog = prog;
	os->init_task.prog_arg = prog_arg;
	os->init_task.time = 0;
}


/**.
Schedules an internal task that is executed in every operation mode and updates
the host lifesign field correctly. There should be one such task for every sending slot
of the node (i.e. for every round in most cases).
@param os Pointer to OS_PROCESS instance.
@param time Schedule time of the task (in macroticks relative to the start of 
	transmission cycle).
@return
	- current position of the task in the scheduler (time ordered list)
	- value -1 when an error occurs (maximum number of tasks exceeded.*/
unsigned os_add_lifesign_update(OS_PROCESS *os, unsigned time)
{
	log(("Adding life-sign task.\n"));
	return os_add_task(os, "lifesign update", time, 0xFFFFFFFF, lifesign_update_prog, (void *)os);
}


/**.
@param os Pointer to OS_PROCESS instance.
@param name All tasks with this name will be removed from scheduler. 
@return Zero if no error occurs, nonzero otherwise. Currently the attempt to remove
	a nonexistent task is not considered to be an error.*/
unsigned os_remove_task_name(OS_PROCESS *os, char *name)
{
	unsigned i = 0;

	log(("Removing task with name '%s'.\n", name));
	while (i < os->task_count) {
		if (strcmp(name, os->tasks[i].name) == 0)
			return os_remove_task_id(os, i);
		i++;
	}
	return 0;
}


/**.
@note This function uses the internal position of a task within a list. This position
	may not be stable when task list is managed, so direct use of this function is
	deprecated - use os_remove_task_name() instead.
@param os Pointer to OS_PROCESS instance.
@param id Number of the task to remove (position within scheduler).
@return Zero if no error occurs, nonzero otherwise. Currently the attempt to remove
	a nonexistent task is not considered to be an error.*/
unsigned os_remove_task_id(OS_PROCESS *os, unsigned id)
{
	unsigned mode = os_get_current_mode(os);
	unsigned next = os->next_task;
	unsigned i;

	log_assert(os->task_count > 1);
	log(("Removing task with ID %d.\n", id));

	if (id >= os->task_count)
		return 0;

	if (id == os->next_task) {
		do {
			next = (next + 1) % os->task_count;
		} while (!enabled_in_mode(mode, os->tasks[next].modes));
		schedule_task(os, next);
	}

	os->task_count--;
	for (i = id; i < os->task_count; i++)
		memcpy(&os->tasks[i], &os->tasks[i+1], sizeof(OS_TASK));
	return 0;
}


/**.
@param os Pointer to OS_PROCESS instance.
@return Nonzero if controller correctly updates his lifesign, zero otherwise.
\todo The OS should implement a default handler for failed controller lifesign.*/
unsigned os_test_controller_lifesign(OS_PROCESS *os)
{
	CNI *cni = os->cni;
	int ret;

	log(("Udating host lifesign.\n"));
	cni_set_host_lifesign(cni, ~cni_get_controller_lifesign(cni) & bit_mask(16));
	ret = (cni_get_controller_lifesign(cni) != os->old_controller_lifesign);
	os->old_controller_lifesign = cni_get_controller_lifesign(cni);

	return ret;
}


/**.
@param os Pointer to OS_PROCESS instance.
@return Zero if CNI status area can be read, nonzero if it is locked by the controller.*/
unsigned os_check_ccf(OS_PROCESS *os)
{
	return (cni_get_ccf(os->cni) & 1);
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Pointer to the CNI structure.*/
CNI *os_get_cni(OS_PROCESS *os)
{
	return os->cni;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Pointer to the MEDL structure.
*/
MEDL *os_get_medl(OS_PROCESS *os)
{
	return &os->controller->medl;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Current time in macroticks.*/
unsigned os_get_time(OS_PROCESS *os)
{
	unsigned dt;

	dt = (os->next_task_time - cni_get_cluster_time(os->cni)) & bit_mask(16);

	return os->tasks[os->next_task].time - dt;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Number of sending slot of the node.*/
unsigned os_get_my_slot(OS_PROCESS *os)
{
	MEDL *medl = &os->controller->medl;

	return logical_name_get_slot_position(&medl->role_list.logical_name);
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Current slot number.*/
unsigned os_get_current_slot(OS_PROCESS *os)
{
	return (c_state_get_medl_position(&os->cni->c_state) - 1) % os->slot_count;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@param mode Number of mode that will be active while the request is made.
@param slot Slot in which the request will be made.
@param idx Number of the requested mode (relative for deferred mode change).
@param rq_type Either 'immediate', 'deferred' or 'clear'.
@return Nonzero if the request is permitted, zero otherwise.*/
unsigned os_get_mode_change_permission(OS_PROCESS *os, unsigned mode, unsigned slot, unsigned idx, unsigned rq_type)
{
	if (rq_type == DEFERRED)
		return medl_mae_get_mcp_d(&os->controller->medl.mode_address_entry[mode][slot], idx);
	if (rq_type == IMMEDIATE)
		return medl_mae_get_mcp_i(&os->controller->medl.mode_address_entry[mode][slot], idx);

	return medl_mae_get_mcp_c(&os->controller->medl.mode_address_entry[mode][slot]);
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Current mode number.*/
unsigned os_get_current_mode(OS_PROCESS *os)
{
	return c_state_get_mode(&os->cni->c_state);
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@param mode Requsted mode - relative number.
@param immediate If set, then the request will be immediate, otherwise deferred.
@return Zero when successfull, nonzero otherwise.*/
unsigned os_request_mode_change(OS_PROCESS *os, unsigned mode, unsigned immediate)
{
	log(("Requesting %s mode change, (mode = %u)\n",
		(immediate) ? "imediate" : "deferred", mode));
	cni_mcr_set_mode(os->cni, mode);
	cni_mcr_set_immediate(os->cni, immediate);
	return 0;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Zero when successfull, nonzero otherwise.*/
unsigned os_cancel_mode_change_request(OS_PROCESS *os)
{
	cni_mcr_set_mode(os->cni, 0);
	cni_mcr_set_immediate(os->cni, 1);
	return 0;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Current number of active nodes.*/
unsigned os_get_node_count(OS_PROCESS *os)
{
	C_STATE *c_state = &os->cni->c_state;
	unsigned i, max;
	unsigned count = 0;

	max = medl_ccp_get_c_state_length(&os->controller->medl.controller_cfg_params);
	max = (max - 2) * 16;

	for (i = 0; i < max; i++)
		count += c_state_get_membership(c_state, i);

	return count;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Pointer to internal EDM counters.*/
OS_EDM_COUNTERS *os_get_edm_counters(OS_PROCESS *os)
{
	return &os->edm_counters;
}


/**.
@param os Pointer to an instance of OS_PROCESS.*/
void os_clear_edm_counters(OS_PROCESS *os)
{
	memset(&os->edm_counters, 0, sizeof(os->edm_counters));
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Number of startups (passes through init state).*/
unsigned os_get_startup_count(OS_PROCESS *os)
{
	return os->startup_count;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Current transmission cycle number.*/
unsigned os_get_transmission_cycle(OS_PROCESS *os)
{
	return os->transmission_cycle;
}


/**.
@param os Pointer to an instance of OS_PROCESS.*/
void os_start_controller(OS_PROCESS *os)
{
	log(("Enabling Controller Start\n"));
	os->start_allowed = 1;
	if (!process_is_planned((CSIM_PROCESS *)os))
		process_activate_delay((CSIM_PROCESS *)os, 0.0);
}


/**.
The logging process does not influence the application.
@param os Pointer to an instance of OS_PROCESS.
@param enable Set to 0 to disable the counting of interrupts, otherwise it will be 
	enabled.*/
void os_set_int_log(OS_PROCESS *os, unsigned enable)
{
	CNI_IE *ie = &os->cni->interrupt_enable;

	if (os->int_log_enabled == enable)
		return;

	os->int_log_enabled = enable;
	if (enable) {
		log(("Enabling logged interrupts: BE, PE, HE, ML, UI1, UI2.\n"));
		cni_ie_set_be(ie, 1);
		cni_ie_set_pe(ie, 1);
		cni_ie_set_he(ie, 1);
		cni_ie_set_ml(ie, 1);
		cni_ie_set_ui1(ie, 1);
		cni_ie_set_ui2(ie, 1);
	} else {
		test_clear(be);
		test_clear(pe);
		test_clear(he);
		test_clear(ml);
		test_clear(ui1);
		test_clear(ui2);
	}
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@param min_cnt Minimum number of active nodes (in membership).
@param mode Relative number of requested mode (always deferred).*/
void os_set_mode_trigger(OS_PROCESS *os, unsigned min_cnt, unsigned mode)
{
	os->min_node_count = min_cnt;
    os->first_app_mode = mode;
}


/**.
@param os Pointer to an instance of OS_PROCESS.*/
void os_restart(OS_PROCESS *os)
{
	log(("Restarting OS and the controller\n"));
	raise_os_error_int(os, OS_ERROR_RESTART);
	os->start_allowed = 1;
	process_activate_delay((CSIM_PROCESS *)os, 0.0);
}


/**.
@param cnt Pointer to EDM counters.
@return Pointer to a static string.*/
char *os_counters2str(OS_EDM_COUNTERS *cnt)
{
	static char s[256];

	s[0] = 0;
	if (cnt->ml) strcat(s, "ml ");
	if (cnt->mi) strcat(s, "mi ");
	if (cnt->mc) strcat(s, "mc ");
	if (cnt->pe_be) strcat(s, "pe_be ");
	if (cnt->pe_ae) strcat(s, "pe_ae ");
	if (cnt->pe_se) strcat(s, "pe_se ");
	if (cnt->pe_me) strcat(s, "pe_me ");
	if (cnt->pe_mc) strcat(s, "pe_mc ");
	if (cnt->pe_cb) strcat(s, "pe_cb ");
	if (cnt->pe_dc) strcat(s, "pe_dc ");
	if (cnt->he_so) strcat(s, "he_so ");
	if (cnt->he_mv) strcat(s, "he_mv ");
	if (cnt->he_ne) strcat(s, "he_ne ");
	if (cnt->be_sc) strcat(s, "be_sc ");
	if (cnt->be_av) strcat(s, "be_av ");
	if (cnt->be_br) strcat(s, "be_br ");
	if (cnt->be_rm) strcat(s, "be_rm ");
	if (cnt->be_rf) strcat(s, "be_rf ");
	if (cnt->be_wd) strcat(s, "be_wd ");
	if (cnt->os_error) strcat(s, "oserr ");
	if (cnt->ti2) strcat(s, "ti2 ");
	if (cnt->ui1) strcat(s, "ui1 ");
	if (cnt->ui2) strcat(s, "ui2 ");

	if (strlen(s) > 0)
		s[strlen(s) - 1] = 0;
	return s;
}


/**.
If the counters contain an OS error, the detailed error code is also included in the
result.
@param os Pointer to an instance of OS_PROCESS.
@return Statically allocated string.*/
char *os_get_edm_str(OS_PROCESS *os)
{
	char *org = os_counters2str(&os->edm_counters);
	static char s[256];

	if (os->edm_counters.os_error) {
		sprintf(s, "%s(%u)", org, (unsigned)os->error_code);
		return s;
	}

	return org;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@param slot Zero-base number of slot.
@return The action time of specified slot (slot 0 always sends at zero time).*/
unsigned os_get_slot_time(OS_PROCESS *os, unsigned slot)
{
	if (slot > os->slot_count)
        	return 0;
        else
        	return os->slot_time[slot];
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@return Duration of TDMA round in macroticks.*/
unsigned os_get_round_length(OS_PROCESS *os)
{
	return os->round_len;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@param round Number of TDMA round that is the base for this request.
@param offset This value is added to the time module TDMA round length, i.e. it will not
	shoft the returned time outside the chosen round.
@return Send time for the specified round modified by an offset.*/
unsigned os_get_send_time(OS_PROCESS *os, unsigned round, int offset)
{
	unsigned round_len = os_get_round_length(os);

	return (os->send_time + round_len + offset) % round_len + round_len * round;
}



/*************************************************************************************/
/**@name Message Writing Functions
 * @{ */
/*************************************************************************************/

/** Set data in message starting from the specified offset. */
void os_set_ofs_data(OS_PROCESS *os, unsigned addr, unsigned ofs, char *buf, unsigned size)
{
	CNI_MAE *mae = cni_get_message_entry(os->cni, addr);
	unsigned i;

	for (i = 0; i < size; i++)
		cni_mae_set_data(mae, i + ofs, (unsigned char)buf[i]);
}

/* Preprocessor generated set functions. */
implement_set_ofs_fn(char, unsigned char)
implement_set_ofs_fn(short, unsigned short)
implement_set_ofs_fn(int, unsigned int)
implement_set_ofs_fn(long, unsigned long)
implement_set_ofs_fn(float, float)
implement_set_ofs_fn(double, double)

/*@}*/
	
/*************************************************************************************/
/**@name Message Reading Functions
 * @{ */
/*************************************************************************************/

/** Retrieves received data, compares both messagess, checks EIF.
 *
 * @return If no valid data are found nonzero code is returned, 0 otherwise. */
unsigned os_get_ofs_data(OS_PROCESS *os, unsigned addr, unsigned ofs, char *buf, unsigned size)
{
	CNI_MAE *mae = cni_get_message_entry(os->cni, addr);
	unsigned i;

	if (!eif_is_valid(os, addr))
		return 1;

	for (i = 0; i < size; i++)
		buf[i] = (char)cni_mae_get_data(mae, i + ofs);

	return 0;
}

/** Functions that retreive data from replicated messages and write the one with
 * correct EIF into 'val'.
 * @return  0 if successfull, nonzero if no valid message was found. */
unsigned os_get_ofs_rep_data(OS_PROCESS *os, const unsigned addr[], unsigned ofs, char *buf, unsigned size)
{
	if (!os_get_ofs_data(os, addr[0], ofs, buf, size))
		return 0;
	return os_get_ofs_data(os, addr[1], ofs, buf, size);
}


/* Preprocessor generated get functions for replicated messages. */
implement_get_ofs_rep_fn(char, unsigned char)
implement_get_ofs_rep_fn(short, unsigned short)
implement_get_ofs_rep_fn(int, unsigned int)
implement_get_ofs_rep_fn(long, unsigned long)
implement_get_ofs_rep_fn(float, float)
implement_get_ofs_rep_fn(double, double)


/*@}*/
	
/*************************************************************************************/
/**@name Interrupt Registering Functions
 * @{ */
/*************************************************************************************/

/* Register interrupt handling routines for every possible condition.
   Only one routine may be registered per interrupt, the old will be replaced. */
implement_add_int_fn(be)
implement_add_int_fn(pe)
implement_add_int_fn(he)
implement_add_int_fn(ml)
implement_add_int_fn(mc)
implement_add_int_fn(mi)
implement_add_int_fn(ui1)
implement_add_int_fn(ui2)
implement_add_int_fn(ti1)
implement_add_int_fn(ti2)
implement_add_int_fn(os_error)
implement_add_int_fn(os_active)


/*@}*/
/*************************************************************************************/
	
/*************************************************************************************/
/**@name Interrupt Enable/Disable Functions
 * @{ */

/* Enable or disable specified interrupt functions: */
implement_enable_int_fn(be)
implement_enable_int_fn(he)
implement_enable_int_fn(pe)
implement_enable_int_fn(ml)
implement_enable_int_fn(mc)
implement_enable_int_fn(mi)
implement_enable_int_fn(ui1)
implement_enable_int_fn(ui2)

/** Enable or disable TI2 interrupt function */
void os_enable_ti2_int(OS_PROCESS *os, unsigned csim_state)
{
	if (csim_state == os->int_table.ti2.enabled)
		return;
	os->int_table.ti2.enabled = csim_state;
	if (!csim_state && os->int_log_enabled)
		return;
	cni_ie_set_ti2(os->cni, csim_state);
}

/* Enable or disable os_error interrupt function */
void os_enable_os_error_int(OS_PROCESS *os, unsigned csim_state)
{
	if (csim_state == os->int_table.os_error.enabled)
		return;
	os->int_table.os_error.enabled = csim_state;
}


/* Enable or disable os_active interrupt function */
void os_enable_os_active_int(OS_PROCESS *os, unsigned csim_state)
{
	if (csim_state == os->int_table.os_active.enabled)
		return;
	os->int_table.os_active.enabled = csim_state;
}


/*@}*/
/*************************************************************************************/

/*************************************************************************************/
/**@name Other Message-Handling Functions
 * @{ */

/**.
After the message is not longer accessed, @c os_ccf_leave() function must be
called to release the lock. The function ensures that the CCF field is set to a valid
value even if it was faulty before - in such case an error is still returned, but it
may be discarded safelly by the application.
@param os Pointer to OS_PROCESS instance.
@param addr Word offset into CNI message area that points to a valid message area entry.
@return Zero if everything is fine, nonzero code otherwise.*/
unsigned os_ccf_enter(OS_PROCESS *os, unsigned addr)
{
	CNI_MAE *mae = cni_get_message_entry(os->cni, addr);
	
	os->old_msg_ccf = cni_mae_increment_ccf(mae, 1);
	if (os->old_msg_ccf & 1)
		return 0;   /* The value is odd and CCF works fine */
	
	/* Here we know, that CCF field contains errorneous value - it is even
	   We can continue by incrementing the value to the next odd, but we should 
	   also tell the user - one fault is seldom alone :-) */
	os->old_msg_ccf = cni_mae_increment_ccf(mae, 1);
	return 1;
}


/**.
The message area entry must be alerady locked by a call to @c os_ccf_enter() function.
This function ensures that the CCF field is set to a valid value even if it was faulty 
before - in such case an error is still returned, but it may be discarded safelly by 
the application.
@param os Pointer to OS_PROCESS instance.
@param addr Word offset into CNI message area that points to a valid message area entry.
@return Zero if everything is fine, nonzero code otherwise.*/
unsigned os_ccf_leave(OS_PROCESS *os, unsigned addr)
{
	CNI_MAE *mae = cni_get_message_entry(os->cni, addr);
	unsigned i;
	
	i = cni_mae_get_ccf(mae);

	if (i & 1)
		cni_mae_increment_ccf(mae, 1);   /* value is odd, increment to even */
	else
		cni_mae_increment_ccf(mae, 2);   /* value is even - increment to next even */
		
	if (i == os->old_msg_ccf)
		return 0;

	/* Here we know, that CCF field is faulty - it sgould be the same as when ..._enter
	   was called. We can continue, but we should also tell the user - one fault is 
	   seldom alone :-) */
	return 1;
}


/**.
@param os Pointer to an instance of OS_PROCESS.
@param addr Word offset within CNI message area.
@return EIF value.*/
CNI_EIF_VALUES os_get_eif(OS_PROCESS *os, unsigned addr)
{
	return (CNI_EIF_VALUES)cni_mae_get_eif(cni_get_message_entry(os->cni, addr));
}


/**.
@param os Pointer to OS_PROCESS instance.
@param addr Word offset into CNI message area that points to a message area entry.*/
void os_validate_eif(OS_PROCESS *os, unsigned addr)
{
	CNI *cni = os->cni;

	log(("Validating EIF.\n"));
	cni_mae_set_eif(cni_get_message_entry(cni, addr), EIF_CORRECT_FRAME);
}


/**.
@param os Pointer to OS_PROCESS instance.
@param addr Word offset into CNI message area that points to a message area entry.*/
void os_invalidate_eif(OS_PROCESS *os, unsigned addr)
{
	CNI *cni = os->cni;

	log(("Invalidating EIF.\n"));
	cni_mae_set_eif(cni_get_message_entry(cni, addr), EIF_NULL_FRAME);
}

/*@}*/
/*************************************************************************************/

/*@}*/
/*@}*/

