/**@addtogroup misc_sim
 * @{*/

/**@addtogroup misc_sim_int  Implementation
 * @{*/


/**@file
 * @brief Usefull functions designed to work with the C-Sim library
 *  (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16 
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 * 
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project \b FIT - Fault Injection for TTA, IST-1999-10748 
 */

#include "ttpc/fl1debug.h"
#include "ttpc/sl1func.h"
#include "csim/csim.h"

/*************************************************************************************/
/* Exported Functions */
/*************************************************************************************/

/**.
\param process Pointer to an instance of C-Sim process.
\return Nonzero code if the process is in planned (or new planned) state.
*/
int process_is_planned(CSIM_PROCESS *process)
{
	CSIM_PROC_STATE pstate;

	log_assert(process != NULL);
	pstate = csim_state(process);
	return (pstate == CSIM_PLANNED || pstate == CSIM_NEW_PLANNED);
}


/**.
\param process Pointer to an instance of C-Sim process.
\return Nonzero code if the process is in active state (currently executed).
*/
int process_is_active(CSIM_PROCESS *process)
{
	log_assert(process != NULL);
	return (process == csim_current());
}


/**.
\param process Pointer to an instance of C-Sim process.
\return Nonzero code if the process is in terminated state (not executable).
*/
int process_is_terminated(CSIM_PROCESS *process)
{
	log_assert(process != NULL);
	return (csim_state(process) == CSIM_TERMINATED);
}


/**.
\param process Pointer to an instance of C-Sim process.
\return Zero if successfull and the process is no more planned, nonzero when the
	process is currently running (active).*/
int process_cancel(CSIM_PROCESS *process)
{
	log_assert(process != NULL);

	if (process_is_active(process))
		return 1;

	if (process_is_planned(process))
		csim_cancel(process);

	return 0;
}

/**.
\param process Pointer to an instance of C-Sim process.
\param time Absolute time value of next schduling point of the process.
\return
	- 0 on success.
	- 1 if process is ACTIVE.
	- 2 if process is TERMINATED.*/
int process_activate_at(CSIM_PROCESS *process, CSIM_TIME time)
{
	log_assert(process != NULL);

	if (process_is_active(process))
		return 1;
	if (process_is_terminated(process))
		return 2;

	csim_reactivate_at(process, time);
	return 0;
}

/**
\param process Pointer to an instance of C-Sim process.
\param time Time value relative to current time that is used to schedule the process.
\return
	- 0 on success
	- 1 if process is ACTIVE 
	- 2 if process is TERMINATED.*/
int process_activate_delay(CSIM_PROCESS *process, CSIM_TIME time)
{
	log_assert(process != NULL);

	if (process_is_active(process))
		return 1;
	if (process_is_terminated(process))
		return 2;

	csim_reactivate_delay(process, time);
	return 0;
}

/*************************************************************************************/

/*@}*/
/*@}*/

