/**@addtogroup bg_sim
 * @{*/

/**@addtogroup bg_sim_int  Implementation
 * @{*/


/**@file
 * @brief C-Sim simulation wrapper for bus guardian (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, University of West Bohemia, Czech Republic
 * @author Jan Jezek, University of West Bohemia, Czech Republic
 * 
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#include "csim/csim.h"
#include "ttpc/sp1busg.h"
#include "ttpc/fp1busg.h"
#include "ttpc/sl1func.h"
#include "ttpc/fl1debug.h"


/************************************************************************************/

/** This macro is a shortcut that makes the logging process througout the
	whole project transparent and unified.*/
#define log(msg) log_bg_csim(msg)

/** Convert BG time units into simulation model time using the clock drift
	specified within @a bg instance.*/
#define macro_to_csim(bg, time) \
	(BUS_GUARDIAN_CLK_SPD * (time) * (1.0 + (bg)->clock_drift))

/** Suspends the execution of current process for @a mt BG macroticks - uses clock
	drift specified within @a bg_model instance.*/
#define bg_hold(bg_model, mt) do { \
	log(("Holding: %0.4f\n", macro_to_csim(bg_model, mt))); \
	csim_hold(macro_to_csim(bg_model, mt)); \
} while(0)


/************************************************************************************/
/* Static Functions */
/************************************************************************************/

/** Set BG arm signal and check for possible errors.*/
static unsigned set_arm_signal(BUS_GUARDIAN *bus_guardian, unsigned value);

/** Program for bus guardian C-Sim process.*/
static void *prog_bus_guardian(void *p_void);

/** Destructor executed automatically by C-Sim when a BG process is terminated.*/
static void bus_guardian_destroy(void *p_void);

/************************************************************************************/

/**.
@param bg Pointer to an instance of BUS_GUARDIAN or a derived structure.
@param value The requested value of bus arm bit.
@return Zero if successfull, nonzero if any error occurs.*/
static unsigned set_arm_signal(BUS_GUARDIAN *bg, unsigned value)
{
	BUS_GUARDIAN_MODEL *bg_model = (BUS_GUARDIAN_MODEL *)bg;
	unsigned old_value = bus_guardian_get_am(bg);

	value &= 1;
	if (bus_guardian_set_am(bg, value))
		return 1;

	if (value == 1 && old_value != 1 && bg->bg_state == BGS_ARM) {
		log(("Arm. Open scheduled at %10.4f.\n",
			macro_to_csim(bg_model, bus_guardian_get_tso(bg)) + csim_time()
		));
		process_activate_delay((CSIM_PROCESS *)bg_model->process, 0);
	}

	return 0;
}


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).
@return Used internaly by C-Sim (only in the POSIX thread implementation).*/
static csim_program(BUS_GUARDIAN_PROCESS, prog_bus_guardian)
	for (;;) {
		switch (p_my->protocol->bg_state) {
			case BGS_INITIALIZE:
				bus_guardian_init_state(p_my->protocol);
				break;
			case BGS_CONFIGURE:
				log(("Configuration enabled.\n"));
				bus_guardian_configure_state(p_my->protocol);
				csim_passivate();
				break;
			case BGS_ARM:
				if (bus_guardian_get_am(p_my->protocol) != 1) {
					log(("Waiting for Bus Arm - arm phase.\n"));
#ifdef ENABLE_BG_MOD
					bg_hold(p_my->model, bus_guardian_get_tap(p_my->protocol));
#else
					csim_passivate();
#endif
				}
				bus_guardian_arm_state(p_my->protocol);
				break;
			case BGS_START_TIMER:
				log(("Slot offset phase.\n"));
				if (bus_guardian_get_tso(p_my->protocol) > 0)
					bg_hold(p_my->model, bus_guardian_get_tso(p_my->protocol));
				if (p_my->protocol->bg_state != BGS_START_TIMER)
					break;
				bus_guardian_start_timer_state(p_my->protocol);
				break;
			case BGS_ENABLE:
				log(("Bus access enabled.\n"));
				if (bus_guardian_get_tsd(p_my->protocol) > 0) {
					bg_hold(p_my->model, bus_guardian_get_tsd(p_my->protocol));
					if (p_my->protocol->bg_state != BGS_ENABLE)
						break;
				}
				bus_guardian_enable_state(p_my->protocol);
				break;
			case BGS_DISABLE:
				log(("Bus access disabled.\n"));
				if (bus_guardian_get_trr(p_my->protocol) > 0) {
					bg_hold(p_my->model, bus_guardian_get_trr(p_my->protocol));
					if (p_my->protocol->bg_state != BGS_DISABLE)
						break;
				}
				bus_guardian_disable_state(p_my->protocol);
				break;
			case BGS_NODE_FAILURE:
				log(("Node Failure.\n"));
				csim_passivate();
		}
	}
csim_end_program


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).*/
static csim_begin_destructor(BUS_GUARDIAN_PROCESS, bus_guardian_destroy)
	free(p_my->model);
csim_end_destructor

/******************************************************************************/
/* Exported Functions */
/************************************************************************************/


/**.
 *
 * This function allocates dynamic memory for a new BG process and model. The
 * memory is initialized with default values and a C-Sim name and destructor
 * are assigned to the process.
 *
 * @param name The internal name of the created BG process (used by C-Sim). It
 *  needs not be unique.
 * @return Pointer to a newly allocated and initialized instance of BG model
 *  (and C-Sim process).*/
BUS_GUARDIAN_MODEL *bus_guardian_create(char *name)
{
	BUS_GUARDIAN_MODEL *bg_model;

	bg_model = (BUS_GUARDIAN_MODEL *)malloc(sizeof(BUS_GUARDIAN_MODEL));
    if (bg_model == NULL)
        return NULL;

    bg_model->process = csim_new_process(BUS_GUARDIAN_PROCESS, prog_bus_guardian);
    if (bg_model->process == NULL)
        return NULL;
        
	bg_model->process->protocol = (BUS_GUARDIAN *)bg_model;
	bg_model->process->model = bg_model;
	csim_set_name((CSIM_DYN_MEM *)bg_model->process, name);
	csim_set_destructor(bg_model->process, bus_guardian_destroy);

	bus_guardian_init((BUS_GUARDIAN *)bg_model);
	bus_guardian_set_clock_drift(bg_model, 0.0);
	bg_model->arm_signal = set_arm_signal;

	return bg_model;
}


/**.
@param bg_model Pointer to an instance of BUS_GUARDIAN_MODEL or derived structure.
@param ctrl Pointer to an instance of CONTROLLER or derived structure.*/
void bus_guardian_attach_controller(BUS_GUARDIAN_MODEL* bg_model, void *ctrl)
{
	bg_model->controller = ctrl;
}


/**.
@param bg_model Pointer to an instance of BUS_GUARDIAN_MODEL or derived structure.
@param drift Value of new clock drift rate in [sec/sec].*/
void bus_guardian_set_clock_drift(BUS_GUARDIAN_MODEL *bg_model, double drift)
{
	bg_model->clock_drift = drift;
}


/*************************************************************************************/

/*@}*/
/*@}*/

