/**@addtogroup chan_sim
 * @{*/

/**@addtogroup chan_sim_int  Implementation
 * @{*/


/**@file
 * @brief C-Sim simulation wrapper for transmission channel (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project \b FIT - Fault Injection for TTA, IST-1999-10748
 */

#include "csim/csim.h"
#include "ttpc/fl1debug.h"
#include "ttpc/sp1chan.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1chan.h"
#include "ttpc/fp1frame.h"
#include "ttpc/sl1func.h"

/************************************************************************************/

/** This macro is a shortcut that makes the logging process througout the
	whole project transparent and unified.*/
#define log(msg) log_channel_csim(msg)

/** Number of bits transfered per time unit (depends on tranmsission speed).*/
#define CHANNEL_BIT_DURATION    ( 1.0 / CHANNEL_SPEED )

/** Length of Start-of-Frame pattern transmitted on the bus (in bits).*/
#define CHANNEL_SOF_LENGTH		8


/************************************************************************************/
/* Static Functions */
/************************************************************************************/

/** Initialize simulation specific parameters of the model.*/
static void channel_model_init(CHANNEL_MODEL *channel);

/** Program for channel model C-Sim process.*/
static void *channel_prog(void *p_void);

/** Destructor executed automatically by C-Sim when a channel process is terminated.*/
static void channel_destroy(void *p_void);


/************************************************************************************/

/**.
@param channel Pointer to an instance of CHANNEL_MODEL or derived structure.*/
static void channel_model_init(CHANNEL_MODEL *channel)
{
	channel_init((CHANNEL *)channel);

	channel->channel_send_frame = channel_send_frame;
}


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).
@return Used internaly by C-Sim (only in the POSIX thread implementation).*/
static csim_program(CHANNEL_PROCESS, channel_prog)
	for (;;) {
		if (!channel_test_bg(p_my->protocol, p_my->protocol->sender)) {
			channel_send_complete(p_my->protocol);
		} else {
			p_my->model->active = 0;
			p_my->model->frame.noise = 0;
			p_my->model->frame.state = FRAME_STATE_NULL;
		}

		csim_passivate();
	}
csim_end_program


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).*/
static csim_begin_destructor(CHANNEL_PROCESS, channel_destroy)
	free(p_my->model);
csim_end_destructor


/************************************************************************************/
/* Exported Functions */
/************************************************************************************/

/**.
This function allocates dynamic memory for a new channel process and model. The memory is initialized with
default values and a C-Sim name and destructor ase assigned to the process.
@param name The internal name of the created channel process (used by C-Sim). It needs not
	be unique.
@return Pointer to a newly allocated and initialized instance of channel model (and C-Sim process).*/
CHANNEL_MODEL *channel_create(char *name)
{
	CHANNEL_MODEL *channel = (CHANNEL_MODEL *)malloc(sizeof(CHANNEL_MODEL));
	CHANNEL_PROCESS *csim_process = csim_new_process(CHANNEL_PROCESS, channel_prog);

	channel->process = csim_process;
	csim_process->protocol = (CHANNEL *)channel;
	csim_process->model = channel;
	csim_set_destructor(csim_process, channel_destroy);
	csim_set_name((CSIM_DYN_MEM *)csim_process, name);

	channel_model_init(channel);

	return channel;
}


/**.
This function uses protocol defined functions from lower layers and adds simulation
of transmission duration. The duration is calculated from bus speed and frame length.
@param channel Pointer to an instance of CHANNEL or derived structure.
@param ctrl Pointer to an instance of CONTROLLER or derived strcture.
@param callback This function is stored and called when the sending is finished
	to notify the sender.
@return Nonzero code if any error occurs, zero otherwise.
@todo Is it neccessary to for this function to be @c extern class?*/
int channel_send_frame(CHANNEL *channel, void *ctrl, PTR_CALLBACK callback)
{
	CHANNEL_MODEL *model = (CHANNEL_MODEL *)channel;
	CSIM_TIME duration;

	if (channel_send_init(channel, ctrl, callback))
		return 1;

	duration = (CHANNEL_SOF_LENGTH + frame_get_length(&channel->frame)) * CHANNEL_BIT_DURATION;

	process_activate_delay((CSIM_PROCESS *)model->process, duration);

	log(("Sending %c-frame (duration %0.2f, arival at %0.2f)\n",
		(channel->frame.header.i_n_bit == FRAME_TYPE_I) ? 'I' : 'N',
		duration,
		csim_time() + duration
	));

	return 0;
}


/**.
@param channel Pointer to an instance of CHANNEL or derived structure.
@param ctrl Pointer to an instance of CONTROLLER or derived strcture (the sender).
@todo This function should be available from @c fp1chan.h, i.e. it should
	by implemented as a virtual function.*/
void channel_cancel_send(CHANNEL *channel, void *ctrl)
{
	if (ctrl == channel->sender && channel->active) {
		channel->frame.state = FRAME_STATE_RECEIVING;

		process_cancel((CSIM_PROCESS *)((CHANNEL_MODEL *)channel)->process);

		if (!channel_test_bg(channel, channel->sender)) {
			channel_send_complete(channel);
		} else {
			channel->active = 0;
			channel->frame.noise = 0;
			channel->frame.state = FRAME_STATE_NULL;
		}

	}
}

/************************************************************************************/

/*@}*/
/*@}*/

