/**@addtogroup cni_sim
 * @{*/

/**@addtogroup cni_sim_int  Implementation 
 * @{*/


/**@file
 * @brief Simulation required initialization of CNI virtual functions
 *  (implementation).
 * @version 1.5
 * @date Last modified 2002-11-16 */

#include "ttpc/fl1debug.h"
#include "ttpc/sp1cni.h"
#include "ttpc/fp1cni.h"
#include "ttpc/fl1func.h"
#include "ttpc/sp1ctrl.h"

/************************************************************************************/

/** This macro is a shortcut that makes the logging process througout the
	whole project transparent and unified.*/
#define log(msg) log_ctrl2_csim(msg)


/************************************************************************************/
/* Static Functions */
/************************************************************************************/

/** Calculate (from model time) and store the local cluster time fielde.*/
static unsigned static_cni_get_cluster_time(CNI *cni);

/** Set @c co field within CNI and possibly wake-up the controller.*/
static unsigned static_cni_set_co(CNI *cni, unsigned value);

/** Set @c ca field within CNI and possibly wake-up the controller.*/
static unsigned static_cni_set_ca(CNI *cni, unsigned value);

/** Set @c timer1 field within CNI and possibly reschedule the timer process.*/
static unsigned static_cni_set_timer1(CNI *cni, unsigned value);

/** Set @c timer2 field within CNI and possibly reschedule the timer process.*/
static unsigned static_cni_set_timer2(CNI *cni, unsigned value);

/** Set @c ti1 field within CNI (interrupt enable) and possibly reschedule the timer process.*/
static unsigned static_cni_ie_set_ti1(CNI *cni, unsigned value);

/** Set @c ti1 field within CNI (interrupt enable) and possibly reschedule the timer process.*/
static unsigned static_cni_ie_set_ti2(CNI *cni, unsigned value);


/************************************************************************************/

/**.
 *
 * @param cni Pointer to an instance of CNI or derived structure.
 * @return The current value of cluster time field.*/
static unsigned static_cni_get_cluster_time(CNI *cni)
{
	((CONTROLLER *)cni->controller)->update_cluster_time((CONTROLLER *)cni->controller);
	return 	cni->cluster_time;
}


/**.
 *
 * This function checks whether the requested write operation changes the current value
 * of @c co field and in such case notifies the controller about the change by
 * calling a function exported by the owner controller.
 * 
 * @param cni Pointer to an instance of CNI or derived structure.
 * @param value This value will be written into the CNI.
 * @return The value actually written into the @c co field.*/
static unsigned static_cni_set_co(CNI *cni, unsigned value)
{
	if (cni->co == value) {
		log(("The Value of CO did not change (0x%4X)\n", value));
		return value;
	}
	cni->co = value & bit_mask(16);
	controller_co_changed((CONTROLLER_MODEL *)cni->controller);

	return cni->co;
}


/**.
 *
 * This function checks whether the requested write operation changes the current value
 * of @c ca field and in such case notifies the controller about the change by
 * calling a function exported by the owner controller.
 *
 * @param cni Pointer to an instance of CNI or derived structure.
 * @param value This value will be written into the CNI.
 * @return The value actually written into the @c ca field.*/
static unsigned static_cni_set_ca(CNI *cni, unsigned value)
{
	if (cni->ca == value)
		return value;
	cni->ca = value & bit_mask(16);
	controller_ca_changed((CONTROLLER_MODEL *)cni->controller);

	return cni->ca;
}


/**.
 *
 * This function checks whether the requested write operation changes the current value
 * of @c timer1 field and in such case notifies the controller about the change by
 * calling a function exported by the owner controller.
 *
 * @param cni Pointer to an instance of CNI or derived structure.
 * @param value This value will be written into the CNI.
 * @return The value actually written into the @c timer1 field.*/
static unsigned static_cni_set_timer1(CNI *cni, unsigned value)
{
	if (cni->timer1 == value)
		return value;
	cni->timer1 = value & bit_mask(16);
	controller_timer_changed((CONTROLLER_MODEL *)(cni->controller), 0, 0);

	return cni->timer1;
}


/**.
 *
 * This function checks whether the requested write operation changes the current value
 * of @c timer2 field and in such case notifies the controller about the change by
 * calling a function exported by the owner controller.
 *
 * @param cni Pointer to an instance of CNI or derived structure.
 * @param value This value will be written into the CNI.
 * @return The value actually written into the @c timer2 field.*/
static unsigned static_cni_set_timer2(CNI *cni, unsigned value)
{
	if (cni->timer2 == value)
		return value;
	cni->timer2 = value & bit_mask(16);
	controller_timer_changed((CONTROLLER_MODEL *)(cni->controller), 1, 0);

	return cni->timer2;
}


/**.
 * 
 * This function checks whether the requested write operation changes the current value
 * of @c ti1 field and in such case notifies the controller about the change by
 * calling a function exported by the owner controller.
 * 
 * @param cni Pointer to an instance of CNI or derived structure.
 * @param value This value will be written into the CNI.
 * @return The value actually written into the @c ti1 field.
 *
 * @todo This function should take as parameter a pointer to @c CNI_IE data structure for
 *  it changes one of its fields, however this structure does not contain
 *  pointer to the owning controller.*/
static unsigned static_cni_ie_set_ti1(CNI *cni, unsigned value)
{
	if (cni->interrupt_enable.ti1 == value)
		return value;
	cni->interrupt_enable.ti1 = value & bit_mask(1);
	controller_timer_changed((CONTROLLER_MODEL *)(cni->controller), 0, 0);

	return cni->interrupt_enable.ti1;
}


/**.
 *
 * This function checks whether the requested write operation changes the current value
 * of @c ti2 field and in such case notifies the controller about the change by
 * calling a function exported by the owner controller.
 *
 * @param cni Pointer to an instance of CNI or derived structure.
 * @param value This value will be written into the CNI.
 * @return The value actually written into the @c ti2 field.
 *
 * @todo This function should take as parameter a pointer to @c CNI_IE data structure for
 *  it changes one of its fields, however this structure does not contain
 *  pointer to the owning controller.*/
static unsigned static_cni_ie_set_ti2(CNI *cni, unsigned value)
{
	if (cni->interrupt_enable.ti2 == value)
		return value;
	cni->interrupt_enable.ti2 = value & bit_mask(1);
	controller_timer_changed((CONTROLLER_MODEL *)(cni->controller), 1, 0);

	return cni->interrupt_enable.ti2;
}


/************************************************************************************/
/* Exported Functions */
/************************************************************************************/


/**.
*/
void cni_module_initialize(void)
{
	cni_get_cluster_time = static_cni_get_cluster_time;
	cni_set_co = static_cni_set_co;
	cni_set_ca = static_cni_set_ca;
	cni_set_timer1 = static_cni_set_timer1;
	cni_set_timer2 = static_cni_set_timer2;
	cni_set_timer1 = static_cni_set_timer1;
	cni_set_timer2 = static_cni_set_timer2;
	cni_ie_set_ti1 = static_cni_ie_set_ti1;
	cni_ie_set_ti2 = static_cni_ie_set_ti2;
}

/************************************************************************************/

/*@}*/
/*@}*/

