/**@addtogroup ctrl_sim
 * @{*/

/**@addtogroup ctrl_sim_int  Implementation
 * @{*/


/**@file
 * @brief C-Sim simulation wrapper for the communication controller
 *  (implementation).
 * @version 1.7
 * @date Last modified 2003-02-15
 * @author Petr Grillinger, Univerzity of West Bohemia, Czech Republic
 * @author Jan Jezek, Univerzity of West Bohemia, Czech Republic
 *
 * @note This file is part of C-Sim model of TTP/C protocol 0.1, 
 *  created as part of EU project FIT - Fault Injection for TTA, IST-1999-10748
 */

#include <math.h>
#include <string.h>

#include "csim/csim.h"
#include "ttpc/sp1ctrl.h"
#include "ttpc/fp1const.h"
#include "ttpc/fp1ctrl.h"
#include "ttpc/fp1cni.h"
#include "ttpc/fp1medl.h"
#include "ttpc/fp1frame.h"
#include "ttpc/sl1func.h"
#include "ttpc/sp1cni.h"
#include "ttpc/sp1busg.h"
#include "ttpc/sp1chan.h"
#include "ttpc/fl1debug.h"


/************************************************************************************/

/** Treshold that determines which double values should be rounded up.
If the fractional part of the number is greater than this value, the value is
considered to be the nearest integral number instead.
@note This constant is used only for clock computations, where the loss of
precission is not significant. */
#define ROUND_UP_TRESHOLD      (1.0 - 1e-9)

/** Time to wait before trying again (in macroticks) when garbage is received.*/
#define CONTROLLER_TIME_TO_RECOGNIZE_BAD_RECEIVE   100

/** This number is added to the hold time for ctrl process if the first attempt to
    wake on macrotick boundary fails.
The simluation should be precise if this number is smaller than 1E-3.*/
#define HOLD_TIME_ADD     (1e-6)

/** Duration of PRP in macroticks.*/
#define PRP_DURATION      ((unsigned)(0.4 * IFG_DURATION))

/** Duration of PSP in macroticks.*/
#define PSP_DURATION      ((unsigned)(0.4 * IFG_DURATION))


/** A shortcut that makes the logging process throughout the whole project transparent 
 * and unified.
 * @param msg This argument is special. It should contain the format specifier as in
 * a call to printf function and optional arguments, but the whole set of parameters
 * must be enclosed in one additional pair of parenthesis, e.g.: @code 
 *   log(("Testing entry number %d (made by %s)\n", 1, "me" )); @endcode */
#define log(msg) log_ctrl2_csim(msg)


/** Suspend execution for the calling C-Sim process for given number of time units.
This function guarantees, that the ctrl will wake-up 'exatly' at macrotick 
boundary, the wait time is modified with the ctrl's drift rate. 
@param cmodel Pointer to an instance of CONTROLLER_MODEL.
@param mt Number of macroticks to wait before resuming execution.*/
#define controller_hold(ctrl, mt)	do { \
	log(("Cluster time = %u, Holding...\n", cni_get_cluster_time(&(ctrl)->cni))); \
	controller_update_cni((CONTROLLER *)(ctrl)); \
	csim_hold(get_hold_time((ctrl), (mt)) + HOLD_TIME_ADD); \
} while (0)


/** Suspend execution for the calling C-Sim process until the given	cluster time. 
This function guarantees, that the ctrl will wake-up 'exatly' at macrotick 
boundary, the wait time is modified with the ctrl's drift rate. 
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param time Value of cluster time to wait for.*/
#define controller_wait_for(ctrl, time)	do { \
	unsigned t; \
	\
	t =	cni_get_cluster_time(&(ctrl)->cni); \
	log(("Cluster time = %u, Holding til %u\n", t, (time))); \
	controller_update_cni((CONTROLLER *)(ctrl)); \
	csim_hold(get_wait_for_time((ctrl), (time))); \
	t = cni_get_cluster_time(&(ctrl)->cni); \
	log(("Cluster time = %u\n", t)); \
	\
	if (t != (time)) { \
		log(("Error: Wait function failed, checking if we can repair the failuire\n")); \
		log(("Macrotick fraction = %0.6f\n", (ctrl)->mloc_data.macrotick_frac)); \
		if (t < (time) || t == 65535) { \
			log(("Yes, the hold was too short, so we just wait some more\n")); \
			csim_hold(get_wait_for_time((ctrl), (time)) + HOLD_TIME_ADD); \
			t = cni_get_cluster_time(&(ctrl)->cni); \
			log(("Cluster time = %u\n", t)); \
		} else { \
			log(("No, we waited for too long. Let the controller do what it can\n")); \
		} \
	} \
} while (0)


/** Suspend the execution of calling C-Sim process until the next PRP start is reached.
@note At the wake time, the C-State time field still contains the previous action time.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.*/
#define wait_for_prp(ctrl) do { \
	log(("Wait for PRP: c-state-time (%u) + slot-duration (%u) - IFG\n", \
		c_state_get_time( &((ctrl)->cni.c_state) ), \
		medl_sce_get_sru_slot_duration( controller_get_current_sce((CONTROLLER *)(ctrl)) ) )); \
	controller_wait_for( (ctrl), \
		(bit_mask(16)) & ( \
			c_state_get_time( &((ctrl)->cni.c_state) ) \
			+ medl_sce_get_sru_slot_duration( controller_get_current_sce( (CONTROLLER *)(ctrl) ) ) \
			- IFG_DURATION \
		) \
	); \
} while (0)


/** Suspend the execution of calling C-Sim process until the next PSP start is reached.
@param ctrl Pointer to an intstance of CONTROLLER_MODEL.*/
#define wait_for_psp(ctrl) \
	controller_wait_for((ctrl), bit_mask(16) & ( \
		c_state_get_time(&(ctrl)->cni.c_state) \
		- PSP_DURATION \
	))


/** Suspend the execution of calling C-Sim process until the next TP start is reached.
@param ctrl Pointer to an intstance of CONTROLLER_MODEL.*/
#define wait_for_tp(ctrl) \
	controller_wait_for((ctrl), bit_mask(16) & ( \
		c_state_get_time(&(ctrl)->cni.c_state) \
	))


/** Log ctrl's internal state (for debuging purposes).*/
#define log_controller_state(ctrl) do { \
	CNI *cni = &(ctrl)->cni; \
	unsigned i, j; \
	char mv[64 + 3 + 1];\
	\
	memset(mv, ' ', sizeof(mv)); \
	for (j = 0; j < 4; j++) \
		for (i = 0; i < 16; i++) \
			mv[j*17 + i] = (cni->c_state.membership[j*16 + i]) ? '1' : '0'; \
	mv[67] = 0; \
	log(("%s A=%u F=%u I=%u, TI1=%u, Timer1=%u\n", \
		cni_get_protocol_state_name(cni->protocol_status.protocol_state), \
		cni->protocol_status.accept_counter, \
		cni->protocol_status.failed_counter, \
		cni->protocol_status.invalid_counter, \
		cni->interrupt_enable.ti1, \
		cni->timer1 \
	)); \
	log(("MV=[%s]\n", mv)); \
} while (0)


/** Check the value of controller's @c co field and change state if not ON.
@param ctrl Pointer to an instance of CONTROLLER or derived structure.
@note This amcro assumes it is run from a loop and thus performs a @c break operation
	to exit this loop by force.*/
#define test_co(ctrl) \
	if (cni_get_co(&(ctrl)->cni) != CONTROLLER_ON) { \
		controller_set_state((ctrl), PS_FREEZE); \
		break; \
	}


/** Check whether a monitoring function is registered and optionally execute the function.
@param cmodel Pointer to an instance of CONTROLLER_MODEL.
@param state Execution state that is passed to the monitoring function (see @c CONTROLLER_EXE_STATE definition).*/
#define monitor(cmodel, state) do { \
	CONTROLLER_MODEL *ctrl = (cmodel); \
	\
	log(("Execution state: %s\n", #state)); \
	ctrl->mloc_data.exe_state = (state); \
	if (ctrl->monitor) \
		ctrl->monitor(ctrl, (state)); \
} while (0)


/************************************************************************************/
/* Static Functions */
/************************************************************************************/

/** Program for watch dog C-Sim process.*/
static void *watch_dog_prog(void *p_void);

/** Program for timer C-Sim process.*/
static void *prog_timer(void *p_void);

/** Program for ctrl model C-Sim process.*/
static void *prog_controller(void *p_void);

/** Destructor executed automatically by C-Sim when a ctrl process is terminated.*/
static void controller_destroy(void *p_void);

/** Reset the watchdog timer to the MEDL-defined value (start new count-down).*/
static void watch_dog_reset(CONTROLLER_MODEL *ctrl);

/** Dynamically create and initialize a new timer process.*/
static TIMER_PROCESS *timer_process_create(char *name);

/** Bind a ctrl model and the given C-Sim timer process together.*/
static void attach_timer_process(CONTROLLER_MODEL *ctrl, unsigned index, TIMER_PROCESS *timer);

/** Initialize simulation model specifiec fields of ctrl.*/
static CONTROLLER_MODEL *controller_model_init(CONTROLLER_MODEL *ctrl_model);

/** This function is called by the channel after a frame is received.*/
static void controller_model_receive_callback(void *ctrl, unsigned channel, CHANNEL_TS ts, FRAME *frame);

/** Transform duration in ctrl's microticks into simulation time units.*/
static CSIM_TIME micro_to_csim(CONTROLLER_MODEL *ctrl, double micro);

/** Transform duration in simulation time units into ctrl's microticks.*/
static double csim_to_micro(CONTROLLER_MODEL *ctrl, CSIM_TIME time);

/** Transform duration in ctrl's macroticks into simulation time units.*/
static CSIM_TIME macro_to_csim(CONTROLLER_MODEL *ctrl, double macro);

/** Transform duration in simulation time units into ctrl's macroticks.*/
static double csim_to_macro(CONTROLLER_MODEL *ctrl, CSIM_TIME time);

/** Compute the wait time (in simulation time units) required to sleep for given number
	of macroticks.*/
static CSIM_TIME get_hold_time(CONTROLLER_MODEL *ctrl, unsigned macroticks);

/** Compute the wait time (in simulation time units) required to sleep until the given
	cluster time.*/
static CSIM_TIME get_wait_for_time(CONTROLLER_MODEL *ctrl, unsigned time);

/** Compute clock correction term that should be applied during given simulation time.*/
static signed get_clock_correction(CONTROLLER_MODEL *ctrl, CSIM_TIME dt);

/** Compute next simulation action time and store it in to ctrl's local data.*/
static void set_next_action_time(CONTROLLER_MODEL *ctrl);

/** Calculate new value of the cluster time and store the value into CNI.*/
static void update_cluster_time(CONTROLLER_MODEL *ctrl);

/** Simulation specific implementation of the freeze protocol state.*/
static int model_freeze_state(CONTROLLER_MODEL *ctrl);

/** Simulation specific implementation of the init protocol state.*/
static void model_init_state(CONTROLLER_MODEL *ctrl);

/** Simulation specific implementation of the listen protocol state (phase 1/2).*/
static int model_prepare_listen_state(CONTROLLER_MODEL *ctrl);

/** Simulation specific implementation of the listen protocol state (phase 2/2).*/
static void model_finish_listen_state(CONTROLLER_MODEL *ctrl);

/** Simulation specific implementation of the cold start protocol state (phase 1/3).*/
static int model_cs_start(CONTROLLER_MODEL *ctrl);

/** Simulation specific implementation of the cold start protocol state (phase 2/3).*/
static void model_cs_prp(CONTROLLER_MODEL *model);

/** Simulation specific implementation of the cold start protocol state (phase 3/3).*/
static int model_cs_end(CONTROLLER_MODEL *ctrl);

/** Simulation specific TP start.*/
static int model_tp_start(CONTROLLER_MODEL *model);

/** Simulation specific PRP implementation (phase 1/2).*/
static int model_prp_start(CONTROLLER_MODEL *model);

/** Simulation specific PRP implementation (phase 2/2).*/
static int model_prp_end(CONTROLLER_MODEL *model);

/** Simulation specific PSP start.*/
static int model_psp_start(CONTROLLER_MODEL *model);

/** Synchronize cluster time using the frame received through the specified channel.*/
static void synchronize_time_on_frame(CONTROLLER_MODEL *ctrl, unsigned channel);

/** Start new clock synchronization interval.*/
static void reset_clock_sync(CONTROLLER_MODEL *ctrl);


/************************************************************************************/

/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL or derived structure.*/
static void watch_dog_reset(CONTROLLER_MODEL *ctrl)
{
	CSIM_TIME dt;

	dt = micro_to_csim(ctrl, medl_isp_get_watchdog_slot_length(&ctrl->medl.implementation_params));
	log(("Watchdog pacified for %0.2lf\n", dt));
	process_activate_delay((CSIM_PROCESS *)ctrl->watch_dog, dt);
}


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).
@return Used internaly by C-Sim (only in the POSIX thread implementation).*/
static csim_program(WATCH_DOG_PROCESS, watch_dog_prog)
	for (;;) {
		controller_run_watch_dog((CONTROLLER *)p_my->controller);
		csim_passivate();		
	}
csim_end_program


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param index Identifier of the timer (currently 0 or 1).
@param timer Pointer to an instance of TIMER_PROCESS.*/
static void attach_timer_process(CONTROLLER_MODEL *ctrl, unsigned index, TIMER_PROCESS *timer)
{
	ctrl->timer[index] = timer;
	timer->controller = (void *)ctrl;
	timer->timer = index;
}


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).
@return Used internaly by C-Sim (only in the POSIX thread implementation).*/
static csim_program(TIMER_PROCESS, prog_timer)
	for (;;) {
		log(("Timer activated\n"));
		controller_raise_timer_interrupt((CONTROLLER *)p_my->controller, p_my->timer);
		csim_passivate();
	}
csim_end_program


/**.
This function allocates dynamic memory for a new timer process. The memory is 
initialized with default values and a given C-Sim name.
@param name The internal name of the created timer process (used by C-Sim). It needs not
	be unique.
@return Pointer to a newly allocated and initialized instance of timer.*/
static TIMER_PROCESS *timer_process_create(char *name)
{
	TIMER_PROCESS *timer;

	timer = csim_new_process(TIMER_PROCESS, prog_timer);
	csim_set_name((CSIM_DYN_MEM *)timer, name);
	timer->timer = 0;

	return timer;
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param micro Time duration in controller's microticks that should be converted.
@return Equal number of simulation time units.*/
static CSIM_TIME micro_to_csim(CONTROLLER_MODEL *ctrl, double micro)
{
	CSIM_TIME x, p, drift, mm;

    if (micro == 0.0)
    	return 0.0; 

    mm = medl_ccp_get_microticks(&ctrl->medl.controller_cfg_params);
	drift = (1.0 + ctrl->clock_drift) * MICRO_PER_MACRO / mm;
	p = (drift - 1.0) / drift;
	x = micro / mm * MACROTICK_DURATION;

	return x - x * p;
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param time Time duration in simulation time units that should be converted.
@return Equal number of controller's microticks.*/
static double csim_to_micro(CONTROLLER_MODEL *ctrl, CSIM_TIME time)
{
	double x, drift, mm;

    if (time == 0.0)
    	return 0.0;

	mm = medl_ccp_get_microticks(&ctrl->medl.controller_cfg_params);
	drift = (1.0 + ctrl->clock_drift) * MICRO_PER_MACRO / mm;
	x = time / MACROTICK_DURATION * mm;

	return x * drift;
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param macro Time duration in macroticks that should be converted.
@return Equal number of simulation time units.*/
static CSIM_TIME macro_to_csim(CONTROLLER_MODEL *ctrl, double macro)
{
	CSIM_TIME x, p, drift, mm;

	if (macro == 0.0)
		return 0.0;

	mm = medl_ccp_get_microticks(&ctrl->medl.controller_cfg_params);
	drift = (1.0 + ctrl->clock_drift) * MICRO_PER_MACRO / mm;
	p = (drift - 1.0) / drift;
	x = macro * MACROTICK_DURATION;

	return x - x * p;
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param time Time duration in simulation time units that should be converted.
@return Equal number of macroticks.*/
static double csim_to_macro(CONTROLLER_MODEL *ctrl, CSIM_TIME time)
{
	CSIM_TIME x, drift, mm;

	if (time == 0.0)
		return 0.0;

	mm = medl_ccp_get_microticks(&ctrl->medl.controller_cfg_params);
	drift = (1.0 + ctrl->clock_drift) * MICRO_PER_MACRO / mm;
	x = time / MACROTICK_DURATION;

	return x * drift;
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param dt Time interval in simulation time units.
@return Clock correction term (in microticks) that should be applied after given time
	interval.*/
static signed get_clock_correction(CONTROLLER_MODEL *ctrl, CSIM_TIME dt)
{
	signed n, c, f;
	CSIM_TIME ft;

	c = ctrl->ploc_data.csct;
	if (c != 0 && dt > 0.0) {
		f = medl_ccp_get_free_running_macroticks(&ctrl->medl.controller_cfg_params) + 1;
		ft = macro_to_csim(ctrl, f);

		n = (unsigned)(dt / ft) + 1;
		if (n < abs(c))
			c = (c > 0) ? n : -n;
	}

	return c;
}


/**.
@note The computed duration of wait time allways ends at macrotick boundary.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param time Cluster time value to wait for (in macroticks).
@return Time to wait in simulation time units.*/
static CSIM_TIME get_wait_for_time(CONTROLLER_MODEL *ctrl, unsigned time)
{
	MLOC_DATA *mloc_data = &ctrl->mloc_data;
	CSIM_TIME dt, ccdt;

	/* Update Cluster Time field */
	update_cluster_time(ctrl);

	/* Compute time difference */
	dt = macro_to_csim(ctrl, 
		(bit_mask(16) & (time - cni_get_cluster_time(&ctrl->cni)))
		- mloc_data->macrotick_frac
	);

	ccdt = csim_time() - mloc_data->clock_correction_time + dt;
	dt -= micro_to_csim(ctrl, get_clock_correction(ctrl, ccdt));

	if (dt < 0.0)
		dt = 0.0;

	log(("Computed Hold Time = %lf, Macrotick fraction = %0.6lf\n", 
		dt, mloc_data->macrotick_frac));
	return dt;
}


/**.
 * 
 * The computed duration of wait time allways ends at macrotick boundary. To wait 
 * for the next macrotick, the right choice of Ca macroticks value is @c 1.
 *
 * @param ctrl Pointer to an instance of CONTROLLER_MODEL.
 * @param time Time interval to wait in macroticks.
 * @return Time to wait in simulation time units.*/
static CSIM_TIME get_hold_time(CONTROLLER_MODEL *ctrl, unsigned macroticks)
{
	MLOC_DATA *mloc_data = &ctrl->mloc_data;
	CSIM_TIME dt, ccdt;

	/* Update Cluster Time field */
	update_cluster_time(ctrl);

	dt = macro_to_csim(ctrl, macroticks - mloc_data->macrotick_frac);
	ccdt = csim_time() - mloc_data->clock_correction_time + dt;
	dt -= micro_to_csim(ctrl, get_clock_correction(ctrl, ccdt));

	if (dt < 0.0)
		dt = 0.0;

	log(("Computed Hold Time: %lf\n", dt));
	return dt;
}


/**
@param ctrl Pointer to an instance of CONTROLLER_MODEL.*/
static void set_next_action_time(CONTROLLER_MODEL *ctrl)
{
	CNI *cni = &ctrl->cni;
	unsigned t = c_state_get_time(&cni->c_state);
	unsigned ct;

	ctrl->mloc_data.action_time = csim_time() + get_wait_for_time(ctrl, t);
	ct = cni_get_cluster_time(cni);
	log(("CState-Time = %d, Cluster-Time = %d,  Next Action Time = %0.4f\n",
		t, ct, ctrl->mloc_data.action_time));
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.*/
static void reset_clock_sync(CONTROLLER_MODEL *ctrl)
{
	ctrl->mloc_data.clock_correction_time = csim_time();
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.*/
static void model_init_state(CONTROLLER_MODEL *ctrl)
{
	controller_init_state((CONTROLLER *)ctrl);
}


/**.
Calls first the protocol layer listen state initialization and the enables all wake-up
flags for the controller.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
return  The code returned by controller_prepare_listen_state().*/
static int model_prepare_listen_state(CONTROLLER_MODEL *ctrl)
{
	int result = controller_prepare_listen_state((CONTROLLER *)ctrl);
	MLOC_DATA *mloc_data = &ctrl->mloc_data;

	if (result == 0) {
		mloc_data->wake_on_co = 1;
		mloc_data->wake_on_ca = 1;
		mloc_data->wake_on_receive = 1;
	}

	return result;
}


/**.
Calls first the protocol layer listen state finishing function and the disables 
no longer required wake-up flags for the controller.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.*/
static void model_finish_listen_state(CONTROLLER_MODEL *ctrl)
{
    MLOC_DATA *mloc_data = &ctrl->mloc_data;

	mloc_data->wake_on_receive = 0;
	mloc_data->wake_on_ca = 0;

	controller_finish_listen_state((CONTROLLER *)ctrl);
	if (cni_ps_get_c_state_valid(&ctrl->cni.protocol_status) == C_STATE_VALID)
		set_next_action_time(ctrl);
}


/**.
This moves the controller's time - it is the only place, where the cluster time field
is directly updated. The time is not updated periodically, instead the value is
computed from the elapsed simulation time when the cluster time field is accessed.
This function also introduces clock drift effect and applies possible clock 
corrections requested by the clock synchronization algorithm.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
\todo Make sure that floating-point arithmetics works (e.g. <code>frac(x) < 1.0</code>)*/
static void update_cluster_time(CONTROLLER_MODEL *ctrl)
{
	CNI *cni = &ctrl->cni;
	PLOC_DATA *ploc_data = &ctrl->ploc_data;
	MLOC_DATA *mloc_data = &ctrl->mloc_data;
	MEDL_CCP *medl_ccp = &ctrl->medl.controller_cfg_params;
	int c;
	unsigned t;
	CSIM_TIME dt, x, ff, fi;

	/* Clock Correction (CC) */
	dt = csim_time() - mloc_data->clock_correction_time;   /* time elapsed since last CC */
	c = get_clock_correction(ctrl, dt);   /* CC term for the elapsed time (dt) */

	ploc_data->csct -= c;       /* reduce the total csct by applied CC term */
	mloc_data->clock_correction_time += abs(c) *
		macro_to_csim(ctrl, medl_ccp_get_free_running_macroticks(medl_ccp));

	/* 'x' is time that elapsed since the last returned macrotick 
	   (not neccessarilly the last whole macrotick)*/
	x = csim_time() - mloc_data->last_clock_access;
	x = csim_to_macro(ctrl, x);   /* add clock drift effect and convert to macroticks */
	x += mloc_data->macrotick_frac;   /* Add unaccounted fractional part of current macrotick */
	x += (double)c / medl_ccp_get_microticks(medl_ccp);   /* apply CC term */

	/* handle double numbers with fract. part close to 1.0:
	   (sometimes the modf() function returns fract. part greater than 1.0) */
	ff = modf(x, &fi);
	if (ff > ROUND_UP_TRESHOLD) {
		fi += 1.0;
		ff = 0.0;
		log(("Double number close to 1 detected (%lf exactly), rounding up to %lf\n",
			x, fi + ff));
	}

	/* update last clock access to current time */
	mloc_data->last_clock_access = csim_time();
	mloc_data->macrotick_frac = ff;

	/* Increment cluster time by the integral part of elapsed macroticks. */
	t = (unsigned)fi + cni->cluster_time;
	if (t > bit_mask(16))
		cni_increment_time_overflow_counter(cni, t >> 16);
	t &= bit_mask(16);
	cni->cluster_time = t;

	/* If CC has just been applied, timer interrupt processes should be replanned. */
	if (c != 0) {
		controller_timer_changed(ctrl, 0, 1);
		controller_timer_changed(ctrl, 1, 1);
	}
}


/**.
First cancels all planned process that belong to the controller, then calls
the protocol defined function for freeze state.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@return  The code returned by controller_freeze_state(). */
static int model_freeze_state(CONTROLLER_MODEL *ctrl)
{
	process_cancel((CSIM_PROCESS *)ctrl->watch_dog);
	process_cancel((CSIM_PROCESS *)ctrl->timer[0]);
	process_cancel((CSIM_PROCESS *)ctrl->timer[1]);
	process_cancel((CSIM_PROCESS *)((BUS_GUARDIAN_MODEL *)(ctrl->bus_guardian))->process);

	return controller_freeze_state((CONTROLLER *)ctrl);
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@param channel Number if channel that transmitted the frame to used for synchronization.*/
static void synchronize_time_on_frame(CONTROLLER_MODEL *ctrl, unsigned channel)
{
	CNI *cni = &ctrl->cni;
	MLOC_DATA *mloc = &ctrl->mloc_data;
	PLOC_DATA *ploc = &ctrl->ploc_data;

	/* Set Cluster Time */
	cni->cluster_time = c_state_get_time(&cni->c_state);
	mloc->last_clock_access = mloc->transmission_start[channel];
	mloc->macrotick_frac = 0.0;
	ploc->csct = 0;
	cni_ps_set_csct(&cni->protocol_status, 0);
	update_cluster_time(ctrl);
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@return Pointer to the initialized structure (the same is @a ctrl_model arg).*/
static CONTROLLER_MODEL *controller_model_init(CONTROLLER_MODEL *ctrl_model)
{
	CONTROLLER *ctrl = (CONTROLLER *)ctrl_model;
	char *s, t[] = "BG - ";
	int i;
	char *ts[2], *wds;

	controller_init(ctrl);
	mloc_data_initialize(&ctrl_model->mloc_data);

	ctrl_model->receive_callback = controller_model_receive_callback;
	ctrl_model->update_cluster_time = (CONTROLLER_FN)update_cluster_time;
	ctrl_model->synchronize_time = (CONTROLLER_INT_FN)synchronize_time_on_frame;
	ctrl_model->reset_clk_correction = (CONTROLLER_FN)reset_clock_sync;

	ctrl_model->clock_drift = 0.0;
	ctrl_model->monitor = NULL;
	for (i = 0; i < 2; i++) {
		ctrl_model->mtd_monitor[i] = NULL;
		ctrl_model->mtd_monitor_arg[i] = NULL;
		ctrl_model->timer[i] = NULL;
	}

	s = (char *)malloc(strlen(t) + strlen(ctrl_model->process->name) + 1);
	strcpy(s, t);
	strcat(s, ctrl_model->process->name);
	ctrl->bus_guardian = (BUS_GUARDIAN *)bus_guardian_create(s);
	bus_guardian_attach_controller((BUS_GUARDIAN_MODEL *)ctrl->bus_guardian, ctrl);
	process_activate_delay((CSIM_PROCESS *)((BUS_GUARDIAN_MODEL *)ctrl->bus_guardian)->process, 0.0);

	ts[0] = (char *)malloc(strlen(ctrl_model->process->name) + 10);
	ts[1] = (char *)malloc(strlen(ctrl_model->process->name) + 10);
	strcpy(ts[0], ctrl_model->process->name);
	strcat(ts[0], ": Timer 1");
	strcpy(ts[1], ctrl_model->process->name);
	strcat(ts[1], ": Timer 2");
	for (i = 0; i < 2; i++)
		attach_timer_process(ctrl_model, i, timer_process_create(ts[i]));

	wds = (char *)malloc(strlen(ctrl_model->process->name) + 16);
	strcpy(wds, ctrl_model->process->name);
	strcat(wds, ": Watch Dog");

	ctrl_model->watch_dog = csim_new_process(WATCH_DOG_PROCESS, watch_dog_prog);
	csim_set_name((CSIM_DYN_MEM *) ctrl_model->watch_dog, wds);
	ctrl_model->watch_dog->controller = ctrl_model;

	return ctrl_model;
}


/**.
@param ctrl_model Pointer to an instance of CONTROLLER_MODEL.
@param channel Numerical indentification of channel that transmitts the frame.
@param ts Current transmission state of the frame.
@param frame Pointer to the current version of received data.*/
static void controller_model_receive_callback(void *ctrl_model, unsigned channel, CHANNEL_TS ts, FRAME *frame)
{
	CONTROLLER_MODEL *ctrl = (CONTROLLER_MODEL *) ctrl_model;
	MLOC_DATA *mloc = &ctrl->mloc_data;
	PLOC_DATA *ploc = &ctrl->ploc_data;
	CSIM_TIME *tr_start = mloc->transmission_start;
	CSIM_TIME dt;

	ploc->frame[channel] = *frame;

	if (ts == CHANNEL_TS_START) {
		tr_start[channel] = csim_time();
		dt = tr_start[channel] - mloc->action_time;


		if (fabs(dt) > RECEIVE_WINDOW_LEN) {
			if (cni_ps_get_c_state_valid(&ctrl->cni.protocol_status) != C_STATE_VALID) {
				/* Synchronization frame (no valid C-State yet) */
				log(("Controller is not synchronized - skipping clock synchronization meassurements.\n"));
				ploc->frame[channel].state = FRAME_STATE_RECEIVING;
			} else {
				/* Invalid frame: Outside the reception window */
				log(("Error: A frame outside the allocated reception window has been "
					"detected; Discarding the frame.\n"));
				ploc->frame[channel].state = FRAME_STATE_NULL;
				controller_stop_receive((CONTROLLER *)ctrl);
			}
		} else {
			/* Valid Frame - store clock synchronization meassurements */
			ploc->frame[channel].state = FRAME_STATE_RECEIVING;
			if (channel == 0)
				ploc->mtd_a = (int)csim_to_micro(ctrl, dt);
			else
				ploc->mtd_b = (int)csim_to_micro(ctrl, dt);

			if (ctrl->mtd_monitor[channel])
				ctrl->mtd_monitor[channel]((CONTROLLER *)ctrl, ctrl->mtd_monitor_arg[channel]);
		}
	} else {
		mloc->transmission_stop[channel] = csim_time();
		if (mloc->wake_on_receive && frame->state == FRAME_STATE_COMPLETE)
			process_activate_delay((CSIM_PROCESS *)ctrl->process, RECEIVE_CALLBACK_DELAY);
	}
}


/**.
Calls protocol defined cold start phase 1 function and additionally enables several
wake-up flags within the controller.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@return The code returned by controller_prepare_cold_start_state().*/
static int model_cs_start(CONTROLLER_MODEL *ctrl)
{
	int result;
	MLOC_DATA *mloc_data = &ctrl->mloc_data;

	result =  controller_prepare_cold_start_state((CONTROLLER *)ctrl);
	if (result == 0) {
		mloc_data->wake_on_ca = 1;
		mloc_data->wake_on_receive = 1;
	}

	return result;
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.*/
static void model_cs_prp(CONTROLLER_MODEL *ctrl)
{
	controller_cold_start_bg_arm((CONTROLLER *)ctrl);
}


/**.
@param ctrl Pointer to an instance of CONTROLLER_MODEL.
@return The code returned by controller_finish_cold_start_state().*/
static int model_cs_end(CONTROLLER_MODEL *ctrl)
{
	int ret;
	CSIM_TIME dt;
	MLOC_DATA *mloc_data = &ctrl->mloc_data;

	mloc_data->wake_on_ca = 0;
	mloc_data->wake_on_receive = 0;

	dt = csim_time() - mloc_data->tmp_model_time;
	controller_run_cold_start_state((CONTROLLER *)ctrl,
		(int)csim_to_macro(ctrl, dt));

	ret = controller_finish_cold_start_state((CONTROLLER *)ctrl);

	if (!ret) {
		controller_set_next_slot((CONTROLLER *)ctrl);
		set_next_action_time(ctrl);
	}
	return ret;
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.
@return The code returned by controller_tp_start().*/
static int model_tp_start(CONTROLLER_MODEL *model)
{
	CSIM_TIME *ts = model->mloc_data.transmission_start;

	ts[0] = ts[1] = csim_time();
	return controller_tp_start((CONTROLLER *)model);
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.
@return
	- 0 if there is no error.
	- 1 if a NBW assertion violation occurs.*/
static int model_prp_start(CONTROLLER_MODEL *model)
{
	CONTROLLER *ctrl = (CONTROLLER *)model;
	
	if (controller_enter_nbw(ctrl))
		return 1;
	
	controller_prp(ctrl);
	controller_perform_cs(ctrl);
	controller_set_next_slot(ctrl);
	set_next_action_time(model);

	return 0;
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.
@return
	- 0 if there is no error.
	- 1 if a NBW assertion violation occurs.*/
static int model_prp_end(CONTROLLER_MODEL *model)
{
	return controller_leave_nbw((CONTROLLER *)model);
}


/**.
Checks whether the current slot is the sending slot of the controller. If not, then
the controller start to listen for incoming frames, otherwise the @c psp_phase field
within local data is set to 1.
@param model Pointer to an instance of CONTROLLER_MODEL.
@return
	- 0 if there is no error.
	- 1 if the listen operation failed.
	- 2 if the controller has been switched off using the @c co field.
\todo This function can be safely moved into functional layer (fp1ctr).*/
static int model_psp_start(CONTROLLER_MODEL *model)
{
	CONTROLLER *ctrl = (CONTROLLER *) model;
	PLOC_DATA *ploc = &ctrl->ploc_data;

	ploc->psp_phase = !controller_psp(ctrl);
	
	if (ploc->psp_phase)
		return controller_enter_nbw(ctrl);

	if (cni_get_co(&ctrl->cni) != CONTROLLER_ON)
		return 2;

	return controller_start_receive(ctrl);
}


/**.
 * 
 * @param p_void Used internally by C-Sim (pointer to internal state of the process).
 * @return Used internaly by C-Sim (only in the POSIX thread implementation).*/
static csim_program(CONTROLLER_PROCESS, prog_controller)
	for (;;) {
		log_controller_state(p_my->model);
		controller_check_state(p_my->protocol); /* check for state changes */

		/* switch by stable p_my->protocol state */
		switch (p_my->protocol->ploc_data.protocol_state) {

		case PS_INIT:
			monitor(p_my->model, CES_INIT);
			model_init_state(p_my->model);
			break;

		case PS_LISTEN:
			test_co(p_my->protocol);
			monitor(p_my->model, CES_LISTEN_INIT);
			if (model_prepare_listen_state(p_my->model))
				break;

			if (controller_start_receive(p_my->protocol)) {
				log(("Invalid frame received - outside the time bounds.\n"));
				monitor(p_my->model, CES_LISTEN_RESTART);
				controller_hold(p_my->model, CONTROLLER_TIME_TO_RECOGNIZE_BAD_RECEIVE);
				test_co(p_my->protocol);
				break;
			}

			if (controller_is_cold_start_allowed(p_my->protocol)) {
				log(("CS allowed: Starting listen timeout.\n"));
				controller_hold(p_my->model, controller_get_listen_timeout(p_my->protocol));
			} else {
				log(("CS disabled: Waiting for I-Frame.\n"));
				controller_update_cni(p_my->protocol);
				csim_passivate();
			}

			test_co(p_my->protocol);
			monitor(p_my->model, CES_LISTEN_FINISH);
			model_finish_listen_state(p_my->model);
			
/*			if (cni_ps_get_c_state_valid(&p_my->protocol->cni.protocol_status) == C_STATE_VALID) {
				log(("Synchronized: waiting for PRP start\n"));
				controller_wait_for(p_my->model, bit_mask(16) & ( 
					c_state_get_time(&p_my->protocol->cni.c_state) 
					- IFG_DURATION));
				controller_check_immediate_ba(p_my->protocol);
			}*/
			break;

		case PS_COLD_START:
			test_co(p_my->protocol);
			monitor(p_my->model, CES_COLD_INIT);
			if (model_cs_start(p_my->model))
				break;

			/**@bug Here perform cold-start-timeout*/

			/* Cold-start: PRP - start of IFG */
			wait_for_prp(p_my->model);
			test_co(p_my->protocol);
			monitor(p_my->model, CES_COLD_PRP);
			model_cs_prp(p_my->model);
			test_co(p_my->protocol);

			if (controller_start_receive(p_my->protocol)) {
				/* There is a garbage on the bus, wait a while to recognize it. */
				log(("Incomplete frame received.\n"));
				monitor(p_my->model, CES_COLD_RESTART);
				controller_hold(p_my->model, CONTROLLER_TIME_TO_RECOGNIZE_BAD_RECEIVE);
				controller_set_state(p_my->protocol, PS_LISTEN);
				break;
			}

			/* The bus is silent, hold on for the cold start timeout period. */
			p_my->model->mloc_data.tmp_model_time = csim_time();
			controller_hold(p_my->model,
				controller_get_cold_start_timeout(p_my->protocol)
				- medl_sce_get_sru_slot_duration(controller_get_current_sce(p_my->protocol))
				+ IFG_DURATION
			);
			test_co(p_my->protocol);

			monitor(p_my->model, CES_COLD_FINISH);
			model_cs_end(p_my->model);
			break;

		case PS_STARTUP:
			/* FIXME: Don't test the Host lifesign field in the first slot in TDMA round */
		case PS_APPLICATION:
		case PS_READY:
		case PS_PASSIVE:
			/* PSP Start */
			test_co(p_my->protocol);
			wait_for_psp(p_my->model);
			test_co(p_my->protocol);
			monitor(p_my->model, CES_PSP);
			model_psp_start(p_my->model);
			test_co(p_my->protocol);

			/* TP Start */
			wait_for_tp(p_my->model);
			test_co(p_my->protocol);
			if (controller_tp_sync_assertion(p_my->protocol))
				break;
			monitor(p_my->model, CES_TP);

			if (p_my->protocol->ploc_data.psp_phase) {
				if (model_tp_start(p_my->model))
					break;
			} else if (controller_is_current_sender(p_my->protocol)) {
				if (p_my->protocol->bus_guardian->arm_signal(p_my->protocol->bus_guardian, 0))
					break;
			}

			watch_dog_reset(p_my->model);
			if (controller_bg_arm(p_my->protocol))
				break;
			controller_raise_tp_sync_interrupt(p_my->protocol);

			/* PRP Start */
			wait_for_prp(p_my->model);
			monitor(p_my->model, CES_PRP);
			test_co(p_my->protocol);
			model_prp_start(p_my->model);
			test_co(p_my->protocol);
			controller_hold(p_my->model, PRP_DURATION);
			test_co(p_my->protocol);
			model_prp_end(p_my->model);

			monitor(p_my->model, CES_IDLE);
			break;

		case PS_AWAIT:
			controller_update_cni(p_my->protocol);
			csim_passivate();
			controller_set_state(p_my->protocol, PS_FREEZE);
			break;

		case PS_SELFTEST:
			controller_update_cni(p_my->protocol);
			csim_passivate();
			controller_set_state(p_my->protocol, PS_FREEZE);
			break;

		case PS_FREEZE:
			if (model_freeze_state(p_my->model))
				csim_passivate();
			break;

		case PS_DOWNLOAD:
			controller_download_state(p_my->protocol);
			break;

		default:
			controller_set_state(p_my->protocol, PS_FREEZE);
			break;
		}
	}
csim_end_program


/**.
@param p_void Used internally by C-Sim (pointer to internal state of the process).*/
static csim_begin_destructor(CONTROLLER_PROCESS, controller_destroy)
	free(p_my->model);
csim_end_destructor


/************************************************************************************/
/* Exported Functions */
/************************************************************************************/

/**.
@param name Internal description of the process to create that is used by C-Sim for
	logging and debugging. Needs not be unique.
@return Pointer to the new instance of CONTROLLER_MODEL.*/
CONTROLLER_MODEL *controller_create(char *name)
{
	CONTROLLER_MODEL *controller_model;

	controller_model = (CONTROLLER_MODEL *)malloc(sizeof(CONTROLLER_MODEL));

	controller_model->process = csim_new_process(CONTROLLER_PROCESS, prog_controller);
	controller_model->process->protocol = (CONTROLLER *)controller_model;
	controller_model->process->model = controller_model;
	csim_set_name((CSIM_DYN_MEM *)controller_model->process, name);
	csim_set_destructor(controller_model->process, controller_destroy);

	return controller_model_init(controller_model);
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.
@param timer Numerical identification of the timer (either 0 or 1).
@param cs This flag denotes that the timer settings did not actually change, instead
	the clock synchronization algorithm modified the clock speed.*/
void controller_timer_changed(CONTROLLER_MODEL *model, int timer, int cs)
{
	CNI *cni = &model->cni;
	unsigned mt;

	timer &= 1;
	if (timer == 0) {
		mt = (cni_get_timer1(cni) - cni_get_cluster_time(cni)) & bit_mask(16);
		if ((cs || mt > 0) && cni_ie_get_ti1(&cni->interrupt_enable))
			process_activate_delay((CSIM_PROCESS *)model->timer[0], get_hold_time(model, mt) + HOLD_TIME_ADD);
		else
			process_cancel((CSIM_PROCESS *)model->timer[0]);
	}
	if (timer == 1) {
		mt = (cni_get_timer2(cni) - cni_get_cluster_time(cni)) & bit_mask(16);
		if ((cs || mt > 0) && cni_ie_get_ti2(&cni->interrupt_enable))
			process_activate_delay((CSIM_PROCESS *)model->timer[1], get_hold_time(model, mt) + HOLD_TIME_ADD);
		else
			process_cancel((CSIM_PROCESS *)model->timer[1]);
	}
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.*/
void controller_co_changed(CONTROLLER_MODEL *model)
{
	CONTROLLER *ctrl = (CONTROLLER *)model;
	CNI *cni = &ctrl->cni;
	unsigned value = cni_get_co(cni);

	process_activate_delay((CSIM_PROCESS *)model->process, 0.0);

	if (value == CONTROLLER_ON && cni_ps_get_protocol_state(&cni->protocol_status) != PS_FREEZE)
		cni_ps_set_protocol_state(&cni->protocol_status, PS_FREEZE);
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.*/
void controller_ca_changed(CONTROLLER_MODEL *model)
{
	CONTROLLER *ctrl = (CONTROLLER *)model;
	unsigned value = cni_get_ca(&ctrl->cni);

	if (value == CONTROLLER_AWAIT && model->mloc_data.wake_on_ca)
		process_activate_delay((CSIM_PROCESS *)model->process, 0.0);
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.
@param monitor_fn Pointer to the monitoring function.*/
void controller_set_monitor(CONTROLLER_MODEL *model, MONITOR_FN monitor_fn)
{
	model->monitor = monitor_fn;
}


/**.
 *
 * @param model Pointer to an instance of CONTROLLER_MODEL.
 * @param mtd_fn Pointer to the monitoring function.
 * @param arg  Argument passed to @a mtd_fn whenever it is called.
 * @param channel Index of a single channel (0 is channel A, 1 channel B.*/
void controller_set_mtd_monitor(CONTROLLER_MODEL *model, CONTROLLER_PTR_FN mtd_fn, void *arg, int channel)
{
	model->mtd_monitor[channel] = mtd_fn;
	model->mtd_monitor_arg[channel] = arg;
}


/**.
@param model Pointer to an instance of CONTROLLER_MODEL.
@param drift Clock drift rate in sec/sec.*/
void controller_set_clock_drift(CONTROLLER_MODEL *model, double drift)
{
	model->clock_drift = drift;
}


/*@}*/
/*@}*/

